/*********************************************************************************
 * dot-net.plugin.c ($Revision: 241 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#define WIN32_LEAN_AND_MEAN
#include "libdotnet.plugin.h"
   
/* ---------------------------------------------------------------------------------------------- 
    Initialize .NET callbacks

    Save handles from the callbacks to the shared memory segment and 
    create a thread to handle incoming events from dtmfbox.

    ev_init   = Callback to be called, when plugin is loaded
    ev_close  = Callback to be called, when plugin is unloaded
    ev_event  = Callback to be called, when a new (phone) event arrives
    ev_output = Callback to be called, when dtmfbox wants to output something    
    no_thread = when not zero then no threading is used
   ---------------------------------------------------------------------------------------------- */
EXPORT int plugin_callback_init(plugin_cb_init ev_init, plugin_cb_close ev_close, plugin_cb_event ev_event, plugin_cb_output ev_output, int no_thread)
{   
    // save callbacks from application
    pl_init = ev_init;
    pl_close = ev_close;
    pl_event = ev_event;
    pl_output = ev_output;        
   
    unlock(1);

    // start as thread or poll events with plugin_callback_event()?
    if(no_thread == 0)
    {        
        application_running = 1;
        t_application = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)&application_thread, 0, 0, NULL);                         
        return (int)(t_application != NULL);
    } 
    else 
    {
        return 1;
    }
}

/* ----------------------------------------------------------------------------------------------  
    Manually raise events, when not started as thread
   ---------------------------------------------------------------------------------------------- */
EXPORT int plugin_callback_event()
{
   int ret = g_status_app;
   application_thread(); 
   return ret;
}

/* ----------------------------------------------------------------------------------------------  
    Destroy .NET callbacks
   ---------------------------------------------------------------------------------------------- */
EXPORT int plugin_callback_close()
{   
    pl_event = NULL;
    pl_output = NULL;
    pl_close = NULL;
    pl_init = NULL;        
    if(t_application) WaitForSingleObject(t_application, 2500);
    t_application = NULL;
    return 0;
}

/* ----------------------------------------------------------------------------------------------  
    Execute dtmfbox command with parameter array (as on console) 

    argc = argument count
    args = arguments 
   ---------------------------------------------------------------------------------------------- */
EXPORT int plugin_execute_command(int argc, char *args[])
{       
    int i;
    int ret = -1;

    lock(0);
    
    if(argc > 255) argc = 255;
    for(i=0; i<argc; i++)
    {
        memset(g_arg[i], 0, 255);
        strncpy_s(g_arg[i], 255, args[i], 255);
    }
    g_argc = argc;    
    
    // Raise event "execute"
    raise(0, 1); 
    lock(0);
    ret = g_ret_id;
    g_ret_id = -3;
    unlock(0);

    return ret;
}

/* ---------------------------------------------------------------------------------------------- 
    (Internal) Plugin is loaded by dtmfbox! 
   ---------------------------------------------------------------------------------------------- */
EXPORT void plugin_init(plugin_exec_command cmd)
{    
    // save callback to execute commands in dtmfbox.exe
    dtmfbox = cmd;  

    // and run thread (wait for commands to execute)...
    unlock(0);
    dtmfbox_running = 1;
    t_dtmfbox = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)&dtmfbox_thread, 0, 0, NULL);    

    // raise event "pl_init()" (3)
    if(pl_init)
    {
        lock(1);
        raise(1, 3);
        lock(1); unlock(1);
    }
}

/* ---------------------------------------------------------------------------------------------- 
    (Internal) Plugin is unloaded by dtmfbox! 
   ---------------------------------------------------------------------------------------------- */
EXPORT void plugin_close() 
{
    // raise event "pl_close()" (4)
    if(pl_close)
    {
        lock(1);
        raise(1, 4);
        lock(1); unlock(1);                
    }
    
    dtmfbox_running = 0;
    application_running = 0;

    // wait till thread(s) ended
    if(t_application) WaitForSingleObject(t_application, 2500);
    if(t_dtmfbox) WaitForSingleObject(t_dtmfbox, 2500);    
    t_dtmfbox = NULL; t_application = NULL;

    // release callbacks
    dtmfbox = NULL;
    pl_event = NULL;
    pl_output = NULL;
    pl_close = NULL;
    pl_init = NULL;    
}

/* ---------------------------------------------------------------------------------------------- 
    (Internal) Callback gets called when a new (phone) event arrives 
   ---------------------------------------------------------------------------------------------- */
EXPORT void plugin_event(char *scriptfile, char *s_event, char *s_direction, char *s_type, int conid_src, int conid_dst, char *no_src, char *no_dst, char *data, int acc_id) 
{    
    if(dtmfbox_running && application_running)
    {
        lock(1);

        strncpy_s(g_scriptfile, sizeof(g_scriptfile), scriptfile, sizeof(g_scriptfile));
        strncpy_s(g_event, sizeof(g_event), s_event, sizeof(g_event));
        strncpy_s(g_direction, sizeof(g_direction), s_direction, sizeof(g_direction));
        strncpy_s(g_type, sizeof(g_type), s_type, sizeof(g_type));
        strncpy_s(g_no_src, sizeof(g_no_src), no_src, sizeof(g_no_src));
        strncpy_s(g_no_dst, sizeof(g_no_dst), no_dst, sizeof(g_no_dst));
        strncpy_s(g_data, sizeof(g_data), data, sizeof(g_data));
        g_accid = acc_id;
        g_conid_src = conid_src;
        g_conid_dst = conid_dst;
        
        // Raise event "phone-event"...
        raise(1, 1);
        lock(1); unlock(1);
    }
}

/* ---------------------------------------------------------------------------------------------- 
    (Internal) Callback gets called, when dtmfbox outputs something (connection id, -call, -list, ...) 
   ---------------------------------------------------------------------------------------------- */
EXPORT void plugin_output(char *output, int pid) 
{ 
    lock(1);

    strncpy_s(g_output, sizeof(g_output), output, sizeof(g_output));
    g_pid = pid;
    
    // Raise event "output"...
    raise(1, 2);
    //lock(1); unlock(1);
}

/* ---------------------------------------------------------------------------------------------- 
    Thread for application.
    Wait for new events/outputs and raise the initialized callbacks.
   ---------------------------------------------------------------------------------------------- */
void application_thread()
{
    //MSG stMsg = { 0 };
    while(application_running)
    {
        switch(g_status_app)
        {
            // New (phone) event arrived
            case 1:                
                if(pl_event) pl_event(g_scriptfile, g_event, g_direction, g_type, g_conid_src, g_conid_dst, g_no_src, g_no_dst, g_data, g_accid);
                unlock(1);
                break;
            
            // Output from dtmfbox.exe
            case 2:
                if(pl_output) pl_output(g_output, g_pid);
                unlock(1);
                break;   

            // Plugin initialized!
            case 3:
                if(pl_init) pl_init();
                unlock(1);
                break;
            
            // Plugin unloaded!
            case 4:
                if(pl_close) pl_close();
                unlock(1);
                break;

            default:                
               Sleep(9);              
        }

        Sleep(1);              
    }

    //unlock(1);    
}

/* ---------------------------------------------------------------------------------------------- 
    Thread for dtmfbox.exe
    Waits for commands and executes them.
   ---------------------------------------------------------------------------------------------- */
void dtmfbox_thread()
{
    while(dtmfbox_running)
    {   
        switch(g_status_dtmfbox)
        {
            case 1:                                     
                // execute dtmfbox command (TODO: make better - currently only 130 parametes at once)
                if(dtmfbox)
                    g_ret_id = dtmfbox(g_argc, 
                        g_arg[0], g_arg[1], g_arg[2], g_arg[3], g_arg[4], g_arg[5], g_arg[6], g_arg[7], g_arg[8], g_arg[9], 
                        g_arg[10], g_arg[11], g_arg[12], g_arg[13], g_arg[14], g_arg[15], g_arg[16], g_arg[17], g_arg[18], g_arg[19],
                        g_arg[20], g_arg[21], g_arg[22], g_arg[23], g_arg[24], g_arg[25], g_arg[26], g_arg[27], g_arg[28], g_arg[29],
                        g_arg[30], g_arg[31], g_arg[32], g_arg[33], g_arg[34], g_arg[35], g_arg[36], g_arg[37], g_arg[38], g_arg[39],
                        g_arg[40], g_arg[41], g_arg[42], g_arg[43], g_arg[44], g_arg[45], g_arg[46], g_arg[47], g_arg[48], g_arg[49],
                        g_arg[50], g_arg[51], g_arg[52], g_arg[53], g_arg[54], g_arg[55], g_arg[56], g_arg[57], g_arg[58], g_arg[59],
                        g_arg[60], g_arg[61], g_arg[62], g_arg[63], g_arg[64], g_arg[65], g_arg[66], g_arg[67], g_arg[68], g_arg[69],
                        g_arg[70], g_arg[71], g_arg[72], g_arg[73], g_arg[74], g_arg[75], g_arg[76], g_arg[77], g_arg[78], g_arg[79],
                        g_arg[80], g_arg[81], g_arg[82], g_arg[83], g_arg[84], g_arg[85], g_arg[86], g_arg[87], g_arg[88], g_arg[89],
                        g_arg[90], g_arg[91], g_arg[92], g_arg[93], g_arg[94], g_arg[95], g_arg[96], g_arg[97], g_arg[98], g_arg[99],
                        g_arg[100], g_arg[101], g_arg[102], g_arg[103], g_arg[104], g_arg[105], g_arg[106], g_arg[107], g_arg[108], g_arg[109],
                        g_arg[110], g_arg[111], g_arg[112], g_arg[113], g_arg[114], g_arg[115], g_arg[116], g_arg[117], g_arg[118], g_arg[119],
                        g_arg[120], g_arg[121], g_arg[122], g_arg[123], g_arg[124], g_arg[125], g_arg[126], g_arg[127], g_arg[128], g_arg[129]
                );
                unlock(0);
                break;
            
            default:
                Sleep(9);
                break;
        }        
        
        Sleep(1);              
    }
     
    //unlock(0);
}

/* ---------------------------------------------------------------------------------------------- 
    Private functions
   ---------------------------------------------------------------------------------------------- */
void lock(int app)
{
    if(app) {
        while(g_status_app != 0) Sleep(1);
        g_status_app = -1;
    }

    if(!app) {
        while(g_status_dtmfbox != 0)  Sleep(1);
        g_status_dtmfbox = -1;    
    }
}

void raise(int app, int value)
{
    if(app) g_status_app = value;
    if(!app) g_status_dtmfbox = value;
}

void unlock(int app)
{
    raise(app, 0);
}

