/*********************************************************************************
 * dot-net.plugin.h ($Revision: 241 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifdef WIN32
#define EXPORT __declspec(dllexport) 
#else
#define EXPORT
#endif

// Callbacks to .NET app (--> delegates)
typedef void (__stdcall *plugin_cb_init)(void);
typedef void (__stdcall *plugin_cb_close)(void);
typedef void (__stdcall *plugin_cb_event)(char *scriptfile, char *s_event, char *direction, char *type, int conid_src, int conid_dst, char *no_src, char *no_dst, char *data, int acc_id);
typedef void (__stdcall *plugin_cb_output)(char *output, int pid);

// Callback to dtmfbox.exe (executing commands)
typedef int (__stdcall *plugin_exec_command)(int argc, ...);

// Initialize callbacks or execute command from .NET app
EXPORT int plugin_callback_init(plugin_cb_init ev_init, plugin_cb_close ev_close, plugin_cb_event ev_event, plugin_cb_output ev_output, int no_thread);
EXPORT int plugin_callback_close();
EXPORT int plugin_execute_command(int argc, char * args[]);

// Required (internal) callbacks for dtmfbox.exe
EXPORT void plugin_init(plugin_exec_command cmd_param);
EXPORT void plugin_close();
EXPORT void plugin_event(char *scriptfile, char *s_event, char *s_direction, char *s_type, int conid_src, int conid_dst, char *no_src, char *no_dst, char *data, int acc_id);
EXPORT void plugin_output(char *output, int pid);

// Internal threads which listens for events and commands
void application_thread();
void dtmfbox_thread();

// Private functions
void lock(int app);
void unlock(int app);
void raise(int app, int value);

//   Shared DataSegment for IPC
//   dtmfbox.exe <-> dtmfbox-net.plugin.dll <-> application.exe 
//
#pragma data_seg(".sdata")            

    // Thread handles    
    HANDLE t_dtmfbox = NULL;                
    HANDLE t_application = NULL;
    int dtmfbox_running = 0;
    int application_running = 0;

    // Callbacks to .NET app
    plugin_cb_init pl_init = NULL;        
    plugin_cb_close pl_close = NULL;
    plugin_cb_event pl_event = NULL;
    plugin_cb_output pl_output = NULL;
    
    // Callback to execute dtmfbox command
    plugin_exec_command dtmfbox = NULL;     
    
    // Shared variables
    char g_arg[256][256] = { "" };          
    char g_output[2048] = "";
    char g_data[2048] = "";
    char g_scriptfile[1024] = "";
    char g_event[256] = "";
    char g_direction[256] = "";
    char g_type[256] = "";
    char g_no_src[256] = "";
    char g_no_dst[256] = "";    
    int g_conid_src = 0;
    int g_conid_dst = 0;
    int g_accid = 0;
    int g_pid = 0;    
    int g_status_app = 0;
    int g_status_dtmfbox = 0;
    int g_argc = 0;    
    int g_ret_id = -3;

#pragma data_seg()
#pragma comment (linker, "/section:.sdata,RWS")

#ifdef __cplusplus
}
#endif
