#include "libfax.plugin.h"

static capifax_con_t 	connection[CAPI_CONNECTIONS];
static int 		        log_level = 2;
static char		        LSI[64];
static char		        header[64];

/* Plugin is loaded! */
EXPORT void plugin_init(plugin_exec_cmd cmd) 
{
    dtmfbox = cmd;  
    memset(connection, 0, sizeof(connection));
    //spandsp_close(-1);
    app_msg("[plugin.fax]: Initialized!");
}

/* Plugin is unloaded! */
EXPORT void plugin_close() 
{
    app_msg("[plugin.fax]: Destroyed!");
    snprintf(LSI, sizeof(LSI), "+00 0000 00000");
    snprintf(header, sizeof(header), "dtmfbox");
    
    spandsp_close(-1);
    dtmfbox = NULL;    
}

/* Callback gets called when a new event arrives */
EXPORT void plugin_event(char *scriptfile, char *s_event, char *s_direction, char *s_type, int conid_src, int conid_dst, char *no_src, char *no_dst, char *data, int acc_id) 
{    

	if(!strcmp(s_event, "CONNECT") && !strcmp(s_direction, "INCOMING") && (!strcmp("CAPI", s_type) || !strcmp("VOIP", s_type)))
	{		
		char file[256];
		snprintf(file, sizeof(file), "./fax-%s_%s.tiff", no_src, no_dst);
		spandsp_close(conid_src);
		spandsp_init(conid_src, "./test.tif", FALSE, 3, FALSE, LSI, header);
        memset(connection[conid_src].buf, 0, sizeof(connection[conid_src].buf));

        if(!strcmp("CAPI", s_type))
            connection[conid_src].type = 0;
        else
            connection[conid_src].type = 1;
	}
		
	if(!strcmp(s_event, "DISCONNECT"))
	{
		spandsp_close(conid_src);
		memset(&connection[conid_src], 0, sizeof(connection[conid_src]));
	}	    
}

/* Callback gets called, when dtmfbox outputs something */
EXPORT void plugin_output(char *output, int pid) 
{ 
    //printf("%s\n", output);    
}

/* Callback gets called, when a new audio frame is available */
EXPORT void plugin_get_frame(char *buf, int buf_size, int conid, int playid, int accid)
{	
    if(connection[conid].running) 
    {
        int ret;

        switch(connection[conid].type) 
        {
            // CAPI
            case 0:
                memcpy(buf, connection[conid].buf, buf_size);
                break;
            
            // VOIP
            case 1:                
                ret = fax_rx(&connection[conid].fax, buf, (buf_size) / 2);
                break;
        }
    }

}

/* Callback gets called, when a new audio frame will be send to the connection */
EXPORT void plugin_put_frame(char *buf, int buf_size, int conid, int recid, int accid)
{
    if(connection[conid].running) 
    {
        int ret;

        switch(connection[conid].type) 
        {
            // CAPI
            case 0:
                ret = fax_rx(&connection[conid].fax, buf, (buf_size) / 2);
                if(!ret) ret = fax_tx(&connection[conid].fax, connection[conid].buf, (buf_size) / 2);
                break;
            
            // VOIP
            case 1:
                ret = fax_tx(&connection[conid].fax, buf, (buf_size) / 2);
                break;
        }
    }
}

/* Callback gets called, when command help is shown */
EXPORT char *plugin_extra_help()
{
    return " dtmfbox fax v0.1";
}

/* 
    This function is for implementing own dtmfbox commands. 
    Return > 0, when the parameter(s) are valid, otherwise 0.    
*/
EXPORT int plugin_extra_commands(int argc, char *arg[], int pid)
{
    return 0;
}

/* --------------------------------------------------------
    spandsp functions
-------------------------------------------------------- */
static int spandsp_init(int con, const char *tiff_file, int sending, int modem, int ecm, const char *LSI, const char *LocalHeaderInfo)
{				
	fax_state_t *fax = NULL;
			
	//con = capi_get_free();
	//if(con == -1) return con;
		
	fax = &connection[con].fax;
	connection[con].running	= 1;
	
	fax_init(&fax->t30_state, sending);
	fax_set_transmit_on_idle (&fax->t30_state, TRUE);	
	
    fax->t30_state.supported_compressions = 0;
	fax->t30_state.supported_compressions |= T30_SUPPORT_NO_COMPRESSION;		
	fax->t30_state.supported_compressions |= T30_SUPPORT_T4_1D_COMPRESSION;
	fax->t30_state.supported_compressions |= T30_SUPPORT_T4_2D_COMPRESSION;
    if(ecm) fax->t30_state.supported_compressions |= T30_SUPPORT_T6_COMPRESSION;
	//fax->t30_state.supported_compressions |= T30_SUPPORT_T43_COMPRESSION;
	//fax->t30_state.supported_compressions |= T30_SUPPORT_T45_COMPRESSION;
	//fax->t30_state.supported_compressions |= T30_SUPPORT_T85_COMPRESSION;
	
	fax->t30_state.supported_resolutions |= T30_SUPPORT_STANDARD_RESOLUTION;
	fax->t30_state.supported_resolutions |= T30_SUPPORT_FINE_RESOLUTION;
	fax->t30_state.supported_resolutions |= T30_SUPPORT_SUPERFINE_RESOLUTION;
	fax->t30_state.supported_resolutions |= T30_SUPPORT_R4_RESOLUTION;
	fax->t30_state.supported_resolutions |= T30_SUPPORT_R8_RESOLUTION;	
	fax->t30_state.supported_resolutions |= T30_SUPPORT_R16_RESOLUTION;
	fax->t30_state.supported_resolutions |= T30_SUPPORT_300_300_RESOLUTION;
	fax->t30_state.supported_resolutions |= T30_SUPPORT_400_400_RESOLUTION;	
	fax->t30_state.supported_resolutions |= T30_SUPPORT_600_600_RESOLUTION;
	fax->t30_state.supported_resolutions |= T30_SUPPORT_1200_1200_RESOLUTION;
	fax->t30_state.supported_resolutions |= T30_SUPPORT_300_600_RESOLUTION;
	fax->t30_state.supported_resolutions |= T30_SUPPORT_400_800_RESOLUTION;
	fax->t30_state.supported_resolutions |= T30_SUPPORT_600_1200_RESOLUTION;	

	fax->t30_state.supported_image_sizes |= T30_SUPPORT_A4_LENGTH;
	fax->t30_state.supported_image_sizes |= T30_SUPPORT_B4_LENGTH;	
	fax->t30_state.supported_image_sizes |= T30_SUPPORT_US_LETTER_LENGTH;
	fax->t30_state.supported_image_sizes |= T30_SUPPORT_US_LEGAL_LENGTH;
	fax->t30_state.supported_image_sizes |= T30_SUPPORT_UNLIMITED_LENGTH;
	fax->t30_state.supported_image_sizes |= T30_SUPPORT_215MM_WIDTH;
	fax->t30_state.supported_image_sizes |= T30_SUPPORT_255MM_WIDTH;
	fax->t30_state.supported_image_sizes |= T30_SUPPORT_303MM_WIDTH;
	
    fax->t30_state.iaf = 0;
	fax->t30_state.iaf |= T30_IAF_MODE_FLOW_CONTROL;	
	//fax->t30_state.iaf |= T30_IAF_MODE_CONTINUOUS_FLOW;	
	//fax->t30_state.iaf |= T30_IAF_MODE_NO_INDICATORS;
	//fax->t30_state.iaf |= T30_IAF_MODE_NO_TCF;
	//fax->t30_state.iaf |= T30_IAF_MODE_NO_FILL_BITS;
/*
    fax->t30_state.supported_t30_features = 0;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_IDENTIFICATION;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_SELECTIVE_POLLING;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_POLLED_SUB_ADDRESSING;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_MULTIPLE_SELECTIVE_POLLING;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_SUB_ADDRESSING;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_TRANSMITTING_SUBSCRIBER_INTERNET_ADDRESS;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_INTERNET_ROUTING_ADDRESS;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_CALLING_SUBSCRIBER_INTERNET_ADDRESS;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_INTERNET_SELECTIVE_POLLING_ADDRESS;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_CALLED_SUBSCRIBER_INTERNET_ADDRESS;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_FIELD_NOT_VALID;
    fax->t30_state.supported_t30_features |= T30_SUPPORT_COMMAND_REPEAT;
*/
	fax->t30_state.supported_modems = 0;	
	if (modem > 0) fax->t30_state.supported_modems |= T30_SUPPORT_V27TER;
	if (modem > 1) fax->t30_state.supported_modems |= T30_SUPPORT_V29;
	if (modem > 2) fax->t30_state.supported_modems |= T30_SUPPORT_V17;

	t30_set_supported_modems (&fax->t30_state, fax->t30_state.supported_modems);	
	
	t30_set_supported_resolutions (&fax->t30_state, fax->t30_state.supported_resolutions);
	t30_set_supported_image_sizes (&fax->t30_state, fax->t30_state.supported_image_sizes);
    t30_set_supported_t30_features(&fax->t30_state, fax->t30_state.supported_t30_features);
    if(ecm) t30_set_ecm_capability (&fax->t30_state, ecm);
    t30_set_supported_compressions (&fax->t30_state, fax->t30_state.supported_compressions);

	// spandsp loglevel
	if(log_level >= 3) {
		fax->logging.level = 0xFFFFFF;
		fax->t30_state.logging.level = 0xFFFFFF;
	}
	
	#if SPANDSP_VERSION == 5
	if (LSI) t30_set_tx_ident (&fax->t30_state, LSI);
	if (LocalHeaderInfo) t30_set_tx_page_header_info (&fax->t30_state, LocalHeaderInfo);
	#else
	if (LSI) t30_set_local_ident (&fax->t30_state, LSI);
	if (LocalHeaderInfo) t30_set_header_info (&fax->t30_state, LocalHeaderInfo);
	#endif
	
	if (sending == TRUE) {
		t30_set_tx_file (&fax->t30_state, tiff_file, -1, -1);
	} 
	else {	
		t30_set_rx_file (&fax->t30_state, tiff_file, -1);
	}
		
	t30_set_phase_b_handler (&fax->t30_state, phase_b_handler, (void*)con);
	t30_set_phase_d_handler (&fax->t30_state, phase_d_handler, (void*)con);
	t30_set_phase_e_handler (&fax->t30_state, phase_e_handler, (void*)con);
		  
	return con;
} 

static int spandsp_close(int con)
{
	int i;	
	
	if(con < 0) {
        for(i=0; i<CAPI_CONNECTIONS; i++) {
            t30_terminate(&connection[i].fax.t30_state);
			fax_release(&connection[i].fax);	
			connection[i].running = 0;
        }
	} else {
        t30_terminate(&connection[con].fax.t30_state);
		fax_release(&connection[con].fax);
		connection[con].running	= 0;
	}
		
	return 0;
}


#if SPANDSP_VERSION == 5
static int phase_b_handler(t30_state_t *s, void *user_data, int result)
#else
static void phase_b_handler(t30_state_t *s, void *user_data, int result)
#endif
{
	t30_stats_t t;
	const char *pfx = "Phase B:";    

	t30_get_transfer_statistics (s, &t);

	dbg(1, "Phase B handler (0x%X) %s", result, t30_frametype(result));
	dbg(1, "%s bit rate %d", pfx, t.bit_rate);
	dbg(1, "%s ECM %s", pfx, (t.error_correcting_mode)  ?  "on"  :  "off");    

	#if SPANDSP_VERSION == 5
	return 0;
	#endif	
}

#if SPANDSP_VERSION == 5
static int phase_d_handler(t30_state_t *s, void *user_data, int result)
#else
static void phase_d_handler(t30_state_t *s, void *user_data, int result)
#endif
{	
	const char *pfx = "Phase D:";
	t30_stats_t t;
	int con_id = (int)(user_data);

	t30_get_transfer_statistics(s, &t);

	dbg(1, "Phase D handler (0x%X) %s", result, t30_frametype(result));
	dbg(1, "%s pages transferred %d", pfx, t.pages_transferred);
	dbg(1, "%s image size %d x %d", pfx, t.width, t.length);    
	dbg(1, "%s bad rows %d", pfx, t.bad_rows);
	dbg(1, "%s longest bad row run %d", pfx, t.longest_bad_row_run);
	dbg(1,  "%s image size %d", pfx, t.image_size);
				
	#if SPANDSP_VERSION == 5
	return 0;
	#endif
}

static void phase_e_handler(t30_state_t *s, void *user_data, int result)
{	
	int con_id = -1;	
	const char *pfx = "Phase E:";
	t30_stats_t t;

	con_id = (int)(user_data);
	connection[con_id].running = 0;

	t30_get_transfer_statistics(s, &t);

	dbg(1, "Phase E handler (0x%X) %s", result, t30_completion_code_to_str(result));
	dbg(1,  "%s pages transferred %d", pfx, t.pages_transferred);
	dbg(1, "%s image resolution %d x %d", pfx, t.x_resolution, t.y_resolution);
	dbg(1, "%s compression type %d", pfx, t.encoding);
	dbg(1, "%s coding method %s", pfx, t4_encoding_to_str(t.encoding));
}


/* --------------------------------------------------------
    other functions
-------------------------------------------------------- */
static void dbg(int level, const char *format, ...)
{		
	if(level <= log_level)
	{
		char output[512];
		va_list args;		
		time_t tim = time(NULL);
		struct tm *now = localtime(&tim);		
				
		va_start(args, format);
		vsnprintf(output, sizeof(output), format, args);
		
		printf("[%s] %s\n", "fax.plugin.so", output);		
		va_end(args);    
	}	
}