#!/bin/bash

TEXT="$1"
SRC_ID="$2"

DTMFBOX="dtmfbox"
ESPEAK="`which espeak`"
if [ "$ESPEAK" = "" ]; then ESPEAK="`which speak`"; fi
ESPEAK_INSTALLED=""		# empty=auto, 0 = not installed (webstream), 1 = installed (use espeak), 2=just beep

if [ -z "$TEXT" ] || [ -z "$SRC_ID" ]; then return 1; fi

if [ -z "$ESPEAK_INSTALLED" ];
then
	ESPEAK_PATH="`echo $ESPEAK | sed -e 's/\(.*\)\/.*/\1/g'`"
	if [ "$ESPEAK" != "" ]; then 
		ESPEAK_INSTALLED="1"; 
	else
		ESPEAK_INSTALLED="0";
	fi	
fi

# check permission
LOCKTEMPDIR=`touch /var/tmp/$SRC_ID-$$.lock 2>&1`
if [ ! -z "$LOCKTEMPDIR" ]; then LOCKTEMPDIR="/tmp"; else LOCKTEMPDIR="/var/tmp"; rm $LOCKTEMPDIR/$SRC_ID-$$.lock; fi

LOCKFILE="$LOCKTEMPDIR/$SRC_ID-espeak.lock"
LOCKCOUNT=`cat $LOCKFILE 2>/dev/null`
if [ -z "$LOCKCOUNT" ]; then let lock_count=0; else let lock_count=$LOCKCOUNT; fi
let lock_count=lock_count+1
echo "$lock_count" > "$LOCKFILE"

FILE="$LOCKTEMPDIR/$SRC_ID-espeak-$$.wav"

cleanup() {
  if [ -p "$FILE" ]; then 
    (
     cat "$FILE" > /dev/null 2>/dev/null
    )&    
    (
      sleep 3
      killall -9 $! 2>/dev/null
      if [ -f "$FILE" ]; then rm -f "$FILE"; fi    
    )&
  fi

  exit 0
}
trap 'cleanup' 1 2 9 15 

$DTMFBOX $SRC_ID -stop play all
if [ $lock_count -ge 2 ];
then
  sleep 1  
fi

LOCKCOUNT_NEW=`cat $LOCKFILE 2>/dev/null`
if [ -z "$LOCKCOUNT_NEW" ]; then let lock_count_new=1; else let lock_count_new=$LOCKCOUNT_NEW; fi
if [ $lock_count_new -le $lock_count ]; 
then 
  rm "$LOCKFILE"  
else
  exit 0;
fi

######### WEBSTREAM ESPEAK (8 khz) #########
if [ "$ESPEAK_INSTALLED" = "0" ];
then
  mkfifo "$FILE" 2>/dev/null
  if [ -p "$FILE" ]; then
  (
    TEXT=`echo "$TEXT" | sed "s/ /%20/g"` 
    wget -q "http://www.v3v.de/speak.php?speech=$TEXT&speed=150&pitch=50&volume=100&lang=en%2Bf1&quality=polyphase&tar=0" -q -O - > "$FILE"
    rm "$FILE" 2>/dev/null
  )&
  fi
  $DTMFBOX $SRC_ID -play "$FILE" mode=stream hz=8000 bufsize=1024 wait_start=100 wait_end=250 >/dev/null
fi

######## INSTALLED ESPEAK (16 khz) #########
if [ "$ESPEAK_INSTALLED" = "1" ];
then
  mkfifo "$FILE" 2>/dev/null
  if [ -p "$FILE" ]; then
  (    
    $ESPEAK -v "en+f1" -p "50" -s "150" -a "100" --stdout "$TEXT" > "$FILE" 
    rm "$FILE" 2>/dev/null
  )& 
  fi
  $DTMFBOX $SRC_ID -play "$FILE" mode=stream hz=22050 bufsize=6400 wait_start=25 wait_end=10 >/dev/null
fi

########## NO ESPEAK - JUST BEEP ###########
if [ "$ESPEAK_INSTALLED" = "2" ];
then
  $DTMFBOX $SRC_ID -tone 400 400 250 2000 32767; 
  $DTMFBOX $SRC_ID -tone 800 800 250 2000 32767; 
  sleep 1;
  $DTMFBOX $SRC_ID -stop tone
fi

cleanup
return 0
