/*********************************************************************************
 * funcs.c ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#undef __DTMFBOX_MENU_PLUGIN_C
#include "menu.plugin.h"
#include "config.h"
#include "menu.h"
#include "funcs.h"

void *call_external_lib(int conid, char *library, char *function, char *args[MAX_ARGS], int arg_len[MAX_ARGS]) //, void *args[MAX_ARGS])
{       
    char s_log[256];
    char *c_args[MAX_ARGS];    
    void *i_args[MAX_ARGS];
    int replaced_type[MAX_ARGS];
    dtmfbox_connection *c = &cons[conid];
    HMODULE hLib = NULL;
    void *ret = NULL;
    int i;
        
    if(library == NULL || function == NULL)
        return NULL;

    // load library
    hLib = LoadLibraryA(library);
    if(hLib)
    {
        void *init = NULL;
        char *error = NULL;

        #ifndef WIN32
            dlerror();
        #endif

        // get procedure address
        init = GetProcAddress(hLib, function);
        
        #ifndef WIN32
            if ((error = dlerror()) != NULL)  {
                init = NULL;
                app_err(error);
            }
        #endif

        if(init)
        {                   
            // handle arguments
            memset(i_args, 0, sizeof(i_args));
            memset(c_args, 0, sizeof(c_args));
            memset(replaced_type, 0, sizeof(replaced_type));

            // walkthrough lib args
            for(i=0; i<MAX_ARGS; i++)
            {   
                char val[512];
                
                if(args[i] == NULL) break;
                snprintf(val, sizeof(val), "%.*s", arg_len[i], args[i]);

                // integer cfg_cfg_replacements
                if(!strcmp(val, "%4%") || !strcmp(val, "%src_id%")) {
                    i_args[i] = ((void*)c->conid_src);
                    replaced_type[i] = 1;
                }

                if(!strcmp(val, "%5%") || !strcmp(val, "%dst_id%")) {
                    i_args[i] = ((void*)c->conid_dst);
                    replaced_type[i] = 1;
                }

                if(!strcmp(val, "%8%") || !strcmp(val, "%acc_id%")) {
                    i_args[i] = ((void*)c->acc_id);
                    replaced_type[i] = 1;
                }

                // copy integer / char* to local variables
                if(!replaced_type[i])
                {
                    // integer
                    if(cfg_isnumeric(val, 0)) {
                        i_args[i] = ((void*)atoi(val));
                        replaced_type[i] = 1;
                    }

                    // char*
                    else 
                    {
                        int len = cfg_strlen(val) + 1024;                    
                        c_args[i] = (char*)malloc(len);                        
                        memset(c_args[i], 0, len);

                        if(len > 1024) {                    
                            memcpy(c_args[i], (char*)val, len-1024);                    
                            if(c_args[i][0] == '"' && c_args[i][len-1024-1] == '"') {
                                memcpy(&c_args[i][0], (char*)val+1, len-1024-2);
                                c_args[i][len-1024-2] = '\0';
                            }
                        }
                    }
                }
                
                // string cfg_cfg_replacements
                if(c_args[i] != NULL && !replaced_type[i])
                    cfg_replace_pattern(c->conid_src, c_args[i], cfg_strlen(c_args[i])+1024-2);
            }
                
            // Status
            snprintf(s_log, sizeof(s_log), "[plugin.menu] Library: %s, Function: %s (acc=%d con=%d)", library, function, c->acc_id+1, c->conid_src);
            app_dbg(s_log);

            // call function (TODO: make better!). max. 32 arguments!
            external_lib = (plugin_exec_lib)init;
            ret = external_lib( (replaced_type[0] == 0 ? c_args[0] : i_args[0]),
                                (replaced_type[1] == 0 ? c_args[1] : i_args[1]),
                                (replaced_type[2] == 0 ? c_args[2] : i_args[2]),
                                (replaced_type[3] == 0 ? c_args[3] : i_args[3]),
                                (replaced_type[4] == 0 ? c_args[4] : i_args[4]),
                                (replaced_type[5] == 0 ? c_args[5] : i_args[5]),
                                (replaced_type[6] == 0 ? c_args[6] : i_args[6]),
                                (replaced_type[7] == 0 ? c_args[7] : i_args[7]),
                                (replaced_type[8] == 0 ? c_args[8] : i_args[8]),
                                (replaced_type[9] == 0 ? c_args[9] : i_args[9]),
                                (replaced_type[10] == 0 ? c_args[10] : i_args[10]),
                                (replaced_type[11] == 0 ? c_args[11] : i_args[11]),
                                (replaced_type[12] == 0 ? c_args[12] : i_args[12]),
                                (replaced_type[13] == 0 ? c_args[13] : i_args[13]),
                                (replaced_type[14] == 0 ? c_args[14] : i_args[14]),
                                (replaced_type[15] == 0 ? c_args[15] : i_args[15]),
                                (replaced_type[16] == 0 ? c_args[16] : i_args[16]),
                                (replaced_type[17] == 0 ? c_args[17] : i_args[17]),
                                (replaced_type[18] == 0 ? c_args[18] : i_args[18]),
                                (replaced_type[19] == 0 ? c_args[19] : i_args[19]),
                                (replaced_type[20] == 0 ? c_args[20] : i_args[20]),
                                (replaced_type[21] == 0 ? c_args[21] : i_args[21]),
                                (replaced_type[22] == 0 ? c_args[22] : i_args[22]),
                                (replaced_type[23] == 0 ? c_args[23] : i_args[23]),
                                (replaced_type[24] == 0 ? c_args[24] : i_args[24]),
                                (replaced_type[25] == 0 ? c_args[25] : i_args[25]),
                                (replaced_type[26] == 0 ? c_args[26] : i_args[26]),
                                (replaced_type[27] == 0 ? c_args[27] : i_args[27]),
                                (replaced_type[28] == 0 ? c_args[28] : i_args[28]),
                                (replaced_type[29] == 0 ? c_args[29] : i_args[29]),
                                (replaced_type[30] == 0 ? c_args[30] : i_args[30]),
                                (replaced_type[31] == 0 ? c_args[31] : i_args[31]));                    
        
            // free memory
            for(i=0; i<MAX_ARGS; i++)
            {
                if(c_args[i] != NULL) free(c_args[i]);
                c_args[i] = NULL; i_args[i] = NULL;
            }
        } else {
            snprintf(s_log, sizeof(s_log), "[plugin.menu] Function %s not found (acc=%d con=%d)!", function, c->acc_id+1, c->conid_src);
            app_err(s_log);
        }
    } else {
        snprintf(s_log, sizeof(s_log), "[plugin.menu] Library %s not found (acc=%d con=%d)!", library, c->acc_id+1, c->conid_src);
        app_err(s_log);
    }

    if(hLib) FreeLibrary(hLib);

    return ret;    
}

int call_external_script(int conid, char *shell_cmd)
{
    char cmd[2048];
    char s_log[2148];
    char app[256];
    char *new_params[MAX_ARGS];    
    char *params[MAX_ARGS];
    int param_len[MAX_ARGS];	 
    int cnt = 0;
    int ret = -1;
    int i = 0;
    dtmfbox_connection *c = &cons[conid];

    if(!strlen(shell_cmd))
        return ret;

    // copy text2speech command
    snprintf(cmd, sizeof(cmd), "%s", shell_cmd);        
    memset(params, 0, sizeof(params));
    memset(param_len, 0, sizeof(param_len));
        
    // cfg_cfg_replace placeholders with real values (parameters)
    cfg_replace_pattern(c->conid_src, cmd, sizeof(cmd));
        
    // Status
    snprintf(s_log, sizeof(s_log), "[plugin.menu] Script: %s (acc=%d con=%d)", cmd, c->acc_id+1, c->conid_src);
    app_dbg(s_log);

    // create a parameter array and save filename in app
#ifdef WIN32
    cnt = search_param(cmd, app, sizeof(app), params, param_len, 0);        // with quotes!
#else
    cnt = search_param(cmd, app, sizeof(app), params, param_len, 1);        // without quotes!
    memcpy(app, app+1, strlen(app)-2);
    app[strlen(app)-2] = '\0';
#endif

    // copy parameters with correct length to new_params (for execv, spawnv)
    for(i=0;i<cnt; i++) 
    {                                
        new_params[i] = (char*)malloc(param_len[i]+1);                
        memset(new_params[i], 0, param_len[i]);                
        memcpy(new_params[i], params[i], param_len[i]);
        new_params[i][param_len[i]] = '\0';                
    }             
    new_params[cnt] = NULL;

    // run script, batch, executable, ...    
    {
        #ifndef WIN32
            // double fork under Linux
            int pid,pid2,died,status;    
            
            if (pid=fork())
            {                                                 
                died = waitpid(pid, &status, 0);
                for(i=0;i<cnt; i++) {
		            free(new_params[i]); 
                    new_params[i] = NULL;                                        
                }
                ret = pid;
            } 
            else 
            {                                   
                // drop priority for child processes
                setpriority(PRIO_PROCESS, 0, +9);
                                
                if (pid2 = fork()) {                    
                    exit(0);
                } 
                else 
                {
                    //setpgid(0, 0);
                    execv(app, new_params);
                }
                
                exit(0);
            }
        #else        
            // under Windows we do not fork() :(            
            ret = (int)spawnvp(P_NOWAITO, app, new_params);            //_P_NOWAITO
        #endif        
    }

    return ret;
}

void call_external(int acc_id, int con_id, menu_t *m)
{
    int ret_script = -1;
    void *ret_lib = NULL;

    // say menu text (say=bla)
    if(m->say != NULL) speak(con_id, m->say);                    
    
    //
    // check menue type and execute commands
    //
    switch(m->type)
    {
        // [menu:...]
        // nothin' to do
        case 1:                        
            break;
        
        // [lib:...]
        // execute external library function
        case 2:                   
            ret_lib = call_external_lib(con_id, m->lib->library, m->lib->function, m->lib->params, m->lib->param_len);
            break;

        // [script:...]
        // execute script or binary
        case 3:
            ret_script = call_external_script(con_id, m->script->cmd);
            break;

        // ...
        default:
            break;
    }

    // redirect after dtmf? 
    // cfg_cfg_replace menu...
    if(m->redirect_link != -1 && acc_id != -1)
    {           
        cons[con_id].status.current_menu = menus[m->redirect_link];

        // new parameters in redirect link? 
        search_param(m->redirect_to, NULL, 0, cons[con_id].status.current_menu->params, cons[con_id].status.current_menu->param_len, 1);                
    } 
    else 
    {
        // ... or just redirect back to last menu, when no redirect= parameter was given
        // lib:, script: only!!
        if(m->redirect_link == -1 && acc_id != -1 && m->type >= 2)
        {
            cons[con_id].status.current_menu = cons[con_id].status.last_menu;
        }
    }
}

void cfg_replace_pattern(int conid, char *str, int size)
{
    char param_val[1024];
    char param[64];
    char s_conid_src[32], s_conid_dst[32], s_accid[32];    
    dtmfbox_connection *c = &cons[conid];
    int i, j;

    snprintf(s_conid_src, sizeof(s_conid_src), "%d", c->conid_src);
    snprintf(s_conid_dst, sizeof(s_conid_dst), "%d", c->conid_dst);
    snprintf(s_accid, sizeof(s_accid), "%d", c->acc_id+1);

    // cfg_cfg_replace events (%1% - %9%)
    cfg_replace(str, str, size, "%1%", c->s_event);
    cfg_replace(str, str, size, "%2%", c->s_direction);
    cfg_replace(str, str, size, "%3%", c->s_type);
    cfg_replace(str, str, size, "%4%", s_conid_src);
    cfg_replace(str, str, size, "%5%", s_conid_dst);
    cfg_replace(str, str, size, "%6%", c->no_src);
    cfg_replace(str, str, size, "%7%", c->no_dst);
    cfg_replace(str, str, size, "%8%", s_accid);
    cfg_replace(str, str, size, "%9%", c->data);
    cfg_replace(str, str, size, "%10%", c->status.dtmf_buffer);
    cfg_replace(str, str, size, "%event%", c->s_event);
    cfg_replace(str, str, size, "%direction%", c->s_direction);
    cfg_replace(str, str, size, "%type%", c->s_type);
    cfg_replace(str, str, size, "%src_id%", s_conid_src);
    cfg_replace(str, str, size, "%dst_id%", s_conid_dst);
    cfg_replace(str, str, size, "%src_no%", c->no_src);
    cfg_replace(str, str, size, "%dst_no%", c->no_dst);
    cfg_replace(str, str, size, "%acc_id%", s_accid);
    cfg_replace(str, str, size, "%data%", c->data);
    cfg_replace(str, str, size, "%dtmf%", c->status.dtmf_buffer);

    // cfg_cfg_replace parameters from command (%$1% - %$32%)    
    if(c->status.current_menu != NULL && c->status.last_menu != NULL)
    {
        for(j=0; j<=1; j++)
        {
            for(i=0; i<MAX_ARGS; i++)
            {   
                menu_t *m = NULL;
                  
                switch(j)
                {
                    // check arguments for the current menue
                    case 0:
                        if(c->status.current_menu->params != NULL)
                            if(c->status.current_menu->params[i] != NULL)
                                m = c->status.current_menu;
                        break;
                    
                    // check arguments for the last menue
                    case 1:
                        if(c->status.last_menu->params != NULL)
                            if(c->status.last_menu->params[i] != NULL)
                                m = c->status.last_menu;
                        break;                
                }

                if(m != NULL)
                {   
                    int len = m->param_len[i];
                    if(len >= sizeof(param_val)) len = sizeof(param_val) - 1;

                    memset(param, 0, sizeof(param));
                    snprintf(param, sizeof(param), "%%$%d%%", i+1);  // %$1% ... %$32%

                    memset(param_val, 0, sizeof(param_val));            
                    memcpy(param_val, m->params[i], len);
                    cfg_replace(str, str, size, param, param_val);
                } 
                else 
                {
                    break;
                }        
            }    
        }
    }
}

EXPORT void speak(int con_id, char *text)
{
    char cmd[4096];
    
    // no text? abort!
    if(!strlen(text)) return;

    // create wave file with espeak or something else (see [text2speech])
    snprintf(cmd, sizeof(cmd), "%s", sets.text2speech.cmd);
    
    // cfg_cfg_replace patterns (%file%, %text%)
    cfg_replace(cmd, cmd, sizeof(cmd), "%text%", text);    
    
    // execute command
    call_external_script(con_id, cmd);               
}

