/*********************************************************************************
 * menu.c ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#undef __DTMFBOX_MENU_PLUGIN_C
#include "menu.plugin.h"
#include "menu.h"
#include "config.h"

// Read in complete menu structure and create links (menu.cfg)
void read_menu()
{   
    char **sections = NULL;
    char s_log[256];
    int cnt_menu = 0;    
    int sec_count = 0;    
    int i;
    
    free_menu();
    
    // get section list
    sections = cfg_get_sections(plugin_menu_cfg, &sec_count, sections);

    // bubble sort section names
    resort:
    for(i=0; i<sec_count-1; i++) 
    {
        if(strcmp(sections[i], sections[i+1]) > 0)
        {
            char *tmp = sections[i];
            sections[i] = sections[i+1];
            sections[i+1] = tmp;
            goto resort;
        }
    }

    // every account has it's on menue...
    for(i=0; i<sec_count; i++)
    {   
        char section[256];        
        int type = 0;
        char *is_menu = NULL;
        char *is_lib  = NULL;
        char *is_script = NULL;
        char *is_action = NULL;

        snprintf(section, sizeof(section), "[%s]", sections[i]);
        is_menu = strstr(section, "menu:");
        is_lib  = strstr(section, "lib:");            
        is_script = strstr(section, "script:");        
        is_action = strstr(section, "action:");
        if(is_menu) type = 1;
        if(is_lib) type = 2;
        if(is_script) type = 3;
        if(is_action) type = 4;

        // Load menu [menu:...], [lib:...], [script:...], [action:...]
        if(type >= 1 && type <= 4) 
        {
            cnt_menu = fill_menu(section, cnt_menu, type);                

            snprintf(s_log, sizeof(s_log), "[plugin.menu] %s loaded (%d entries)", sections[i], menus[cnt_menu-1]->menu->count);
            app_dbg2(s_log);       
        }
        // Error!
        else 
        {
            snprintf(s_log, sizeof(s_log), "[plugin.menu] Unable to recognize section %s!", sections[i]);
            app_err(s_log);       
        }
    }
    
    // link all menues
    link_menu(cnt_menu);

    // free section list
    for(i=0; i<sec_count; i++) {
        if(sections[i] != NULL) {
            free(sections[i]); sections[i] = NULL;
        }
    }
    if(sections) free(sections); 
    sections = NULL;

    snprintf(s_log, sizeof(s_log), "[plugin.menu] %d entries loaded!", menu_count);
    app_dbg(s_log);       

}


// fill menu structure
// start with [menu:main] and read in all other tags recursively (menu:, lib:, script:)
int fill_menu(char *section, int pos, int type)
{       
    menu_t *m;
    unsigned i;
        
    menus = (menu_t**)realloc(menus, sizeof(menu_t*) * (pos+1));
    
    // allocate memory for menu 
    menus[pos] = malloc(sizeof(menu_t));    
    memset(menus[pos], 0, sizeof(menu_t));
    m = menus[pos];    
        
    // always allocate memory for the key_tree
    m->menu = malloc(sizeof(key_tree_t));
    memset(m->menu, 0, sizeof(key_tree_t));   

    // optional allocate memory for [script:...], [lib:...] and [action:...]
    if(type == 2) m->lib = malloc(sizeof(menu_lib_t));
    if(type == 3) m->script = malloc(sizeof(menu_script_t));
    if(type == 4) m->action = malloc(sizeof(menu_action_t));
    if(m->lib) memset(m->lib, 0, sizeof(menu_lib_t));      
    if(m->script) memset(m->script, 0, sizeof(menu_script_t));
    if(m->action) memset(m->action, 0, sizeof(menu_action_t));

    // fetch all keys from section    
    snprintf(m->menu->section, sizeof(m->menu->section), "%s", section);    
    m->menu->keys = cfg_get_entries(plugin_menu_cfg, 0, section, &m->menu->count, m->menu->keys);
    m->menu->values = cfg_get_entries(plugin_menu_cfg, 1, section, &m->menu->count, m->menu->values);
    m->menu_id = pos;
    m->type = type;
    m->redirect_link = -1;
    pos++;

    // Bubblesort keys/values by length (longest first)
    // This has to be done, because when more dtmf characters are typed in, the correct 
    // menu may not be found. 
    // eg. When you type 10#, and 0# is also defined, 0# may be used instead of 10#.
    // This sort routine corrects that behaviour.
    resort:
    for(i=0; i<m->menu->count-1; i++)
    {
        if(strlen(m->menu->keys[i]) < strlen(m->menu->keys[i+1]))
        {
            char *tmp = m->menu->keys[i];
            m->menu->keys[i] = m->menu->keys[i+1];
            m->menu->keys[i+1] = tmp;

            tmp = m->menu->values[i];
            m->menu->values[i] = m->menu->values[i+1];
            m->menu->values[i+1] = tmp;

            goto resort;
        }
    }

    return pos;
}

int search_link(char *key, int menu_count)
{
    char key2[256];
    int ret = -1;
    int x;
    
    snprintf(key2, sizeof(key2), "[%s]", key);

    // search for id to redirect    
    for(x=0; x<menu_count; x++)
    {
        if(menus[x] == NULL) break;
        if(menus[x]->menu == NULL) break;

        if(!strcmp(key, menus[x]->menu->section) ||
           !strcmp(key2, menus[x]->menu->section))
        {
            ret = x;   
            break;
        }
    }

    return ret;
}

// link menues and submenues together (menus[x]->menu->link[y]);
void link_menu(unsigned count)
{    
    unsigned i, j;

    // link entries with menues
    for(i=0; i<count; i++)
    {
        menu_t *m = menus[i];
        if(m->menu != NULL)
        {
            // get link id
            for(j=0; j<m->menu->count; j++)
            {
                char section[256];    
               
                // create section name without parameter
                search_param(m->menu->values[j], section, sizeof(section), NULL, NULL, 1);
                             
                // search for menu-id to goto
                m->menu->link = (int*)realloc(m->menu->link, sizeof(int)*(j+1));
                m->menu->link[j] = search_link(section, count);
            }
            
            // link other config values
            for(j=0; j<m->menu->count; j++)
            {
                // say something?
                if(!strcmp(m->menu->keys[j], "say")) 
                    m->say = m->menu->values[j];                

                // redirect?
                if(!strcmp(m->menu->keys[j], "redirect") && m->menu->values[j] != NULL) 
                {                      
                    char link_name[256];
                    
                    // create section name without parameter
                    search_param(m->menu->values[j], link_name, sizeof(link_name), NULL, NULL, 1);                    

                    m->redirect_to = m->menu->values[j];
                    m->redirect_link = search_link(link_name, count);                        
                } 

                // type specific?
                switch(m->type)
                {
                    // [menu:...] 
                    case 1:
                        break;

                    // [lib:...] 
                    case 2:
                        // parse library name (*.so, *.dll)
                        if(!strcmp(m->menu->keys[j], "library"))                        
                            m->lib->library = m->menu->values[j];                        

                        // parse function name
                        if(!strcmp(m->menu->keys[j], "function") && m->params == NULL) 
                        {                      
                            char function_name[256];
                                                        
                            m->params = m->action->params;
                            m->param_len = m->action->param_len;
                            
                            m->lib->function = m->menu->values[j];

                            // parse arguments
                            search_param(m->lib->function, function_name, (int)sizeof(function_name), m->lib->params, m->lib->param_len, 0);
                            
                            // parse new function name (without parameters)
                            if(strcmp(function_name, m->menu->values[j]));
                            {
                                memset(m->menu->values[j], 0, strlen(function_name)+1);
                                memcpy(m->menu->values[j], function_name, strlen(function_name));
                            }
                        }                        
                        break;

                    // [script:...] 
                    case 3:
                        if(!strcmp(m->menu->keys[j], "cmd"))
                            m->script->cmd = m->menu->values[j];                                                     
                        break;

                    // [action:...] 
                    case 4:                        
                        {
                            char *key = m->menu->keys[j];
                            char *val = m->menu->values[j];

                            if(key != NULL && val != NULL)
                            {                
                                if(!strcmp(key, "event")) m->action->s_event = val;
                                if(!strcmp(key, "direction")) m->action->s_direction = val;
                                if(!strcmp(key, "type")) m->action->s_type = val;
                                if(!strcmp(key, "src_no")) m->action->s_nosrc = val;
                                if(!strcmp(key, "dst_no")) m->action->s_nodst = val;
                                if(!strcmp(key, "src_id")) m->action->s_idsrc = val;
                                if(!strcmp(key, "dst_id")) m->action->s_iddst = val;
                                if(!strcmp(key, "acc_id")) m->action->s_accid = val;
                                if(!strcmp(key, "data")) m->action->s_data = val;
                                if(!strcmp(key, "dtmf")) m->action->s_dtmf = val;
                                if(!strcmp(key, "action")) m->action->s_action = val;
                            }
                        }

                        if(cfg_isnumeric(m->action->s_accid, 0)) m->action->i_accid = atoi(m->action->s_accid);
                        if(cfg_isnumeric(m->action->s_idsrc, 0)) m->action->i_idsrc = atoi(m->action->s_idsrc);
                        if(cfg_isnumeric(m->action->s_iddst, 0)) m->action->i_iddst = atoi(m->action->s_iddst);                        

                        // search parameters
                        if(m->action->s_action != NULL && m->params == NULL) 
                        {
                            char function_name[256];
                            m->params = m->action->params;
                            m->param_len = m->action->param_len;
                            search_param(m->action->s_action, function_name, sizeof(function_name),  m->action->params, m->action->param_len, 0);
                            memcpy(m->action->s_action, function_name, cfg_strlen(function_name));
                            m->action->s_action[cfg_strlen(function_name)] = '\0';
                        }
                        break; 
                    
                    // ...
                    default:
                        break;
                }
            }
        }
    }   

    // save menu count
    menu_count = count;
}

// Free memory that was allocated for the menu
void free_menu()
{    
    unsigned i, j;

    for(i=0; i<menu_count; i++)
    {
        menu_t *m = menus[i];
        
        if(m == NULL) 
            continue;

        // free m->menu
        if(m->menu != NULL)
        {
            for(j=0; j<m->menu->count; j++) {
                if(m->menu->keys[j] != NULL) free(m->menu->keys[j]);
                if(m->menu->values[j] != NULL) free(m->menu->values[j]);

                m->menu->keys[j] = NULL;
                m->menu->values[j] = NULL;
            }
            if(m->menu->keys) free(m->menu->keys); 
            if(m->menu->values) free(m->menu->values); 
            if(m->menu->link) free(m->menu->link); 
            m->menu->keys = NULL; m->menu->values = NULL; m->menu->link = NULL;

            free(m->menu); 
            m->menu = NULL; 
        }

        // free m->script
        if(m->script != NULL)
        {
            free(m->script); m->script = NULL;
        }

        // free m->lib
        if(m->lib != NULL)
        {
            free(m->lib); m->lib = NULL;
        }

        // free m->action
        if(m->action)
        {
            free(m->action); m->action = NULL;
        }

        // free m
        free(menus[i]);
        menus[i] = NULL;
    }            

    // free menu matrix
    if(menus != NULL) free(menus); 
    menus = NULL; menu_count = 0;
}


int search_param(char *menu_value, char *section, int section_size, char *params[MAX_ARGS], int param_len[MAX_ARGS], int remove_squares)
{    
    char *param_start = NULL;
    char *param_end = NULL;
    char *param_split = NULL;    
    int i;    
    int end = 0;
    
    if(section) memset(section, 0, section_size);    

    param_start = strstr(menu_value, "(");
    param_end = menu_value + cfg_strlen(menu_value) - 1;
    
    if(param_start != NULL && param_end[0] == ')' && params != NULL && param_len != NULL)
    {
        int len = (int)(param_start - menu_value);
        if(len > section_size) len = section_size - 1;
        param_start++;

        memset(param_len, 0, sizeof(int) * MAX_ARGS - 1);

        // create section name
        if(section)
        {
            if(remove_squares)
            {
                section[0] = '[';
                memcpy(section+1, menu_value, param_start - menu_value - 1);
                section[strlen(section)] = ']';
            } 
            else 
            {
                memcpy(section, menu_value, param_start - menu_value - 1);                
            }
        }

        // get args
        for(i=0; i<MAX_ARGS; i++)
        {            
            int do_break = 0;
            int is_string = 0;
            int is_unknown = 0;
            int pos = 0;
            int x;
	    
            while(!do_break)
            {
                if(param_start+pos >= param_end && (param_start+pos != NULL)) {
                    pos++;
                    if(is_string) pos++;
                    break;
                }

                switch(param_start[pos])
                {                    
                    case ',':
                        if(is_string == 0) do_break = 1;                        
                        break;

                    case '"':
                        if(is_unknown == 0) {
                            if(is_string) {
                                if(param_start[pos-1] == '\\')
                                    break;
                                else
                                    is_string++;
                            } else {
                                is_string++;
                            }
                        }
                        if(is_string == 2) do_break = 1;                           
                        break;

                    case ' ':
                        if(!is_string && !is_unknown) { param_start++; pos--; };
                        break;
                    
                    default:
                        if(!is_string) is_unknown = 1;
                        break;
                }
                
                pos++;
            }
                        
            param_split = param_start + pos;
            if(param_split)
            {
		        while((char)*param_start == ' ') param_start++;					
		        if((char)*param_start == '"') is_string++;
		
                if(is_string && remove_squares) {
                    params[i] = param_start+1;		    
		            param_len[i] = (int)(param_split - param_start)-2;
                } else {
                    params[i] = param_start;
		            param_len[i] = (int)(param_split - param_start);
		        }
						
		        for(x=param_len[i]-1; x>=0; x--)
		        {
                    if((char)params[i][x] == ',') { param_len[i]--; }
                    if((char)params[i][x] == ' ') { param_len[i]--; }
                    else if((char)params[i][x] == ')') { param_len[i]--;  }
                    else if((char)params[i][x] == '\0') { param_len[i]--; }
			        else if((char)params[i][x] == '"' && remove_squares && is_string) { param_len[i]--; break; }
			        else break;
		        }							
							
                param_start=param_split;                
            }
               
	    
	        end++;
            if(!strlen(param_start)) break;
	        
            if(is_string)
            {             
                pos = 0;
                while(1)
                {
                    if(param_start+pos >= param_end && (param_start+pos != NULL))
                        break;
                    
                    if(param_start[pos] == ',' || param_start[pos] == ')' || param_start[pos] == '"') {
                        param_start += pos + 1;
                        break;
                    } else {
                        break;
                    }

                    pos++;
                }
            }                        
        }
    } 
    else 
    {   
        if(section)
        {   if(remove_squares) {
                if(param_start && param_end)
                    snprintf(section, section_size, "[%.*s]", param_start-menu_value, menu_value);
                else
                    snprintf(section, section_size, "[%s]", menu_value);
            }
            else {
                if(param_start && param_end)
                    snprintf(section, section_size, "%.*s", param_start-menu_value, menu_value);
                else
                    snprintf(section, section_size, "%s", menu_value);
            }
        }
    }

    // remove leading ")"    
    if(params != NULL && param_len != NULL && end > 0) {
        if(params[end-1] != NULL)
        {
            if(!strcmp(params[end-1], ")") && param_len[end] == 0) {
                params[end-1] = NULL;
                end--;
            }
        }
    }

    return end;
}
