/*********************************************************************************
 * capi_ctrl.c ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "capi_ctrl.c"

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Initialize CAPI
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
int capi_init()
{
#if DTMFBOX_USE_CAPI

    CAPI_REGISTER_ERROR ErrorCode;
    unsigned int i;
    unsigned int bchannels = 0;

    app_msg(THIS_FILE, "Init CAPI ...");

	// Test register...
	#if DTMFBOX_WIN32        
        oCAPI.AppId = CAPI20_REGISTER(CAPI_BUFFERSIZE, 1, CAPI_BUFFERCNT, CAPI_PACKETS, &ErrorCode);
    #else
        ErrorCode = CAPI20_REGISTER(1, CAPI_BUFFERCNT, CAPI_PACKETS, &oCAPI.AppId);
    #endif
	if(ErrorCode) {        
        oCAPI.AppId = (unsigned short)-1;
        capi_info(ErrorCode);
        return -2;	
	}

    // Fetch profile
    {
        struct capi_profile capiprofile;

        CAPI20_GET_PROFILE(0, (unsigned char *)&capiprofile);
        app_dbg(THIS_FILE, "CAPI controller: %d", capiprofile.ncontroller);

        for(i=1; i<=capiprofile.ncontroller; i++)
        {
           struct capi_profile ctrl_profile;           
           CAPI20_GET_PROFILE(i, (unsigned char*)&ctrl_profile);
           bchannels += ctrl_profile.nbchannel;

           app_dbg(THIS_FILE, "Found controller #%d with %d B-channel(s)", i, ctrl_profile.nbchannel);           
        }
    }
    
    // Register with new bchannel count
    CAPI20_RELEASE(oCAPI.AppId);
    oCAPI.AppId = (unsigned)-1;

    #if DTMFBOX_WIN32        
        oCAPI.AppId = CAPI20_REGISTER(CAPI_BUFFERSIZE, bchannels, CAPI_BUFFERCNT, CAPI_PACKETS, &ErrorCode);
    #else
        ErrorCode = CAPI20_REGISTER(bchannels, CAPI_BUFFERCNT, CAPI_PACKETS, &oCAPI.AppId);
    #endif

    if(ErrorCode) {
        // CAPI error!
        oCAPI.AppId = (unsigned short)-1;
        capi_info(ErrorCode);
        return -2;
    }	

   	app_msg(THIS_FILE, "CAPI registered (ApplID: %d, B-Channels: %d)", oCAPI.AppId, bchannels);
    
    // Initialize alaw table buffer (for faster encode)
    create_table_buffer(oSET.capi.ulaw);

    app_msg(THIS_FILE, "CAPI initialized!");

#endif

    return 0;
}


// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Deinitialize CAPI
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
int capi_close()
{
#if DTMFBOX_USE_CAPI
    int i;

    app_msg(THIS_FILE, "Destroy CAPI ...");
    if(oCAPI.AppId != (unsigned short)-1)
    {
        for (i=0; i<oSET.max_connections; i++) 
        {         
          if(oCONS[i].info->type == CAPI) 
          {
              // disconnect
              _cmsg CMSG1;
              DISCONNECT_REQ(&CMSG1, oCONS[i].capi->AppId, 0, oCONS[i].capi->PLCI, NULL, NULL, NULL, NULL);
          }
        }

        CAPI20_RELEASE(oCAPI.AppId);
        oCAPI.AppId = (unsigned short)-1;
        
    } else {
        app_err(THIS_FILE, "CAPI AppID == -1 !!");
    }
    
    pj_thread_sleep(250);
    app_msg(THIS_FILE, "CAPI destroyed!");

#endif

    return 0;
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//
// Listen-CAPI & Loop
//
// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_listen_req()
{
    #if DTMFBOX_USE_CAPI

    long info_mask = 0x3ff;
    struct capi_profile capiprofile;
    int start_pos = 1;
    int end_pos = 1;
    int i, j;

    CAPI20_GET_PROFILE(0, (unsigned char *)&capiprofile);
    
    // CAPI available?
    if(oCAPI.AppId != (unsigned short)-1)
    {        
        oCAPI.SetupListen = 0;

        // Walkthrough controllers from settings
        for(i=0; i<CAPI_CONTROLLERS; i++)
        {
            if(oSET.capi.controller.ctrl[i] != 0)
            {
                // listen to all controllers (=-1)
                if(oSET.capi.controller.ctrl[i] == -1)
                {                    
                    start_pos = 1;
                    end_pos = capiprofile.ncontroller;                    
                }
                // listen to specified controller (>0)
                else 
                {                    
                    start_pos = oSET.capi.controller.ctrl[i];
                    end_pos = oSET.capi.controller.ctrl[i];
                }
                
                for(j=start_pos; j<=end_pos; j++) {
                    LISTEN_REQ(oCAPI.cMsg, oCAPI.AppId, 0, j, info_mask, 0x1FFF03FF, 0, NULL, NULL);
                    app_msg(THIS_FILE, "(0x%x) - Listen to CAPI controller %d ...", info_mask, j);
                    oCAPI.SetupListen++;
                }
            }
        }
    }

    #endif
}

int capi_listen()
{ 
    #if DTMFBOX_USE_CAPI

    struct timeval tv;
    unsigned int Info;
    int error = 0;

    // CAPI available?
    if(oCAPI.AppId != (unsigned short)-1)
    {
        // Request Listen:
        capi_listen_req();

        app_dbg(THIS_FILE, "Start CAPI listen loop ..."); 

        // CAPI-Listen-Loop:
        oGLB.closing = 0;
        while(!oGLB.closing && oCAPI.AppId != (unsigned short)-1)
        {
            tv.tv_sec = 1;
            tv.tv_usec = 0;

            #if DTMFBOX_WIN32        
                    error = CAPI_WAIT_FOR_SIGNAL(oCAPI.AppId);                
            #else
                    error = CAPI20_WaitforMessage(oCAPI.AppId, &tv);
            #endif

            if(error == 0)
            {
                #if DTMFBOX_WIN32
                    switch(Info = CAPI_GET_CMSG(oCAPI.cMsg, oCAPI.AppId))
                #else
                    switch(Info = capi_get_cmsg(oCAPI.cMsg, oCAPI.AppId))
                #endif
                {
                    case 0x0000:
                        switch (oCAPI.cMsg->Subcommand)
                        {
                            // Confirmation
                            case CAPI_CONF:
                                capi_conf();                        
                                break;

                            // Indication
                            case CAPI_IND:
                                capi_ind();                        
                                break;

                            default:
                                pj_thread_sleep(1);                                                        
                                break;
                        }
                        break;
                    
                    default:
                        pj_thread_sleep(1);                                            
                        break;                     
                }
            }
        }

        app_dbg(THIS_FILE, "CAPI listen loop stopped!"); 
    }
    else
    {
        // we do not have CAPI - just idle loop...
        oGLB.closing = 0;
        while(!oGLB.closing)
            pj_thread_sleep(1000);
    }

#else

    // we do not have CAPI - just idle loop...
    oGLB.closing = 0;
    while(!oGLB.closing)
        pj_thread_sleep(1000);

#endif

    return 0;
}


// CAPI indication events
void capi_ind()
{
    #if DTMFBOX_USE_CAPI

    // Get Msg-Nr. & Controller
    oCAPI.iMsgNr = oCAPI.cMsg->Messagenumber;
    oCAPI.controller = oCAPI.cMsg->adr.adrController & 0x7f;  // convert from __u32-to-int
    
    switch(oCAPI.cMsg->Command)
    {       
      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      // CAPI_CONNECT - Connect
      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      case CAPI_CONNECT:
        oCAPI.PLCI = CONNECT_IND_PLCI(oCAPI.cMsg);        
        capi_event_connect();
        break;

      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      // CAPI_CONNECT_ACTIVE - Active
      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      case CAPI_CONNECT_ACTIVE:
        oCAPI.PLCI = CONNECT_IND_PLCI(oCAPI.cMsg);
        oCAPI.NCCI = 0;   
        capi_event_connect_active();
        break;

      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      // CAPI_CONNECT_B3 - Data connect
      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      case CAPI_CONNECT_B3:
        oCAPI.NCCI = CONNECT_B3_IND_NCCI(oCAPI.cMsg);
        oCAPI.PLCI = oCAPI.NCCI & 0x0000ffff;        
        capi_event_connect_b3();
        break;

      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      // CAPI_CONNECT_B3_ACTIVE - Data active
      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      case CAPI_CONNECT_B3_ACTIVE:
        oCAPI.NCCI = CONNECT_B3_IND_NCCI(oCAPI.cMsg);
        oCAPI.PLCI = oCAPI.NCCI & 0x0000ffff;        
        capi_event_connect_b3_active();
        break;

      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      // CAPI_DATA_B3 - Data - receive/send
      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      case CAPI_DATA_B3:
        oCAPI.NCCI = CONNECT_B3_IND_NCCI(oCAPI.cMsg);
        oCAPI.PLCI = oCAPI.NCCI & 0x0000ffff;
        capi_event_data_b3();        
        break;

      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      // CAPI_FACILITY - Facility (DTMF)
      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      case CAPI_FACILITY:
        oCAPI.NCCI = CONNECT_B3_IND_NCCI(oCAPI.cMsg);
        oCAPI.PLCI = oCAPI.NCCI & 0x0000ffff;        
        capi_event_facility();
        break;

      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      // CAPI_INFO - Info
      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      case CAPI_INFO:         
        oCAPI.PLCI = CONNECT_IND_PLCI(oCAPI.cMsg);        
        capi_event_info();
        break;

      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      // CAPI_DISCONNECT_B3 - Disconnect data
      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      case CAPI_DISCONNECT_B3:
        oCAPI.NCCI = CONNECT_B3_IND_NCCI(oCAPI.cMsg);
        oCAPI.PLCI = oCAPI.NCCI & 0x0000ffff;        
        capi_event_disconnect_b3();
        break;

      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      // CAPI_DISCONNECT - Disconnect
      // ------------------------------------------------------------------------------------------------------------------------------------------------------
      case CAPI_DISCONNECT:        
        oCAPI.PLCI = CONNECT_IND_PLCI(oCAPI.cMsg);         
        capi_event_disconnect();
        break;
    }

    #endif
}

// CAPI confirmation events
void capi_conf()
{
    #if DTMFBOX_USE_CAPI
    int i;

    oCAPI.iMsgNr = oCAPI.cMsg->Messagenumber;
    oCAPI.controller = oCAPI.cMsg->adr.adrController & 0x7f;  // convert from __u32-to-int
    

    switch(oCAPI.cMsg->Command)
    {
        case CAPI_MANUFACTURER:
            oCAPI.PLCI = CONNECT_IND_PLCI(oCAPI.cMsg);
            app_dbg2(THIS_FILE, "[ctrl=%d, plci=%d, ncci=%d] CNF: CAPI_MANUFACTURER", oCAPI.controller, oCAPI.PLCI, oCAPI.NCCI);
            capi_info(oCAPI.cMsg->Info);            
            break;

        case CAPI_FACILITY:
            oCAPI.NCCI = CONNECT_B3_IND_NCCI(oCAPI.cMsg);
            oCAPI.PLCI = oCAPI.NCCI & 0x0000ffff;
            app_dbg2(THIS_FILE, "[ctrl=%d, plci=%d, ncci=%d] CNF: CAPI_FACILITY", oCAPI.controller, oCAPI.PLCI, oCAPI.NCCI);
            capi_info(oCAPI.cMsg->Info);
            break;

        case CAPI_LISTEN:
            oCAPI.SetupListen--;
            app_dbg2(THIS_FILE, "[ctrl=%d] CNF: CAPI_LISTEN", oCAPI.controller);
            capi_info(oCAPI.cMsg->Info);
            break;

        case CAPI_ALERT:
            oCAPI.PLCI = CONNECT_IND_PLCI(oCAPI.cMsg);
            app_dbg2(THIS_FILE, "[ctrl=%d, plci=%d] CNF: CAPI_ALERT", oCAPI.controller, oCAPI.PLCI);
            capi_info(oCAPI.cMsg->Info);
            break;

        case CAPI_DATA_B3:
            oCAPI.NCCI = CONNECT_B3_IND_NCCI(oCAPI.cMsg);
            oCAPI.PLCI = oCAPI.NCCI & 0x0000ffff;            
            break;

        case CAPI_INFO:
            oCAPI.PLCI = CONNECT_IND_PLCI(oCAPI.cMsg);
            app_dbg2(THIS_FILE, "[ctrl=%d, plci=%d] CNF: CAPI_INFO", oCAPI.controller, oCAPI.PLCI);
            capi_info(oCAPI.cMsg->Info);
            break;

        case CAPI_CONNECT:
            oCAPI.PLCI = CONNECT_IND_PLCI(oCAPI.cMsg);
            app_dbg2(THIS_FILE, "[ctrl=%d, plci=%d] CNF: CAPI_CONNECT", oCAPI.controller, oCAPI.PLCI);
            capi_info(oCAPI.cMsg->Info);

            // Get connection by fake PLCI (set by CONNECT_REQ)
            i = connection_by_plci(-2);

            // CALL (OUTGOING):
            if(oCAPI.cMsg->Info == 0)
            {
                if(i!=-1)
                {
                    if(oCONS[i].capi->controller == oCAPI.controller)
                    {   
                        oCONS[i].capi->PLCI = oCAPI.PLCI;
                        oCONS[i].capi->NCCI = (unsigned)-2;
                        oCONS[i].capi->iMsgNr = oCAPI.iMsgNr;
                        oCONS[i].capi->controller = oCAPI.controller;
                        oCONS[i].capi->AppId = oCAPI.AppId;
                        oCONS[i].capi->cMsg = oCAPI.cMsg;                        
                        oCONS[i].info->type = CAPI;
                        oCONS[i].info->status = 1;
                        oCONS[i].info->active = 0;
                        
                        // Start tone thread?
                        if(oCONS[i].in_conf[0] != -1 && oSET.audio.earlyb3_ring)
                        {
                            // no client connections!! (Registrar-Mode)
                            if(oCONS[oCONS[i].in_conf[0]].info->active)
                            {
                                int args[2];
                                args[0] = oCONS[i].in_conf[0];
                                args[1] = i;                    
                                pj_thread_create(oCONS[i].pools->con_pool, "fake_ring_%p", (pj_thread_proc*)&port_tone_ring_thread, (void*)&args, 0, (unsigned)NULL, &oCONS[i].tone_thread);
                                pj_thread_sleep(25);
                            }
                        }
    
                        // SCRIPT: "CONNECT" - OUTGOING
                        script_event_add(CAPI, CONNECT, oCONS[i].script->file, i, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");
                    }
                    
                }
            }
            else
            {
                if(i != -1)
                    reset_connection(&oCONS[i]);
            }
            break;

        case CAPI_CONNECT_B3:
            oCAPI.NCCI = CONNECT_B3_IND_NCCI(oCAPI.cMsg);
            oCAPI.PLCI = oCAPI.NCCI & 0x0000ffff;

            i = connection_by_plci(oCAPI.PLCI);
            if(i==-1)
                i = connection_by_ncci(oCAPI.NCCI);
            if(i==-1)
                i = connection_by_ncci(-2);

            app_dbg2(THIS_FILE, "[ctrl=%d, plci=%d, ncci=%d] CNF: CAPI_CONNECT_B3", oCAPI.controller, oCAPI.PLCI, oCAPI.NCCI);
            capi_info(oCAPI.cMsg->Info);

            // CALL (OUTGOING):
            if(i != -1)
            {
                if(oCONS[i].capi->controller == oCAPI.controller) {
                    oCONS[i].capi->NCCI = oCAPI.NCCI;
                    oCONS[i].info->active = 1;
                }
            }
            break;

        default:
            //app_dbg(THIS_FILE, "(CNF, CTRL %d, PLCI: %d) - MISC (%d)", oCAPI.controller, oCAPI.PLCI, oCAPI.cMsg->Command);
            //capi_info(oCAPI.cMsg->Info);
            break;
    }

    #endif
}

