/*********************************************************************************
 * capi_events.c ($Revision: 250 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "capi_events.c"

void capi_event_connect()
{
    #if DTMFBOX_USE_CAPI

        char info[256];
        _cmsg CMSG1;
        int i;
        int acc_id = -1;
        char CallingPartyNumber[MAX_URI_LENGTH];
        char CalledPartyNumber[MAX_URI_LENGTH];
        pj_status_t status = PJ_SUCCESS;  
        
        // Seach for a free connection
        DTMFBOX_ACC_LOCK();
        i = connection_by_plci(oCAPI.PLCI);
        if(i == -1)
            i = connection_by_free();
        if(i == -1) {
            DTMFBOX_ACC_UNLOCK();
            return;
        }
        oCONS[i].info->type = CAPI;
        DTMFBOX_ACC_UNLOCK();

        // Fill connection array        
        oCONS[i].capi->PLCI = oCAPI.PLCI;        
        oCONS[i].capi->iMsgNr = oCAPI.iMsgNr;
        oCONS[i].capi->controller = oCAPI.controller;
        oCONS[i].capi->AppId = oCAPI.AppId;
        oCONS[i].capi->cMsg = oCAPI.cMsg;            
        oCONS[i].script->dtmf_pos = 0;
        oCONS[i].script->delimiter[0] = oSET.delimiter[0];  

        // From internal?
        if(acc_id == -1)
        {            
            capi_get_source_no(oCAPI.cMsg, CalledPartyNumber, 0);
            acc_id = account_by_number(CalledPartyNumber);
            if(acc_id == -1) {                   
            } else {
                oCONS[i].info->direction |= OUTGOING;
                capi_get_target_no(oCAPI.cMsg, CallingPartyNumber, 1);
                capi_get_source_no(oCAPI.cMsg, CalledPartyNumber, 1);
            }
        }

        // From external?
        if(acc_id == -1) {
            capi_get_target_no(oCAPI.cMsg, CalledPartyNumber, 0);        
            acc_id = account_by_number(CalledPartyNumber);
            if(acc_id == -1) {                                
            } else {
                oCONS[i].info->direction |= INCOMING;
                                
                capi_get_source_no(oCAPI.cMsg, CallingPartyNumber, 1);
                capi_get_target_no(oCAPI.cMsg, CalledPartyNumber, 1);
            }
        }

        // Account not found?
        if(acc_id == -1) 
        {
            capi_get_source_no(oCAPI.cMsg, CallingPartyNumber, 1);
            capi_get_target_no(oCAPI.cMsg, CalledPartyNumber, 1);

            if(!oSET.capi.allow_unknown_acc)
            {
                app_msg(THIS_FILE, "Ignoring connection for unknown account. CalledParty=%s, CallingParty=%s", CalledPartyNumber, CallingPartyNumber);
                capi_send_connect_resp(&oCONS[i], 1);   // reject connect
                reset_connection(&oCONS[i]);
                goto hell;                
            } else {                
                oCONS[i].info->direction |= INCOMING;
            }
        } 
        
        oCONS[i].info->account_id = acc_id;
      
        // Init capi port
        status = port_capi_init(&oCONS[i]);
        if(status != PJ_SUCCESS) {
            port_capi_close(&oCONS[i]);
            capi_send_connect_resp(&oCONS[i], 1);   // reject connect
            reset_connection(&oCONS[i]);
            app_err(THIS_FILE, "(%d) Unable to create capi port!", status);
            goto hell;
        }

        pj_memset(oCONS[i].script->dtmf, 0, sizeof(oCONS[i].script->dtmf));
        strncpy(oCONS[i].script->file, oSET.scriptfile, sizeof(oCONS[i].script->file));

        // save phone no.
        // when targetno is "unknown" (and phone call: outgoing) - it will be replaced by the dial-string            
        strncpy(oCONS[i].info->no_source, CalledPartyNumber, sizeof(oCONS[i].info->no_source));
        strncpy(oCONS[i].info->no_target, CallingPartyNumber, sizeof(oCONS[i].info->no_target));

        // Registration status (ok)
        if(acc_id != -1)
        {
            oSET.accounts[acc_id].status = 1;
            strcpy(oSET.accounts[acc_id].status_reason, "(200) - OK");
        }
         
        app_dbg(THIS_FILE, "%s (IND) CAPI_CONNECT", connection_print_info(&oCONS[i], info, sizeof(info)));
        
        if(oCONS[i].info->direction & INCOMING)
        {            
            unsigned char bchaninfo[3];
            bchaninfo[0] = 0x2;
            bchaninfo[1] = 0x0;
            bchaninfo[2] = 0x0;

            // Only ring on incoming controller...
            // *Ring*
            {   
                
                ALERT_REQ(&CMSG1, oCONS[i].capi->AppId, 0, oCONS[i].capi->PLCI,
                              NULL,       // BChannelinformation
                              NULL,       // Keypadfacility
                              NULL,       // Useruserdata
                              NULL,       // Facilitydataarray 
                              NULL);               
            }

            // Registrar-Mode: Ring all clients
            if(acc_id != -1)
            {
                int x, y;                    
                int client_acc_id = -1;
                int VOIP_CONNECTION;

                for(x=0; x<oSET.voip.max_clients; x++)
                {
                    for(y=0; y<MAX_ACCOUNTS; y++)
                    {
                        client_acc_id = oGLB.clients[x].account_id[y];
                        if(client_acc_id == acc_id)
                        {                                
                            char SourceNo[MAX_URI_LENGTH];
                            char TargetNo[MAX_URI_LENGTH];
                            char ip[MAX_URI_LENGTH];       
                            int port;
                                                        
                            //pj_ansi_snprintf(ip, sizeof(ip), "%.*s", oGLB.clients[x].addr.slen, oGLB.clients[x].addr.ptr);                                
                            //pj_ansi_snprintf(TargetNo, sizeof(TargetNo), "sip:%s@%.*s", oSET.accounts[client_acc_id].registrar_username, oGLB.clients[x].addr.slen, oGLB.clients[x].addr.ptr);
                            pj_ansi_snprintf(ip, sizeof(ip), "%s", oGLB.clients[x].addr);                                
                            pj_ansi_snprintf(TargetNo, sizeof(TargetNo), "sip:%s@%s", oSET.accounts[client_acc_id].registrar_username, oGLB.clients[x].addr);

                            // public or private ip?
                            // change host SourceNo, so that media flow is correctly (private/public ip)                            
                            /*                            
                            if(!ip_is_public(ip))
                                pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "sip:%s@%s:%d", oCONS[i].info->no_target, oGLB.ip_addr_local, oGLB.udp_port_local);
                            else {
                            */
                                //pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "sip:%s@%s:%d", oCONS[i].info->no_target, oGLB.ip_addr_global, oGLB.udp_port_global);
                            
                                status = voip_resolve_transport_address(ip, &port);
                                if(status != PJ_SUCCESS)
                                {
                                    app_err(THIS_FILE, "(%d) Unable to resolve transport address", status);
                                    break;
                                }

                                pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "sip:%s@%s:%d", oCONS[i].info->no_target, ip, port);
                            //}                            
                            VOIP_CONNECTION = voip_registrar_connect_clients(&oCONS[i], SourceNo, TargetNo, acc_id, x, VOIP, oCONS[i].info->direction | CLIENT);
                            break;
                        }
                    }
                }
            }
        } 
     
        // SCRIPT: "CONNECT" - INCOMING
        script_event_add(CAPI, CONNECT, oSET.scriptfile, i, oCONS[i].in_conf[0], CalledPartyNumber, CallingPartyNumber, "");
        
        // SCRIPT: "EARLY" - OUTGOING
        if(!oCONS[i].capi->SetupEarly && (oCONS[i].info->direction & OUTGOING) && strcmp(CallingPartyNumber, "unknown")) {
            script_event_add(CAPI, EARLY, oCONS[i].script->file, oCONS[i].id, oCONS[i].in_conf[0], CalledPartyNumber, CallingPartyNumber, "");
            oCONS[i].capi->SetupEarly = 1;
        }

hell:
    return;           
    
    #endif
}

void capi_event_connect_active()
{
    #if DTMFBOX_USE_CAPI

        char info[256];
        _cmsg CMSG1;
        int i;
        
        // Search for connection
        DTMFBOX_ACC_LOCK();
        i = connection_by_plci(oCAPI.PLCI);
        if(i == -1)
            i = connection_by_free();
        if(i==-1) {
            DTMFBOX_ACC_UNLOCK();
            return;
        }
        oCONS[i].info->type = CAPI;
        oCONS[i].capi->PLCI = oCAPI.PLCI;
        DTMFBOX_ACC_UNLOCK();

        app_dbg2(THIS_FILE, "%s (IND) CAPI_CONNECT_ACTIVE", connection_print_info(&oCONS[i], info, sizeof(info)));

        // Active respond
        CONNECT_ACTIVE_RESP(&CMSG1, oCAPI.AppId, oCAPI.iMsgNr, CONNECT_IND_PLCI(oCAPI.cMsg));

        // Request B3 on initial outgoing connection
        if(oCONS[i].info->direction & OUTGOING) //&& oCONS[i].capi->NCCI == -2)        
            CONNECT_B3_REQ(&CMSG1, oCAPI.AppId, oCAPI.iMsgNr, CONNECT_IND_PLCI(oCAPI.cMsg), 0);

    #endif
}

void capi_event_connect_b3()
{
    #if DTMFBOX_USE_CAPI

        char info[256];
        _cmsg CMSG1;
        int i;
        
        // Get connection by PLCI (set NCCI!)
        i = connection_by_plci(oCAPI.PLCI);
        if(i==-1)
            return;
        
        app_dbg2(THIS_FILE, "%s (IND) CAPI_CONNECT_B3", connection_print_info(&oCONS[i], info, sizeof(info)));

        // Respond B3 ...
        CONNECT_B3_RESP(&CMSG1, oCAPI.AppId, oCAPI.iMsgNr, oCAPI.NCCI, 0, (_cstruct)NULL);
            
        // Save                
        oCONS[i].info->type = CAPI;
        oCONS[i].capi->NCCI = oCAPI.NCCI;
        oCONS[i].capi->PLCI = oCAPI.PLCI;        
        oCONS[i].capi->hdl = 0;            
        oCONS[i].info->status = 2; 

    #endif
}

void capi_event_connect_b3_active()
{
    #if DTMFBOX_USE_CAPI
        
        char info[256];
        _cmsg CMSG1;
        int i, j;

        // B3 active respond
        CONNECT_B3_ACTIVE_RESP(&CMSG1, oCAPI.AppId, oCAPI.iMsgNr, oCAPI.NCCI);

        // Search for connection
        i = connection_by_ncci(oCAPI.NCCI);
        if(i == -1)
            i = connection_by_plci(oCAPI.PLCI);

        if(i!=-1)
        {
            // fill connection-variables                        
            oCONS[i].capi->NCCI = oCAPI.NCCI;            
            oCONS[i].capi->AppId = oCAPI.AppId;                        
            oCONS[i].capi->hdl = 0;            
            oCONS[i].info->type = CAPI;
            oCONS[i].info->status = 2;
            oCONS[i].info->active = 1;
            oCONS[i].capi->SetupInband = 0;

            app_dbg(THIS_FILE, "%s (IND) CAPI_CONNECT_B3_ACTIVE", connection_print_info(&oCONS[i], info, sizeof(info)));

            // Wait for ringing thread
            if(oCONS[i].tone_thread != NULL && oCONS[i].ports->tone != NULL) {
                pj_thread_join(oCONS[i].tone_thread);
                pj_thread_destroy(oCONS[i].tone_thread);
                oCONS[i].tone_thread = NULL;
            }
            
        	for(j=0; j<INCONF_PORTS; j++)
            {
                int in_conf = oCONS[i].in_conf[j];
                
                if(in_conf != -1) 
                {         
                    // a client hooked up (Registrar-Mode)
                    if(oCONS[i].info->direction & CLIENT || (oCONS[in_conf].info->direction & CLIENT && oCONS[i].info->active))
                    {                        
                        if(oCONS[in_conf].info->type == CAPI)
                            capi_confirm_call(&oCONS[in_conf]);
                        
                        if(oCONS[in_conf].info->type == VOIP && !(oCONS[in_conf].info->direction & INCOMING))
                            voip_confirm_call(&oCONS[in_conf]);
                    }

                    // disconnect all inactive clients (Registrar-Mode)
                    if((oCONS[in_conf].info->direction & CLIENT)) {
                        voip_registrar_disconnect_clients(&oCONS[in_conf], 1);
                    } 
                      
                    // reconnect conference partners 
                    mixer_conf_port_connect(&oCONS[i], &oCONS[in_conf], 0);
                    mixer_conf_port_connect(&oCONS[in_conf], &oCONS[i], 0);                                
                }
            }
            
            // Initialize tone port            
            if(oCONS[i].ports->tone == NULL)            
                port_tone_init(&oCONS[i], 0, 0, 0, 0, 0);
        
            // Enable DTMF
            capi_facility_dtmf(1);

            // SCRIPT: "CONFIRMED"            
            script_event_add(CAPI, CONFIRMED, oCONS[i].script->file, i, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");

            if(oCONS[i].info->auto_hookupdown)
                script_auto_hookup(i);
        }

    #endif
}

void capi_event_facility()
{
    #if DTMFBOX_USE_CAPI
        
        char info[256];
        _cmsg CMSG1;
        _cstruct *parm;
        int i = -1;

        // Respond
        FACILITY_RESP(&CMSG1, oCAPI.AppId, 0, oCAPI.PLCI, FACILITY_IND_FACILITYSELECTOR(oCAPI.cMsg), FACILITY_IND_FACILITYINDICATIONPARAMETER(oCAPI.cMsg));

        i = connection_by_ncci(oCAPI.NCCI);
        if(i == -1)
            i = connection_by_plci(oCAPI.PLCI);
        else
            return; 

        app_dbg2(THIS_FILE, "%s (IND) CAPI_FACILITY", connection_print_info(&oCONS[i], info, sizeof(info)));

    	parm = (_cstruct*)FACILITY_IND_FACILITYINDICATIONPARAMETER(oCAPI.cMsg);

        // Check FacilitySelector
        switch(FACILITY_IND_FACILITYSELECTOR(oCAPI.cMsg))
        {
            // DTMF received
            case 0x0001:
                capi_get_dtmf((unsigned char)(FACILITY_IND_FACILITYINDICATIONPARAMETER(oCAPI.cMsg)[1]), 0, connection_by_plci(oCAPI.PLCI));
                break;

            /*
            // Supplementary Services
            case 0x0003:
                temp = (unsigned int)(((unsigned int)parm[0]) | ((unsigned int)parm[2] << 8));
        		reason = (unsigned int)parm[3];

                // Retrieve
                if(temp == 0x0203) 
                {
                    if(reason)
                    {
                    }
                    else
                    {
                    }
                }
                
                // Hold
                if(temp == 0x0202)
                {
                    if(reason)
                    {
                    }
                    else
                    {
                    }
                }
                break;  
            */
        }

    #endif
}

void capi_event_info()
{
    #if DTMFBOX_USE_CAPI

        _cmsg CMSG1;
        char info_element[128];
        char status[512];
        char info[256];
        int conid = -1;
        int i;

        // Respond to INFO
        INFO_RESP(&CMSG1, oCAPI.AppId, oCAPI.iMsgNr, INFO_IND_PLCI(oCAPI.cMsg));

        // Get 128 bytes of InfoElement
        pj_memset(info_element, 0, sizeof(info_element));
        for(i = 0; i < sizeof(info_element); i++) {
	        info_element[i] = INFO_IND_INFOELEMENT(oCAPI.cMsg)[i];
	    }
        
        conid = connection_by_plci(oCAPI.PLCI);
        if(conid != -1) 
        {
            pj_ansi_snprintf(status, sizeof(status), "(IND) CAPI_INFO (0x%04x) -", INFO_IND_INFONUMBER(oCAPI.cMsg));

            switch(INFO_IND_INFONUMBER(oCAPI.cMsg))
            {
                case 0x0008:    // Cause                
                    app_dbg2(THIS_FILE, "%s CAUSE (0x%02x)", status, info_element[1] & 0x7f);                                
                    break;
                case 0x0014:    // Call state
                    app_dbg2(THIS_FILE, "%s CALL STATE (0x%02x)", status, info_element[0]);
                    break;
                case 0x0018:    // Channel Identification                                               
                    app_dbg2(THIS_FILE, "%s CHANNEL IDENTIFICATION (0x%02x)", status, info_element[0]);                                   
                    // unhold the call (get B3 again)
                    if(oCONS[conid].capi->SetupInband)                    
                        CONNECT_B3_REQ(&CMSG1, oCAPI.AppId, oCAPI.iMsgNr, oCONS[conid].capi->NCCI, 0);                    
                    break;
                case 0x001c:    // Facility Q.932
                    app_dbg2(THIS_FILE, "%s FACILITY", status);
                    break;
                case 0x001e:    // Progress Indicator
                    app_dbg2(THIS_FILE, "%s PROGRESS INDICATOR (0x%02x)", status, info_element[0]);                
                    switch(info_element[2] & 0x7F)
                    {
                        case 0x01:
                            break;
                        case 0x02:
                            break;
                        case 0x03:
                            break;
                        case 0x04:
                            break;
                        case 0x05:
                            break;
                        case 0x08:  
                            break;
                    }
                    break;
                case 0x0027:    // Notification Indicator
                    switch((unsigned int)info_element[0])
                    {
                        case 0:
                            app_dbg2(THIS_FILE, "%s NI - CALL SUSPENDED (%d)", status, info_element[0]);
                            break;
                        case 1:
                            app_dbg2(THIS_FILE, "%s NI - CALL RESUMED (%d)", status, info_element[0]);
                            break;
                        case 2:
                            app_dbg2(THIS_FILE, "%s NI - BEARER SERVICE CHANGED (%d)", status, info_element[0]);
                            break;
                        default:
                            app_dbg2(THIS_FILE, "%s NI - UNKNOWN (%d)", status, info_element[0]);
                            break;
                    }
                    break;
                case 0x0028:    // Display
                    app_dbg2(THIS_FILE, "%s DISPLAY (%.*s)", status, (unsigned int)info_element[0], &info_element[1]);
                    break;
                case 0x0029:    // DateTime
                    app_dbg2(THIS_FILE, "%s DATE/TIME (%02d/%02d/%02d %02d:%02d)", status, info_element[0], info_element[1], info_element[2], info_element[3], info_element[4]);
                    break;    
                case 0x0074:    // Redirecting Number
                    break;
                case 0x00a1:    // Sending complete
                    break;
                case 0x4000:    // Charge in Units
                    break;
                case 0x4001:    // Charge in Currency
                    break;
                case 0x8001:    // Alerting
                    app_dbg2(THIS_FILE, "%s ALERTING", status);           
                    if(!oCONS[conid].capi->SetupEarly && (oCONS[conid].info->direction & OUTGOING)) {
                        script_event_add(CAPI, EARLY, oCONS[conid].script->file, oCONS[conid].id, oCONS[conid].in_conf[0], oCONS[conid].info->no_source, oCONS[conid].info->no_target, "");
                        oCONS[conid].capi->SetupEarly = 1;
                    }
                    break;
                case 0x8002:    // Call Proceeding                
                    break;
                case 0x8003:    // Progress                            
                    if(!oCONS[conid].capi->SetupEarly && (oCONS[conid].info->direction & OUTGOING)) {
                        script_event_add(CAPI, EARLY, oCONS[conid].script->file, oCONS[conid].id, oCONS[conid].in_conf[0], oCONS[conid].info->no_source, oCONS[conid].info->no_target, "");
                        oCONS[conid].capi->SetupEarly = 1;
                    }                    
                    app_dbg2(THIS_FILE, "%s PROGRESS", status);
                    break;
                case 0x8005:    // Setup
                    app_dbg2(THIS_FILE, "%s SETUP", status);
                    break;
                case 0x8007:    // Connect
                    app_dbg2(THIS_FILE, "%s CONNECT", status);
                    break;
                case 0x800d:    // Setup ACK
                    app_dbg2(THIS_FILE, "%s SETUP ACK", status);                    
                    oCONS[conid].capi->SetupAck = 1;                
                    break;
                case 0x800f:    // Connect ACK
                    app_dbg2(THIS_FILE, "%s CONNECT ACK", status);
                    break;
                case 0x8045:    // Disconnect  
                    app_dbg2(THIS_FILE, "%s DISCONNECT", status);                    
                    capi_hangup(&oCONS[conid]);                    
                    break;
                case 0x804d:    // Release
                    app_dbg2(THIS_FILE, "%s RELEASE", status);
                    break;
                case 0x805a:    // Release Complete
                    app_dbg2(THIS_FILE, "%s RELEASE COMPLETE", status);
                    break;
                case 0x8062:    // Facility
                    app_dbg2(THIS_FILE, "%s FACILITY", status);
                    break;
                case 0x806e:    // Notify
                    app_dbg2(THIS_FILE, "%s NOTIFY", status);
                    break;
                case 0x807b:    // Information
                    app_dbg2(THIS_FILE, "%s INFORMATION", status);
                    break;
                case 0x807d:    // Status
                    app_dbg2(THIS_FILE, "%s STATUS", status);
                    break;
                default:        // Unknown
                    app_dbg2(THIS_FILE, "%s UNKNOWN INFO", status);
                    break;
            }
      
            // (real) Early B3 for CAPI
            /*
            //if(oSET.audio.earlyb3_ring)
            {
                int x;
                x = (info_element[2] & 0x7F);

                // Receive dial-tone before connect !
                if((oCAPI.cMsg->InfoNumber == 0x18)) // 0x8001 0x18 // 0x1e
                {                
                    if(x == 0x08)
                    {
                        app_dbg(THIS_FILE, "(IND, CTRL %d, PLCI: %d) - Early-B3...", oCAPI.controller, oCAPI.PLCI);
                        CONNECT_B3_REQ(&CMSG1, oCAPI.AppId, oCAPI.iMsgNr, oCAPI.PLCI, 0);
                    }
                }

                // Hangup
                if(oCAPI.cMsg->InfoNumber == 0x08)
                {
                    i = connection_by_plci(oCAPI.PLCI);
                    if(i != -1)
                        capi_hangup(&oCONS[conid]);
                }
            }
            */
        
            // Internal controller        
            // we received a dial string (DTMF in InfoElement)?        
            if((oCONS[conid].info->direction & OUTGOING) && oCONS[conid].capi->SetupAck)
            {
                // Receive dial string
                if(oCAPI.cMsg->InfoNumber == 0x70 || oCAPI.cMsg->InfoNumber == 0x2c)
                {
                    int c;
                    int len = (int)oCAPI.cMsg->InfoElement[0];

                    // With keypad - 0x2c
                    int start_pos = 1;
                    int end_pos = 0;

                    // Without keypad - 0x70
                    if(oCAPI.cMsg->InfoNumber == 0x70)
                    {
                        start_pos = 2;
                        end_pos = 1;
                    }

                    // Get connection id
                    c = connection_by_plci(oCAPI.PLCI);
                    if(c != -1)
                    {
                        // if targetno is "unknown" empty it!
                        if(!strcmp(oCONS[c].info->no_target, "unknown"))
                            pj_memset(oCONS[c].info->no_target, 0, sizeof(oCONS[c].info->no_target));

                        // dialed number and hooked up after that
                        if(!strlen(oCONS[c].info->no_dialed) && (len - end_pos) > 1)
                        {
                            // replace target number with dialstring
                            strncat(oCONS[c].info->no_target, (const char*)(oCAPI.cMsg->InfoElement + start_pos), len-end_pos);
                            
                            // extract each DTMF from dial string and call script
                            for(i=0; i<len-end_pos; i++)
                            {
                                app_dbg3(THIS_FILE, "%s (IND) Dialed %c", connection_print_info(&oCONS[conid], info, sizeof(info)), (char)oCAPI.cMsg->InfoElement[i+start_pos]);
                                capi_get_dtmf(oCAPI.cMsg->InfoElement[i+start_pos], 1, c);
                            }
                        }

                        // hooked up and then dialing number(s)...
                        if(len-end_pos == 1)
                        {
                            // save target no. and call script
                            oCONS[c].info->no_target[strlen(oCONS[c].info->no_target)] = oCAPI.cMsg->InfoElement[start_pos];
                            oCONS[c].info->no_target[strlen(oCONS[c].info->no_target)] = '\0';

                            app_dbg3(THIS_FILE, "%s (IND) Dialing %c", connection_print_info(&oCONS[conid], info, sizeof(info)), (char)oCAPI.cMsg->InfoElement[start_pos]);
                            capi_get_dtmf(oCAPI.cMsg->InfoElement[start_pos], 1, c);
                        }
                    }
                }
            }     
        }
     #endif
}

void capi_event_data_b3()
{
    #if DTMFBOX_USE_CAPI
        
        int i;
        
        // Get connection by NCCI
        i = connection_by_ncci(oCAPI.NCCI);
        
        //app_dbg3(THIS_FILE, "(IND, CTRL %d, PLCI: %d, NCCI: %d) - CAPI_DATA_B3", oCON.controller, oCAPI.PLCI, oCAPI.NCCI);

        if(i != -1)
        {
            // Set connection information
            oCONS[i].capi->iMsgNr = oCAPI.iMsgNr;
            oCONS[i].capi->controller = oCAPI.controller;
            
            // We put capi_streaming into an own thread.
            // Otherwise there would be audio-stutters while streaming to CAPI custom_port (reverse_channel)            
            capi_streaming(&oCONS[i]);
        }        

    #endif
}


void capi_event_disconnect_b3()
{
    #if DTMFBOX_USE_CAPI

        char info[256];
        _cmsg CMSG1;
        int i, j;

        // Respond disconnect
        DISCONNECT_B3_RESP(&CMSG1, oCAPI.AppId, oCAPI.iMsgNr, oCAPI.NCCI);

        i = connection_by_ncci(oCAPI.NCCI);
        if(i != -1) 
        {        
            app_dbg(THIS_FILE, "%s (IND) CAPI_DISCONNECT_B3", connection_print_info(&oCONS[i], info, sizeof(info)));

            // disconnect all active media ports to conference partners.
            // The media ports gets reconnected in capi_event_connect_b3_active() when call is hold            
            for(j=0;j<INCONF_PORTS; j++)
            {
                if(oCONS[i].in_conf[j] != -1)
                {                
                    mixer_conf_port_disconnect(&oCONS[i], &oCONS[oCONS[i].in_conf[j]], 0);
                    mixer_conf_port_disconnect(&oCONS[oCONS[i].in_conf[j]], &oCONS[i], 0);                    
                }
            }

            // inband (call hold)?
            oCONS[i].capi->SetupInband = 1;

            // SCRIPT: "UNCONFIRMED"            
            script_event_add(CAPI, UNCONFIRMED, oCONS[i].script->file, i, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");
        }

    #endif
}

void capi_event_disconnect()
{
    #if DTMFBOX_USE_CAPI

        _cmsg CMSG1;
        int i;

        // Disconnect respond
        DISCONNECT_RESP(&CMSG1, oCAPI.AppId, oCAPI.iMsgNr, oCAPI.PLCI);

        // Stop playing/recording
        oCAPI.hdl = 0;

        // Disconnected due error?
        capi_info(DISCONNECT_IND_REASON(oCAPI.cMsg));

        // Close connection
        i = connection_by_plci(oCAPI.PLCI);
        if(i != -1)
        {         
            char info[256];

            app_dbg(THIS_FILE, "%s (IND) CAPI_DISCONNECT", connection_print_info(&oCONS[i], info, sizeof(info)));

            // SCRIPT: "DISCONNECT"
            snprintf(oCONS[i].script->dtmf, sizeof(oCONS[i].script->dtmf), "%d", DISCONNECT_IND_REASON(oCAPI.cMsg));

            script_event_add(CAPI, DISCONNECT, oCONS[i].script->file, oCONS[i].id, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, oCONS[i].script->dtmf);
            voip_registrar_disconnect_clients(&oCONS[i], 0);

            if(oCONS[i].info->auto_hookupdown)
                script_auto_hookdown(i);

            reset_connection(&oCONS[i]);
        }
       

    #endif
}
