/*********************************************************************************
 * capi_funcs.c ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "capi_funcs.c"

// Static pointer to table buffer (alaw/ulaw/pcm)
static pj_int8_t    *_linear16_2_law = &linear16_2_law[32768];
static pj_uint16_t  *_law_2_linear16 = &law_2_linear16[0];

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Make a CAPI call
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
int capi_call(connection_t *mCON, char *no_source, char *no_target, int controller, int call_anonymous, char *scriptfile, char *delimiter)
{
    #if DTMFBOX_USE_CAPI

    _cmsg cmsg;
    connection_t *CUR_CON = (connection_t*)mCON;
    unsigned char called_party_number[70];
    unsigned char calling_party_number[70];       
    unsigned int err;        
    pj_status_t status;
    int i;
    
    // CAPI installed?
    if(oCAPI.AppId == (unsigned short)-1) {
        capi_info(0x1009); 
        reset_connection(CUR_CON);
        return 0;
    }

    // Create capi_port
    status = port_capi_init(CUR_CON);
    if(status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create CAPI port!", status);
        reset_connection(CUR_CON);
        return 0;
    }
        
    // copy parameters to struct     
    if(!isnumeric(no_source) && CUR_CON->info->account_id != -1)
        pj_ansi_snprintf(CUR_CON->info->no_source, sizeof(CUR_CON->info->no_source), "%s", oSET.accounts[CUR_CON->info->account_id].number);    
    else
        pj_ansi_snprintf(CUR_CON->info->no_source, sizeof(CUR_CON->info->no_source), "%s", no_source);

    pj_ansi_snprintf(CUR_CON->info->no_target, sizeof(CUR_CON->info->no_target), "%s", no_target);    
    pj_ansi_snprintf(CUR_CON->script->file, sizeof(CUR_CON->script->file), "%s", scriptfile);
    pj_ansi_snprintf(CUR_CON->script->delimiter, sizeof(CUR_CON->script->file), "%s", delimiter);
    CUR_CON->capi->controller = controller;
    CUR_CON->info->status = 1; // CONNECTING
    CUR_CON->info->direction |= OUTGOING;

    // Pre-Connect conference partner, if there is one     
    for(i=0; i<INCONF_PORTS; i++)
    {
      if(CUR_CON->in_conf[i] != -1)
      {
          mixer_conf_port_connect(CUR_CON, &oCONS[CUR_CON->in_conf[i]], 1);
          mixer_conf_port_connect(&oCONS[CUR_CON->in_conf[i]], CUR_CON, 1);
      }
    }
    
    // Calling-Party Number
    pj_memset(calling_party_number, 0x00, sizeof(calling_party_number));    
    calling_party_number[0] = (unsigned char)(2 + strlen(CUR_CON->info->no_source));        
    calling_party_number[1] = 0x00; //0x00
    calling_party_number[2] = 0x80;
    if(call_anonymous) calling_party_number[2] = 0xA0;
    strncpy((char *)&calling_party_number[3], CUR_CON->info->no_source, strlen(CUR_CON->info->no_source));

    // Called-Party Number
    pj_memset(called_party_number, 0x00, sizeof(called_party_number));
    called_party_number[0] = (unsigned char)(1 + strlen(CUR_CON->info->no_target));
    called_party_number[1] = 0x80; // 0x80
    strncpy((char *)&called_party_number[2], CUR_CON->info->no_target, strlen(CUR_CON->info->no_target));
       
    app_dbg(THIS_FILE, "(CTRL %d, SRC: %s, DST: %s) - CONNECT_REQ", CUR_CON->capi->controller, CUR_CON->info->no_source, CUR_CON->info->no_target);
    
    // Request connect
    err = CONNECT_REQ(&cmsg, 
                        oCAPI.AppId,
                        0,
                        CUR_CON->capi->controller,                  // Controller
                        16,                                         // CIP-value (16 = Telephony)
                        (unsigned char *)called_party_number,       // Target No.
                        (unsigned char *)calling_party_number,      // Source No.
                        NULL,                                       // CalledPartySubaddress
                        NULL,                                       // CallingPartySubaddress,
                        1,                                          // B1 transp
                        1,                                          // B2 transp
                        0,                                          // B3 transp
                        NULL,                                       // B1configuration
                        NULL,                                       // B2configuration
                        NULL,                                       // B3configuration
                        NULL,                                       // BC
                        NULL,                                       // LLC
                        NULL,                                       // HLC
                        NULL,                                       // B-channel information
                        NULL,                                       // Keypadfacility
                        NULL,                                       // Userdata
                        NULL,                                       // Facilitydataarray
                        NULL,                                       // ...
                        NULL);                                      // SendingComplete       
    
        
    // Error?
    if(err)
    {
        app_err(THIS_FILE, "(%d) Unable to send CONNECT_REQ!", err);
        capi_info(err);
        reset_connection(CUR_CON);        
        return 0;
    }        

    #else
    PJ_UNUSED_ARG(delimiter);
    PJ_UNUSED_ARG(scriptfile);
    PJ_UNUSED_ARG(call_anonymous);
    PJ_UNUSED_ARG(controller);
    PJ_UNUSED_ARG(no_target);
    PJ_UNUSED_ARG(no_source);
    PJ_UNUSED_ARG(mCON);
    #endif

    return 1;
}

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// CAPI Hangup
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_hangup(connection_t *mCON)
{
    #if DTMFBOX_USE_CAPI

    _cmsg CMSG1;
    connection_t *CUR_CON = (connection_t*)mCON;
        
    if(CUR_CON->info->type != CAPI)
        return;
    
    CUR_CON->info->active = 0;

    DISCONNECT_REQ(&CMSG1, CUR_CON->capi->AppId, 0, CUR_CON->capi->PLCI, NULL, NULL, NULL, NULL);    

    #else
    PJ_UNUSED_ARG(mCON);
    #endif
}

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// CAPI confirm call
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_confirm_call(connection_t *mCON)
{
    #if DTMFBOX_USE_CAPI

    connection_t *CUR_CON = (connection_t*)mCON;

    if(CUR_CON->info->type != CAPI || CUR_CON->info->status == 2)
        return;
   
    // Alert (outgoing)
    if(CUR_CON->info->direction == OUTGOING)
        capi_send_alert_req(CUR_CON);

    // Respond connect (confirm)
    capi_send_connect_resp(CUR_CON, 0);

    // set "CONFIRMED"
    CUR_CON->info->status = 2;

    #else
    PJ_UNUSED_ARG(mCON);
    #endif
}

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Enable/disable CAPI DTMF (1=on, 0=off)
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_facility_dtmf(int enable)
{
    #if DTMFBOX_USE_CAPI
        
    _cmsg CMSG1;
    _cbyte fac[9];

    fac[0]=0x08;   // 0x08 (msg length)
    fac[1]=(_cbyte)enable; // DTMF ON: 0x01, DTMF OFF: 0x02
    fac[2]=0x00;   // NULL
    fac[3]=0x30;   // CAPI_DTMF_DURATION (0x30)
    fac[4]=0x00;   // NULL
    fac[5]=0x30;   // CAPI_DTMF_DURATION (0x30)
    fac[6]=0x00;   // NULL
    fac[7]=0x00;   // NULL
    fac[8]=0x00;   // NULL

    if(enable)
        app_dbg2(THIS_FILE, "Enable DTMF (ncci=%d)", oCAPI.NCCI);
    else
        app_dbg2(THIS_FILE, "Disable DTMF (ncci=%d)", oCAPI.NCCI);

    FACILITY_REQ(&CMSG1, oCAPI.AppId, 0, oCAPI.NCCI, 1, (unsigned char *)fac);

    #else
    PJ_UNUSED_ARG(enable);
    #endif
}

void capi_hold_call(connection_t* mCON, int hold)
{
    #if DTMFBOX_USE_CAPI
    
    char info[256];
    _cmsg CMSG1;
    _cbyte fac[9];
    connection_t *CUR_CON = (connection_t*)mCON;

    fac[0] = 3;	    // len 
    fac[1] =  (_cbyte)(0x0003-hold)  & 0xFF;
    fac[2] = (_cbyte)((0x0003-hold)  >> 8) & 0xFF;
    fac[3] = 0;	

    if(hold)
        app_dbg(THIS_FILE, "%s Place on hold", connection_print_info(CUR_CON, info, sizeof(info)));
    else
        app_dbg(THIS_FILE, "%s Place on retrieve", connection_print_info(CUR_CON, info, sizeof(info)));

    FACILITY_REQ(&CMSG1, oCAPI.AppId, 0, CUR_CON->capi->NCCI, 3, (unsigned char *)fac);

    #else
    PJ_UNUSED_ARG(mCON);
    PJ_UNUSED_ARG(hold);
    #endif
}

void capi_display_message(connection_t* mCON, char *displaytext)
{
    #if DTMFBOX_USE_CAPI
    
    char info[256];    
    _cbyte fac[62+3];
    char text[62];
    _cmsg CMSG1;
    int max_len = 31;
    connection_t *CUR_CON = (connection_t*)mCON;
    
    // max 31 chars !
    if(strlen(displaytext) < 31)
        max_len = strlen(displaytext);
        
    pj_memset(text, 0x00, sizeof(text));    
    pj_memset(fac, 0x0, sizeof(fac));
    strncpy(text, displaytext, max_len);

    fac[0]=(_cbyte)strlen(text)+2;                      // Complete length
    fac[1]=0x28;                                        // Display Element
    fac[2]=(_cbyte)strlen(text);                        // Length of Message (max. 31!)
    pj_memcpy(fac + 3, text, strlen(text));             // Text

    app_dbg(THIS_FILE, "%s Send Display-Message \"%s\"", connection_print_info(CUR_CON, info, sizeof(info)), text);
    
    INFO_REQ(&CMSG1, CUR_CON->capi->AppId, 0, CUR_CON->capi->PLCI, (unsigned char*)"", (unsigned char*)"", (unsigned char*)"", (unsigned char*)"", fac, NULL);
    
    #else
    PJ_UNUSED_ARG(mCON);
    PJ_UNUSED_ARG(displaytext);
    #endif
}

void capi_send_alert_req(connection_t* mCON)
{
    #if DTMFBOX_USE_CAPI
    
    _cmsg CMSG1;
    connection_t *CUR_CON = (connection_t*)mCON;
    
    // Send alert
    ALERT_REQ(&CMSG1, 
              CUR_CON->capi->AppId, 
              0, 
              CUR_CON->capi->PLCI,
              NULL,       // BChannelinformation
              NULL,       // Keypadfacility
              NULL,       // Useruserdata
              NULL,       // Facilitydataarray 
              NULL);       


    #else
    PJ_UNUSED_ARG(mCON);
    #endif
}

void capi_send_connect_resp(connection_t* mCON, int reject)
{
    #if DTMFBOX_USE_CAPI

    _cmsg CMSG1;
    connection_t *CUR_CON = (connection_t*)mCON;
    
    // Respond connect
    CONNECT_RESP(&CMSG1,
                 CUR_CON->capi->AppId,
                 0,           // MsgNr.
                 CUR_CON->capi->PLCI,
                (_cword)reject,      // Reject call ?
                 1,           // B1 Protocol
                 1,           // B2 Protocol
                 0,           // B3 Protocol
                 NULL,        // B1configuration
                 NULL,        // B2configuration
                 NULL,        // B3configuration
                 NULL,        // Globalconfiguration
                 NULL,        // ConnectedNumber
                 NULL,        // ConnectedSubaddress
                 NULL,        // LLC
                 NULL,        // BChannelinformation
                 NULL,        // Keypadfacility
                 NULL,        // Useruserdata
                 NULL);       // Facilitydataarray

    #else
    PJ_UNUSED_ARG(mCON);
    PJ_UNUSED_ARG(reject);
    #endif

}

#if DTMFBOX_USE_CAPI

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Get the calling party number on CAPI_CONNECT
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_get_source_no(_cmsg *CMSG, char number[MAX_URI_LENGTH], int add_prefix)
{    
    unsigned char *x=CONNECT_IND_CALLINGPARTYNUMBER(CMSG);
    unsigned int len = 0;

    pj_memset(number, 0, sizeof(number));

    if(x == NULL) {
        x = INFO_IND_INFOELEMENT(oCAPI.cMsg);
        if(x != NULL)
            len = (int)x[0];
    }
    else {
        len = *CONNECT_IND_CALLINGPARTYNUMBER(CMSG);
    }
    
    if(len <= 1)
    {
        strcpy(number, "unknown");
    }
    else
    {
        if(len>MAX_URI_LENGTH)
            len = MAX_URI_LENGTH-1;

        // Put (inter)national country code in front of the number
        if(add_prefix)
        {
            switch(x[1] & 112)
            {
                case 32:
                strcat(number, oSET.capi.int_prefix);
                break;

                case 64:
                strcat(number, oSET.capi.int_prefix);
                strcat(number, oSET.capi.nat_prefix);
                break;
            }
        }

        // get number
        if(x[2] & 128)
        {
           number[strlen(number)+x[0]-1]=0;
           number[strlen(number)+x[0]-2]=0;
           pj_memcpy(number+strlen(number),x+3,(size_t)(x[0]-2));
        }
        else
        {
           number[strlen(number)+x[0]]=0;
           number[strlen(number)+x[0]-1]=0;
           pj_memcpy(number+strlen(number),x+2,(size_t)(x[0]-1));
        }
    }

    if(!strlen(number))
        strcpy(number, "anonymous");    
}

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Get the called party number on CAPI_CONNECT
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_get_target_no(_cmsg *CMSG, char number[MAX_URI_LENGTH], int add_prefix)
{    
    unsigned char *x=CONNECT_IND_CALLEDPARTYNUMBER(CMSG);
    unsigned int len = 0;

    pj_memset(number, 0, sizeof(number));

    if(x == NULL) {
        x = INFO_IND_INFOELEMENT(oCAPI.cMsg);
        if(x != NULL)
            len = (int)x[0];
    }
    else {
        len = *CONNECT_IND_CALLEDPARTYNUMBER(CMSG);
        if(CONNECT_IND_CALLEDPARTYNUMBER(CMSG)[0] == 0)
            len = 0;
    }

    if(len <= 1)
    {
        strcpy(number, "unknown");
    }
    else
    {
        if(len>MAX_URI_LENGTH)
            len = MAX_URI_LENGTH-1;

        // Put (inter)national country code in front of the number
        if(add_prefix)
        {
            switch(x[1] & 112)
            {
                case 32:
                strcpy(number, oSET.capi.int_prefix);
                break;

                case 64:
                strcpy(number, oSET.capi.int_prefix);
                strcat(number, oSET.capi.nat_prefix);
                break;
            }
        }

        // get number
        number[strlen(number)+(size_t)x[0]]=0;
        number[strlen(number)+(size_t)x[0]-1]=0;
        pj_memcpy(number+strlen(number),x+2,(size_t)(x[0]-1));
    }

    if(!strlen(number))
        strcpy(number, "anonymous");    
}

#endif


// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Check for DTMF characters and execute script
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_get_dtmf(unsigned char dtmf, int ddi, int con_id)
{
    #if DTMFBOX_USE_CAPI
    
    int i;
    int dtmf_pos = 0;
    enum SCRIPT_EVENTS s_event = DTMF;

    // abort when no dtmf received is empty
    if((int)dtmf == 0)
        return;

    // check if it is really a DTMF signal
    //if((int)dtmf >= 0 && (int)dtmf <= 255)
    //{
        if(!isdigit((int)dtmf)) {
            if(dtmf != '#' && dtmf != '*')
                return;
        }
    //}

    if(oCAPI.AppId != (unsigned short)-1 && con_id >= 0 && con_id <= oSET.max_connections)
    {
        char info[256];

        dtmf_pos = oCONS[con_id].script->dtmf_pos;
        if(ddi) { s_event = DDI; }

        app_msg(THIS_FILE, "%s Incoming DTMF (%c)", connection_print_info(&oCONS[con_id], info, sizeof(info)), (unsigned char)(dtmf));

        // Send DTMF-Character to each conference partner
        for(i=0; i<INCONF_PORTS; i++)
        {
            int conf_id = oCONS[con_id].in_conf[i];
            if(conf_id != -1)
            {
                if(oCONS[conf_id].info->type == VOIP)
                    voip_send_dtmf(dtmf, conf_id);

                if(oCONS[conf_id].info->type == CAPI)                
                    capi_send_dtmf(dtmf, conf_id);    
            }
        }

        // Save DTMF and call script when necessary        
        oCONS[con_id].script->dtmf[dtmf_pos] = (char)dtmf;                
        oCONS[con_id].script->dtmf_pos++;
        
        if((char)dtmf == oCONS[con_id].script->delimiter[0] ||
           (dtmf_pos >= sizeof(oCONS[con_id].script->dtmf)-1) ||
           oCONS[con_id].script->delimiter[0] == '\0')
        {            
            script_event_add(CAPI, s_event, oCONS[con_id].script->file, oCONS[con_id].id, oCONS[con_id].in_conf[0], oCONS[con_id].info->no_source, oCONS[con_id].info->no_target, oCONS[con_id].script->dtmf);
            oCONS[con_id].script->dtmf_pos = 0;
            pj_memset(oCONS[con_id].script->dtmf, 0, sizeof(oCONS[con_id].script->dtmf));
        }
    }
    else
    {
        app_err(THIS_FILE, "Connection not found (%d)!", con_id);
    }

    #else
    PJ_UNUSED_ARG(con_id);
    PJ_UNUSED_ARG(ddi);
    PJ_UNUSED_ARG(dtmf);
    #endif
}

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// Send a DTMF digit 
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_send_dtmf(unsigned char dtmf, int con_id)
{
    #if DTMFBOX_USE_CAPI

    if(oCAPI.AppId != (unsigned short)-1)
    {
        char info[256];        
        char fac[32];
        _cmsg CMSG1;

        fac[0]=0x08;             // 0x08 (msg length)
        fac[1]=0x03;             // Send DTMF
        fac[2]=0x00;             // NULL (DWORD)
        fac[3]=0x30;             // CAPI_DTMF_DURATION
        fac[4]=0x00;             // NULL (DWORD)
        fac[5]=0x30;             // CAPI_DTMF_DURATION
        fac[6]=0x00;             // NULL (DWORD)
        fac[7]=0x01;             // length of dtmf digits
        fac[8]=dtmf;             // dtmf digit
        
        FACILITY_REQ(&CMSG1, oCONS[con_id].capi->AppId, 0, oCONS[con_id].capi->NCCI, 1, (unsigned char *)fac);

        app_dbg(THIS_FILE, "%s Outgoing DTMF (%d)", connection_print_info(&oCONS[con_id], info, sizeof(info)), (unsigned char)(dtmf));
    }

    #else
    PJ_UNUSED_ARG(con_id);
    PJ_UNUSED_ARG(dtmf);
    #endif
}

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// CAPI streaming (wave <-> voip <-> capi <-> snd)
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_streaming(connection_t *mCON)
{
    #if DTMFBOX_USE_CAPI

    connection_t *CUR_CON = (connection_t*)mCON;        
    pj_uint8_t alaw_buffer[CAPI_PACKETS];                               // the alaw_buffer for current connection
    pj_int16_t wave_buffer[CAPI_PACKETS];                               // the mixed wave_buffer             
    pjmedia_frame frame;
    pj_int16_t *wave;
    pj_uint8_t *alaw;
    _cword len = CAPI_PACKETS;
    pj_status_t status;    
    int i;
    
    if(CUR_CON->info->active == 1)
    {
        // --------------------------------
        // SEND
        // --------------------------------
        // --> receive alaw from CAPI and send to conference bridge
        len = DATA_B3_IND_DATALENGTH(CUR_CON->capi->cMsg);
        pj_memcpy(alaw_buffer, (pj_uint8_t*)DATA_B3_IND_DATA(CUR_CON->capi->cMsg), len);

        // Quickly respond DATA_B3...             
        DATA_B3_RESP(CUR_CON->capi->cMsg, CUR_CON->capi->AppId, CUR_CON->capi->iMsgNr, CUR_CON->capi->NCCI, DATA_B3_IND_DATAHANDLE(CUR_CON->capi->cMsg));
                
        frame.buf = wave_buffer;
        frame.size = sizeof(wave_buffer);
        frame.type = PJMEDIA_FRAME_TYPE_AUDIO;
        pj_memset(wave_buffer, 0, sizeof(wave_buffer));
        
        wave = (pj_int16_t*) wave_buffer;
        alaw = (pj_uint8_t*) alaw_buffer;        

        // convert (alaw/ulaw -> pcm)
        for (i=0; i!=len; ++i, ++wave) 
            *wave = _law_2_linear16[(pj_uint8_t)alaw[i]];          
        
        // send samples to conference bridge
        status = pjmedia_port_put_frame(CUR_CON->ports->custom, &frame);
        
        // --------------------------------
        // RECEIVE 
        // -------------------------------- 
        // <-- receive (mixed) pcm audio from conference bridge and send to CAPI
        pj_memset(alaw_buffer, 0, sizeof(alaw_buffer));                
        status = pjmedia_port_get_frame(CUR_CON->ports->custom, &frame);                                
        if(status == PJ_SUCCESS)
        {   
            wave = (pj_int16_t*) wave_buffer;
            alaw = (pj_uint8_t*) alaw_buffer;

            // convert (pcm -> alaw/ulaw)           
            for (i=0; i!=CAPI_PACKETS; ++i, ++alaw) 
                *alaw = _linear16_2_law[(pj_int16_t)wave[i]];          
        } 
        
        // --> send samples to CAPI
        #if DTMFBOX_WIN32
        DATA_B3_REQ(CUR_CON->capi->cMsg, CUR_CON->capi->AppId, CUR_CON->capi->iMsgNr, CUR_CON->capi->NCCI, (_cdword)alaw_buffer, CAPI_PACKETS, CUR_CON->capi->hdl, 0);
        #else
        DATA_B3_REQ(CUR_CON->capi->cMsg, CUR_CON->capi->AppId, CUR_CON->capi->iMsgNr, CUR_CON->capi->NCCI, (void*)alaw_buffer, CAPI_PACKETS, CUR_CON->capi->hdl, 0);
        #endif

        // increase hdl
        CUR_CON->capi->hdl++;               
    }

    //app_dbg3(THIS_FILE, "DATA_B3_REQ - Ctrl: %d - PLCI: %d, NCCI: %d, Hdl: %d, Len %d, Frame-Size: %d", CUR_CON->controller, CUR_CON->capi->PLCI, CUR_CON->capi->NCCI, CUR_CON->hdl, len, DATA_B3_IND_DATALENGTH(CUR_CON->capi->cMsg));

    #else
    PJ_UNUSED_ARG(mCON);
    #endif
}

// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
// CAPI info/error messages
// ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void capi_info(long info)
{
    char capi_err[128];

    pj_ansi_snprintf(capi_err, sizeof(capi_err), "CAPI 0x%x -", info);

    switch(info)
    {
        case 0x0001:
            app_err(THIS_FILE, "%s NCPI not supported by current protocol", capi_err);
            break;
        case 0x0002:
            app_err(THIS_FILE, "%s Flags not supported by current protocol, flags ignored", capi_err);
            break;
        case 0x0003:
            app_err(THIS_FILE, "%s Alert already sent by another application", capi_err);
            break;

        // CAPI_Register messages
        case 0x1001:
            app_err(THIS_FILE, "%s Too many applications", capi_err);
            break;
        case 0x1002:
            app_err(THIS_FILE, "%s Logical Block size too small; must be at least 128 bytes", capi_err);
            break;
        case 0x1003:
            app_err(THIS_FILE, "%s Buffer exceeds 64 kbytes", capi_err);
            break;
        case 0x1004:
            app_err(THIS_FILE, "%s Message buffer size too small, must be at least 1024 bytes", capi_err);
            break;
        case 0x1005:
            app_err(THIS_FILE, "%s Max. number of logical connections not supported", capi_err);
            break;
        case 0x1006:
            app_err(THIS_FILE, "%s reserved (unknown error)", capi_err);
            break;
        case 0x1007:
            app_err(THIS_FILE, "%s The message could not be accepted because of an internal busy condition", capi_err);
            break;
        case 0x1008:
            app_err(THIS_FILE, "%s OS Resource error (out of memory?)", capi_err);
            break;
        case 0x1009:
            app_err(THIS_FILE, "%s CAPI not installed", capi_err);
            break;
        case 0x100A:
            app_err(THIS_FILE, "%s Controller does not support external equipment", capi_err);
            break;
        case 0x100B:
            app_err(THIS_FILE, "%s Controller does only support external equipment", capi_err);
            break;

        // Error information concerning message exchange functions 
        case 0x1101:
            app_err(THIS_FILE, "%s Illegal application number", capi_err);
            break;
        case 0x1102:
            app_err(THIS_FILE, "%s Illegal command or subcommand, or message length less than 12 octets", capi_err);
            break;
        case 0x1103:
            app_err(THIS_FILE, "%s The message could not be accepted because of a queue full condition", capi_err);
            break;
        case 0x1104:
            app_err(THIS_FILE, "%s Queue overflow: a message was lost!!", capi_err);
            break;
        case 0x1105:
            app_err(THIS_FILE, "%s Unknown notification parameter", capi_err);
            break;
        case 0x1106:
            app_err(THIS_FILE, "%s The message could not be accepted because on an internal busy condition", capi_err);
            break;
        case 0x1107:
            app_err(THIS_FILE, "%s OS resource error (out of memory?)", capi_err);
            break;
        case 0x1108:
            app_err(THIS_FILE, "%s CAPI not installed", capi_err);
            break;
        case 0x1109:
            app_err(THIS_FILE, "%s Controller does not support external equipment", capi_err);
            break;
        case 0x110A:
            app_err(THIS_FILE, "%s Controller does only support external equipment", capi_err);
            break;
        case 0x110B:
            app_err(THIS_FILE, "%s Controller does only support external equipment", capi_err);
            break;

        // error information concerning resource / coding problems 
        case 0x2001:
            app_err(THIS_FILE, "%s Message not supported in current state", capi_err);
            break;
        case 0x2002:
            app_err(THIS_FILE, "%s Illegal Controller / PLCI / NCCI", capi_err);
            break;
        case 0x2003:
            app_err(THIS_FILE, "%s Out of PLCI", capi_err);
            break;
        case 0x2004:
            app_err(THIS_FILE, "%s Out of NCCI", capi_err);
            break;
        case 0x2005:
            app_err(THIS_FILE, "%s Out of LISTEN", capi_err);
            break;
        case 0x2007:
            app_err(THIS_FILE, "%s llegal message parameter coding", capi_err);
            break;

        // error information concerning requested services 
        case 0x3001:
            app_err(THIS_FILE, "%s B1 protocol not supported", capi_err);
            break;
        case 0x3002:
            app_err(THIS_FILE, "%s B2 protocol not supported", capi_err);
            break;
        case 0x3003:
            app_err(THIS_FILE, "%s B3 protocol not supported", capi_err);
            break;
        case 0x3004:
            app_err(THIS_FILE, "%s B1 protocol parameter not supported", capi_err);
            break;
        case 0x3005:
            app_err(THIS_FILE, "%s B2 protocol parameter not supported", capi_err);
            break;
        case 0x3006:
            app_err(THIS_FILE, "%s B3 protocol parameter not supported", capi_err);
            break;
        case 0x3007:
            app_err(THIS_FILE, "%s B protocol combination not supported", capi_err);
            break;
        case 0x3008:
            app_err(THIS_FILE, "%s NCPI not supported", capi_err);
            break;
        case 0x3009:
            app_err(THIS_FILE, "%s CIP Value unknown", capi_err);
            break;
        case 0x300A:
            app_err(THIS_FILE, "%s Flags not supported (reserved bits)", capi_err);
            break;
        case 0x300B:
            app_err(THIS_FILE, "%s Facility not supported", capi_err);
            break;
        case 0x300C:
            app_err(THIS_FILE, "%s Data length not supported by current protocol", capi_err);
            break;
        case 0x300D:
            app_err(THIS_FILE, "%s Reset procedure not supported by current protocol", capi_err);
            break;
        case 0x300E:
            app_err(THIS_FILE, "%s TEI assignment failed or supplementary service not supported", capi_err);
            break;
        case 0x3010:
            app_err(THIS_FILE, "%s Request not allowed in this state", capi_err);
            break;

        // informations about the clearing of a physical connection
    	case 0x3301:
	        app_err(THIS_FILE, "%s Protocol error layer 1 (broken line or B-channel removed by signalling protocol)", capi_err);
            break;
        case 0x3302:
	        app_err(THIS_FILE, "%s Protocol error layer 2", capi_err);
            break;
        case 0x3303:
	        app_err(THIS_FILE, "%s Protocol error layer 3", capi_err);
            break;
        case 0x3304:
	        app_err(THIS_FILE, "%s Another application got that call", capi_err);
            break;

        // disconnect errors / reasons
        case 0x3481:
            app_err(THIS_FILE, "%s Unallocated (unassigned) number", capi_err);        
            break;
        case 0x3482:
	        app_err(THIS_FILE, "%s No route to specified transit network", capi_err);
            break;
        case 0x3483:
	        app_err(THIS_FILE, "%s No route to destination", capi_err);
            break;
        case 0x3486:
	        app_err(THIS_FILE, "%s Channel unacceptable", capi_err);
            break;
        case 0x3487:
	        app_err(THIS_FILE, "%s Call awarded and being delivered in an established channel", capi_err);
            break;
        case 0x3490:
	        app_err(THIS_FILE, "%s Normal call clearing", capi_err);
            break;
        case 0x3491:
	        app_err(THIS_FILE, "%s User busy", capi_err);
            break;
        case 0x3492:
	        app_err(THIS_FILE, "%s No user responding", capi_err);
            break;
        case 0x3493:
	        app_err(THIS_FILE, "%s No answer from user (user alerted)", capi_err);
            break;
        case 0x3495:
	        app_err(THIS_FILE, "%s Call rejected", capi_err);
            break;
        case 0x3496:
	        app_err(THIS_FILE, "%s Number changed", capi_err);
            break;
        case 0x349A:
	        app_err(THIS_FILE, "%s Non-selected user clearing", capi_err);
            break;
        case 0x349B:
	        app_err(THIS_FILE, "%s Destination out of order", capi_err);
            break;
        case 0x349C:
	        app_err(THIS_FILE, "%s Invalid number format", capi_err);
            break;
        case 0x349D:
	        app_err(THIS_FILE, "%s Facility rejected", capi_err);
            break;
        case 0x349E:
	        app_err(THIS_FILE, "%s Response to STATUS ENQUIRY", capi_err);
            break;
        case 0x349F:
	        app_err(THIS_FILE, "%s Normal, unspecified", capi_err);
            break;
        case 0x34A2:
	        app_err(THIS_FILE, "%s No circuit / channel available", capi_err);
            break;
        case 0x34A6:
	        app_err(THIS_FILE, "%s Network out of order", capi_err);
            break;
        case 0x34A9:
	        app_err(THIS_FILE, "%s Temporary failure", capi_err);
            break;
        case 0x34AA:
	        app_err(THIS_FILE, "%s Switching equipment congestion", capi_err);
            break;
        case 0x34AB:
	        app_err(THIS_FILE, "%s Access information discarded", capi_err);
            break;
        case 0x34AC:
	        app_err(THIS_FILE, "%s Requested circuit / channel not available", capi_err);
            break;
        case 0x34AF:
	        app_err(THIS_FILE, "%s Resources unavailable, unspecified", capi_err);
            break;
        case 0x34B1:
	        app_err(THIS_FILE, "%s Quality of service unavailable", capi_err);
            break;
        case 0x34B2:
	        app_err(THIS_FILE, "%s Requested facility not subscribed", capi_err);
            break;
        case 0x34B9:
	        app_err(THIS_FILE, "%s Bearer capability not authorized", capi_err);
            break;
        case 0x34BA:
	        app_err(THIS_FILE, "%s Bearer capability not presently available", capi_err);
            break;
        case 0x34BF:
	        app_err(THIS_FILE, "%s Service or option not available, unspecified", capi_err);
            break;
        case 0x34C1:
	        app_err(THIS_FILE, "%s Bearer capability not implemented", capi_err);
            break;
        case 0x34C2:
	        app_err(THIS_FILE, "%s Channel type not implemented", capi_err);
            break;
        case 0x34C5:
	        app_err(THIS_FILE, "%s Requested facility not implemented", capi_err);
            break;
        case 0x34C6:
	        app_err(THIS_FILE, "%s Only restricted digital information bearer capability is available", capi_err);
            break;
        case 0x34CF:
	        app_err(THIS_FILE, "%s Service or option not implemented, unspecified", capi_err);
            break;
        case 0x34D1:
	        app_err(THIS_FILE, "%s Invalid call reference value", capi_err);
            break;
        case 0x34D2:
	        app_err(THIS_FILE, "%s Identified channel does not exist", capi_err);
            break;
        case 0x34D3:
	        app_err(THIS_FILE, "%s A suspended call exists, but this call identity does not", capi_err);
            break;
        case 0x34D4:
	        app_err(THIS_FILE, "%s Call identity in use", capi_err);
            break;
        case 0x34D5:
	        app_err(THIS_FILE, "%s No call suspended", capi_err);
            break;
        case 0x34D6:
	        app_err(THIS_FILE, "%s Call having the requested call identity has been cleared", capi_err);
            break;
        case 0x34D8:
	        app_err(THIS_FILE, "%s Incompatible destination", capi_err);
            break;
        case 0x34DB:
	        app_err(THIS_FILE, "%s Invalid transit network selection", capi_err);
            break;
        case 0x34DF:
	        app_err(THIS_FILE, "%s Invalid message, unspecified", capi_err);
            break;
        case 0x34E0:
	        app_err(THIS_FILE, "%s Mandatory information element is missing", capi_err);
            break;
        case 0x34E1:
	        app_err(THIS_FILE, "%s Message type non-existent or not implemented", capi_err);
            break;
        case 0x34E2:
	        app_err(THIS_FILE, "%s Message not compatible with call state or message type non-existent or not implemented", capi_err);
            break;
        case 0x34E3:
	        app_err(THIS_FILE, "%s Information element non-existent or not implemented", capi_err);
            break;
        case 0x34E4:
	        app_err(THIS_FILE, "%s Invalid information element contents", capi_err);
            break;
        case 0x34E5:
	        app_err(THIS_FILE, "%s Message not compatible with call state", capi_err);
            break;
        case 0x34E6:
	        app_err(THIS_FILE, "%s Recovery on timer expiry", capi_err);
            break;
        case 0x34EF:
	        app_err(THIS_FILE, "%s Protocol error, unspecified", capi_err);
            break;
        case 0x34FF:
	        app_err(THIS_FILE, "%s Interworking, unspecified", capi_err);
            break;
    }
}
