/*********************************************************************************
 * config.c ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "config.c"

// Read INI values by section and key
char* get_value(char* section, char* key, char* value)
{
    char line[1024];
    char* findsec = strdup(section);
    char* findkey = strdup(key);
    FILE* fp;
    char* pos;
    int insection;
    int found;
    unsigned int i;

    // Open file
    fp = fopen(oGLB.configfile, "r");

    // error - cannot read file!
    if(fp == NULL) 
    {
        free(findkey); free(findsec);
        findkey = NULL; findsec = NULL;
        return NULL;
    }

    //convert section & key to uppercase
    for(i=0; i<strlen(findsec); i++)
        findsec[i] = (char)toupper(findsec[i]);
    for(i=0; i<strlen(findkey); i++)
        findkey[i] = (char)toupper(findkey[i]);

    insection = 0;
    found = 0;

    // Each line
    while(fgets(line, sizeof(line), fp) != NULL)
    {
        // remove comments
        pos = strstr(line, "//");
        while(pos != NULL) {
            int no_comment = 0;
            if(pos - line != 0)
            {
                if(line[(pos - line)-1] == 92) {
                    no_comment = 1;
                    memcpy(line + ((pos - line)-1), line + ((pos - line)), strlen(pos));                    
                }
            }

            if(!no_comment)
                pos[0] = '\0';

            pos = strstr(pos+2, "//");
        }

        // remove CR
        pos = strchr(line, '\n');
        if(pos != NULL)
            pos[0] = '\0';

        // trim
        trim(line);

        // copy return value
        if(found==0)
            strcpy(value, line);

        // convert line to uppercase
        for(i=0; i<strlen(line); i++)
            line[i] = (char)toupper(line[i]);

        if(insection == 0)
        {
            // enter section
            pos = strstr(line, findsec);
            if(pos!=NULL)
            {
                insection = 1;
            }
        } else {
            // leave section
            if(line[0] == '[')
            {
                insection = 0;
            }
        }

        if(insection == 1)
        {
            pos = strstr(line, findkey);
            if(pos!=NULL)
            {
                // key found!!
                found = 1;
                strcpy(value, value + strlen(findkey));
                break;
            }
        }
    }

    // Close file
    fclose(fp);

    // nothing found?
    if(found == 0)
        value[0] = '\0';

    // output configuration (loglevel 5)
    if(strlen(value))
        if(!strcmp("voip_pass=", key))
            app_dbg3(THIS_FILE, "%s - %s*******", section, key);
        else
            app_dbg3(THIS_FILE, "%s - %s%s", section, key, value);

    // put settings to environment    
    if(strlen(findsec) > 2 && strlen(findsec) < 32)
    {            
        char env[512];
        char tmp[32];

        pj_memcpy(findsec, findsec+1, strlen(findsec)-2);
        pj_memcpy(tmp, findsec, strlen(findsec)-2);
        tmp[strlen(findsec)-2] = '\0';

        // make unique environment variable name
        if(!strcmp(tmp, "GENERAL") || !strcmp(tmp, "VOIP") || !strcmp(tmp, "CAPI"))
            snprintf(env, sizeof(env), "DTMFBOX_%s%s", findkey, value);
        else
            snprintf(env, sizeof(env), "DTMFBOX_%s_%s%s", tmp, findkey, value);

        // maybe too much
        //putenv(env);
    }

    // free
    free(findkey); free(findsec);
    findkey = NULL; findsec = NULL;

    return value;
}

// split values 
// val_in       = input value
// val_out      = output value (also return)
// size        = sizeof(val_out)
// split_char   = splitting char
char *get_sep(char *val_in, char *val_out, int size, char split_char)
{
    char *p;    
    //char ret_val[256];
    
    memset(val_out, 0, size);

    p = strchr(val_in, split_char);
    if(p != NULL)
    {
        int psize = p - val_in;
        if(psize >  size)
            psize = sizeof(val_out);

        strncpy(val_out, val_in, psize);
        val_out[psize] = '\0';

        strcpy(val_in, p+1);        
    }
    else
    {        
        if(strlen(val_in))
        {
            strncpy(val_out, val_in, strlen(val_in));    
            val_out[strlen(val_in)] = '\0';

            strcpy(val_in, "");
        } 
        else
        {
            strcpy(val_out, "");
        }
    }    

    return trim(val_out);
}

// Right-Trim
char* rtrim(char *str)
{
    char* p = str + strlen(str);
    while(isspace(*--p));
        *(p + 1) = '\0';
    
    return str;
}

// Left-Trim
char* ltrim(char *str)
{
    char *p1 = str;
    char *p2 = p1;
    int ok = 0;
    do
    {
        if (!isspace(*p1) || ok)
        {            
            *p2 = *p1;
            *p2++;
            ok = 1;
        }
    } while (*p1++ != '\0');

    return str;
}

char *trim(char *str)
{
    return rtrim(ltrim(str));
}

// Check, that string is numerical
int isnumeric(char *str)
{
    unsigned int i;
    int ret = strlen(str);
    
    for(i=0; i<strlen(str); i++)
    {
        //if((int)str[i] >= 0 && (int)str[i] <= 255)
        //{
            if(!isdigit((int)str[i]))
            {
                ret = 0;
                break;
            }
        //}
    }

    return ret;
}

// Replace a string
char* replace(const char *in, char *out, int size, const char* find, const char* replace)
{
	const char* p;
    char *ret;
    char *tmp;
    int pos = 0;

    ret = (char*)malloc(size);
    tmp = (char*)malloc(size);

    if (out == NULL || ret == NULL || tmp == NULL) {
        if(ret) free(ret);
        if(tmp) free(tmp);
        tmp = NULL; ret = NULL;
		return NULL;
    } else {
        memset(ret, 0, size);
        memset(tmp, 0, size);
        strncpy(tmp, in, size);
    }
	
    p = tmp;
    do
    {
        if(p!=NULL) p = strstr(p, find);
        if(p == NULL) break;
        pos = (int)(strlen(in) - strlen(p));

        // copy first part
        if (strcmp(p, in)) {
		    strncpy(ret, tmp, p-tmp);
        }

        // replace with replace-string and strcat the rest
	    strcat(ret + (p-tmp), replace);
	    strcat(ret, p + strlen(find));

        // save as new text
        strncpy(tmp, ret, size);
        
        // research from beginning
        p = tmp;
        memset(ret, 0, size);
        
    } while(p != NULL);

    strncpy(out, tmp, size);
    free(ret); free(tmp);
    tmp = NULL; ret = NULL;

	return out;
}

// Check if entry= in [key] is on or off (returns 1 or 0)
int config_isonoff(char *key, char *entry)
{
    char value[1024];
    
    get_value(key, entry, value);
    value[1] = '\0';
    if(!strcmp(value, "1") || !strcmp(value, "on") || !strcmp(value, "yes")) return 1;
    else return 0;
}

void init_config(char *current_executable)
{    
    #if (DTMFBOX_WIN32 == 0)
    char *TmpDirs[3] = { "/tmp", "/var/lock", "/var/tmp" };    
    FILE *fp = NULL;
    int i;
    int iStart, iEnd;
    #endif

    char *pDir = NULL;        
    
    // Change to executable directory
    snprintf(oGLB.current_exe, sizeof(oGLB.current_exe), "%s", current_executable);
    #if DTMFBOX_WIN32
        pDir = strrchr(oGLB.current_exe, '\\');
    #else
        pDir = strrchr(oGLB.current_exe, '/');
    #endif
    if(pDir != NULL)
    {        
        strncpy(oGLB.workdir, oGLB.current_exe, pDir - oGLB.current_exe + 1);
        oGLB.workdir[pDir - oGLB.current_exe + 1] = '\0';        
        chdir((const char*)oGLB.workdir);                
    }

    // Set default configuration
    pj_memset(&oSET, 0, sizeof(oSET));            
    oSET.max_connections = MAX_CONS;
    oSET.adjust_priority = 0;    
    oSET.scriptfile[0] = '\0';
    oSET.plugins[0] = '\0';
    oSET.delimiter[0] = '\0';
    oSET.logging.loglevel = 3;    
    oSET.logging.show_conid = 1;    
    oSET.logging.show_accid = 1;
    oSET.logging.show_src = 0;
    oSET.logging.show_dst = 0;
    oSET.logging.show_type = 1;
    oSET.voip.udp_port = 5060;
    oSET.voip.rtp_start = 4000;
    oSET.voip.rtp_max = MAX_TRANSPORTS;
    oSET.voip.max_clients = 8;    
    oSET.voip.register_interval = 3600;    
    oSET.voip.registrar = 0;    
    oSET.voip.allow_unknown_acc = 0;
    oSET.voip.keep_alive = 180;    
    oSET.voip.stun[0] = '\0';
    oSET.voip.stun_port = 3478;
    oSET.voip.use_vad = 0;
    oSET.voip.ice = 0;
    oSET.use_capi = 0;
    oSET.use_voip = 0;    
    oSET.capi.nat_prefix[0] = '\0';
    oSET.capi.int_prefix[0] = '\0';
    oSET.capi.ulaw = 0;
    oSET.capi.allow_unknown_acc = 0;
    oSET.audio.resampler_quality_conf = PJMEDIA_CONF_USE_LINEAR;
    oSET.audio.resampler_quality = PJMEDIA_RESAMPLE_USE_LINEAR;
    oSET.audio.null_sound = 1;
    oSET.audio.bidirectional_sound = 0;
    oSET.audio.snd_play = -1;
    oSET.audio.snd_rec = -1;    
    oSET.audio.snd_ec_tail = 0;
    oSET.audio.con_ec_tail = 0;
    oSET.audio.snd_clockrate = SOUND_CLOCKRATE;    
    oSET.audio.mixer_rx_volume = 100;
    oSET.audio.mixer_tx_volume = 100;
    oSET.audio.mixer_talkover = 0;
    oSET.audio.earlyb3_ring = 1;
    oSET.audio.conf_clockrate = CONF_CLOCKRATE;        
    oCAPI.AppId = (unsigned short)-1;    

    // Check access for pidfile
    // It is required, that all users write/read the same pidfile. 
    // So first check, if there is already a dtmfbox.pid available. 
    #if (DTMFBOX_WIN32 == 0)
    iStart = 0; iEnd = 3;
    for(i=iStart; i<iEnd; i++)
    {
        char tmp[256];
        snprintf(tmp, sizeof(tmp), "%s/dtmfbox-%d.pid", TmpDirs[i], getpid());
        fp = fopen(tmp, "r+");
        if(fp != NULL)
        {
            snprintf(oSET.pid_file, sizeof(oSET.pid_file), "%s/dtmfbox.pid", TmpDirs[i]);
            iStart = i; 
            iEnd = i+1;
            fclose(fp);
            break;
        }
    }

    for(i=iStart; i<iEnd; i++)
    {
        char tmp[256];
        snprintf(tmp, sizeof(tmp), "%s/dtmfbox-%d.pid~", TmpDirs[i], getpid());
        fp = fopen(tmp, "w");
        if(fp != NULL) {
            snprintf(oSET.pid_file, sizeof(oSET.pid_file), "%s/dtmfbox.pid", TmpDirs[i]);
            iStart = i;
            iEnd = i+1;
            fclose(fp);            
            remove(tmp);
            break;
        }
    }
    if(!strlen(oSET.pid_file))
    {
        printf("Error: Unable to create pidfile in \"/tmp\", \"/var/lock\", \"/var/tmp\" or \"/var/run\"\n");
        exit(1);
    }

    // Check access for pipe dir
    for(i=iStart; i<iEnd; i++)
    {
        char tmp[256];
        snprintf(tmp, sizeof(tmp), "%s/dtmfbox-%d.pipe~", TmpDirs[i], getpid());
        fp = fopen(tmp, "w");
        if(fp != NULL) {
            snprintf(oSET.pipe_dir, sizeof(oSET.pipe_dir), "%s", TmpDirs[i]);
            fclose(fp);
            remove(tmp);
            break;
        }
    }
    if(!strlen(oSET.pipe_dir))
    {
        printf("Error: Unable to create pidfile in \"/tmp\", \"/var/lock\" or \"/var/tmp\"\n");
        exit(1);
    }
    
    #else
    
    snprintf(oSET.pipe_dir, sizeof(oSET.pipe_dir), "%s", DTMFBOX_PIPES);
    snprintf(oSET.pid_file, sizeof(oSET.pid_file), "%s", DTMFBOX_PIDFILE);
    
    #endif

}

// Read configuration file and fill APP_SETTINGS struct
int read_config()
{
    char value[1024];
    char key[32];
    int i;

    // Can the file be opened?
    if(get_value("", "", value) == NULL)
        return -1;

    // ------------------------------------
    // [general] section
    // ------------------------------------
    strcpy(key, "[general]");

    // Scriptfile (*.bat or *.exe)
    strncpy(oSET.scriptfile, get_value(key, "scriptfile=", value), sizeof(oSET.scriptfile));

    // Plugins (*.dll, *.so). Comma-Seperated.
    strncpy(oSET.plugins, get_value(key, "plugins=", value), sizeof(oSET.plugins));

    // Delimiter
    get_value(key, "delimiter=", value);
    if(strstr(value, "asterisk") != NULL) oSET.delimiter[0] = '*';
    else if(strstr(value, "poundkey") != NULL) oSET.delimiter[0] = '#';
    else if(strstr(value, "none") != NULL) oSET.delimiter[0] = '\0';
    else oSET.delimiter[0] = value[0];

    // adjust priority
    oSET.adjust_priority  = config_isonoff(key, "adjust_priority=");
    
    // Max. connections
    get_value(key, "max_connections=", value);    
    if(strlen(value) && isnumeric(value)) oSET.max_connections = atoi(value);    
    if(oSET.max_connections > MAX_CONS) oSET.max_connections = MAX_CONS;
    
    // ------------------------------------
    // [logging] section
    // ------------------------------------
    strcpy(key, "[logging]");

    // Loglevel
    get_value(key, "loglevel=", value);
    value[1] = '\0';
    if(isnumeric(value)) oSET.logging.loglevel = atoi(value);
        
    oSET.logging.show_sender = config_isonoff(key, "show_sender=");
    oSET.logging.show_type = config_isonoff(key, "show_type=");
    oSET.logging.show_conid = config_isonoff(key, "show_con=");
    oSET.logging.show_accid = config_isonoff(key, "show_acc=");
    oSET.logging.show_src = config_isonoff(key, "show_source=");
    oSET.logging.show_dst = config_isonoff(key, "show_destination=");
    oSET.logging.show_ctrl = config_isonoff(key, "show_controller=");
    oSET.logging.show_plci = config_isonoff(key, "show_plci=");
    oSET.logging.show_ncci = config_isonoff(key, "show_ncci=");

    // ------------------------------------
    // [audio] section
    // ------------------------------------
    strcpy(key, "[audio]");

    // Conference bridge clockrate?
    get_value(key, "conf_clockrate=", value);    
    if(isnumeric(value)) oSET.audio.conf_clockrate = atoi(value);
    else oSET.audio.conf_clockrate = CONF_CLOCKRATE;

    // Use null-sound or soundcard?
    oSET.audio.null_sound = config_isonoff(key, "null_sound=");
    if(oSET.audio.null_sound != 1) oSET.max_connections++;

    // Bidirectional or unidirectional sound?
    oSET.audio.bidirectional_sound = config_isonoff(key, "bidirectional_sound=");

    // Soundcard player id
    get_value(key, "snd_play=", value);
    if(isnumeric(value)) oSET.audio.snd_play = atoi(value);
    else oSET.audio.snd_play = -1;

    // Soundcard recorder id
    get_value(key, "snd_rec=", value);
    if(isnumeric(value)) oSET.audio.snd_rec = atoi(value);
    else oSET.audio.snd_rec = -1;

    // Soundcard frequency
    get_value(key, "snd_clockrate=", value);
    if(isnumeric(value)) oSET.audio.snd_clockrate = atol(value);
    else oSET.audio.snd_clockrate = SOUND_CLOCKRATE;

    // Resampler quality (0=high, 1=middle, 2=low)
    get_value(key, "resampler_quality=", value);
    if(!strcmp(value, "0")) {
        oSET.audio.resampler_quality = 0;
        oSET.audio.resampler_quality_conf = 0;
    }
    if(!strcmp(value, "1")) {
        oSET.audio.resampler_quality = PJMEDIA_RESAMPLE_USE_SMALL_FILTER;
        oSET.audio.resampler_quality_conf = PJMEDIA_CONF_SMALL_FILTER;
    }
    if(!strcmp(value, "2")) {
        oSET.audio.resampler_quality = PJMEDIA_RESAMPLE_USE_LINEAR;
        oSET.audio.resampler_quality_conf = PJMEDIA_CONF_USE_LINEAR;
    }

    // Soundcard echo canceller tail length
    get_value(key, "snd_ec_tail=", value);
    if(isnumeric(value)) oSET.audio.snd_ec_tail = atol(value);
    else oSET.audio.snd_ec_tail = 0;

    // Connection echo canceller tail length
    get_value(key, "con_ec_tail=", value);
    if(isnumeric(value)) oSET.audio.con_ec_tail = atol(value);
    else oSET.audio.con_ec_tail = 0;

    // TX-Audio level
    get_value(key, "mixer_tx_volume=", value);
    if(isnumeric(value)) oSET.audio.mixer_tx_volume = atoi(value);
    else oSET.audio.mixer_tx_volume = MIX_DEF_TX_VOLUME;

    // RX-Audio level
    get_value(key, "mixer_rx_volume=", value);
    if(isnumeric(value)) oSET.audio.mixer_rx_volume = atoi(value);
    else oSET.audio.mixer_rx_volume = MIX_DEF_RX_VOLUME;

    // Talkover active by default?
    oSET.audio.mixer_talkover = config_isonoff(key, "mixer_talkover=");

    // use Early-B3 connect for CAPI?
    oSET.audio.earlyb3_ring = config_isonoff(key, "earlyb3_ring=");

    // ------------------------------------
    // [voip] section
    // ------------------------------------
    strcpy(key, "[voip]");

    // VoIP on/off
    oSET.voip_active = config_isonoff(key, "voip=");
  
    // Allow unknown voip accounts??
    oSET.voip.allow_unknown_acc = config_isonoff(key, "voip_allow_unknown=");

    // Act as VoIP-registrar?
    oSET.voip.registrar = config_isonoff(key, "voip_registrar=");

    // VoIP realm
    strncpy(oSET.voip.realm, get_value(key, "voip_realm=", value), sizeof(oSET.voip.realm));

    // VoIP UDP port
    get_value(key, "voip_udp_port=", value);
    if(strlen(value)) oSET.voip.udp_port = atol(value);

    // VoIP RTP start port
    get_value(key, "voip_rtp_start=", value);
    if(strlen(value) && isnumeric(value)) oSET.voip.rtp_start = (pj_uint16_t)atol(value);

    // VoIP RTP start port
    get_value(key, "voip_rtp_max=", value);
    if(strlen(value) && isnumeric(value)) oSET.voip.rtp_max = (int)atoi(value);
    if(oSET.voip.rtp_max > oSET.max_connections) oSET.voip.rtp_max = oSET.max_connections;

    // VoIP max. clients who can connect (Registrar-Mode)
    get_value(key, "voip_max_clients=", value);
    if(strlen(value) && isnumeric(value)) oSET.voip.max_clients = atoi(value);

    // VoIP Re-Register interval
    get_value(key, "voip_register_interval=", value);
    if(strlen(value) && isnumeric(value)) oSET.voip.register_interval = atol(value);

    // VoIP keep-alive interval
    get_value(key, "voip_keep_alive=", value);
    if(strlen(value) && isnumeric(value)) oSET.voip.keep_alive = atol(value);

    // STUN-Server + port:
    strncpy(oSET.voip.stun, get_value(key, "voip_stun=", value), sizeof(oSET.voip.stun));
    get_value(key, "voip_stun_port=", value);
    if(strlen(value) && isnumeric(value)) oSET.voip.stun_port = atol(value);    

    // Use ICE?
    if(strlen(oSET.voip.stun))
        oSET.voip.ice = config_isonoff(key, "voip_ice=");

    // internal/external ip/hostname
    strncpy(oSET.voip.ip_addr, get_value(key, "voip_ip_addr=", value), sizeof(oSET.voip.ip_addr));
    strncpy(oSET.voip.hostname, get_value(key, "voip_ip_addr=", value), sizeof(oSET.voip.hostname));

    // Use VAD (silence detector)?
    oSET.voip.use_vad = config_isonoff(key, "voip_use_vad=");

    //  Nameserver (max. 4 nameservers = xxx.xxx.xxx.xxx,yyy.yyy.yyy.yyy,...)
    {
        char tmp[1024];
        int cnt = 0;
                
        pj_memset(oSET.voip.nameserver, 0, sizeof(oSET.voip.nameserver));

        get_value(key, "voip_nameserver=", value);        
        get_sep(value, tmp, sizeof(tmp), ',');
        while(cnt < 4)
        {            
            strncpy(oSET.voip.nameserver[cnt], tmp,  sizeof(*&oSET.voip.nameserver[cnt]));
            cnt++;            

            get_sep(value, tmp, sizeof(tmp), ',');
        }
    }
        
    // codec order (6 codecs = pcmu,pcma,ilbc,gsm,l16,speex)
    {
        char tmp[1024];
        int cnt = 0;
                
        pj_memset(oSET.voip.codec_order, 0, sizeof(oSET.voip.codec_order));

        get_value(key, "voip_codecs=", value);
        get_sep(value, tmp, sizeof(tmp), ',');
        while(cnt < 6)
        {            
            strncpy(oSET.voip.codec_order[cnt], tmp, sizeof(*&oSET.voip.codec_order[cnt]));
            cnt++;            
            get_sep(value, tmp, sizeof(tmp), ',');
        }
    }

    // default codec, when none selected
    if(!strlen(oSET.voip.codec_order[0]))
        strcpy(oSET.voip.codec_order[0], "g711");
        

    // ------------------------------------
    // [capi] section
    // ------------------------------------
    strcpy(key, "[capi]");

    // capi_controller=1,2,...,*
    // comma seperated (max. 5!)
    oSET.capi.controller.device = CAPI;
    {
        char tmp[1024];
        int cnt = 0;
                
        pj_memset(oSET.capi.controller.ctrl, 0, sizeof(oSET.capi.controller.ctrl));

        get_value(key, "capi_controller=", value);        
        get_sep(value, tmp, sizeof(tmp), ',');
        while(cnt < CAPI_CONTROLLERS)
        {           
            oSET.capi.controller.ctrl[cnt] = 0;

            if(isnumeric(tmp))
            {
                oSET.capi.controller.ctrl[cnt] = atoi(tmp);                            
            } else {
                if(!strcmp(value, "*")) oSET.capi.controller.ctrl[cnt] = -1;
            }
            
            cnt++;
            get_sep(value, tmp, sizeof(tmp), ',');
        }
    }

    // National & international prefix
    strncpy(oSET.capi.nat_prefix, get_value(key, "capi_nat_prefix=", value), sizeof(oSET.capi.nat_prefix));
    strncpy(oSET.capi.int_prefix, get_value(key, "capi_int_prefix=", value), sizeof(oSET.capi.int_prefix));

    // A-law or U-law conversation?
    oSET.capi.ulaw = config_isonoff(key, "capi_ulaw=");

    // Allow unknown voip accounts??
    oSET.capi.allow_unknown_acc = config_isonoff(key, "capi_allow_unknown=");

    // ------------------------------------
    // [accx] section (1-10)
    // ------------------------------------
    oSET.use_voip = 0;
    oSET.use_capi = 0;

    for(i=0; i<MAX_ACCOUNTS; i++)
    {
        snprintf(key, sizeof(key), "[acc%d]", i+1);

        // Account active?
        oSET.accounts[i].active = config_isonoff(key, "active=");

        // Is this a VoIP or CAPI account?
        get_value(key, "type=", value);
        if(strstr(value, "voip"))
        {            
            oSET.accounts[i].controller.device = VOIP;

            // VoIP
            if(oSET.accounts[i].active) 
                oSET.use_voip = 1;
        }
        else
        {         
            oSET.accounts[i].controller.device = CAPI;

            // CAPI
            if(oSET.accounts[i].active)
                oSET.use_capi = 1;
        }

        // Misc. VoIP settings (Registrar, Username, Password, Realm, ...)
        strncpy(oSET.accounts[i].name, get_value(key, "name=", value), sizeof(oSET.accounts[i].name));
        strncpy(oSET.accounts[i].number, get_value(key, "number=", value), sizeof(oSET.accounts[i].number));

        // Doutgoing controller for CAPI
        get_value(key, "capi_controller_out=", value);
        if(isnumeric(value)) oSET.accounts[i].controller.ctrl_out = atoi(value);
        else oSET.accounts[i].controller.ctrl_out = oSET.capi.controller.ctrl[0];   // first Listen-Controller as default, when not defined

        // VoIP settings
        strncpy(oSET.accounts[i].username, get_value(key, "voip_user=", value), sizeof(oSET.accounts[i].username));
        strncpy(oSET.accounts[i].password, get_value(key, "voip_pass=", value), sizeof(oSET.accounts[i].password));
        strncpy(oSET.accounts[i].realm, get_value(key, "voip_realm=", value), sizeof(oSET.accounts[i].realm));
        strncpy(oSET.accounts[i].proxy, get_value(key, "voip_proxy=", value), sizeof(oSET.accounts[i].proxy));
        strncpy(oSET.accounts[i].contact, get_value(key, "voip_contact=", value), sizeof(oSET.accounts[i].contact));
        strncpy(oSET.accounts[i].registrar, get_value(key, "voip_registrar=", value), sizeof(oSET.accounts[i].registrar));
        strncpy(oSET.accounts[i].id, get_value(key, "voip_id=", value), sizeof(oSET.accounts[i].id));
        oSET.accounts[i].do_not_register = config_isonoff(key, "voip_do_not_register=");

        // Registrar-Mode settings (active, username, password)
        oSET.accounts[i].registrar_active = config_isonoff(key, "registrar_active=");
        if(oSET.accounts[i].registrar_active) {
            strncpy(oSET.accounts[i].registrar_username, get_value(key, "registrar_user=", value), sizeof(oSET.accounts[i].registrar_username));
            strncpy(oSET.accounts[i].registrar_password, get_value(key, "registrar_pass=", value), sizeof(oSET.accounts[i].registrar_password));
        }

        // Set status pending
        oSET.accounts[i].status = 0;
        oSET.accounts[i].status_reason[0] = '\0';
        oSET.accounts[i].status_nextreg = 0;
        oSET.accounts[i].status_failcnt = 0;
    }

    // Check values, correct them if needed...
    if(oSET.voip.max_clients < 1 || oSET.voip.max_clients > MAX_CLIENTS)
        oSET.voip.max_clients = 8;
    //if(oSET.voip.rtp_start < 1 || oSET.voip.rtp_start > 65535)
      //  oSET.voip.rtp_start = 4000;
    if(oSET.voip.udp_port < 1 || oSET.voip.udp_port > 65535)
        oSET.voip.udp_port = 5060;
    if(oSET.voip.register_interval < 1)
        oSET.voip.rtp_start = 3600;
    if(oSET.voip.keep_alive < 0 || oSET.voip.keep_alive > 3600)
        oSET.voip.keep_alive = 180;
    if(oSET.audio.mixer_rx_volume > 200 || oSET.audio.mixer_rx_volume < 0)
        oSET.audio.mixer_rx_volume = MIX_DEF_RX_VOLUME;
    if(oSET.audio.mixer_tx_volume > 200 || oSET.audio.mixer_tx_volume < 0)
        oSET.audio.mixer_tx_volume = MIX_DEF_TX_VOLUME;
    if(oSET.audio.conf_clockrate < 8000 || oSET.audio.conf_clockrate > 44100)
        oSET.audio.conf_clockrate = CONF_CLOCKRATE;
    if(oSET.audio.snd_clockrate < 8000 || oSET.audio.snd_clockrate > 44100)
        oSET.audio.conf_clockrate = SOUND_CLOCKRATE;
    
    oSET.audio.conf_samplerate = oSET.audio.conf_clockrate * 20 / 1000;

    // Compiler-Flags defaults
    #if(DTMFBOX_USE_ICE == 0)
       oSET.voip.ice = 0;
    #endif

    #if(DTMFBOX_USE_VOIP == 0)
       oSET.voip_active = 0;
       oSET.voip.ice = 0;
    #endif

    return 0;
}

