/*********************************************************************************
 * funcs.c ($Revision: 260 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "funcs.c"

// strip user part from a sip-uri
char *sip_strip_user(char *sip)
{
    char*pos;
    char tmp[MAX_URI_LENGTH];

    pj_memset(tmp, 0, sizeof(tmp));

    // remove sip:
    pos = strstr(sip, "sip:");
    if(pos!=NULL)            
        pj_memcpy(tmp, pos+4, strlen(pos));
    else
        pj_memcpy(tmp, sip, strlen(sip));

    // remove @.*
    pos = strstr(tmp, "@");
    if(pos!=NULL)    
        pos[0] = '\0';

    // remove + prefix
    if(strlen(tmp))
    {
        if(tmp[0] == '+')
        {
            int len=strlen(tmp)-1;
            pj_memcpy(tmp, tmp+1, len);
            tmp[len] = '\0';
        }
    }

    strcpy(sip, tmp);

    return sip;
}

// Strip host part from a sip-uri
char *sip_strip_host(char *sip)
{
    char*pos;
    char tmp[MAX_URI_LENGTH];

    pj_memset(tmp, 0, sizeof(tmp));

    pos = strstr(sip, "sip:");
    if(pos!=NULL)            
        pj_memcpy(tmp, pos+4, strlen(pos)-4);
    else
        pj_memcpy(tmp, sip, strlen(sip));

    pos = strstr(tmp, "@");
    if(pos==NULL)
    {
        strcpy(sip, "");
        return sip;
    }
    else
    {
        pj_memcpy(tmp, pos+1, strlen(pos)-1);
        tmp[strlen(pos)-1] = '\0';
    }

    pos = strstr(tmp, ">");
    if(pos!=NULL)
        pos[0] = '\0';

    strcpy(sip, tmp);

    return sip;
}

// Get account-id by username
int account_by_username(char* username)
{
    int i;

    // Search through all accounts
    for(i=0; i<MAX_ACCOUNTS; i++)
    {
        // Check if account settings match
        if(strlen(oSET.accounts[i].registrar_username) > 0 &&
          oSET.accounts[i].active == 1 && 
          oSET.accounts[i].registrar_active == 1)
        {
            // Username found?
            if(strcmp(oSET.accounts[i].registrar_username, username) == 0)
            {
                return i;
            }
        }
    }

    // account not found :(
    return -1;
}

// Get account-id by Number
int account_by_number(char* number)
{
    int i;

    // Search through all accounts
    for(i=0; i<MAX_ACCOUNTS; i++)
    {
        // Check if account settings match
        if(strlen(oSET.accounts[i].number) > 0 &&
          oSET.accounts[i].active == 1)
        {
            // Number found (number, as in settings)!
            if(!strcmp(oSET.accounts[i].number, number))
                return i;
        }
    }

    // number not found :(
    return -1;
}


// Get connection-id by NCCI
int connection_by_ncci(long NCCI)
{
    int i;
    int ret = -1;

    for(i=0; i<oSET.max_connections; i++)
    {
        if(oCONS[i].capi->NCCI == (unsigned long)NCCI)
        {
            ret = i;
            break;
        }
    }
    
    //app_dbg3(THIS_FILE, "connection_by_ncci - NCCI: %d, Connection: %d", NCCI, ret);

    return ret;
}

// Get connection-id by PLCI
int connection_by_plci(long PLCI)
{
    int i;
    int ret = -1;

    for(i=0; i<oSET.max_connections; i++)
    {
        if((unsigned long)oCONS[i].capi->PLCI == (unsigned long)PLCI)
        {
            ret = i;
            break;
        }
    }

    //app_dbg3(THIS_FILE, "connection_by_plci - PLCI: %d, Connection: %d", PLCI, ret);

    return ret;
}

// Get connection-id by invite session (VoIP)
int connection_by_invite(pjsip_inv_session *inv)
{
    int ret = -1;

    #if DTMFBOX_USE_VOIP
        int i;

        // Invite-Session suchen
        for(i=0; i<oSET.max_connections; i++)
        {
            if(oCONS[i].rtp != NULL) {
                if(oCONS[i].rtp->g_inv == inv)
                {
                    ret = i;
                    break;
                }
            }
        }
        
        //app_dbg3(THIS_FILE, "connection_by_invite - Connection: %d", ret);

    #else
    PJ_UNUSED_ARG(inv);
    #endif

    return ret;
}

// Get a free transport
int transport_by_free()
{
    int ret = -1;
    
    #if DTMFBOX_USE_VOIP
        int i;
        for(i=0; i<oSET.voip.rtp_max; i++)
        {
            if(oRTP[i]->g_inv == NULL)
            {
                ret = i;
                break;
            }
        }
    #endif

    return ret;
}

// Get next free playing queue (save playing count in play_cnt);
int play_by_free(int con_id, int *play_cnt)
{    
    int queue = -1;
    int queue_cnt = 0;
    int i;

    // get new playing queue and count
    for(i=0; i<INCONF_PORTS; i++)
    {
        if(sh_actions->play_rec.play_thread[con_id][i] == -1 &&
             oCONS[con_id].ports->play_resample[i] == NULL && 
             oCONS[con_id].ports->play_stream[i] == NULL &&
             queue == -1)
           
        {
            queue = i;            
        }
        
        if(sh_actions->play_rec.play_thread[con_id][i] != -1 ||
           oCONS[con_id].ports->play_resample[i] != NULL ||
           oCONS[con_id].ports->play_stream[i] != NULL)       
        {
           queue_cnt++;
        }            
    }

    *play_cnt = queue_cnt;

    if(queue == -1) 
        return -1;
    
    return queue;
}

// Get a free connection-id
int connection_by_free()
{
    int i;
    int ret = -1;

    for(i=0; i<oSET.max_connections; i++)
    {
        if(oCONS[i].info->type == FREE)
        {
            ret = i;
            break;
        }
    }
    
    if(ret == -1)
        app_err(THIS_FILE, "connection_by_free - Connection: %d - no free connections left. Max: %d!", ret, oSET.max_connections);
    else
        app_dbg3(THIS_FILE, "connection_by_free - Connection: %d", ret);

    return ret;
}

// Get client-table id by address (VoIP-Registrar)
int client_by_address(char *client_addr)
{
    int i;
    int ret = -1;

    // walkthrough accounts
    for(i=0; i<oSET.voip.max_clients; i++)
    {
        // already in list?
        //if(!pj_strcmp(client_addr, &oGLB.clients[i].addr))
        if(!strcmp(client_addr, oGLB.clients[i].addr))
        {
            ret = i;
            break;
        }
    }

    return ret;
}

// Get a free client-id (VoIP-Registrar)
int client_by_free()
{
    int i;
    int ret = -1;
    //pj_str_t client_addr = pj_str("");

    // walkthrough accounts
    for(i=0; i<oSET.voip.max_clients; i++)
    {
        // already in list?
        if(!strcmp("", oGLB.clients[i].addr))
        {
            ret = i;
            break;
        }
    }

    return ret;
}


// Get client-account id by address and account_id (VoIP-Registrar/Accounts)
int client_account_by_account(int client_id, int account_id)
{
    int i;
    int ret = -1;

    // walkthrough accounts
    for(i=0; i<MAX_ACCOUNTS; i++)
    {
        if(oGLB.clients[client_id].account_id[i] == account_id)
        {
            ret = i;
            break;
        }
    }

    return ret;
}

// Get free client-account id (VoIP-Registrar/Accounts)
int client_account_by_free(int client_id)
{
    int i;
    int ret = -1;

    // walkthrough accounts
    for(i=0; i<MAX_ACCOUNTS; i++)
    {
        if(oGLB.clients[client_id].account_id[i] == -1)
        {
            ret = i;
            break;
        }
    }

    return ret;
}

// Initialize connection struct and array
void init_connections()
{    
    int i, j;
     
    // Alloc memory for structs
    oCONS = pj_pool_zalloc(oGLB.pool, sizeof(connection_t) * oSET.max_connections);    
    for(i=0; i<oSET.max_connections; i++) {        
        oCONS[i].capi = pj_pool_zalloc(oGLB.pool, sizeof(connection_capi_t));        
        oCONS[i].pools = pj_pool_zalloc(oGLB.pool, sizeof(connection_pools_t));
        oCONS[i].ports = pj_pool_zalloc(oGLB.pool, sizeof(connection_ports_t));
        oCONS[i].slots = pj_pool_zalloc(oGLB.pool, sizeof(connection_slots_t));
        oCONS[i].mixer = pj_pool_zalloc(oGLB.pool, sizeof(connection_mixer_t));
        oCONS[i].info = pj_pool_zalloc(oGLB.pool, sizeof(connection_info_t));
        oCONS[i].script = pj_pool_zalloc(oGLB.pool, sizeof(connection_script_t));                    
    }
    
    // Alloc memory for transports
    oRTP = pj_pool_zalloc(oGLB.pool, sizeof(connection_transport_t*) * oSET.voip.rtp_max);
    for(i=0; i<oSET.voip.rtp_max; i++)    
        oRTP[i] = pj_pool_zalloc(oGLB.pool, sizeof(connection_transport_t));
   
    app_msg(THIS_FILE, "Allocate %d connections with %d RTP/RTCP transports", oSET.max_connections, oSET.voip.rtp_max);

    // Reset temporary CAPI object
    oCAPI.cMsg = &oCAPI.cMsgInst;
    oCAPI.AppId = 0;
    oCAPI.iMsgNr = 0;
    oCAPI.PLCI = 0;
    oCAPI.NCCI = 0;
    oCAPI.controller = 0;
    oCAPI.hdl = 0;

    // Reset connection array 
    for(i=0; i < oSET.max_connections; i++)
    {           
        oCONS[i].id = i;                
        for(j=0; j<INCONF_PORTS; j++)        
            oCONS[i].slots->play[j] = -1;
        
        reset_connection(&oCONS[i]);
    }    

    // Create a global User-Agent Header for SIP ;)
    {
        char useragent_value[64];
        const pj_str_t useragent_name = pj_str("User-Agent");            
        pj_str_t useragent;
                
        pj_ansi_snprintf(useragent_value, sizeof(useragent_value), "%s %s", APPNAME, APPVER);
        useragent = pj_str(useragent_value);
        oGLB.useragent_hdr = pjsip_generic_string_hdr_create(oGLB.pool, &useragent_name, &useragent);
    }

}

// Reset connection
void reset_connection(connection_t *mCON)
{
    connection_t *CUR_CON = (connection_t*)mCON;
    pj_status_t status;
    int i;
        
    // Do not reset soundcard !!
    if(CUR_CON->info->type == SCRD)
        return;
    
    DTMFBOX_ACC_LOCK();

    // IMPORTANT: MAKE SURE, THAT EVERYTHING STOPS!
    // --------------------------------------------    
    CUR_CON->info->active = 0;
    CUR_CON->info->status = 0;
    // --------------------------------------------

    // Cancel script timer
    pjsip_endpt_cancel_timer(oGLB.endpt, &CUR_CON->script->timer);
        
    // Close pipe
    if(CUR_CON->script->pipe_id != -1) {
        script_event_run();     // flush remaining events
        pipe_close(CUR_CON->script->pipe_id);
    }

    // Close recorder and player (if not already done)        
    for(i=0; i<INCONF_PORTS; i++) 
    {
        int play_id = -1;
        port_filestream_player_recorder(CUR_CON, DEVNULL, "", 0, 0, 0, 0, 0, &play_id);        
    }        
    {
        int play_id = 0;
        port_filestream_player_recorder(CUR_CON, "", DEVNULL, 0, 0, 0, 0, 0, &play_id);
        //port_filestream_close(CUR_CON, 0, 1);
    }

    // Destroy Tone-Thread
    if(CUR_CON->tone_thread != NULL) {
        pj_thread_join(CUR_CON->tone_thread);
        pj_thread_destroy(CUR_CON->tone_thread);
        CUR_CON->tone_thread = NULL;            
    }
    
    // Destroy Tone-Port       
    if(CUR_CON->ports->tone != NULL)
        port_tone_close(CUR_CON);   
    
    // Close CAPI-Port
    if(CUR_CON->ports->custom != NULL && !strncmp(CUR_CON->ports->custom->info.name.ptr, "capi port", 9))
        status = port_capi_close(CUR_CON);

    // Close VoIP-Port
    if(CUR_CON->ports->custom != NULL && !strncmp(CUR_CON->ports->custom->info.name.ptr, "voip port", 9))
        status = port_voip_close(CUR_CON);

    // Reset CONNECTION struct    
    CUR_CON->info->account_id = -1;
    CUR_CON->info->client_id = -1;
    CUR_CON->info->direction = NONE;
    CUR_CON->info->auto_hookupdown = 0;
    CUR_CON->capi->PLCI = 0;
    CUR_CON->capi->NCCI = 0;
    CUR_CON->capi->hdl = 0;        
    CUR_CON->capi->SetupInband = 0;
    CUR_CON->capi->SetupAck = 0;
    CUR_CON->capi->SetupListen = 0;
    CUR_CON->capi->SetupEarly = 0;    
    CUR_CON->rtp = NULL;
    CUR_CON->in_conf_cnt = 0;    
    CUR_CON->tone_thread = NULL;
    CUR_CON->ports->media = NULL;        
    CUR_CON->ports->rec_resample = NULL;
    CUR_CON->ports->rec_stream = NULL;                
    CUR_CON->ports->tone = NULL;    
    CUR_CON->slots->media = -1;
    CUR_CON->slots->rec = -1;    
    CUR_CON->slots->tone = -1;    
    CUR_CON->mixer->in_talkover = 0;
    CUR_CON->script->dtmf_pos = 0;    
    CUR_CON->script->waitevent = 0;    
    CUR_CON->script->pipe_id = -1;
    pj_memset(CUR_CON->info->no_source, 0, sizeof(CUR_CON->info->no_source));
    pj_memset(CUR_CON->info->no_target, 0, sizeof(CUR_CON->info->no_target));
    pj_memset(CUR_CON->info->no_dialed, 0, sizeof(CUR_CON->info->no_dialed));
    pj_memset(CUR_CON->script->file, 0, sizeof(CUR_CON->script->file));
    pj_memset(CUR_CON->script->delimiter, 0, sizeof(CUR_CON->script->delimiter));
    pj_memset(CUR_CON->script->dtmf, 0, sizeof(CUR_CON->script->dtmf));    

    // Reset mixer
    for(i=0; i<N_CON_PORTS; i++)
    {
        CUR_CON->mixer->is_background[i] = 0;
        CUR_CON->mixer->is_talkover[i] = 0;
        CUR_CON->mixer->talkover_active[i] = 0;
    }

    // Reset ports
    for(i=0; i<INCONF_PORTS; i++)
    {        
        CUR_CON->ports->play_stream[i] = NULL;
        CUR_CON->ports->play_resample[i] = NULL;        
        CUR_CON->slots->play[i] = -1;        
        CUR_CON->in_conf[i] = -1;
    }
    
    // Release connection pool
    if(CUR_CON->pools->con_pool != NULL)
    {            
        pj_pool_release(CUR_CON->pools->con_pool);
        CUR_CON->pools->con_pool = NULL;
    }

    // Create connection memory pool
    if(CUR_CON->pools->con_pool == NULL)
    {
        CUR_CON->pools->con_pool = pj_pool_create(&oGLB.cp.factory,      // pool factory
                                        "connection_pool (%p)",         // pool name
                                        POOL_SIZE_CON,                  // init size
                                        POOL_INCR_CON,                  // increment size
                                        NULL                            // callback on error
                                        );
    } 

    // Free to use
    CUR_CON->info->type = FREE;    

    DTMFBOX_ACC_UNLOCK();
}

char *connection_print_info(connection_t *CUR_CON, char *info, int size)
{    
    char src[MAX_URI_LENGTH];
    char dst[MAX_URI_LENGTH];    
    char type[32];    
    char plci[16];
    char ncci[16];
    char ctrl[16];
    char conid[16];
    char accid[16];

    memset(info, 0, size);
    if(CUR_CON == NULL) return info;
    if(CUR_CON->info == NULL) return info;

    memset(conid, 0, sizeof(conid));
    memset(accid, 0, sizeof(accid));
    memset(type, 0, sizeof(type));    
    memset(src, 0, sizeof(src));
    memset(dst, 0, sizeof(dst));
    memset(ctrl, 0, sizeof(ctrl));
    memset(ncci, 0, sizeof(ncci));
    memset(plci, 0, sizeof(plci));
    
    if(oSET.logging.show_conid)
        snprintf(conid, sizeof(conid), " conid=%d", CUR_CON->id);

    if(strlen(CUR_CON->info->no_source) && oSET.logging.show_src)
        snprintf(src, sizeof(src), " src=%s", CUR_CON->info->no_source);
    if(strlen(CUR_CON->info->no_target) && oSET.logging.show_dst)
        snprintf(dst, sizeof(dst), " trg=%s", CUR_CON->info->no_target);
    
    if(oSET.logging.show_accid)
        snprintf(accid, sizeof(accid), " accid=%d", CUR_CON->info->account_id >= 0 ? CUR_CON->info->account_id + 1 : -1);

    if(oSET.logging.show_type) 
    {
        switch(CUR_CON->info->type)
        {
            case CAPI:
                snprintf(type, sizeof(type), " type=capi/%s", CUR_CON->info->direction == INCOMING ? "inc" : "out");
                break;
            
            case VOIP:
                snprintf(type, sizeof(type), " type=voip/%s", CUR_CON->info->direction == INCOMING ? "inc" : "out");
                break;

            case SCRD:            
                snprintf(info, size, "[type=scrd]");
                return info;
                break;
            
            default:
                snprintf(type, sizeof(type), " type=err!");            
                break;
        }
    }

    if(CUR_CON->info->type == CAPI)
    {
        if(oSET.logging.show_plci) snprintf(plci, sizeof(plci), " plci=%d", CUR_CON->capi->PLCI);
        if(oSET.logging.show_ncci) snprintf(ncci, sizeof(ncci), " ncci=%d", CUR_CON->capi->NCCI);
        if(oSET.logging.show_ctrl) snprintf(ctrl, sizeof(ctrl), " ctrl=%d", CUR_CON->capi->controller);                        
    }

    snprintf(info, size, "%s%s%s%s%s%s%s%s", type, conid, accid, src, dst, ctrl, plci, ncci);
    
    if(strlen(info)) {
        info[0] = '['; info[strlen(info)] = ']';        
    }

    return info;
}

// Application debug messages 3
void app_dbg3(const char *sender, const char *format, ...)
{
#if PJ_LOG_MAX_LEVEL >= 4

    if(oGLB.daemonized && oGLB.file_log == NULL)
        return;

    if(oSET.logging.loglevel >= 4)
    {
        char output[LOG_BUFFERSIZE];
        va_list args;

        va_start(args, format);
        vsnprintf(output, sizeof(output), format, args);
        PJ_LOG(4, (sender, "%s", output));
        va_end(args);
    }

#endif
}

// Application debug messages 2
void app_dbg2(const char *sender, const char *format, ...)
{
#if PJ_LOG_MAX_LEVEL >= 3

    if(oGLB.daemonized && oGLB.file_log == NULL)
        return;

    if(oSET.logging.loglevel >= 3)
    {
        char output[LOG_BUFFERSIZE];
        va_list args;

        va_start(args, format);
        vsnprintf(output, sizeof(output), format, args);
        PJ_LOG(3, (sender, "%s", output));
        va_end(args);
    }

#endif
}

// Application debug messages
void app_dbg(const char *sender, const char *format, ...)
{    
#if PJ_LOG_MAX_LEVEL >= 2

    if(oGLB.daemonized && oGLB.file_log == NULL)
        return;

    if(oSET.logging.loglevel >= 2)
    {
        char output[LOG_BUFFERSIZE];
        va_list args;

        va_start(args, format);
        vsnprintf(output, sizeof(output), format, args);
        PJ_LOG(2, (sender, "%s", output));
        va_end(args);    
    }

#endif
}

// Application error messages
void app_err(const char *sender, const char *format, ...)
{    
#if PJ_LOG_MAX_LEVEL >= 1

    if(oGLB.daemonized && oGLB.file_log == NULL)
        return;

    if(oSET.logging.loglevel >= 1)
    {
        char output[LOG_BUFFERSIZE];
        va_list args;

        va_start(args, format);
        vsnprintf(output, sizeof(output), format, args);
        PJ_LOG(1, (sender, "%s", output));
        va_end(args);
    }

#endif
}

// Application messages
void app_msg(const char *sender, const char *format, ...)
{
#if PJ_LOG_MAX_LEVEL >= 1

    if(oGLB.daemonized && oGLB.file_log == NULL)
        return;

    if(oSET.logging.loglevel >= 1)
    {
        char output[LOG_BUFFERSIZE];
        va_list args;

        va_start(args, format);
        vsnprintf(output, sizeof(output), format, args);
        PJ_LOG(1, (sender, "%s", output));
        va_end(args);
    }
#endif
}

// Log callback (write to stdout or file)
void pj_log_func_cb(int level, const char *data, int len)
{
    // write to stdout:
    if(!oGLB.daemonized)
        pj_log_write(level, data, len);

    if(oGLB.file_log != NULL)
    {
        fwrite(data, len, 1, oGLB.file_log);
        fflush(oGLB.file_log);
    }    
}

// Dynamically adjust process priority (high/low) on active connections
void adjust_priority()
{
    int i;
    int all_free = 1;
    
    for(i=0; i<oSET.max_connections; i++) {
        if(oCONS[i].info->type != FREE && oCONS[i].info->type != SCRD) {
            
            // active connection --> higher process priority
            if(oGLB.process_boosted == 0)
                boost_priority(1);                
            
            all_free = 2;
            break;
        }        
    }

    // no active connections --> lower process priority
    if(all_free == 1 && oGLB.process_boosted == 1)
        boost_priority(0);
}

// Boost process priority (set_priority: 1=high, 0=low)
void boost_priority(int set_priority)
{   
    if(set_priority)
        app_dbg3(THIS_FILE, "Set high priority");
    else
        app_dbg3(THIS_FILE, "Set low priority");

    #if DTMFBOX_WIN32
        if(set_priority)
        {
            // higher priority
            SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
            SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);
            oGLB.process_boosted = 1;
        }
        else {
            // lower priority
            SetPriorityClass(GetCurrentProcess(), NORMAL_PRIORITY_CLASS);
            SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
            oGLB.process_boosted = 0;
        }
    #else        
        if(set_priority)
        {
            // higher priority
            setpriority(PRIO_PROCESS, 0, -10);
            oGLB.process_boosted = 1;
        }
        else {
            // lower priority
            setpriority(PRIO_PROCESS, 0, +10);
            oGLB.process_boosted = 0;
        }

    #endif
}

