/*********************************************************************************
 * mixer.h ($Revision: 250 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/

//#define SILENCE_THRESHOLD   75    // 16bit signed
#define SILENCE_THRESHOLD   30      // 16bit signed
#define TALKOVER_VOLUME     10      // 0..128 = 0%-100%

#define F_BIAS 12                   // 2^12 
#define VOL_EXPONENT 3.0

#define PLID_REC_TX     INCONF_PORTS+0
#define PLID_CON_RX     INCONF_PORTS+1
#define PLID_CON_TX     INCONF_PORTS+2

typedef struct 
{
    pjmedia_port             *mother_port;

    int                      notalkover_rx_volume;    // 0 if not in talkover, value otherwise
    int 		             target_tx_volume;
    int 		             target_rx_volume;
    long 		             fade_tx_frames_remain;
    long 		             fade_rx_frames_remain;
    //int                      talkover_active;  // in talkover port: indicates if vol > SILENCE_THRESHOLD
                                              // in background port: indicates if talkover is active (i.e. volume down)
    int                      talkover_downfade;
    int                      talkover_fade_active;  // only for background port

    int			             infade;
    int                      locked;

    long                     tx_volume;    // current tx_volume
    long                     rx_volume;    // current rx_volume

} mixer_port_data;

void script_mixer_control(int play_id, int rec_id);

long vol_scale(int vol);
long vol_lin2log(long vol);
long vol_log2lin(long vol);

int mixer_check_volume(int vol);
void mixer_convert_ms_to_frames(mixer_port_data *mixdata, long *msfr);
int mixer_init_fade(mixer_port_data *mixdata, int target_vol, int fadetime, int is_rx);
void recompute_talkover_status(int con_id);
void set_talkover_active(int con_id, int play_id, int value);

void do_mixing(mixer_port_data *mixdata, pjmedia_frame *frame, int con_id, int play_id, int rec_id);

pj_status_t mixer_port_get_frame(pjmedia_port *port, pjmedia_frame *frame, int con_id, int play_id, int dn_status);
pj_status_t mixer_port_put_frame(pjmedia_port *port, pjmedia_frame *frame, int con_id, int rec_id);


