/*********************************************************************************
 * options.c ($Revision: 265 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "options.c"

// Parameter parser (first instance)
int param_parser(int argc, char *arg[])
{
    FILE *fp;
    int i;
    
    oGLB.daemonized = 0;
    oGLB.logfile[0] = '\0';
    oGLB.configfile[0] = '\0';

    for(i=1; i<argc; i++)
    {
        // daemonize application?
        if(!strcmp(arg[i], "-daemon"))
            oGLB.daemonized = 1;

        // start application?
        if(!strcmp(arg[i], "-start"))
            oGLB.foreground = 1;

        // logfile?
        if(!strcmp(arg[i], "-log"))
            if(i!=argc-1)
                strncpy(oGLB.logfile, arg[i+1], sizeof(oGLB.logfile));

        // config-file?
        if(!strcmp(arg[i], "-cfg"))
            if(i!=argc-1) 
                strncpy(oGLB.configfile, arg[i+1], sizeof(oGLB.configfile));
            
        // show help?
        if(!strcmp(arg[i], "-?") || !strcmp(arg[i], "--?") || !strcmp(arg[i], "-help") || !strcmp(arg[i], "--help"))
        {
            print_help();
            return -1;
        }
    }
    
    if(strlen(oGLB.configfile)) 
    {
        // Check if given configuration file exists
        fp = fopen(oGLB.configfile, "r");
        if(fp == NULL) {            
            memset(oGLB.configfile, 0, sizeof(oGLB.configfile));
        } else {
            fclose(fp);
        }
    }

    if(!strlen(oGLB.configfile))    
    {
        // Search for configuration                
        #if (DTMFBOX_WIN32 == 1)
        
        // workdir/dtmfbox.cfg?
        snprintf(oGLB.configfile, sizeof(oGLB.configfile), "%s\\%s", oGLB.workdir, DTMFBOX_CFGFILE);
        fp = fopen(oGLB.configfile, "r");
        if(fp != NULL) goto found;
        
        #else

        // /etc/dtmfbox.cfg ?
        snprintf(oGLB.configfile, sizeof(oGLB.configfile), "/etc/%s", DTMFBOX_CFGFILE);
        fp = fopen(oGLB.configfile, "r");
        if(fp != NULL) goto found;

        // workdir/dtmfbox.cfg ?
        snprintf(oGLB.configfile, sizeof(oGLB.configfile), "%s/%s", oGLB.workdir, DTMFBOX_CFGFILE);
        fp = fopen(oGLB.configfile, "r");
        if(fp != NULL) goto found;

        // ./dtmfbox.cfg ?
        snprintf(oGLB.configfile, sizeof(oGLB.configfile), "./%s", DTMFBOX_CFGFILE);
        fp = fopen(oGLB.configfile, "r");
        if(fp != NULL) goto found;

        #endif

        found:
        if(fp != NULL) 
            fclose(fp);
        else {
            //printf("Error: Unable to find configuration!\n");
            //exit(1);
        }

    }

    return 0;
}

// Parameter parser (second instance / plugins)
int param_parser_plugin(int argc, char *arg[], int pid)
{
    int i;
    int max_args = argc;
    if(max_args > 256) max_args = 256;
    
    PJ_UNUSED_ARG(pid);

    // Fill info for event
    sh_actions->parse_plugin = 1;
    sh_actions->argc = max_args;
    for(i=0;i<max_args; i++) 
        snprintf(sh_actions->args[i], sizeof(sh_actions->args[i]), "%s", arg[i]);        

    // raise event to call plugin commands
    sh_actions->queue.ready = 1;
    while(sh_actions->queue.ready == 1) usleep(1);
    i = sh_actions->parse_plugin;

    return i;
}


// Parameter parser (second instance / control daemon)
int param_parser_daemon(int argc, char *arg[], int pid)
{    
    int play_thread = 0;
    int my_queue = 0;
    int con = -1;    
    int waitevent = 0;
    int pipe_id = -1;    
    int pipe_id_global = -1;
    int play_queue = 0;    
    int ret = -1;
    int plugin_executed_cmd = 0;

    if(argc <= 1)
    {
        // show daemon-help and plugin-help
        print_help_daemon();
        return -1;        
    }
        
    // Allocate shared memory (sh_actions) - Windows:
    if(open_shared_memory())
        exit(1);
    else
        sh_actions = sh_actions_shmem;

    //if(pid != getpid())
    {
        // Get a queue number and increment queue        
        my_queue = sh_actions->queue.max_queue;
        queue_counter(+1);

        // pending actions? Then wait ...
        while(sh_actions->queue.in_use == 1 && sh_actions->queue.cur_queue != my_queue)
            usleep(1);

        // ok, it's our turn...
        sh_actions->queue.in_use = 1;
        sh_actions->queue.cur_pid = getpid();        
    }

    // First, parse the command line to the loaded plugins.
    if(pid != getpid())
        plugin_executed_cmd = param_parser_plugin(argc, arg, pid);

    // After that, parse the default commands.
    // These are the integrated commands, like -call, -play, -hook, etc.
    if(param_parser_daemon_options(argc, arg, &play_thread, &pipe_id))
        return 0;

    // No valid arguments (plugin/integrated)? Then show help!
    //
    if(pid != getpid() && (plugin_executed_cmd <= 0 && !sh_actions->valid_commands))
    {   
        // show daemon-help and plugin-help
        print_help_daemon();
        
        // next command ...
        queue_counter(-1);        
        sh_actions->queue.in_use = 0;    
        close_shared_memory(0);
        return -1;
    }

    // Plugin already executed the command? Then return!
    // Otherwise the console is used...
    if(plugin_executed_cmd > 0)
    {                   
        queue_counter(-1); // TEST
        sh_actions->parse_plugin = 0;
        sh_actions->queue.in_use = 0;                    
        return -1;
    }

    // Create global pipe (output to second instance)
    if(pid != getpid())
    {
        char name[32];
        snprintf(name, sizeof(name), "dtmfbox_out_%d", sh_actions->queue.cur_pid);
        pipe_id_global = pipe_create(name, 0, 1, 0);
    }

    // initialize return values and save some infos
    sh_actions->retval.play_id = -1;
    con = sh_actions->connection;    
    waitevent = sh_actions->script.waitevent;

    // When we call param_parser_daemon from a plugin, we can call script_do_actions() directly!
    if(pid == getpid()) {
        ret = script_do_actions();        
                        
        reset_actions();  
        queue_counter(-1);
        sh_actions->queue.in_use = 0;
        return ret;
    }

    // This will execute the commands in the shared memory variable 'sh_actions'
    // give thread some time to run script_do_actions() ...    
    sh_actions->queue.ready = 1;
    while(sh_actions->queue.ready == 1 ) usleep(1);

    if(con != -1 && !play_thread)
    {   
        // Get playing queue
        play_queue = sh_actions->retval.play_id;

        // next command...
        queue_counter(-1);
        sh_actions->queue.in_use = 0;

        // Wait until playing ends (-play)?
        if(play_queue != -1)
        {
            while(sh_actions->play_rec.play_thread[con][play_queue] != -1) //&& playing == 1)
                usleep(1);                         
        }

        // block till next event/endless?
        if(waitevent != 0 && sh_actions->con_info[con].status > 0)
        {
            char buf[2048];
            int len = 1;            
            while(len > 0)
            {
                len = pipe_read(pipe_id, buf, sizeof(buf));
                if(len > 0) 
                {                    
                    printf("%s", buf);
                    fflush(stdout);
                     
                    if(waitevent == 1)
                        break;
                } 
                else 
                {                    
                    if(waitevent == 2 && (sh_actions->con_info[con].status == 3 || sh_actions->con_info[con].status == 0))
                        break;
                    else
                        len=1;
                }
            }

            pipe_close(pipe_id);
        }
    } 
    else 
    {
        // next command...
        queue_counter(-1);
        sh_actions->queue.in_use = 0;
    }

    // RETURN VALUES (show values to console output):
    if(pipe_id_global != -1)
    {
        char buf[2048];
        int len = 1;

        while(len > 0) 
        {
            pj_memset(buf, 0, sizeof(buf));
            len = pipe_read(pipe_id_global, buf, sizeof(buf));

            if(len >= 0) {
                printf("%s", buf);
                fflush(stdout);

                // return play_id, connection_id, etc...
                // so output must be numerical
                if(ret == -1 && len <= 3) {
                    if(isnumeric(buf))
                        ret = atoi(buf);
                    else
                        ret = -2;
                }                                
            }
            
        }
        
        pipe_close(pipe_id_global);
    }

    // release sh_actions 
    close_shared_memory(0);
    return ret;
}


int param_parser_daemon_options(int argc, char *arg[], int *play_thread, int *pipe_id)
{
    char options[MAX_SUBOPTIONS][MAX_SUBOPTIONS_LENGTH];
    char sub_values[MAX_SUBOPTIONS][MAX_SUBOPTIONS_LENGTH];
    char sub_name[MAX_SUBOPTIONS_LENGTH];    
    int option_name, option_type;    
    int pos = 1;
    int sub_pos = 0;
    int len = 0;
    int sub_len = 0;
    int i;
    int id = -1;

    // Validate Connection-ID
    //
    // If -1, a new connection should be created (connected with none / default)
    // If >-1, a new call should be connected with another connection.    
    if(isnumeric(arg[1]))
    {
        id = atoi(arg[1]);
        if(strcmp(arg[1], "0") && id == 0)
            id = -1;        
    } 
    sh_actions->connection = id;

    //
    // Get values from command line
    //
    while(len != -1)
    {        
        char **opts = (char**)&options;
        char **vals = (char**)&sub_values;

        // get main option (with -)
        len = get_arg_opt(argc, arg, &pos, &option_name, &option_type, opts);
        sub_len = 0;
        sub_pos = 0;
        
        switch(option_name)
        {
            // ---------------------------------------------
            // -con
            // ---------------------------------------------
            case OPT_CON:
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                if(sub_len != -1) {
                    if(isnumeric(sub_values[0]))
                        sh_actions->connection = atoi(sub_values[0]);   
                }
                break;

            // ---------------------------------------------
            // -play
            // ---------------------------------------------
            case OPT_PLAY:
                // -playthread, -playstreamthread ? set thread on!
                if(option_type == 1 || option_type == 3)
                    *play_thread = -1;
                
                while(sub_len != -1)
                {
                    sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                    if(sub_len != -1)
                    {                        
                        // "", file=
                        if(!strcmp(sub_name, "") || !strcmp(sub_name, "file"))
                        {   
                            // help for -play
                            if(!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?"))
                            {
                                print_usage_play();
                                goto usage_end;
                            }
                            else
                            {                                
                                strncpy(sh_actions->play_rec.play_file, sub_values[0], sizeof(sh_actions->play_rec.play_file));

                                // WAVE as default
                                sh_actions->play_rec.wave = 1;
                                sh_actions->play_rec.raw = 0;                                
                                sh_actions->play_rec.speex = 0;
                                sh_actions->play_rec.rtp = 0;
                            }

                            // -playstream, -playstreamthread ? set clockrate!
                            if(option_type == 2 || option_type == 3 && sub_len == 2) {
                                if(isnumeric(sub_values[1]))
                                    sh_actions->play_rec.stream_clockrate = atoi(sub_values[1]);
                                else
                                    sh_actions->play_rec.stream_clockrate = 8000;

                                 // RAW
                                 sh_actions->play_rec.raw = 1;
                                 sh_actions->play_rec.speex = 0;                                 
                                 sh_actions->play_rec.wave = 0;                                    
                                 sh_actions->play_rec.rtp = 0;
                            }                                                        
                        }
                        // hz=
                        if(!strcmp(sub_name, "hz"))
                        {
                            if(isnumeric(sub_values[0]))
                                sh_actions->play_rec.stream_clockrate = atoi(sub_values[0]);
                            else
                                sh_actions->play_rec.stream_clockrate = 8000;

                            if(sh_actions->play_rec.stream_clockrate <= 0 || sh_actions->play_rec.stream_clockrate > 44100)
                                sh_actions->play_rec.stream_clockrate = 8000;

                             // RAW
                             sh_actions->play_rec.raw = 1;
                             sh_actions->play_rec.speex = 0;                                 
                             sh_actions->play_rec.wave = 0;                                    
                             sh_actions->play_rec.rtp = 0;
                        }                        
                        // mode=
                        if(!strcmp(sub_name, "mode"))
                        {
                            for(i=0; i<sub_len; i++)
                            {
                                if(!strcmp(sub_values[i], "thread"))
                                    *play_thread = -1;
                                
                                if(!strcmp(sub_values[i], "stream") || !strcmp(sub_values[i], "raw"))
                                {
                                    if(sh_actions->play_rec.stream_clockrate == 0)
                                        sh_actions->play_rec.stream_clockrate = 8000;

                                    // RAW
                                    sh_actions->play_rec.raw = 1;
                                    sh_actions->play_rec.wave = 0;
                                    sh_actions->play_rec.speex = 0;
                                    sh_actions->play_rec.rtp = 0;
                                } 

                                if(!strcmp(sub_values[i], "speex")) 
                                {
                                    // SPEEX
                                    sh_actions->play_rec.speex = 1;
                                    sh_actions->play_rec.raw = 0;
                                    sh_actions->play_rec.wave = 0;                                    
                                    sh_actions->play_rec.rtp = 0;
                                }

                                if(!strcmp(sub_values[i], "rtp")) 
                                {
                                    // RTP/RTCP
                                    sh_actions->play_rec.rtp = 1;
                                    sh_actions->play_rec.speex = 0;
                                    sh_actions->play_rec.raw = 0;
                                    sh_actions->play_rec.wave = 0;
                                }
                            }
                        }
                        
                        // rtp=
                        if(!strcmp(sub_name, "rtp") || strcmp(sub_name, "rtp_port"))
                        {
                            if(isnumeric(sub_values[0]))
                                sh_actions->play_rec.rtp_port = atoi(sub_values[0]);
                        }

                        // bufsize=
                        if(!strcmp(sub_name, "buffersize") || !strcmp(sub_name, "bufsize"))
                        {
                            if(isnumeric(sub_values[0]))
                                sh_actions->play_rec.stream_buffersize = atol(sub_values[0]);
                            else
                                sh_actions->play_rec.stream_buffersize = 0;
                        }

                        // wait_start=
                        if(!strcmp(sub_name, "wait_start") || !strcmp(sub_name, "waitstart"))
                        {
                            if(isnumeric(sub_values[0]))
                                sh_actions->play_rec.wait_start = atoi(sub_values[0]);
                            else
                                if(!strcmp(sub_values[0], "endless"))
                                    sh_actions->play_rec.wait_start = -1;
                                else
                                    sh_actions->play_rec.wait_start = 0;
                        }

                        // wait_end=
                        if(!strcmp(sub_name, "wait_end") || !strcmp(sub_name, "waitend"))
                        {
                            if(isnumeric(sub_values[0]))
                                sh_actions->play_rec.wait_end = atoi(sub_values[0]);
                            else
                                if(!strcmp(sub_values[0], "endless"))
                                    sh_actions->play_rec.wait_end = -1;
                                else
                                    sh_actions->play_rec.wait_end = 0;
                        }

                        // timeout= (both: waitstart= and waitend=)
                        if(!strcmp(sub_name, "wait") || !strcmp(sub_name, "timeout"))
                        {
                            if(isnumeric(sub_values[0])) {
                                sh_actions->play_rec.wait_start = atoi(sub_values[0]);
                                sh_actions->play_rec.wait_end = atoi(sub_values[0]);
                            }
                            else {
                                if(!strcmp(sub_values[0], "endless")) {
                                    sh_actions->play_rec.wait_start = -1;
                                    sh_actions->play_rec.wait_end  = -1;
                                }
                                else {
                                    sh_actions->play_rec.wait_start = 0;
                                    sh_actions->play_rec.wait_end = 0;

                                }
                            }
                        }
                    }
                }                
                break;

                
            // ---------------------------------------------
            // -record
            // ---------------------------------------------
            case OPT_RECORD:
                while(sub_len != -1)
                {
                    sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                    if(sub_len != -1)
                    {
                        // "", file=
                        if(!strcmp(sub_name, "") || !strcmp(sub_name, "file")) 
                        {
                            // help for -record
                            if(!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?"))
                            {
                                print_usage_record();
                                goto usage_end;
                            }
                            else
                            {
                                strncpy(sh_actions->play_rec.rec_file, sub_values[0], sizeof(sh_actions->play_rec.rec_file));
                                
                                // WAVE as default
                                sh_actions->play_rec.wave = 1;
                                sh_actions->play_rec.raw = 0;                                
                                sh_actions->play_rec.speex = 0;
                                sh_actions->play_rec.rtp = 0;
                            }
                        
                            // -recordstream ? set clockrate!
                            if(option_type == 1 && sub_len == 2) {
                                if(isnumeric(sub_values[1]))
                                    sh_actions->play_rec.stream_clockrate = atoi(sub_values[1]);
                                else
                                    sh_actions->play_rec.stream_clockrate = 8000;

                                // RAW
                                sh_actions->play_rec.raw = 1;
                                sh_actions->play_rec.wave = 0;                                
                                sh_actions->play_rec.speex = 0;
                                sh_actions->play_rec.rtp = 0;
                            }
                        }

                        if(!strcmp(sub_name, "hz")) 
                        {                        
                            if(isnumeric(sub_values[0])) 
                                sh_actions->play_rec.stream_clockrate = atoi(sub_values[0]);
                            else
                                sh_actions->play_rec.stream_clockrate = 8000;

                            if(sh_actions->play_rec.stream_clockrate <= 0 || sh_actions->play_rec.stream_clockrate > 44100)
                                sh_actions->play_rec.stream_clockrate = 8000;

                            // RAW
                            sh_actions->play_rec.raw = 1;
                            sh_actions->play_rec.wave = 0;                                
                            sh_actions->play_rec.speex = 0;
                            sh_actions->play_rec.rtp = 0;
                        }
                        
                        // rtp=
                        if(!strcmp(sub_name, "rtp") || strcmp(sub_name, "rtp_port"))
                        {
                            if(isnumeric(sub_values[0]))
                                sh_actions->play_rec.rtp_port = atoi(sub_values[0]);
                        }

                        if(!strcmp(sub_name, "mode"))
                        {
                            for(i=0; i<sub_len; i++)
                            {
                                if(!strcmp(sub_values[i], "stream") || !strcmp(sub_values[i], "raw"))
                                {
                                    if(sh_actions->play_rec.stream_clockrate == 0)
                                        sh_actions->play_rec.stream_clockrate = 8000;

                                    // RAW
                                    sh_actions->play_rec.raw = 1;
                                    sh_actions->play_rec.wave = 0;                                
                                    sh_actions->play_rec.speex = 0;
                                    sh_actions->play_rec.rtp = 0;
                                }      

                                if(!strcmp(sub_values[i], "speex")) 
                                {                                    
                                    // SPEEX
                                    sh_actions->play_rec.speex = 1;
                                    sh_actions->play_rec.raw = 0;
                                    sh_actions->play_rec.wave = 0;
                                    sh_actions->play_rec.rtp = 0;
                                }

                                if(!strcmp(sub_values[i], "rtp")) 
                                {
                                    // RTP/RTCP
                                    sh_actions->play_rec.rtp = 1;
                                    sh_actions->play_rec.speex = 0;
                                    sh_actions->play_rec.raw = 0;
                                    sh_actions->play_rec.wave = 0;
                                }
                            }                        
                        }
                    }
                }
                break;

            // ---------------------------------------------
            // -mixer
            // ---------------------------------------------
            case OPT_MIXER:

                // defaults:
                sh_actions->mixer.fade_duration = 1000;

                while(sub_len != -1)
                {   
                    sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                    if(sub_len != -1)
                    {
                        if (!strcmp(sub_name, ""))
                        {
                            if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                            {
                                print_usage_mixer();
                                goto usage_end;
                            }
                        
                            if (isnumeric(sub_values[0]) && sub_len >= 1) 
                            {
                               // play_id has been specified                              
                               sh_actions->mixer.play_id = atoi(sub_values[0]);                               
                            } 
                            else 
                            {                                 
                               // rx, set incoming volume for connection
                               if(!strcmp(sub_values[0], "rx") || !strcmp(sub_values[0], "incoming")) 
                                  sh_actions->mixer.play_id = PLID_CON_RX;
                               
                               // tx, set outgoing volume for connection
                               else if(!strcmp(sub_values[0], "tx") || !strcmp(sub_values[0], "outgoing")) 
                                  sh_actions->mixer.rec_id = PLID_CON_TX;
                            }
                        }

                        if(!strcmp(sub_name, "volume") || !strcmp(sub_name, "vol"))
                        {
                            sh_actions->mixer.options = sh_actions->mixer.options | OPT_MIX_VOLUME;
                            if(isnumeric(sub_values[0]))
                                sh_actions->mixer.volume = atoi(sub_values[0]);
                        }

                        if(!strcmp(sub_name, "fade"))
                        {
                            sh_actions->mixer.options = sh_actions->mixer.options | OPT_MIX_FADE;
                            if(isnumeric(sub_values[0]))
                                sh_actions->mixer.fade_volume = atoi(sub_values[0]);
                        }

                        if(!strcmp(sub_name, "duration") || !strcmp(sub_name, "fadetime"))
                        {
                            sh_actions->mixer.options = sh_actions->mixer.options | OPT_MIX_FADE_DURATION;
                            if(isnumeric(sub_values[0]))
                                sh_actions->mixer.fade_duration = atoi(sub_values[0]);
                        }

                        if(!strcmp(sub_name, "background") || !strcmp(sub_name, "bg"))
                        {
                            if(!strcmp(sub_values[0], "on"))
                                sh_actions->mixer.options = sh_actions->mixer.options | OPT_MIX_BACKGROUND_ON;
                            else
                                sh_actions->mixer.options = sh_actions->mixer.options | OPT_MIX_BACKGROUND_OFF;
                        }

                        if(!strcmp(sub_name, "talkover"))
                        {
                            if(!strcmp(sub_values[0], "on"))
                                sh_actions->mixer.options = sh_actions->mixer.options | OPT_MIX_TALKOVER_ON;
                            else
                                sh_actions->mixer.options = sh_actions->mixer.options | OPT_MIX_TALKOVER_OFF;
                        }

                        // for connection port: rx or tx dir.
                        if(!strcmp(sub_name, "direction") || !strcmp(sub_name, "dir"))
                        {
                            if (sh_actions->mixer.play_id == -1) {
                                if(!strcmp(sub_values[0], "rx") || !strcmp(sub_values[0], "incoming"))
                                    sh_actions->mixer.play_id = PLID_CON_RX;
                                if(!strcmp(sub_values[0], "tx") || !strcmp(sub_values[0], "outgoing"))
                                    sh_actions->mixer.rec_id = PLID_CON_TX;
                            }
                        }
                    }
                }
                break;

            // ---------------------------------------------
            // -tone
            // ---------------------------------------------
            case OPT_TONE:                
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                if(sub_len != -1)
                {
                    if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                    {
                        print_usage_tone();
                        goto usage_end;
                    }
                    else
                    {                        
                        sh_actions->tone.active = 1;
                        sh_actions->tone.volume = 16384;    // default volume

                        if(strlen(sub_values[0]) && isnumeric(sub_values[0]))
                            sh_actions->tone.freq1 = (short)atol(sub_values[0]);
                        if(strlen(sub_values[1]) && isnumeric(sub_values[1]))
                            sh_actions->tone.freq2 = (short)atol(sub_values[1]);
                        if(strlen(sub_values[2]) && isnumeric(sub_values[2]))
                            sh_actions->tone.on_msec = (short)atol(sub_values[2]);
                        if(strlen(sub_values[3]) && isnumeric(sub_values[3]))
                            sh_actions->tone.off_msec = (short)atol(sub_values[3]);
                        if(strlen(sub_values[4]) && isnumeric(sub_values[4]))
                            sh_actions->tone.volume = (short)atol(sub_values[4]);
                    }
                }
                break;

            // ---------------------------------------------
            // -stop
            // ---------------------------------------------
            case OPT_STOP:
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                if(sub_len != -1)
                {
                    if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                    {
                        print_usage_stop();
                        goto usage_end;
                    }
                    else
                    {
                        if(!strcmp(sub_values[0], "play"))
                        {
                            sh_actions->stop.option = 1;
                            sh_actions->stop.play_id = -1;
                                                    
                            // stop a specified stream?                        
                            if (isnumeric(sub_values[1])) 
                               sh_actions->stop.play_id = atoi(sub_values[1]);

                            // stop all streams?
                            if(!strcmp(sub_values[1], "all"))
                               sh_actions->stop.play_id = -2;
                        }

                        if(!strcmp(sub_values[0], "record"))                    
                            sh_actions->stop.option = 2;
                        
                        if(!strcmp(sub_values[0], "tone"))                    
                            sh_actions->stop.option = 4;

                        if(!strcmp(sub_values[0], "daemon")) {
                            printf("Stop daemon!\n");                    
                            sh_actions->stop.option = 3;
                        }     
                        if(!strcmp(sub_values[0], "timer")) {
                            sh_actions->timer.timeout = -1;

                            // stop a specified timer?                        
                            if (isnumeric(sub_values[1])) 
                               strncpy(sh_actions->timer.optionstring, sub_values[1], sizeof(sh_actions->timer.optionstring));
                            else 
                               memset(sh_actions->timer.optionstring, 1, sizeof(sh_actions->timer.optionstring));
                               //pj_bzero(sh_actions->timer.optionstring, sizeof(sh_actions->timer.optionstring));
                        }
                        if(!strcmp(sub_values[0], "waitevent")) {
                            sh_actions->stop.option = 5;
                        }     
                    }
                }
                break;

            // ---------------------------------------------
            // -scriptfile
            // ---------------------------------------------
            case OPT_SCRIPTFILE:
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                if(sub_len != -1) {
                    if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                    {
                        print_usage_scriptfile();
                        goto usage_end;
                    }
                    else
                        strncpy(sh_actions->script.file, sub_values[0], sizeof(sh_actions->script.file));
                }
                break;

            // ---------------------------------------------
            // -bridge
            // ---------------------------------------------
            case OPT_BRIDGE:
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                if(sub_len != -1)
                {
                    if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                    {
                        print_usage_bridge();
                        goto usage_end;
                    }
                    else
                    {
                        if(!strcmp(sub_name, "on"))                    
                          if(isnumeric(sub_values[0]))
                              sh_actions->bridge.connect = atoi(sub_values[0]);                    

                        if(!strcmp(sub_name, "off"))                    
                          if(isnumeric(sub_values[0]))
                              sh_actions->bridge.disconnect = atoi(sub_values[0]);                    
                       }
                }
                break;

            // ---------------------------------------------
            // -delimiter
            // ---------------------------------------------
            case OPT_DELIMITER:
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                if(sub_len != -1)
                {
                    if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                    {
                        print_usage_delimiter();
                        goto usage_end;
                    }
                    else
                    {
                        if(!strcmp(sub_values[0], "poundkey"))
                            sh_actions->script.delimiter[0] = '#';
                        else if(!strcmp(sub_values[0], "asterisk"))
                            sh_actions->script.delimiter[0] = '*';
                        else if(!strcmp(sub_values[0], "none"))
                            sh_actions->script.delimiter[0] = '_';
                        else if(!strcmp(sub_values[0], ""))
                            sh_actions->script.delimiter[0] = '_';
                        else
                            strncpy(sh_actions->script.delimiter, sub_values[0], 1);
                       }
                }
                break;

            // ---------------------------------------------
            // -waitevent
            // ---------------------------------------------
            case OPT_WAITEVENT:
                {                    
                    sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                    if(sub_len != -1)
                    {
                        if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                        {
                            print_usage_waitevent();
                            goto usage_end;
                        }
                        else
                        {
                            if(!strcmp(sub_name, "mode")) {
                                if(!strcmp(sub_values[0], "endless") || !strcmp(sub_values[0], "more"))
                                    sh_actions->script.waitevent = 2;          
                                else
                                    sh_actions->script.waitevent = 1;      
                            }
                        }
                    } else {
                        sh_actions->script.waitevent = 1;
                    }
                                                                                
                    if(*pipe_id == -1 && sh_actions->script.waitevent > 0) 
                    {
                        char name[32];
                        snprintf(name, sizeof(name), "dtmfbox_waitevent_%d", sh_actions->queue.cur_pid);
                        *pipe_id = pipe_create(name, 0, 1, 0);
                        if(*pipe_id == -1)
                            sh_actions->script.waitevent = 0;
                    }
                }
                break;

            // ---------------------------------------------
            // -dtmf
            // ---------------------------------------------
            case OPT_DTMF:
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                if(sub_len != -1)
                {
                    if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                    {
                        print_usage_dtmf();
                        goto usage_end;
                    }
                    else
                    {
                        if(!strcmp(sub_values[0], "poundkey"))
                            sh_actions->dtmf[0] = '#';
                        else if(!strcmp(sub_values[0], "asterisk"))
                            sh_actions->dtmf[0] = '*';
                        else if(!strcmp(sub_values[0], "none"))
                            sh_actions->dtmf[0] = '_';
                        else if(!strcmp(sub_values[0], ""))
                            sh_actions->dtmf[0] = '_';
                        else
                            strncpy(sh_actions->dtmf, sub_values[0], 1);            
                     }
                }
                break;

            // ---------------------------------------------
            // -text
            // ---------------------------------------------
            case OPT_TEXT:
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                if(sub_len != -1)
                {
                    if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                    {
                        //print_usage_dtmf();
                        goto usage_end;
                    }
                    else
                    {
                        int max_len = 31;
                        if(strlen(sub_values[0]) < 31)
                            max_len = strlen(sub_values[0]);

                        pj_memset(sh_actions->text, 0, sizeof(sh_actions->text));
                        strncpy(sh_actions->text, sub_values[0], max_len);            
                    }
                }
                break;

            // ---------------------------------------------
            // -hook
            // ---------------------------------------------
            case OPT_HOOK:
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                if(sub_len != -1)
                {
                    if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                    {
                        print_usage_hook();
                        goto usage_end;
                    }
                    else
                    {
                        if(!strcmp(sub_values[0], "up") || !strcmp(sub_values[0], "on"))
                            sh_actions->hook = 1;
                        if(!strcmp(sub_values[0], "down") || !strcmp(sub_values[0], "off"))
                            sh_actions->hook = 2;
                        if(!strcmp(sub_values[0], "hold"))
                            sh_actions->hook = 3;
                        if(!strcmp(sub_values[0], "unhold"))
                            sh_actions->hook = 4;                        
                        if(!strcmp(sub_values[0], "reject"))
                            sh_actions->hook = 5;
                        if(!strcmp(sub_values[0], "alert"))
                            sh_actions->hook = 6;
                        if(!strcmp(sub_values[0], "auto"))
                            sh_actions->hook = 7;
                        if(!strcmp(sub_values[0], "manual"))
                            sh_actions->hook = 8;
                    }
                }
                break;

            // ---------------------------------------------
            // -call
            // ---------------------------------------------
            case OPT_CALL: 
                while(sub_len != -1)
                {
                    sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                    if(sub_len != -1)
                    {
                        if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                        {
                            print_usage_call();
                            goto usage_end;
                        }
                        else
                        {
                            if(!strcmp(sub_name, ""))
                            {
                                if(sub_len >= 1)                           
                                    strncpy(sh_actions->call.number, sub_values[0], sizeof(sh_actions->call.number));
                                if(sub_len >= 2)
                                    strncpy(sh_actions->call.target, sub_values[1], sizeof(sh_actions->call.target));
                                if(sub_len >= 3)
                                    if(isnumeric(sub_values[2]))
                                        sh_actions->call.capi_controller = atoi(sub_values[2]);                            
                            }

                            // src=
                            if(!strcmp(sub_name, "src") || !strcmp(sub_name, "source") || !strcmp(sub_name, "from"))
                                strncpy(sh_actions->call.number, sub_values[0], sizeof(sh_actions->call.number));
                            
                            // dst=
                            if(!strcmp(sub_name, "dst") || !strcmp(sub_name, "destination") || !strcmp(sub_name, "to"))
                                strncpy(sh_actions->call.target, sub_values[0], sizeof(sh_actions->call.target));

                            // ctrl=
                            if(!strcmp(sub_name, "ctrl") || !strcmp(sub_name, "controller"))
                                if(isnumeric(sub_values[0]))
                                    sh_actions->call.capi_controller = atoi(sub_values[0]);

                            // mode=
                            if(!strcmp(sub_name, "mode"))
                            {
                                for(i=0; i<sub_len; i++)
                                {
                                    if(!strcmp(sub_values[i], "anonymous"))
                                        sh_actions->call.anonymous = 1;
                                }
                            }
                        }
                    }
                }
                break;

            // ---------------------------------------------
            // -list
            // ---------------------------------------------
            case OPT_LIST: 
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                if(sub_len != -1)
                {
                    if(!strcmp(sub_values[0], "all"))                    
                        sh_actions->list = 2;

                    if(!strcmp(sub_values[0], "clients"))                    
                        sh_actions->list = 3;

                    if(!strcmp(sub_values[0], "accounts"))                    
                        sh_actions->list = 4;                    

                    if(!strcmp(sub_values[0], "plugins"))                    
                        sh_actions->list = 5;                    
                }
                else
                    sh_actions->list = 1;                                    

                break;

            // ---------------------------------------------
            // -timer
            // ---------------------------------------------
            case OPT_TIMER: 
                while(sub_len != -1)
                {
                    sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 1);
                    if(sub_len != -1)
                    {                       
                        if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                        {
                            print_usage_timer();
                            goto usage_end;
                        }
                        else
                        {           
                            if(!strcmp(sub_name, ""))
                            {
                                if(isnumeric(sub_values[0]))
                                    sh_actions->timer.timeout = atol(sub_values[0]);

                                if(sub_len >= 1)
                                    strncpy(sh_actions->timer.optionstring, sub_values[1], sizeof(sh_actions->timer.optionstring) - 2);
                            }
                            
                            if(!strcmp(sub_name, "mode"))
                            {                            
                                if(!strcmp(sub_values[0], "msec"))
                                    sh_actions->timer.mode = 1;
                            }                            
                        }                        
                    }
                }
                break;

            // ---------------------------------------------
            // -env
            // ---------------------------------------------
            case OPT_ENV: 
                sub_len = get_sub_opt(len, opts, &sub_pos, sub_name, vals, 0);
                
                if(sub_len != -1)
                {
                    if (!strcmp(sub_values[0], "help") || !strcmp(sub_values[0], "?")) 
                    {
                        print_usage_environment();
                        goto usage_end;
                    }
                    
                    if(strlen(sub_name) == 0 && sub_len == 1)
                    {
                        strncpy(sh_actions->script.env_key, sub_values[0], sizeof(sh_actions->script.env_key));
                        snprintf(sh_actions->script.env_val, sizeof(sh_actions->script.env_val), "%s", DEVNULL);
                    }

                    if(strlen(sub_name) && sub_len > 0)
                    {                                                                       
                        strncpy(sh_actions->script.env_key, sub_name, sizeof(sh_actions->script.env_key));
                        snprintf(sh_actions->script.env_val, sizeof(sh_actions->script.env_val), "%s", sub_values[0]);
                    }                                 
                }              
                break;

            // ---------------------------------------------
            // -version
            // ---------------------------------------------
            case OPT_VERSION: 
                print_version();
                goto usage_end;
                break;

        }        
    }


    // Has valid commands, ready to go?
    sh_actions->valid_commands = !(sh_actions->list == 0 &&
                                   sh_actions->hook == 0 &&
                                   sh_actions->stop.option == 0 &&
                                   sh_actions->mixer.options == 0 &&
                                   sh_actions->mixer.play_id == -1 &&
                                   sh_actions->timer.timeout == 0 &&
                                   sh_actions->bridge.connect == -1 &&
                                   sh_actions->bridge.disconnect == -1 &&
                                   sh_actions->tone.active == 0 &&       
                                   sh_actions->script.waitevent == 0 &&    
                                   !strcmp(sh_actions->script.env_key, DEVNULL) && 
                                   !strcmp(sh_actions->play_rec.rec_file, DEVNULL) &&
                                   !strcmp(sh_actions->play_rec.play_file, DEVNULL) &&
                                   !strcmp(sh_actions->call.target, DEVNULL) &&
                                   !strcmp(sh_actions->script.file, DEVNULL) &&
                                   !strcmp(sh_actions->script.delimiter, "-") &&
                                   !strcmp(sh_actions->text, DEVNULL) &&
                                   !strcmp(sh_actions->dtmf, DEVNULL));

    return 0;

usage_end:        
    queue_counter(-1);
    reset_actions();
    sh_actions->queue.in_use = 0;
    close_shared_memory(0);
    return 1;
}

void queue_counter(int increment)
{   
    sh_actions->queue.in_queue += increment;

    if(increment > 0) {
        sh_actions->queue.max_queue++;
        if(sh_actions->queue.max_queue >= SCRIPT_MAX_QUEUE) sh_actions->queue.max_queue = 0;
    } else {
        sh_actions->queue.cur_queue++;    
        if(sh_actions->queue.cur_queue >= SCRIPT_MAX_QUEUE) sh_actions->queue.cur_queue = 0;
    }    
}

// --------------------------------------------------------------------------------------------
// Parameter functions (get options and suboptions):
// --------------------------------------------------------------------------------------------

// get main parameter (-play, -stop, -record, -scriptfile ...)
static int get_arg_opt(int argc, char *arg[], int *pos, int *option_name, int *option_type, char **options)
{
    char tmp_name[MAX_SUBOPTIONS_LENGTH];
    char *val = (char*)&options[0];
    int i, j, k;
    int ret = -1;
        
    *option_name = -1;
    *option_type = -1;
    pj_memset(val, 0, MAX_SUBOPTIONS*MAX_SUBOPTIONS_LENGTH);
        
    for(i=*pos; i<argc; i++) 
    {
        for(j=0; j<sizeof(par_options) / sizeof(struct get_options); j++) 
        {
            pj_ansi_snprintf(tmp_name, sizeof(tmp_name), "%s=", par_options[j].name);
            if(!strcmp(par_options[j].name, arg[i]) || !strcmp(tmp_name, arg[i]))
            {
                *option_name = par_options[j].val;
                *option_type = par_options[j].type;
                ret = 0;

                for(k=i+1; k<argc; k++) {
                    if(arg[k][0] != '-') 
                    {                 
                        int len = MAX_SUBOPTIONS_LENGTH;
                        if(strlen(arg[k]) < MAX_SUBOPTIONS_LENGTH) len = strlen(arg[k]);

                        pj_memcpy(val, arg[k], len);                                                
                        val += MAX_SUBOPTIONS_LENGTH;
                        
                        ret++;
                        if(ret == MAX_SUBOPTIONS)
                            break;

                    } else {
                        break;
                    }
                }

                break;
            }
        }

        if(ret != -1)
            break;
    }

    *pos += (ret + 1);

    return ret;
}

// get sub parameter (suboptions)
static int get_sub_opt(int subopt_cnt, char **options, int *subopt_pos, char *name, char **values, int split_subvalues)
{
    char *val = (char*)&options[0];
    char *val2 = (char*)&values[0];
    int i;
    int ret = -1;
    int is_not_separated = 0;
    
    val += ((*subopt_pos) * MAX_SUBOPTIONS_LENGTH);
    pj_memset(val2, 0, MAX_SUBOPTIONS * MAX_SUBOPTIONS_LENGTH);
    pj_memset(name, 0, MAX_SUBOPTIONS_LENGTH);

    for(i=*subopt_pos; i<subopt_cnt; i++, val+=MAX_SUBOPTIONS_LENGTH)
    {
        // is suboption separated by '=' ?
        char *pos = strchr(val, '=');
        if(pos != NULL)
        {   
            if(is_not_separated) {
                break;
            }
            
            // save name of suboption
            pj_memcpy(name, val, pos-val);  
            name = trim(name);
            ret = 0;           

            // get suboption parameters
            {                           
                char *pos_val = strchr(val, ',');
    
                // separated by ',' ?
                if(pos_val != NULL && split_subvalues)
                {       
                    char tmp[MAX_SUBOPTIONS_LENGTH];

                    pj_memset(tmp, 0, sizeof(tmp));    
                    pj_memcpy(tmp, pos+1, sizeof(tmp));
                    
                    while(strlen(tmp) > 0)
                    {           
                        get_sep(pos+1, tmp, sizeof(tmp), ',');                                                
                        if(strlen(tmp) == 0)
                            break;
                        
                        pj_memcpy(val2, tmp, strlen(tmp));
                        val2 += MAX_SUBOPTIONS_LENGTH;
                        ret++;

                        if(ret == MAX_SUBOPTIONS_LENGTH)
                            break;                
                    }
                }
                // only one value
                else
                {
                    if(strlen(pos) > 0)
                    {
                        pj_memcpy(val2, pos+1, strlen(pos)-1);                        
                        ret++;
                    }
                }
            }
            break;
        }

        // suboption is not separated. save until a separated option occurs.
        else
        {
            is_not_separated = 1;
            pj_memcpy(val2, val, strlen(val));
            val2 += MAX_SUBOPTIONS_LENGTH;
            ret++;
        }
    }

    *subopt_pos = i + !(is_not_separated);
    ret += is_not_separated;
    
    return ret;
}


// --------------------------------------------------------------------------------------------
// HELP and USAGE:
// --------------------------------------------------------------------------------------------

// Header
static void print_usage_header(char *title)
{
   printf(
   "-------------------------------------------------------------------------------\n"
   " %s %s\n"
   "-------------------------------------------------------------------------------\n",
   APPNAME, title); 
}

// Help (first instance)
static void print_help()
{
    print_usage_header(APPVER);

    printf(" %s [OPTIONS]\n", APPNAME);
    printf("\n");
    printf(" -start                  Start normal\n");
    printf(" -daemon                 Start as daemon \n");
    printf(" -cfg [file]             Configuration file\n");
    printf(" -log [file]             Output to logfile\n");
    printf("-------------------------------------------------------------------------------\n");    
}

// Help (second instance)
static void print_help_daemon()
{
    print_usage_header(APPVER);

    printf(" %s [CONNECTION] [OPTIONS]\n", APPNAME);
    printf("\n");

    #if DTMFBOX_HELP_USAGE
    printf(" For command help, type: \"-command help\". eg.: \"-call help\", \"-play help\"...)\n\n");
    #endif

    printf(" [CONNECTION]\n");
    printf(" 0-X                            Connection (to answer, play file ... )\n\n");

    printf(" [OPTIONS]\n");
    printf(" -hook [up|down|hold|...]       Answer, hangup, hold, unhold or alert a call\n");
    printf(" -call [acc] [no|sip] [ctl]     Call no. or sip (12345 or 12345@host.net)\n");
    printf(" -play [file] [options]         Play wave file or stream (fifo/raw)\n");
    printf(" -mixer [play_id] [options]     Mixer control\n");
    printf(" -record [file] [hz] [options]  Record wavefile (8khz, 16bit, mono) or stream\n");    
    printf(" -tone [f1] [f2] [m1] [m2] [v]  Play tone: freq1, freq2, on_ms, off_ms, volume\n");       
    printf(" -dtmf [0-9|*|#]                Send a dtmf\n");        
    printf(" -text \"text\"                   Send a display message (CAPI only!)\n");        
    printf(" -bridge [on=x|off=x]           Connect/disconnect audio (bridge)\n");    
    printf(" -scriptfile [file]             Change scriptfile\n");
    printf(" -delimiter [0-9|*|#|none]      Change delimiter: 0-9, asterisk, poundkey, none\n");
    printf(" -waitevent [options]           Wait for event and output to console\n");
    printf(" -timer [timeout] \"opt str.\"    Set script timer\n");
    printf(" -env variable=\"value\"          Set global environment variable\n");
    printf(" -stop [type]                   Stop: play, record, tone, waitevent, daemon\n");    
    printf(" -list [type]                   List: connections, clients, accounts, plugins\n");
    printf(" -version                       Show version\n");
    printf("-------------------------------------------------------------------------------\n");    

    //printf(" -recordstream [file] [hz]      Record raw data to file/fifo (16bit, mono)\n");
    //printf(" -playthread [file]             Play wave file (in background)\n");
    //printf(" -playstream [file] [hz]        Play raw data from file/fifo (16bit, mono)\n");
    //printf(" -playstreamthread [file] [hz]  Play raw data (in background) (16bit, mono)\n");
    //printf(" -stop record                   Stop recording\n");
    //printf(" -stop tone                     Stop tone\n");
    //printf(" -stop daemon                   Stop application\n");
}


static void print_version()
{        
    printf("-------------------------------------------------------------------------------\n");    
    #if DTMFBOX_WIN32
    printf(" dtmfbox ....................................:  %s (Win32)\n", APPVER);
    #else
    printf(" dtmfbox ....................................:  %s (Linux)\n", APPVER);
    #endif
    printf(" PJSIP ......................................:  %s\n", PJ_VERSION);
    printf("-------------------------------------------------------------------------------\n");            
    printf(" DTMFBOX_USE_CAPI ...........................:  %d\n", DTMFBOX_USE_CAPI);
    printf(" DTMFBOX_USE_VOIP ...........................:  %d\n", DTMFBOX_USE_VOIP);
    printf(" DTMFBOX_USE_ICE ............................:  %d\n", DTMFBOX_USE_ICE);
    printf(" DTMFBOX_USE_SOUND ..........................:  %d\n", DTMFBOX_USE_SOUND);        
    printf("-------------------------------------------------------------------------------\n");    
    printf(" PJMEDIA_HAS_G711_CODEC .....................:  %d\n", PJMEDIA_HAS_G711_CODEC);    
    printf(" PJMEDIA_HAS_L16_CODEC ......................:  %d\n", PJMEDIA_HAS_L16_CODEC);
    printf(" PJMEDIA_HAS_GSM_CODEC ......................:  %d\n", PJMEDIA_HAS_GSM_CODEC);
    printf(" PJMEDIA_HAS_ILBC_CODEC .....................:  %d\n", PJMEDIA_HAS_ILBC_CODEC);
    printf(" PJMEDIA_HAS_SPEEX_CODEC ....................:  %d\n", PJMEDIA_HAS_SPEEX_CODEC);
    printf(" PJMEDIA_HAS_SPEEX_AEC ......................:  %d\n", PJMEDIA_HAS_SPEEX_AEC);
    printf(" PJMEDIA_SOUND_IMPLEMENTATION ...............:  %d\n", PJMEDIA_SOUND_IMPLEMENTATION);
    printf("-------------------------------------------------------------------------------\n");
    printf(" PJ_LOG_MAX_LEVEL ...........................:  %d\n", PJ_LOG_MAX_LEVEL);
    printf(" PJ_HAS_FLOATING_POINT ......................:  %d\n", PJ_HAS_FLOATING_POINT);
    printf(" PJ_OS_HAS_CHECK_STACK ......................:  %d\n", PJ_OS_HAS_CHECK_STACK);
    printf(" PJ_ENABLE_EXTRA_CHECK ......................:  %d\n", PJ_ENABLE_EXTRA_CHECK);
    printf("-------------------------------------------------------------------------------\n");
    printf(" MAX ACCOUNTS ...............................:  %d\n", MAX_ACCOUNTS);
    printf(" MAX CONNECTIONS ............................:  %d\n", MAX_CONS);
    printf(" MAX CONFERENCE PARTNERS PER CONNECTION .....:  %d\n", INCONF_PORTS);
    printf(" MAX PORTS IN CONFERENCE BRIDGE .............:  %d\n", CONF_PORTS);    
    printf(" CLOCKRATE/BITS/CHANNELS ....................:  %d/%d/%d\n", oSET.audio.conf_clockrate, CONF_BITS, CONF_CHANNELS);
    printf("-------------------------------------------------------------------------------\n");    
}

static void print_usage_mixer(void) 
{
    #if DTMFBOX_HELP_USAGE
    
    print_usage_header("- mixer usage:");

    printf(
    " syntax:\n"
    "  -mixer [play_id (optional)] commandline\n\n"

    " When [play_id] is 'tx' or 'rx', the mixer function is applied to a\n"
    " connection!\n\n"

    " volume is a number from 0..200 specifing the volume in percent\n\n"
    
    " commandline (more than one option can be specified):\n"
    "   * fade to target volume over the next t milliseconds: \n"
    "      fade=<target volume> duration=<fade time t in milliseconds>\n"
    "   * set current volume to target volume:\n"
    "      volume=<target volume>\n"
    "   * set a stream as background (i.e. music) (for auto-talkover):\n"
    "      background=<on|off>\n"
    "      bg=<on|off>\n"
    "   * set a stream as foreground (\"talkover-voice\") (for auto-talkover):\n"
    "      talkover=<on|off>\n"
    "     NOTE: either talkover OR background can be set, not both!\n"
    
    // not implemented yet
    //"   * directly set internal rx_volume: (range 0..32767, 4096=100%)\n"
    //"   ( rxvol=<target volume> {sets volume without a short fade} )\n"

    "\n"
    " the first file player is initialized with the following defaults:\n"
    "   bg=on volume=100\n"
    " all following players have the following defaults:\n"
    "   talkover=on volume=100\n"
    "\n");

    #endif
}

static void print_usage_call(void) 
{
    #if DTMFBOX_HELP_USAGE
    
    print_usage_header("- call usage:");
    
    printf(    
    " syntax:\n"
    "  -call [acc-no|acc-id] [dest-no|sip] [controller] [options]\n\n"    

    " options:\n"
    "  src=xxxxxx              - Account number or id that should make the call\n"
    "  dst=xxxxxx              - Destination number or SIP, to be called\n"
    "  ctrl=x                  - CAPI controller to use (optional)\n"
    "  mode=anonymous          - Hide number\n\n"

    " example:\n"
    "  * initiate a CAPI call (use controller 3):\n"
    "    %s -call src=123456 dst=5555555 ctrl=3\n"    
    "     or:\n"
    "    %s -call 123456 5555555 3\n\n"    

    "  * initiate a VoIP call and connect it with connection 0:\n"
    "    %s 0 -call src=123456 dst=5555555@anyhost.com\n"
    "\n"
    , APPNAME, APPNAME, APPNAME);


    #endif
}

static void print_usage_hook(void) 
{
    #if DTMFBOX_HELP_USAGE

    print_usage_header("- hook usage:");

    printf(    
    " syntax:\n"
    "  -hook [up|down|hold|unhold|reject|alert|auto|manual]\n\n"    

    "  Answer, hangup, hold, unhold, reject or alert a connection.\n\n"
    "  \"-hook auto\" automatically confirms & disconnects conference partners.\n"
    "  \"-hook manual\" disables auto-confirm/-disconnect. Default is \"manual\".\n\n"
    
    " example:\n"
    "  * answer connection 0:\n"
    "    %s 0 -hook up\n"
    "\n"
    , APPNAME);

    #endif
}

static void print_usage_play(void) 
{
    #if DTMFBOX_HELP_USAGE
    
    print_usage_header("- play usage:");

    printf(    
    " syntax:\n"
    "  -play [file|ip:port] [opt]       - Play wave file/stream/ip by options\n"    
    "  -playthread [file]               - Play wave file in background\n"
    "  -playstream [stream] [hz]        - Play stream at samplerate (hz=8000-44100)\n"
    "  -playstreamthread [stream] [hz]  - Play stream at samplerate in background\n\n"

    " options:\n"    
    "  mode=wave                        - Play a wave file (default)\n"
    "  mode=stream                      - Play a stream/fifo (raw data)\n"
    "  mode=speex                       - Play a speex file (8000hz/nb/no header!)\n"
    "  mode=thread                      - Play in background\n"    
    "  rtp=x                            - Local port for RTP streaming\n"
    "  hz=x                             - Samplerate that should be used (stream)\n"
    "  bufsize=x                        - Buffersize in bytes (stream)\n"
    "  wait_start=[x|endless]           - Wait x frames before timeout (stream start)\n"
    "  wait_end=[x|endless]             - Wait x frames before timeout (stream end)\n"
    "  wait=[x|endless]                 - Start & end timeout in frames (stream)\n\n"    

    " example:\n"
    "  * play a speex file (in background) to connection 0:\n"
    "    %s 0 -play test.spx mode=thread,speex\n"
    "     or:\n"
    "    %s 0 -playthread test.spx mode=speex\n\n"

    "  * play a stream in foreground at 22050Hz to connection 1:\n"
    "    %s 1 -play /var/fifo mode=stream hz=22050\n"
    "     or:\n"
    "    %s 1 -playstream /var/fifo 22050\n\n"
    
    "  * stream from RTP/RTCP transport to local port 5001:\n"
    "    %s 1 -play 192.168.178.10:16384 rtp=5001\n"    
    "\n"
    , APPNAME, APPNAME, APPNAME, APPNAME, APPNAME);

    #endif
}

static void print_usage_record(void) 
{
    #if DTMFBOX_HELP_USAGE
    
    print_usage_header("- record usage:");
    
    printf(    
    " syntax:\n"
    "  -record [file|ip:port] [opts]   - Record wave file or stream by options\n"        
    "  -recordstream [stream] [hz]     - Record stream at samplerate (hz=8000-44100)\n\n"
    
    " options:\n"    
    "  mode=wave                       - Record a wave file (default)\n"
    "  mode=stream                     - Record to stream/fifo (raw data)\n"
    "  mode=speex                      - Record a speex file (8000hz/nb/no header!)\n"
    "  rtp=x                           - Local port for RTP/RTCP streaming\n"
    "  hz=xxxxx                        - Samplerate that should be used (stream)\n\n"
    
    " example:\n"
    "  * record a wave file (connection 0):\n"
    "    %s 0 -record test.wav\n\n"

    "  * record a stream at 22050Hz (connection 1):\n"
    "    %s 1 -record /var/fifo mode=stream hz=22050\n"
    "     or:\n"
    "    %s 1 -recordstream /var/fifo 22050\n\n"

    "  * record to RTP/RTCP transport from local port 5003:\n"
    "    %s 1 -record 192.168.178.10:16384 rtp=5003\n"

    "\n"
    , APPNAME, APPNAME, APPNAME, APPNAME);

    #endif
}

static void print_usage_tone(void) 
{
    #if DTMFBOX_HELP_USAGE

    print_usage_header("- tone usage:");
    
    printf(
        " syntax:\n"
        "  -tone [freq1] [freq2] [on_msec] [off_msec] [volume]\n\n"

        "  Play a tone\n\n"
                
        "  freq1               - First frequency\n"
        "  freq2               - Second frequency\n"
        "  on_msec             - On duration (in msec)\n"
        "  off_msec            - Off duration (in msec)\n"
        "  volume              - Tone volume (0-32767)\n\n"

        " example:\n"
        "  * play tone to connection 0:\n"
        "    %s 0 -tone 440 440 2000 4000 30000\n"
        "\n"
    , APPNAME);

    #endif
}

static void print_usage_bridge(void) 
{
    #if DTMFBOX_HELP_USAGE
    
    print_usage_header("- bridge usage:");
    
    printf(    
    " syntax:\n"
    "  -bridge [on=x] [off=x]\n\n"
    
    "  Bridge or unbridge connections\n\n"

    " options:\n"
    "  on=x         -   Bridge two calls (x is the connection ID)\n"
    "  off=x        -   Unbridge two calls (x is the connection ID)\n\n"

    " example:\n"
    "  * audio should be streamed from connection 0 to 1:\n"
    "    %s 0 -bridge on=1\n"        
    "\n"
    , APPNAME);

    #endif
}

static void print_usage_delimiter(void) 
{
    #if DTMFBOX_HELP_USAGE

    print_usage_header("- delimiter usage:");

    printf(    
    " syntax:\n"
    "  -delimiter [0-9|poundkey|asterisk|none]\n\n"

    "  Change delimiter\n\n"
    
    "  poundkey         - Call script when receiving #\n"
    "  asterisk         - Call script when receiving *\n"
    "  0 - 9            - Call script when receiving the selected number\n"
    "  none             - Immediately call script on DTMF (no delimiter)\n"
    "\n"
    );

    #endif
}

static void print_usage_scriptfile(void) 
{
    #if DTMFBOX_HELP_USAGE

    print_usage_header("- scriptfile usage:");

    printf(    
    " syntax:\n"
    "  -scriptfile [file]\n\n"

    "  Change scriptfile\n\n"
    
    "  scriptfile         - The scriptfile that should be called on events\n"
    "\n"
    );

    #endif
}

static void print_usage_dtmf(void) 
{
    #if DTMFBOX_HELP_USAGE

    print_usage_header("- dtmf usage:");

    printf(    
    " syntax:\n"
    "  -dtmf [0-9|poundkey|asterisk]\n\n"

    "  Send a DTMF to a connection\n\n"
    
    "  poundkey         - Send #\n"
    "  asterisk         - Send *\n"
    "  0 - 9            - Send number\n"
    "\n"
    );        

    #endif
}

static void print_usage_timer(void) 
{
    #if DTMFBOX_HELP_USAGE

    print_usage_header("- timer usage:");

    printf(    
    " syntax:\n"
    "  -timer [timeout] \"option1 option2 ...\" [mode]\n\n"

    "  Schedule a timer. Scriptfile for the current connection\n"
    "  will be called after <timeout> seconds.\n\n"

    "  timeout : timeout in sec (or msec, see [mode])\n\n"

    "  optionX : these parameters will be passed to the script\n"
    "            when it is called after the timeout\n"
    "            \"TIMER\" is appended as first parameter. The\n"
    "            optionX will start at the ninth parameter.\n\n"

    "  mode    : mode=sec   - timeout in seconds\n"
    "            mode=msec  - timeout in milliseconds\n\n"

    "  IMPORTANT: use \\\" inside option string for quotes!\n"
    "\n"
    );        

    #endif
}

static void print_usage_environment(void) 
{
    #if DTMFBOX_HELP_USAGE

    print_usage_header("- env usage:");

    printf(    
    " syntax:\n"
    "  -env variable=\"value\"\n\n"

    "  Set or unset a global environment variable that is\n"
    "  available in all scripts.\n\n"

    "  options : just set the variables you need in the script,\n"
    "            by the following syntax:\n"
    "            dtmfbox 0 -env MY_VAR=\"Hello!\"\n\n"

    "  attention: use \" quotes around the value!\n"
    "\n"
    );        

    #endif
}

static void print_usage_waitevent(void) 
{
    #if DTMFBOX_HELP_USAGE
    
    print_usage_header("- waitevent usage:");

    printf(    
    " syntax:\n"
    "  -waitevent [options]\n\n"

    "  Output event, instead of calling scriptfile\n\n"

    " options:\n"
    "  mode=endless         -  Block events until stop or disconnect\n"
    "\n"
    );        

    #endif
}

static void print_usage_stop(void) 
{
    #if DTMFBOX_HELP_USAGE
    
    print_usage_header("- stop usage:");

    printf(    
    " syntax:\n"
    "  -stop [type]\n\n"

    "  Stop playing, recording, tone, daemon\n\n"

    " type:\n"
    "  play [play_id] [all] -   Stop playing. When no play_id is specified, the last\n"
    "                           file/stream in the queue will be stopped.\n"
    "  record               -   Stop recording a file/stream\n"
    "  tone                 -   Stop tone\n"
    "  timer                -   Stop timer\n"
    "  waitevent            -   Stop wait event block\n"
    "  daemon               -   Stop daemon\n"
    "\n"
    );        

    #endif
}

