/*********************************************************************************
 * pipe.c ($Revision: 250 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "pipe.c"

int pipe_init()
{
    pj_memset(oPIPE, 0, sizeof(oPIPE));
    return 0;
}

int pipe_create(char *pipe_name, int wait, int read, int write)
{    
    int pipe_id = -1;
    int i;
    int pipe_no_wait = 0;
    
    #if (DTMFBOX_WIN32 == 1)
        PJ_UNUSED_ARG(write);
        PJ_UNUSED_ARG(read);    
    #else
        int pipe_read_write = 0;
    #endif

    // search free pipe
    for(i=0; i<PIPE_MAX; i++) 
    {
        #if (DTMFBOX_WIN32 == 1)
        if(oPIPE[i].hPipe == NULL) 
        #else
        if(oPIPE[i].hPipe == 0)
        #endif
        {
            pipe_id = i;
            break;
        }
    }    
    if(pipe_id == -1)
    {
        app_err(THIS_FILE, "No more free pipes left (max. %d)!", PIPE_MAX);
        return -1;
    }
    
    #if DTMFBOX_WIN32        

        if(!strstr(pipe_name, "\\"))
            snprintf(oPIPE[pipe_id].name, sizeof(oPIPE[pipe_id].name), "%s\\%s", oSET.pipe_dir, pipe_name);
        else
            snprintf(oPIPE[pipe_id].name, sizeof(oPIPE[pipe_id].name), "%s", pipe_name);    

        // block?
        if(wait)
            pipe_no_wait = PIPE_WAIT;
        else
            pipe_no_wait = PIPE_NOWAIT;
        
        // create pipe
        oPIPE[pipe_id].hPipe = CreateNamedPipe(oPIPE[pipe_id].name,
            PIPE_ACCESS_DUPLEX,         // Beide Richtungen
            PIPE_TYPE_BYTE | pipe_no_wait, 
            4, 8192, 8192,              
                                         
            4000,                       // Timeout in Millisekunden
            NULL);                      // Sicherheitsdesk

        if(oPIPE[pipe_id].hPipe == INVALID_HANDLE_VALUE) {
            app_err(THIS_FILE, "Unable to create named pipe %s!", oPIPE[pipe_id].name);
            return -1;            
        }        
    #else

        if(!strstr(pipe_name, "/"))
            snprintf(oPIPE[pipe_id].name, sizeof(oPIPE[pipe_id].name), "%s/%s", oSET.pipe_dir, pipe_name);
        else
            snprintf(oPIPE[pipe_id].name, sizeof(oPIPE[pipe_id].name), "%s", pipe_name);    
       
        if(wait)
            pipe_no_wait = 0;
        else
            pipe_no_wait = O_NONBLOCK;

        if(write && read)
            pipe_read_write = O_RDWR;
        else if(!write && read)
            pipe_read_write = O_RDONLY;
        else if(write && !read)
            pipe_read_write = O_WRONLY;
        else
            pipe_read_write = O_RDWR;

        umask(0);
        mkfifo(oPIPE[pipe_id].name, 0666);
                
        if((oPIPE[pipe_id].hPipe = open(oPIPE[pipe_id].name, pipe_read_write | pipe_no_wait)) == -1) {
            app_err(THIS_FILE, "Unable to open named pipe %s!", oPIPE[pipe_id].name);
            oPIPE[pipe_id].hPipe = 0;
            return -1;            
        }

    #endif
        
    oPIPE[pipe_id].block = wait;

    return pipe_id;
}

int pipe_open(char *pipe_name, int wait, int read, int write)
{    
    int pipe_id = -1;
    int i;
            
    #if (DTMFBOX_WIN32 == 1)
        PJ_UNUSED_ARG(write);
        PJ_UNUSED_ARG(read);    
    #else
        int pipe_no_wait = 0;
        int pipe_read_write = 0;
    #endif

    // search free pipe
    for(i=0; i<PIPE_MAX; i++) 
    {
        #if DTMFBOX_WIN32 == 1
        if(oPIPE[i].hPipe == NULL) 
        #else
        if(oPIPE[i].hPipe == 0) 
        #endif
        {
            pipe_id = i;
            break;
        }
    }    
    if(pipe_id == -1)
    {
        app_err(THIS_FILE, "No more free pipes left (max. %d)!", PIPE_MAX);
        return -1;
    }

    // open pipe
    #if DTMFBOX_WIN32

        if(!strstr(pipe_name, "\\"))
            snprintf(oPIPE[pipe_id].name, sizeof(oPIPE[pipe_id].name), "%s\\%s", oSET.pipe_dir, pipe_name);
        else
            snprintf(oPIPE[pipe_id].name, sizeof(oPIPE[pipe_id].name), "%s", pipe_name);    
               
        oPIPE[pipe_id].hPipe = CreateFile(oPIPE[pipe_id].name,
                        GENERIC_WRITE | GENERIC_READ,
                        0,
                        NULL,
                        OPEN_EXISTING,
                        FILE_ATTRIBUTE_NORMAL,
                        (HANDLE)NULL);

        if(oPIPE[pipe_id].hPipe == INVALID_HANDLE_VALUE) 
        {
            app_err(THIS_FILE, "Unable to open named pipe %s!", oPIPE[pipe_id].name);            
            memset(oPIPE[pipe_id].name, 0, sizeof(oPIPE[pipe_id].name));
            oPIPE[pipe_id].hPipe = NULL;
            return -1;            
        }        
    #else
        if(!strstr(pipe_name, "/"))
            snprintf(oPIPE[pipe_id].name, sizeof(oPIPE[pipe_id].name), "%s/%s", oSET.pipe_dir, pipe_name);
        else
            snprintf(oPIPE[pipe_id].name, sizeof(oPIPE[pipe_id].name), "%s", pipe_name);    
        
        if(wait)
            pipe_no_wait = 0;
        else
            pipe_no_wait = O_NONBLOCK;
        
        if(write && read)
            pipe_read_write = O_RDWR;
        else if(!write && read)
            pipe_read_write = O_RDONLY;
        else if(write && !read)
            pipe_read_write = O_WRONLY;
        else
            pipe_read_write = O_RDWR;

        if((oPIPE[pipe_id].hPipe = open(oPIPE[pipe_id].name, pipe_read_write | O_NONBLOCK, 0)) == -1) {
            app_err(THIS_FILE, "Unable to open named pipe %s!", oPIPE[pipe_id].name);            
            memset(oPIPE[pipe_id].name, 0, sizeof(oPIPE[pipe_id].name));
            oPIPE[pipe_id].hPipe = 0;
            return -1;            
        }
    #endif

    oPIPE[pipe_id].block = wait;

    return pipe_id;    
}


void pipe_close(int pipe_id)
{
    #if DTMFBOX_WIN32
    if(oPIPE[pipe_id].hPipe != NULL)
    #else
    if(oPIPE[pipe_id].hPipe != 0)
    #endif
    {
        #if DTMFBOX_WIN32
            CloseHandle(oPIPE[pipe_id].hPipe);            
            oPIPE[pipe_id].hPipe = NULL;
        #else
            close(oPIPE[pipe_id].hPipe);
            oPIPE[pipe_id].hPipe = 0;
            unlink(oPIPE[pipe_id].name);            
            remove(oPIPE[pipe_id].name);            
        #endif

        memset(oPIPE[pipe_id].name, 0, sizeof(oPIPE[pipe_id].name));        
        oPIPE[pipe_id].block = 0;
    }
}

int pipe_read(int pipe_id, char *buf, int len)
{
    int read_bytes = 0;

    #if DTMFBOX_WIN32
    if(oPIPE[pipe_id].hPipe != NULL)
    #else
    if(oPIPE[pipe_id].hPipe != 0)
    #endif
    {
        #if DTMFBOX_WIN32
            ReadFile(oPIPE[pipe_id].hPipe, buf, len, (LPDWORD)&read_bytes, 0);        
            if(read_bytes >= 0)
                buf[read_bytes] = '\0';
            else
                buf[0] = '\0';
        
        #else

            fd_set rfd;
            struct timeval timeout;
            int ready = -1;

            timeout.tv_sec = 0;         
            timeout.tv_usec = 100;        // 100 usec

            FD_ZERO(&rfd);
            FD_SET(oPIPE[pipe_id].hPipe, &rfd);

            ready = select(oPIPE[pipe_id].hPipe + 1, &rfd, NULL, NULL, &timeout);
            if(ready > 0)
                read_bytes = read(oPIPE[pipe_id].hPipe, buf, len);                

            if(read_bytes >= 0) {
                buf[read_bytes] = '\0';                                
            } else {
                buf[0] = '\0';
            }


        #endif
    }

    return read_bytes;
}

int pipe_write(int pipe_id, char *buf, int len)
{
    int write_bytes = 0;

    #if DTMFBOX_WIN32
    if(oPIPE[pipe_id].hPipe != NULL)
    #else
    if(oPIPE[pipe_id].hPipe != 0)
    #endif
    {
        #if DTMFBOX_WIN32
            WriteFile(oPIPE[pipe_id].hPipe, buf, len, (LPDWORD)&write_bytes, 0);                    
        #else   
            fd_set wfd;
            struct timeval timeout;
            int ready = -1;

            timeout.tv_sec = 0;         
            timeout.tv_usec = 100;        // 100 usec
            FD_ZERO(&wfd);
            FD_SET(oPIPE[pipe_id].hPipe, &wfd);
            ready = select(oPIPE[pipe_id].hPipe + 1, NULL, &wfd, NULL, &timeout);
            if(ready > 0)
                write_bytes = write(oPIPE[pipe_id].hPipe, buf, len);
            
            if(write_bytes < 0)
                write_bytes = 0;
        #endif
    }

    return write_bytes;
}
