/*********************************************************************************
 * plugins.c ($Revision: 250 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "plugins.c"

void plugins_init()
{
    char c_plugins[2048];
    char c_plugin[256];
    int i;
    
    plugin_count = 0;
    pj_memcpy(c_plugins, oSET.plugins, sizeof(c_plugins));
    pj_memset(plugin_proc_event, 0, sizeof(plugin_proc_event));
    pj_memset(plugin_proc_output, 0, sizeof(plugin_proc_output));
    pj_memset(plugin_proc_get_frame, 0, sizeof(plugin_proc_get_frame));
    pj_memset(plugin_proc_put_frame, 0, sizeof(plugin_proc_put_frame));
    pj_memset(plugin_proc_extra_commands, 0, sizeof(plugin_proc_extra_commands));    
    pj_memset(plugin_proc_extra_help, 0, sizeof(plugin_proc_extra_help));    

    for(i=0; i<MAX_PLUGINS; i++)
    {                   
        get_sep(c_plugins, c_plugin, sizeof(c_plugin), ',');
        if(!strlen(c_plugin)) break;

        plugins[plugin_count] = LoadLibraryA(c_plugin);
        if(plugins[plugin_count] != NULL)
        {
            plugin_init plug_init = (plugin_init)GetProcAddress(plugins[plugin_count], "plugin_init");            
            if(plug_init)  
            {
                plugin_proc_event[plugin_count] = (plugin_event)GetProcAddress(plugins[plugin_count], "plugin_event");
                plugin_proc_output[plugin_count] = (plugin_output)GetProcAddress(plugins[plugin_count], "plugin_output");
                plugin_proc_get_frame[plugin_count] = (plugin_get_frame)GetProcAddress(plugins[plugin_count], "plugin_get_frame");
                plugin_proc_put_frame[plugin_count] = (plugin_put_frame)GetProcAddress(plugins[plugin_count], "plugin_put_frame");
                plugin_proc_extra_commands[plugin_count] = (plugin_extra_commands)GetProcAddress(plugins[plugin_count], "plugin_extra_commands");
                plugin_proc_extra_help[plugin_count] = (plugin_extra_help)GetProcAddress(plugins[plugin_count], "plugin_extra_help");
                plugin_count++;
                
                plug_init(((plugin_exec_command)(&plugin_execute_cmd)));
                app_msg(THIS_FILE, "#%d - Plugin \"%s\" loaded!", plugin_count, c_plugin);                            
            }
            else 
            {                
                FreeLibrary(plugins[plugin_count]);
                app_err(THIS_FILE, "Cannot find plugin_init() function in plugin \"%s\"!", c_plugin);
            }
        } else {
            app_err(THIS_FILE, "Unable to load plugin \"%s\"!", c_plugin);
        }
    }
}

void plugins_close()
{
    int i;    
    plugin_close plug_close;

    pj_memset(plugin_proc_event, 0, sizeof(plugin_proc_event));
    pj_memset(plugin_proc_output, 0, sizeof(plugin_proc_output));
    pj_memset(plugin_proc_get_frame, 0, sizeof(plugin_proc_get_frame));
    pj_memset(plugin_proc_put_frame, 0, sizeof(plugin_proc_put_frame));
    pj_memset(plugin_proc_extra_commands, 0, sizeof(plugin_proc_extra_commands));    
    pj_memset(plugin_proc_extra_help, 0, sizeof(plugin_proc_extra_help));    

    for(i=0; i<plugin_count; i++)
    {      
        if(plugins[i] != NULL) 
        {
            plug_close = (plugin_close)GetProcAddress(plugins[i], "plugin_close");            
            if(plug_close) plug_close();            
            FreeLibrary(plugins[i]);            
            plug_close = NULL;
        }
    }
    
    plugin_count = 0;
}


int plugin_execute_cmd(int argc, ...)
{   
    char *arg[MAX_SUBOPTIONS];    
    int i = 0;
    int j = 0;
    int ret = -1;
    
    va_list args;    
    va_start(args, argc);
    
    memset(arg, 0, sizeof(arg));

    arg[0] = (char*)malloc(MAX_SUBOPTIONS_LENGTH);
    memset(arg[0], 0, MAX_SUBOPTIONS_LENGTH-1);
    i++;
    argc++;

    // create arguments...
    do 
    {
        char *s = va_arg(args, char*);        
        if(s == NULL) break;

        arg[i] = (char*)malloc(MAX_SUBOPTIONS_LENGTH);
        memset(arg[i], 0, MAX_SUBOPTIONS_LENGTH-1);                
        memcpy(arg[i], s, MAX_SUBOPTIONS_LENGTH-1);        
        i++;            
    } 
    while(i < MAX_SUBOPTIONS_LENGTH && argc > i);    
    
    // ... just output log? (-log1 to -log5)
    if(argc>2) 
    {
         for(j=1; j<=5; j++)
        {
            char loglevel[32];
            snprintf(loglevel, sizeof(loglevel), "-log%d", j);
            if(!strcmp(arg[1], loglevel))
            {
                switch(j) {
                    case 1: app_msg(THIS_FILE, "%s", arg[2]); break;
                    case 2: app_err(THIS_FILE, "%s", arg[2]); break;
                    case 3: app_dbg(THIS_FILE, "%s", arg[2]); break;
                    case 4: app_dbg2(THIS_FILE, "%s", arg[2]); break;
                    case 5: app_dbg3(THIS_FILE, "%s", arg[2]); break;
                }                
                goto end;
            }
        }
    }

    // ... or execute command?
    ret = param_parser_daemon(i, arg, getpid());

end:
    va_end(args);
    for(j=0; j<i; j++) {
        free(arg[j]);
        arg[j] = NULL;
    }
    
    return ret;
}

void plugin_raise_event(script_event_t script_event, char *s_event, char *s_direction, char *s_type, char *no_src, char *no_dst, char *data)
{
    int i;

    for(i=0; i<plugin_count; i++)
    {      
        plugin_event p_event = *((plugin_event*)(&plugin_proc_event[i]));

        if(p_event != NULL) 
            p_event(script_event.scriptfile, 
                            s_event, 
                            s_direction, 
                            s_type, 
                            script_event.conid_src, 
                            script_event.conid_dst, 
                            no_src, 
                            no_dst, 
                            data, 
                            script_event.acc_id+1);        
    }
}

void plugin_raise_output(char *output, int pid)
{
    int i;

    if(pid == 0)
    {
        for(i=0; i<plugin_count; i++)
        {      
            plugin_output p_output = *((plugin_output*)(&plugin_proc_output[i]));
                    
            if(p_output != NULL) 
                p_output(output, pid);
        }
    }
}

void plugin_raise_get_frame(char *buf, int buf_size, int conid, int playid, int accid)
{
    int i;

    for(i=0; i<plugin_count; i++)
    {      
        plugin_get_frame p_get = *((plugin_get_frame*)(&plugin_proc_get_frame[i]));

        if(p_get != NULL) 
            p_get(buf, buf_size, conid, playid, accid+1);
    }
}

void plugin_raise_put_frame(char *buf, int buf_size, int conid, int recid, int accid)
{
    int i;

    for(i=0; i<plugin_count; i++)
    {      
        plugin_put_frame p_put = *((plugin_put_frame*)(&plugin_proc_put_frame[i]));

        if(p_put != NULL) 
            p_put(buf, buf_size, conid, recid, accid+1);
    }
}

int plugin_raise_extra_commands(int argc, char *arg[], int pid)
{
    int i;
    int ret = 0;

    for(i=0; i<plugin_count; i++)
    {      
        plugin_extra_commands p_cmd = *((plugin_extra_commands*)(&plugin_proc_extra_commands[i]));

        if(p_cmd != NULL) 
            ret += p_cmd(argc, arg, pid);
    }

    return ret;
}

void plugin_raise_extra_help(int pipe_id)
{
    char tmp[128];
    char *line = "-------------------------------------------------------------------------------\n";
    char *line2 = "===============================================================================\n";
    char *no_help = " * plugin help not available *";
    int i;

    snprintf(tmp, sizeof(tmp), " %s %s - %d plugin(s) loaded\n", APPNAME, APPVER, plugin_count);        
    pipe_write(pipe_id, line, strlen(line));
    pipe_write(pipe_id, tmp, strlen(tmp));
    pipe_write(pipe_id, line, strlen(line));
    pipe_write(pipe_id, "\n", 1);

    pipe_write(pipe_id, line2, strlen(line2));

    for(i=0; i<plugin_count; i++)
    {      
        plugin_extra_help p_help = *((plugin_extra_help*)(&plugin_proc_extra_help[i]));
       
        if(p_help!= NULL) 
        {
            char *help_text = p_help();
            pipe_write(pipe_id, help_text, strlen(help_text));
        } 
        else 
        {
            pipe_write(pipe_id, no_help, strlen(no_help));
        }
        
        pipe_write(pipe_id, "\n", 1);                    
        pipe_write(pipe_id, line2, strlen(line2));
    }       
}
