/*********************************************************************************
 * port_capi.c ($Revision: 250 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "port_capi.c"

pj_status_t port_capi_init(connection_t *mCON)
{
    connection_t *CUR_CON = (connection_t*)mCON;    
    pj_status_t status = PJ_SUCCESS;
    
    PJ_ASSERT_RETURN(CUR_CON != NULL, 1);    
    PJ_ASSERT_RETURN(CUR_CON->ports->custom == NULL, 2);    
    
    status = port_capi_create(CUR_CON->pools->con_pool, CUR_CON, &CUR_CON->ports->custom);
    if(status != PJ_SUCCESS)    
        app_err(THIS_FILE, "(%d) Unable to create CAPI port!", status);
    
    return status;
}

pj_status_t port_capi_close(connection_t *mCON)
{
    connection_t *CUR_CON = (connection_t*)mCON;    
    portdata_capi *pdata;  
    pj_status_t status = PJ_SUCCESS;

    PJ_ASSERT_RETURN(CUR_CON != NULL, 1);    
    PJ_ASSERT_RETURN(CUR_CON->ports->custom != NULL, 2);    
    
    pdata = CUR_CON->ports->custom->port_data.pdata;    

    // disconnect this media_slot from each connection
    mixer_conf_port_disconnect_all(CUR_CON->slots->media);
        
    #if (DTMFBOX_USE_NULL_PORT == 1)
        // disconnect audioless loopback
        pjmedia_conf_disconnect_port(oGLB.conf.bridge, CUR_CON->slots->media, oGLB.audio.flow_null_slot);
        pjmedia_conf_disconnect_port(oGLB.conf.bridge, oGLB.audio.flow_null_slot, CUR_CON->slots->media);    
    #endif

    // destroy echo-canceller
    if(pdata->echo_canceller != NULL)
    {
        pjmedia_port_destroy(pdata->echo_canceller);        
        pdata->echo_canceller = NULL;
    }

    // remove media-port from conference
    if(CUR_CON->slots->media != -1)
    {        
        status = pjmedia_conf_remove_port(oGLB.conf.bridge, CUR_CON->slots->media);
        pj_assert(status == PJ_SUCCESS);
        CUR_CON->slots->media = -1;
    }

    // destroy media-port
    if(CUR_CON->ports->media != NULL)
    {
        status = pjmedia_port_destroy(CUR_CON->ports->media);
        if(status != PJ_SUCCESS)    
            app_err(THIS_FILE, "(%d) Unable to destroy media port!", status);
        else
            CUR_CON->ports->media = NULL;
    }

    // destroy reverse channel port
    if(pdata->reverse_ch_port != NULL)
    {
        status = pjmedia_port_destroy(pdata->reverse_ch_port);
        if(status != PJ_SUCCESS)    
            app_err(THIS_FILE, "(%d) Unable to destroy reverse channel port!", status);
        else
            pdata->reverse_ch_port = NULL;
    }
/*
    // destroy resample port
    if(pdata->resample_port != NULL) 
    {
        status = pjmedia_port_destroy(pdata->resample_port);
        if(status != PJ_SUCCESS)    
            app_err(THIS_FILE, "(%d) Unable to destroy resample port!", status);
        else
            pdata->resample_port = NULL;
    }
*/
    // destroy capi-port
    if(CUR_CON->ports->custom != NULL)
    {
        status = pjmedia_port_destroy(CUR_CON->ports->custom);
        if(status != PJ_SUCCESS)
            app_err(THIS_FILE, "(%d) Unable to destroy CAPI port!", status);
        else
            CUR_CON->ports->custom = NULL;
    }

    return status;
}

pj_status_t port_capi_create(pj_pool_t *pool, connection_t *mCON, pjmedia_port **p_port)
{   
    connection_t *CUR_CON = (connection_t*)mCON;    
    pjmedia_port *port;
    portdata_capi *pdata;    
    pj_status_t status = PJ_SUCCESS;
    unsigned count;    
    
    PJ_ASSERT_RETURN(pool, PJ_EINVAL);

    port = pj_pool_zalloc(pool, sizeof(pjmedia_port));
    PJ_ASSERT_RETURN(port != NULL, PJ_ENOMEM);
    
    // Fill port info
    port->info.bits_per_sample = CONF_BITS;
    port->info.channel_count = CONF_CHANNELS;
    port->info.encoding_name = pj_str("pcm");
    port->info.has_info = 1;
    port->info.name = pj_str("capi port");
    port->info.need_info = 0;
    port->info.pt = 0xFF;
    port->info.clock_rate = oSET.audio.conf_clockrate;
    port->info.samples_per_frame = port->info.clock_rate * 20 / 1000;
    port->info.bytes_per_frame = port->info.samples_per_frame * (CONF_BITS / 8);
    port->info.type = PJMEDIA_TYPE_AUDIO;
    
    // Initialize callbacks
    port->put_frame = &port_capi_put_frame;
    port->get_frame = &port_capi_get_frame;

    // Create capi_port pdata 
    port->port_data.pdata = pdata = pj_pool_zalloc(pool, sizeof(portdata_capi));
    
    // Fill capi_port pdata
    count = port->info.samples_per_frame / CONF_CHANNELS;
    pdata->con_id = CUR_CON->id;   
    pdata->mixdata.tx_volume = vol_lin2log(vol_scale(oSET.audio.mixer_tx_volume));
    pdata->mixdata.rx_volume = vol_lin2log(vol_scale(oSET.audio.mixer_rx_volume));  
    pdata->mixdata.mother_port = port;

    // Create a splitcomb port and a reverse channel
    // The reverse_channel is for the manual putframe/getframe operation in capi_streaming().
    // The splitcomb port (CUR_CON->ports->media) is added to conference and will be connected with other media ports.
    status = pjmedia_splitcomb_create(pool, oSET.audio.conf_clockrate, CONF_CHANNELS, oSET.audio.conf_samplerate, CONF_BITS, 0, &CUR_CON->ports->media);
    pj_assert(status == PJ_SUCCESS);    
    status = pjmedia_splitcomb_create_rev_channel(pool, CUR_CON->ports->media, 0, 0, &pdata->reverse_ch_port);
    pj_assert(status == PJ_SUCCESS);
    
    // When clockrate from conference bridge differs 8000hz, create a resample port (CAPI connections)!
    // When this connection is a soundcard, do not resample!
    if(oSET.audio.conf_clockrate != CAPI_CLOCKRATE && CUR_CON->info->type == CAPI)    
        status = pjmedia_resample_port_create(pool, pdata->reverse_ch_port, CAPI_CLOCKRATE, oSET.audio.resampler_quality, &pdata->resample_port);
    else
        pdata->resample_port = pdata->reverse_ch_port;

    // Use echo canceller, when con_ec_tail > 0 (and not soundcard!)
    if(CUR_CON->info->type == CAPI && oSET.audio.con_ec_tail > 0) 
    {
        // create echo port (connect with media_port)
        status = pjmedia_echo_port_create(pool, CUR_CON->ports->media, oSET.audio.con_ec_tail, 0, 0, &pdata->echo_canceller); 

        // Set echo canceller as media_slot
        if(status == PJ_SUCCESS)
        {
            char c_name[32];
            pj_str_t name;

            pj_ansi_snprintf(c_name, sizeof(c_name), "capi port - ec [%d]", CUR_CON->id);
            name = pj_str(c_name);

            status = pjmedia_conf_add_port(oGLB.conf.bridge, pool, pdata->echo_canceller, &name, (unsigned int*)&CUR_CON->slots->media);
            pj_assert(status == PJ_SUCCESS);
        }
    }    
    // Do not use echo canceller, set media-port (splitcomb) as media-slot
    else
    {        
        char c_name[32];
        pj_str_t name;

        pj_ansi_snprintf(c_name, sizeof(c_name), "capi port [%d]", CUR_CON->id);
        name = pj_str(c_name);

        status = pjmedia_conf_add_port(oGLB.conf.bridge, pool, CUR_CON->ports->media, &name, (unsigned int*)&CUR_CON->slots->media);
        pj_assert(status == PJ_SUCCESS);    
    }

    #if (DTMFBOX_USE_NULL_PORT == 1)
        // connect audioless loopback
        pjmedia_conf_connect_port(oGLB.conf.bridge, CUR_CON->slots->media, oGLB.audio.flow_null_slot, 0);
        pjmedia_conf_connect_port(oGLB.conf.bridge, oGLB.audio.flow_null_slot, CUR_CON->slots->media, 0);
    #endif

    *p_port = port;

    return status;
}

static pj_status_t port_capi_get_frame(pjmedia_port *port, pjmedia_frame *frame)
{   
    portdata_capi *pdata = port->port_data.pdata;    
    pj_int16_t *samples = frame->buf;
    pjmedia_frame frame_new;
    pj_status_t status;
            
    frame_new.buf = samples;
    frame_new.size = frame->size;
    frame_new.type = frame->type;
    frame_new.timestamp = frame->timestamp;

    // Get frames from reverse channel        
    status = pjmedia_port_get_frame(pdata->resample_port, &frame_new);          
    
    // Not connected with any port?
    if(frame_new.type == PJMEDIA_FRAME_TYPE_NONE)
    {
        frame_new.type = PJMEDIA_FRAME_TYPE_AUDIO;
        pj_memset(frame_new.buf, 0, frame_new.size);
    }

    // Mixer callback    
    if(port)
    {
        pdata = port->port_data.pdata;
        mixer_port_get_frame(port, &frame_new, pdata->con_id, -1, status);        
    }

    return status;
}

static pj_status_t port_capi_put_frame(pjmedia_port *port, const pjmedia_frame *frame)
{    
    portdata_capi *pdata = port->port_data.pdata;
    pj_int16_t *samples = frame->buf;
    pjmedia_frame frame_new;
    
    frame_new.buf = samples;    
    frame_new.size = frame->size;
    frame_new.type = frame->type;
    frame_new.timestamp = frame->timestamp;
    
    // Mixer callback        
    mixer_port_put_frame(port, &frame_new, pdata->con_id, -1);

    // Send frames to reverse channel
    pdata = port->port_data.pdata;      
    return pjmedia_port_put_frame(pdata->resample_port, &frame_new);                    
}
