/*********************************************************************************
 * port_filestream.h ($Revision: 260 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/

#pragma once
#define _CRT_SECURE_NO_DEPRECATE 1
#define _CRT_NONSTDC_NO_DEPRECATE 1

#define N_FILESTREAM_BUFFERS 5

typedef struct
{   
    pjmedia_port            *file_wave_rec_port;    // Wavefile recorder
    pjmedia_port            *file_wave_play_port;   // Wavefile player
    
    #if DTMFBOX_WIN32                               // Raw file (stream)
        FILE                *fp_in; 
        FILE                *fp_out;        
    #else
        int                 fp_in;
        int                 fp_out;
    #endif
        
    mixer_port_data         mixdata;                // Mixer

    #if DTMFBOX_SPEEX_ENCODER
    speex_converter_data    speex;                  // Speex encoder/decoder
    #endif

    pj_int16_t              *buf;                   // stream input buffer
    unsigned                buf_size;               // in bytes
    unsigned long           buf_write;              // current write position
    unsigned long           buf_read;               // current read position
    int                     stream_started;         // stream status
    int                     stream_ended;           // stream ended (prevent buffer underrun for several time)
    int                     eof_count;              // count EOF and stream some 0-bytes...    
    int                     wait_start;             // Start-Timeout in msec
    int                     wait_end;               // End-Timeout in msec

    int                     play_id;                // current filestream id in connection_t
    int                     con_id;                 // current connection id 

    // "Hooked" resampler callbacks
    pj_status_t             (* resampled_cb_get_frame)(pjmedia_port *this_port, pjmedia_frame *frame); 
    pj_status_t             (* resampled_cb_put_frame)(pjmedia_port *this_port, pjmedia_frame const *frame); 
    
} portdata_filestream;

pj_status_t port_filestream_player_recorder(connection_t* mCON, char* in_file, char* out_file, int clockrate, int speex, long buffer_size, int wait_start, int wait_end, int *play_id);
pj_status_t port_filestream_close_all_eof();

static pj_status_t port_filestream_init(connection_t *mCON, char *in_file, char *out_file, int clockrate, int speex, long buffer_size, int wait_start, int wait_end, int *play_id);
static pj_status_t port_filestream_close(connection_t *mCON, int play_queue, int close_playport);
static pj_status_t port_filestream_create(pj_pool_t *pool, pjmedia_port **p_port, char *in_file, char *out_file, int clockrate, int speex, long buffer_size, int wait_start, int wait_end, int con_id, int play_id);
static pj_status_t port_filestream_eof(pjmedia_port *port, void* usr_data);

static pj_status_t port_filestream_on_destroy(pjmedia_port *port);
static pj_status_t port_filestream_get_frame(pjmedia_port *port, pjmedia_frame *frame);
static pj_status_t port_filestream_put_frame(pjmedia_port *port, const pjmedia_frame *frame);
static pj_status_t port_filestream_resampled_get_frame(pjmedia_port *port, pjmedia_frame *frame);
static pj_status_t port_filestream_resampled_put_frame(pjmedia_port *port, const pjmedia_frame *frame);

#if(DTMFBOX_WIN32 == 0)
// Ringbuffer functions (Linux/Unix only)
static unsigned long read_in_ringbuffer(int fp_in, pj_int16_t *buf, unsigned size, unsigned long wr, unsigned long rd);
static unsigned long copy_out_ringbuffer(pj_int16_t *buf, unsigned size, unsigned long wr, unsigned long rd, pj_int16_t *outbuf, unsigned outsize);
static pj_status_t port_filestream_get_buffer(pjmedia_port *port, pjmedia_frame *frame);
static pj_status_t port_filestream_fill_buffer(pjmedia_port *port);
#endif
