/*********************************************************************************
 * port_null.c ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "port_null.c"

#if (DTMFBOX_USE_NULL_PORT == 1)

pj_status_t port_null_init()
{    
    pjmedia_port *port;
    pj_status_t status = PJ_SUCCESS;
        
    PJ_ASSERT_RETURN(oGLB.pool, PJ_EINVAL);

    oGLB.audio.flow_null_port = pj_pool_zalloc(oGLB.pool, sizeof(pjmedia_port));
    PJ_ASSERT_RETURN(oGLB.audio.flow_null_port != NULL, PJ_ENOMEM);
    
    // Fill port info
    port = oGLB.audio.flow_null_port;

    port->info.bits_per_sample = CONF_BITS;
    port->info.channel_count = CONF_CHANNELS;
    port->info.encoding_name = pj_str("pcm");
    port->info.has_info = 1;
    port->info.name = pj_str("null port");
    port->info.need_info = 0;
    port->info.pt = 0xFF;
    port->info.clock_rate = oSET.audio.conf_clockrate;
    port->info.samples_per_frame = port->info.clock_rate * 20 / 1000;
    port->info.bytes_per_frame = port->info.samples_per_frame * (CONF_BITS / 8);
    port->info.type = PJMEDIA_TYPE_AUDIO;

    // Initialize callbacks
    port->put_frame = &port_null_put_frame;
    port->get_frame = &port_null_get_frame;
    
    // Add global null port to conference
    status = pjmedia_conf_add_port(oGLB.conf.bridge, oGLB.pool, oGLB.audio.flow_null_port, &port->info.name, &oGLB.audio.flow_null_slot);
    pj_assert(status == PJ_SUCCESS);    
       
    return status;
}

pj_status_t port_null_close() 
{
    pj_status_t status = PJ_SUCCESS;

    if(oGLB.audio.flow_null_slot != (unsigned int)-1) {
        status = mixer_conf_port_disconnect_all(oGLB.audio.flow_null_slot);
        pjmedia_conf_remove_port(oGLB.conf.bridge, oGLB.audio.flow_null_slot);
        oGLB.audio.flow_null_slot = (unsigned int)-1;
    }

    if(oGLB.audio.flow_null_port != NULL) {
        status = pjmedia_port_destroy(oGLB.audio.flow_null_port);
        oGLB.audio.flow_null_port = NULL;
    }

    return status;
}

static pj_status_t port_null_get_frame(pjmedia_port *port, pjmedia_frame *frame)
{
    int byte = 0;   
    PJ_UNUSED_ARG(port);

    pj_memset(frame->buf, byte, frame->size);
    frame->type = PJMEDIA_FRAME_TYPE_AUDIO;    
    return PJ_SUCCESS;
}

static pj_status_t port_null_put_frame(pjmedia_port *port, const pjmedia_frame *frame)
{    
    PJ_UNUSED_ARG(port);
    PJ_UNUSED_ARG(frame);
    return PJ_SUCCESS;
}

#endif
