/*********************************************************************************
 * type_connection.h ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/

// --------------------------------------------------------------------------------------------
// connection_t sub-types
// --------------------------------------------------------------------------------------------
typedef enum connection_devtype_e
{
    FREE        = 0,                                            // Free connection (not used)
    CAPI        = 1,                                            // CAPI (libcapi, (ISDN)
    VOIP        = 2,                                            // VoIP (pjsip)
    SCRD        = 3                                             // Soundcard (pjsip, soundport)
} 
connection_devtype_e;

typedef enum connection_dir_e
{
    NONE        =   0,                                          // unknown state
    INCOMING    =   1,                                          // Incoming connection
    OUTGOING    =   2,                                          // Outgoing connection
    CLIENT      =   4                                           // Client connection (Registrar-Mode)
}
connection_dir_e;

typedef enum connection_status_e
{
    STATUS_INACTIVE      =   0,
    STATUS_CONNECTED     =   1,
    STATUS_EARLY         =   2,    
    STATUS_CONFIRMED     =   3,
    STATUS_DISCONNECTED  =   4
}
connection_status_e;

typedef struct connection_pools_t                               // Connection memory pools
{
    pj_pool_t           *con_pool;                              // Memory pool for connection lifetime
    pj_pool_t           *play;                                  // Memory pool for playing (filestream.c)
    pj_pool_t           *rec;                                   // Memory pool for recording (filestream.c)

} 
connection_pools_t;

typedef struct connection_capi_t
{
   #if DTMFBOX_USE_CAPI
    _cmsg               cMsgInst;                               // cMsg Instance
    _cmsg               *cMsg;                                  // cMsg Object
    _cword              iMsgNr;                                 // Current Msg-No.
    _cword              hdl;                                    // Count each packet    
   #else
    int                 cMsgInst;
    int                 *cMsg;
    int                 iMsgNr;
    unsigned long       hdl;    
   #endif

   #if DTMFBOX_WIN32
   unsigned short       AppId;                                  // Application ID   
   #else
   unsigned             AppId;                                  // Application ID   
   #endif

   unsigned int         controller;                             // Controller
   unsigned long int    PLCI;                                   // PLCI
   unsigned long int    NCCI;                                   // NCCI   
   int                  SetupAck;                               // Received SetupAck?
   int                  SetupEarly;                             // Received Early?
   int                  SetupInband;                            // Received Inband Progress Indicator?
   int                  SetupListen;                            // Received Listen?
} 
connection_capi_t;

typedef struct connection_transport_t
{
    pjmedia_transport   *g_med_transport;                       // Media stream transport
    pjmedia_session     *g_med_session;                         // Call's media session.
    pjmedia_sock_info   g_med_skinfo;                           // Socket info for media
    pjsip_inv_session   *g_inv;                                 // Invite session    

} connection_transport_t;

typedef struct connection_ports_t
{
    pjmedia_port        *media;                                 // Media-Port (for conference)
    pjmedia_port        *custom;                                // Custom-Port (port_capi.c, port_soundcard.c, ...)
    pjmedia_port        *tone;                                  // Tone-Generator port (fake ringing, -tone)
    pjmedia_port        *rec_resample;                          // Recorder-Port (resample port that is connected over conference bridge)
    pjmedia_port        *rec_stream;                            // Recorder-Port (private/stream)
    pjmedia_port        *play_resample[INCONF_PORTS];           // Player-Port (resample port that is connected over conference bridge)     
    pjmedia_port        *play_stream[INCONF_PORTS];             // Player-Port (private/stream)
}
connection_ports_t;

typedef struct connection_slots_t
{
    int                 media;                                  // Media-Slot (voice) (ports.media)
    int                 tone;                                   // Tone-Generator slot (fake ringing, -tone) (ports.tone)
    int                 play[INCONF_PORTS];                     // Player-Slot   (ports.play_resample[x])
    int                 rec;                                    // Recorder-Slot (ports.rec_resample)
}
connection_slots_t;

typedef struct connection_mixer_t                               // Mixer infos (Talkover or Background port)
{
  int                   is_background[N_CON_PORTS];          
  int                   is_talkover[N_CON_PORTS];
  int                   talkover_active[N_CON_PORTS];
  int                   in_talkover;
} 
connection_mixer_t;

typedef struct connection_info_t
{
    int                 active;                                 // currently sending/receiving audio?
    int                 status;                                 // status: 0=inactive (free), 1=connecting, 2=confirmed, 3=disconnecting
    connection_devtype_e type;                                  // VOIP, CAPI or SCRD connection?    
    int                 account_id;                             // Account-ID used by this connection
    int                 client_id;                              // Client-ID used by this connection (Registrar-Mode)
    connection_dir_e    direction;                              // Incoming, outgoing or client connection
    int                 auto_hookupdown;                        // Automatically confirm & disconnect conference partners
    char                no_target[MAX_URI_LENGTH];              // Target phone number
    char                no_source[MAX_URI_LENGTH];              // Source phone number
    char                no_dialed[MAX_URI_LENGTH];              // Currently dialed no (CAPI_INFO / internal controller)
}
connection_info_t;

typedef struct connection_script_t
{
    char                file[256];                              // Scriptfile
    char                dtmf[256];                              // DTMF-Signals received (max. 256 chars)
    char                timer_scriptfile[256];                  // Timer optionstring
    char                timer_optionstring[512];                // Timer optionstring    
    char                delimiter[2];                           // Delimiter    
    int                 dtmf_pos;                               // DTMF-Position (when using delimiter)
    int                 waitevent;                              // Do not call script, just print output to second instance (block) when 1
    int                 pipe_id;

    pj_timer_entry      timer;                                  // Script timer entry    
}
connection_script_t;


// --------------------------------------------------------------------------------------------
// connection_t main type 
// --------------------------------------------------------------------------------------------
typedef struct connection_t
{   
    unsigned                id;                                 // Array-ID (0-x) for this connection
                
    connection_capi_t       *capi;                              // Capi         
    connection_pools_t      *pools;                             // Memory pools for connection
    connection_ports_t      *ports;                             // Ports    
    connection_slots_t      *slots;                             // Slots
    connection_mixer_t      *mixer;                             // Mixer infos
    connection_info_t       *info;                              // Connection infos
    connection_script_t     *script;                            // Script infos

    connection_transport_t  *rtp;                               // Transport

    int                     in_conf[INCONF_PORTS];              // In conference with - id array
    int                     in_conf_cnt;                        // In conference with - count    
    
    pj_thread_t             *tone_thread;                       // Thread (port_tone_ring_thread)           
} 
connection_t;

