/*********************************************************************************
 * type_settings.h ($Revision: 250 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/

// --------------------------------------------------------------------------------------------
// settings_t sub-types
// --------------------------------------------------------------------------------------------

typedef struct settings_controller_t
{
    connection_devtype_e    device;                             // Device-Type (VOIP/CAPI)
    int                     ctrl[CAPI_CONTROLLERS];             // Controllers for LISTEN_REQ    
    int                     ctrl_out;                           // Default outgoing controller (Registrar-Mode, -call)
} 
settings_controller_t;

typedef struct settings_voip_t
{    
    int                     allow_unknown_acc;                  // Allow unknown accounts?
    int                     registrar;                          // Act as registrar?
    char                    realm[128];                         // Own realm
    char                    ip_addr[128];                       // Public or local ip/host (temporary/updated value)    
    char                    hostname[128];                      // Public or internal ip/host (remembered value)
    char                    stun[128];                          // STUN server (optional)
    long                    stun_port;                          // STUN server port (optional)
    int                     ice;                                // Use ICE (optional, needs STUN!)
    char                    nameserver[4][128];                 // 4 nameservers (optional)    
    long                    udp_port;                           // UDP Server Port (global/changed by STUN)    
    pj_uint16_t             rtp_start;                          // RTP Start-Port    
    int                     rtp_max;                            // RTP Start-Port    
    int                     max_clients;                        // Registrar-Mode: max. no. of clients, that can connect 
    long                    register_interval;                  // Re-Register interval (in sec)        
    long                    keep_alive;                         // Keep-Alive send (ACK) (in sec)
    int                     use_vad;                            // Use silcence detector (VAD)?
    char                    codec_order[6][32];                 // Codec order    
}
settings_voip_t;

typedef struct settings_capi_t
{
    char                    int_prefix[12];                     // International prefix (country code)
    char                    nat_prefix[12];                     // National prefix
    int                     ulaw;                               // A-law (0 default) or U-law (1)
    int                     allow_unknown_acc;                  // Allow unknown accounts?

    settings_controller_t   controller;                         // CAPI-Controller    
}
settings_capi_t;

typedef struct settings_audio_t
{
    int                     conf_clockrate;                     // Conference bridge clockrate 
    int                     conf_samplerate;                    // Conference samplerate    
    int                     null_sound;                         // No audio = 1 
    int                     bidirectional_sound;                // Bidirectional or unidirectional sound?
    int                     snd_play;                           // Soundcard ID (play)
    int                     snd_rec;                            // Soundcard ID (record)
    long                    snd_clockrate;                      // Soundcard clockrate
    int                     resampler_quality;                  // Resampler Quality (0=high, 1=middle, 2=low)
    int                     resampler_quality_conf;             // Resampler Quality (0=high, 1=middle, 2=low)

    long                    snd_ec_tail;                        // Sound echo canceller tail length
    long                    con_ec_tail;                        // Connection echo canceller tail length    

    int                     mixer_tx_volume;                    // Default TX volume
    int                     mixer_rx_volume;                    // Default RX volume
    int                     mixer_talkover;                     // Talkover active by default?
    
    int                     earlyb3_ring;                       // Use fake ring?
}
settings_audio_t;

typedef struct settings_account_t                               // Accounts/Number Settings (dtmfbox.cfg)
{
    pjsip_regc              *regc;                              // regc struct (registrar registration)
    pjsip_cred_info         cred;                               // credential info
    
    settings_controller_t   controller;                         // controller struct for this account

    char                    name[128];                          // account description
    char                    number[128];                        // no. for this account
    int                     active;                             // account active?

    char                    username[128];                      // VOIP: Username
    char                    password[128];                      // VOIP: Passwort
    char                    realm[128];                         // VOIP: Server-Realm
    char                    registrar[128];                     // VOIP: Registrar
    char                    proxy[MAX_URI_LENGTH];              // VOIP: Proxy (optional)
    char                    contact[MAX_URI_LENGTH];            // VOIP: Contact-Header (optional)
    char                    id[MAX_URI_LENGTH];                 // VOIP: ID (optional)    
    int                     do_not_register;                    // VOIP: do not register (only use for outgoing calls)?

    int                     registrar_active;                   // Registrar-Mode: client can login?
    char                    registrar_username[128];            // Registrar-Mode: Username
    char                    registrar_password[128];            // Registrar-Mode: Password
    
    pjsip_route_hdr         route_set;                          // Route-set (proxy)


    int                     status;                             // Status: 0=Pending, 1=Registration Ok!, 2=Registration Error!
    int                     status_publish;                     // Publish status
    char                    status_reason[256];                 // SIP-Status reason (text)
    long                    status_nextreg;                     // Next registration (in sec)
    long                    status_failcnt;                     // Registration fail count
} 
settings_account_t;

typedef struct settings_logging_t
{
    int                     loglevel;                           // Loglevel (0-5)
    int                     show_sender;                        // Show sender
    int                     show_type;                          // Show type
    int                     show_conid;                         // Show connection id
    int                     show_accid;                         // Show account id
    int                     show_src;                           // Show source number
    int                     show_dst;                           // Show target number
    int                     show_ctrl;                          // Show Ctrl.
    int                     show_plci;                          // Show PLCI
    int                     show_ncci;                          // Show NCCI
}
settings_logging_t;

// --------------------------------------------------------------------------------------------
// settings_t main-type
// --------------------------------------------------------------------------------------------

typedef struct settings_t                                       // Application settings (dtmfbox.cfg)
{
    // [general] section    
    char                    plugins[2048];                      // Plugin(s) to call (*.dll, *.so)
    char                    scriptfile[256];                    // Scriptfile to call (*.bat, *.exe)    
    char                    pid_file[256];                      // Pid file
    char                    pipe_dir[256];                      // Pipe directory (Linux only)
    char                    delimiter[2];                       // delimiter for DTMF (1-9, *, #)        
    int                     adjust_priority;                    // Dynamically adjust priority
    int                     max_connections;                    // Max. connections
    
    int                     voip_active;                        // VoIP active?
    int                     use_voip;                           // Use VoIP?
    int                     use_capi;                           // Use CAPI?
    
    // other sections
    settings_logging_t      logging;                            // [logging] section
    settings_audio_t        audio;                              // [audio] section
    settings_capi_t         capi;                               // [capi] section
    settings_voip_t         voip;                               // [voip] section
    settings_account_t      accounts[MAX_ACCOUNTS];             // [accX] section 
} 
settings_t;
