/*********************************************************************************
 * voip_ctrl.c ($Revision: 260 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "voip_ctrl.c"

int voip_init()
{    
    pj_sockaddr_in p_pub_addr;                      // for stun (udp/sip)    
    pj_status_t status;   
    int i, j;        

    #if DTMFBOX_USE_VOIP
       if(oSET.voip_active)    
           app_msg(THIS_FILE, "Init VoIP ...");
    #endif
     
    // Create DNS resolver 
    dns_resolver_create();

    // Create UDP server socket (default: 5060)
    if(udp_server_create(&p_pub_addr))
        return -1;

    // Initialize media endpoint !
    status = pjmedia_endpt_create(&oGLB.cp.factory, NULL, 1, &oGLB.med_endpt);
    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
   
    // Create conference bridge 
    status = pjmedia_conf_create(oGLB.pool, CONF_PORTS, oSET.audio.conf_clockrate, CONF_CHANNELS, oSET.audio.conf_samplerate, CONF_BITS, oSET.audio.resampler_quality_conf | PJMEDIA_CONF_NO_DEVICE, &oGLB.conf.bridge);
    PJ_ASSERT_RETURN(status == PJ_SUCCESS, status);
        
    // Create null port (for master port)
    status = pjmedia_null_port_create(oGLB.pool, oSET.audio.conf_clockrate, CONF_CHANNELS, oSET.audio.conf_samplerate, CONF_BITS, &oGLB.audio.master_null_port); 
    PJ_ASSERT_RETURN(status == PJ_SUCCESS, status);
    
    app_msg(THIS_FILE, "Conference bridge created [%d/%d/%d]", oSET.audio.conf_clockrate, CONF_BITS, CONF_CHANNELS);

    // Initialize connection array
    init_connections();
    
    // Initialize codecs (G711...)
    voip_codecs_init();
    
    // Add null port
    #if (DTMFBOX_USE_NULL_PORT == 1)
    port_null_init();
    #endif

    // Initialize sound devices (as first connection)
    i = connection_by_free();
    status = 1;
    if(i != -1) 
        status = port_soundcard_init(&oCONS[i]);     

    // Get conference master port
    oGLB.conf.bridge_port = pjmedia_conf_get_master_port(oGLB.conf.bridge);    

    // Create master port         
    status = pjmedia_master_port_create(oGLB.pool, oGLB.audio.master_null_port, oGLB.conf.bridge_port, 0, &oGLB.conf.master_port);    
    PJ_ASSERT_RETURN(status == PJ_SUCCESS, status);

    // Start master port
    status = pjmedia_master_port_start(oGLB.conf.master_port);
    PJ_ASSERT_RETURN(status == PJ_SUCCESS, status);

    // Create media transport (RTP/RTCP)
    if(oSET.voip_active)
    {        
        #if DTMFBOX_USE_VOIP
            #if DTMFBOX_USE_ICE

                if(oSET.voip.ice)
                    status = ice_media_transports_create(&p_pub_addr);
                else
                    status = udp_media_transports_create(&p_pub_addr);

            #else
                    status = udp_media_transports_create(&p_pub_addr);
            #endif

            if(status != PJ_SUCCESS)
            {
                app_err(THIS_FILE, "(%d) Unable to create media transports!", status);
                return -1;
            }
        #endif
    }
    
    // Init client list (Registrar-Mode)
    for(i=0; i<MAX_CLIENTS; i++)
    {        
        //oGLB.clients[i].addr = pj_str("");
        pj_memset(oGLB.clients[i].addr, 0, sizeof(oGLB.clients[i].addr));        
        oGLB.clients[i].timeout = 0;
    
        for(j=0; j<MAX_CONTACTS; j++)
            pj_memset(oGLB.clients[i].contacts[j], 0, MAX_URI_LENGTH);

        for(j=0; j<MAX_ACCOUNTS; j++)        
            oGLB.clients[i].account_id[j] = -1;           
    }

    // Start voip_listen() thread    
    status = pj_thread_create(oGLB.pool, "voip_listen_%p", (pj_thread_proc*)&voip_listen, (void*)NULL, (pj_size_t)0, (unsigned)NULL, &oGLB.voip_listen_thread);    
    if(status != PJ_SUCCESS)
    {
        app_err(THIS_FILE, "(%d) Unable to start VoIP listen loop!", status);
        return -1;
    }
    
    #if DTMFBOX_USE_VOIP
        if(oSET.voip_active)    
            app_msg(THIS_FILE, "VoIP initialized!");    
    #endif

    return 0;
}


// Create DNS resolver (max. 4 nameservers).
// Every request will also resolve SRV server records
pj_status_t dns_resolver_create()
{
    pj_status_t status = PJ_SUCCESS;

    #if DTMFBOX_USE_VOIP
        #if PJSIP_HAS_RESOLVER

        if(strlen(oSET.voip.nameserver[0]) || 
           strlen(oSET.voip.nameserver[1]) ||
           strlen(oSET.voip.nameserver[2]) ||
           strlen(oSET.voip.nameserver[3]))
        {
            pj_str_t dns_server[4];
            int i;
            int dns_server_cnt = 0;
            
            // Read nameservers from settings
            for(i=0; i<4; i++)
            {
                if(strlen(oSET.voip.nameserver[i]))
                {
                    dns_server[i] = pj_str(oSET.voip.nameserver[i]);
                    dns_server_cnt++;
                }
            }

            // Create DNS resolver 
	        status = pjsip_endpt_create_resolver(oGLB.endpt, &oGLB.dns_resolver);
	        if (status != PJ_SUCCESS) {
	            app_err(THIS_FILE, "(%d) Error creating resolver", status);
	            return status;
	        }
            
            // Set nameservers
            status = pj_dns_resolver_set_ns(oGLB.dns_resolver, dns_server_cnt, dns_server, NULL);
	        if (status != PJ_SUCCESS) {
	            app_err(THIS_FILE, "(%d) Error setting nameserver", status);
	            return status;
	        }

            // Bind DNS resolver to endpoint
            status = pjsip_endpt_set_resolver(oGLB.endpt, oGLB.dns_resolver);
	        if (status != PJ_SUCCESS) {
	            app_err(THIS_FILE, "(%d) Error setting DNS resolver", status);
	            return status;
	        }
        }

        #endif
    #endif

    return status;
}


pj_status_t udp_server_create(pj_sockaddr_in *p_pub_addr) //pj_sockaddr_in **m_p_pub_addr)
{
    #if DTMFBOX_USE_VOIP

    pj_hostent he;
    pj_sockaddr ip_addr;       
    pj_in_addr *ip_addr_in = &ip_addr.ipv4.sin_addr;
    pjsip_host_port a_name;    
    pj_sockaddr_in bound_addr;    
    pj_str_t address, hostname; 
    pj_status_t status;
    char ip[32];
    
    if(oSET.voip_active)
    {
        // Resolve STUN Server
        oGLB.stun_status = PJ_EUNKNOWN;
        stun_server_resolve(1);

        // Init bound address to ANY
        pj_memset(&bound_addr, 0, sizeof(bound_addr));
        bound_addr.sin_addr.s_addr = PJ_INADDR_ANY;

        // Get local ip
        if(pj_gethostip(pj_AF_INET(), &ip_addr) == PJ_SUCCESS) {
            pj_memset(oGLB.ip_addr_local, 0, sizeof(oGLB.ip_addr_local));
            pj_memset(oGLB.ip_addr_global, 0, sizeof(oGLB.ip_addr_global));            
            strncpy(oGLB.ip_addr_local, pj_inet_ntoa(*ip_addr_in), sizeof(oGLB.ip_addr_local));
            strncpy(oGLB.ip_addr_global, pj_inet_ntoa(*ip_addr_in), sizeof(oGLB.ip_addr_global));
            oGLB.udp_port_local = oSET.voip.udp_port;
            oGLB.udp_port_global = oSET.voip.udp_port;
        }
        else {
            app_err(THIS_FILE, "Cannot resolve local IP!");
            return -1;
        }
       
        // Set IP address, if not already done
        if(strlen(oSET.voip.ip_addr) == 0) {
            strncpy(oSET.voip.ip_addr, oGLB.ip_addr_local, sizeof(oSET.voip.ip_addr));            
        }
        else
        {            
            pj_str_t addr = pj_str(oSET.voip.ip_addr);
            pj_hostent he;

            status = pj_gethostbyname(&addr, &he);
            if(status == PJ_SUCCESS)
            {                   
                if(!strlen(oGLB.ip_addr_local))
                    strncpy(oGLB.ip_addr_local, pj_inet_ntoa(*(pj_in_addr*)he.h_addr), sizeof(oGLB.ip_addr_local));

                strncpy(oGLB.ip_addr_global, pj_inet_ntoa(*(pj_in_addr*)he.h_addr), sizeof(oGLB.ip_addr_global));
            }
        }

        // Create UDP server socket
        status = pj_sock_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, &oGLB.sock);
        if (status != PJ_SUCCESS) {
          app_err(THIS_FILE, "(%d) UDP socket() error", status);
          return status;
        }

        // Bind UDP server socket to port
        status = pj_sock_bind_in(oGLB.sock, pj_ntohl(bound_addr.sin_addr.s_addr), (pj_uint16_t)oSET.voip.udp_port);        
        if (status != PJ_SUCCESS) {
          app_err(THIS_FILE, "(%d) UDP bind() error. Maybe Port %d already in use?", status, oSET.voip.udp_port);
          pj_sock_close(oGLB.sock);
          return status;
        }

        // Change ip of UDP socket
        if(strlen(oSET.voip.ip_addr))
        {
            pj_str_t ip = pj_str(oGLB.ip_addr_global);               
            status = pj_sockaddr_in_set_str_addr(&bound_addr, &ip);
            if(status != PJ_SUCCESS) {
                app_err(THIS_FILE, "(%d) Unable to resolve bound address. Use ANY!", status);
                bound_addr.sin_addr.s_addr = PJ_INADDR_ANY;
            }
        }
    
        // Use STUN-Server
        if(strlen(oSET.voip.stun))
        {                 
             pj_str_t stun1 = pj_str(oSET.voip.stun);
             pj_str_t stun2 = pj_str(oSET.voip.stun);
                      
             stun_server_resolve(1);

             if (oGLB.stun_srv.addr.sa_family != 0) {
	            pj_ansi_strcpy(ip, pj_inet_ntoa(oGLB.stun_srv.ipv4.sin_addr));
	            stun1 = pj_str(ip);
                stun2 = pj_str(ip);                
             }

             app_msg(THIS_FILE, "Resolve IP from %s:%d ...", oSET.voip.stun, oSET.voip.stun_port);             
             status = pjstun_get_mapped_addr(&oGLB.cp.factory, 1, &oGLB.sock,
                             &stun1, 
                             pj_ntohs(oGLB.stun_srv.ipv4.sin_port),
                             &stun2, 
                             pj_ntohs(oGLB.stun_srv.ipv4.sin_port),
                             p_pub_addr);

             if (status != PJ_SUCCESS) 
             {            
                app_err(THIS_FILE, "(%d) STUN resolve error", status);
                
                // Error! Get local IP
                if(pj_gethostip(pj_AF_INET(), &ip_addr) == PJ_SUCCESS)
                {
                    p_pub_addr->sin_addr = *ip_addr_in;
                    p_pub_addr->sin_port = (pj_uint16_t)oGLB.udp_port_local;
                    strncpy(oGLB.ip_addr_global, pj_inet_ntoa(*ip_addr_in), sizeof(oGLB.ip_addr_global));            
                } else {
                    app_err(THIS_FILE, "Cannot resolve IP!");
                    return -1;
                }

             }

             if (status == PJ_SUCCESS)
             {
                 // Set new ip + port for udp socket                  
                 oGLB.udp_port_global = pj_ntohs(p_pub_addr->sin_port);
                 strncpy(oGLB.ip_addr_global, pj_inet_ntoa(p_pub_addr->sin_addr), sizeof(oGLB.ip_addr_global));
             }
        }
        // Search IP (no STUN)
        else
        {
            // Get IP from hostname (Paramter: voip_ip_addr=hostname)
            if(strlen(oSET.voip.hostname))
            {
                app_msg(THIS_FILE, "Resolve IP from host %s ...", oSET.voip.hostname);
                hostname = pj_str(oSET.voip.hostname);

                if(pj_gethostbyname(&hostname, &he) == PJ_SUCCESS)
                {
                    p_pub_addr->sin_addr = *(pj_in_addr*)he.h_addr;  
                    p_pub_addr->sin_port = (pj_uint16_t)oGLB.udp_port_global;
                    strncpy(oGLB.ip_addr_global, pj_inet_ntoa(*(pj_in_addr*)he.h_addr), sizeof(oGLB.ip_addr_global));
                } 
                else 
                {        
                    // Error! Get local IP
                    if(pj_gethostip(pj_AF_INET(), &ip_addr) == PJ_SUCCESS)
                    {
                        p_pub_addr->sin_addr = *ip_addr_in;
                        strncpy(oGLB.ip_addr_global, pj_inet_ntoa(*ip_addr_in), sizeof(oGLB.ip_addr_global));            
                    } else {
                        app_err(THIS_FILE, "Cannot resolve IP!");
                        return -1;
                    }
                }            
            } 
            // Get Host-IP
            else 
            {
                app_msg(THIS_FILE, "Resolve local IP ...");

                if(pj_gethostip(pj_AF_INET(), &ip_addr) == PJ_SUCCESS)
                {
                    p_pub_addr->sin_addr = *ip_addr_in;
                    strncpy(oGLB.ip_addr_global, pj_inet_ntoa(*ip_addr_in), sizeof(oGLB.ip_addr_global));            
                } else {
                    app_err(THIS_FILE, "Cannot resolve IP!");
                    return -1;
                }
            }        
        }

        address = pj_str(oGLB.ip_addr_global); 
        app_msg(THIS_FILE, "Use IP: %s", oGLB.ip_addr_global);    
            
        // Create UDP-Server (default port: 5060)
        {                                
            a_name.host = pj_str(oGLB.ip_addr_global);
            a_name.port = (pj_uint16_t)oGLB.udp_port_global;
            
            status = pjsip_udp_transport_attach(oGLB.endpt, oGLB.sock, &a_name, 1, &oGLB.transport);
            if (status != PJ_SUCCESS) {
                app_err(THIS_FILE, "(%d) Unable to start UDP transport!", status);
                return -1;
            } else {
                app_msg(THIS_FILE, "UDP server listening on port %d", oGLB.udp_port_global);
            }
        }

        // Initialize transaction layer 
        // Hash tables, etc.
        status = pjsip_tsx_layer_init_module(oGLB.endpt);
        PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);

        // Initialize UA layer 
        // Dialog hash tables, etc.
        status = pjsip_ua_init_module(oGLB.endpt, NULL);
        PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);

        // Initialize PRACK support
        status = pjsip_100rel_init_module(oGLB.endpt);
        PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);

        // Initialize invite session module
        // These callbacks will be called on incoming requests, media session state, etc.
        {
            pjsip_inv_callback inv_dtmfbox;
            pj_bzero(&inv_dtmfbox, sizeof(inv_dtmfbox));

            inv_dtmfbox.on_media_update = &call_on_media_update;
            inv_dtmfbox.on_new_session = &call_on_forked;
            inv_dtmfbox.on_state_changed = &call_on_state_changed;
            
            // Initialize session invite module 
            status = pjsip_inv_usage_init(oGLB.endpt, &inv_dtmfbox);
            PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);

            // Register modules
            status = pjsip_endpt_register_module(oGLB.endpt, &mod_dtmfbox);
            PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
            
            status = pjsip_endpt_register_module(oGLB.endpt, &mod_dtmfbox_registrar);
            PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
        }


        // Add endpoint capabilities (INFO, OPTIONS, etc) for this UA
        {   
            pj_str_t allowed[] = { {"INFO", 4}, {"REGISTER", 8 } };  //  //{"INVITE", 6}, {"ACK",3}, {"BYE",3}, {"CANCEL",6},  {"OPTIONS", 7}, 
            pj_str_t accepted = { "application/sdp", 15 };

            // Register supported methods
            pjsip_endpt_add_capability(oGLB.endpt, &mod_dtmfbox, PJSIP_H_ALLOW, NULL, PJ_ARRAY_SIZE(allowed), allowed);
            
            // Register "application/sdp" in ACCEPT header
            pjsip_endpt_add_capability(oGLB.endpt, &mod_dtmfbox, PJSIP_H_ACCEPT, NULL, 1, &accepted);            
        }
    }

    #else
    PJ_UNUSED_ARG(p_pub_addr);
    #endif

    return PJ_SUCCESS;
}


pj_status_t voip_codecs_init()
{    
    pj_status_t status = PJ_SUCCESS;

    #if DTMFBOX_USE_VOIP    
    //unsigned int count = 1;
    int i;

    if(oSET.voip_active)
    {
        // Codecs:
        for(i=0; i<=5; i++)
        {
            //pj_str_t codec_id = pj_str(oSET.voip.codec_order[i]);

            // iLBC codec
            #if defined(PJMEDIA_HAS_ILBC_CODEC) && PJMEDIA_HAS_ILBC_CODEC!=0            
                if(!strcmp(oSET.voip.codec_order[i], "ilbc")) 
                {
                    pjmedia_codec_ilbc_init(oGLB.med_endpt, 20);
                    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
                }
            #endif

            // GSM codec
            #if defined(PJMEDIA_HAS_GSM_CODEC) && PJMEDIA_HAS_GSM_CODEC!=0            
                if(!strcmp(oSET.voip.codec_order[i], "gsm")) {
                    pjmedia_codec_gsm_init(oGLB.med_endpt);
                    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
                }
            #endif

            // PCMU/PCMA (G.711) codec 
            #if defined(PJMEDIA_HAS_G711_CODEC) && PJMEDIA_HAS_G711_CODEC!=0                        
                if(!strcmp(oSET.voip.codec_order[i], "g711")) {                    
                    status = pjmedia_codec_g711_init(oGLB.med_endpt);
                    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
                }
            #endif
          
            // L16 codec       
            #if defined(PJMEDIA_HAS_L16_CODEC) && PJMEDIA_HAS_L16_CODEC!=0                        
                if(!strcmp(oSET.voip.codec_order[i], "l16")) {
                    pjmedia_codec_l16_init(oGLB.med_endpt, 0);
                    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);              
                }
            #endif

            // Speex codec       
            #if defined(PJMEDIA_HAS_SPEEX_CODEC) && PJMEDIA_HAS_SPEEX_CODEC!=0                        
                if(!strcmp(oSET.voip.codec_order[i], "speex")) {
                    pjmedia_codec_speex_init(oGLB.med_endpt, 0, -1, -1);
                    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
                }
            #endif
        }
    }
    /*
    // First reset codec priority
    {
        pj_str_t codec_id = pj_str("");
        pjmedia_codec_mgr_set_codec_priority(pjmedia_endpt_get_codec_mgr(oGLB.med_endpt), &codec_id, PJMEDIA_CODEC_PRIO_DISABLED);
    }

    // Set codec priority
    for(i=5; i<=0; --i) 
    {
        if(strlen(oSET.voip.codec_order[i]))
        {                
            pj_str_t codec_id = pj_str(oSET.voip.codec_order[i]);
            pjmedia_codec_mgr_set_codec_priority(pjmedia_endpt_get_codec_mgr(oGLB.med_endpt), &codec_id, PJMEDIA_CODEC_PRIO_NEXT_HIGHER);
        }
    }
    */
    #endif

    return status;
}

#if DTMFBOX_USE_VOIP

pj_status_t stun_server_resolve(int wait)
{
    pj_str_t stun_adr;
    pj_hostent he;

    if(oGLB.stun_status == PJ_EUNKNOWN) 
    {
        // Initialize STUN configuration
        pj_stun_config_init(&oGLB.stun_cfg, &oGLB.cp.factory, 0,  pjsip_endpt_get_ioqueue(oGLB.endpt), pjsip_endpt_get_timer_heap(oGLB.endpt));

        oGLB.stun_status = PJ_EPENDING;

        if(strlen(oSET.voip.stun))
        {
            // Init STUN socket
            stun_adr = pj_str(oSET.voip.stun);
            oGLB.stun_status = pj_sockaddr_in_init(&oGLB.stun_srv.ipv4, &stun_adr, (pj_uint16_t)oSET.voip.stun_port);

            if (oGLB.stun_status != PJ_SUCCESS) {
    		    oGLB.stun_status = pj_gethostbyname(&stun_adr, &he);

                if (oGLB.stun_status == PJ_SUCCESS) {
		            pj_sockaddr_in_init(&oGLB.stun_srv.ipv4, NULL, 0);
		            oGLB.stun_srv.ipv4.sin_addr = *(pj_in_addr*)he.h_addr;
		            oGLB.stun_srv.ipv4.sin_port = pj_htons((pj_uint16_t)oSET.voip.stun_port);
		        }
            }
        }
        else
        {
            oGLB.stun_status = PJ_SUCCESS;
        }

        return oGLB.stun_status;
    }
    else if(oGLB.stun_status == PJ_EPENDING) {
        if(wait)
        {
            while(oGLB.stun_status == PJ_EPENDING)
                voip_handle_events(10);
        }

        return oGLB.stun_status;
    }
    else
    {
        return oGLB.stun_status;
    }
  
}


#if DTMFBOX_USE_ICE

static void on_ice_complete(pjmedia_transport *tp, 
			    pj_ice_strans_op op,
			    pj_status_t result)
{
    PJ_UNUSED_ARG(result);
    PJ_UNUSED_ARG(op);
    PJ_UNUSED_ARG(tp);
}

pj_status_t ice_media_transports_create(pj_sockaddr_in *p_pub_addr)
{    
    char stunip[PJ_INET6_ADDRSTRLEN];
    pj_ice_strans_cfg ice_cfg;
    pj_status_t status;
    int i;        

    PJ_UNUSED_ARG(p_pub_addr);

    // Resolve stun ip 
    stun_server_resolve(1);

    pj_ice_strans_cfg_default(&ice_cfg);
    pj_stun_config_init(&ice_cfg.stun_cfg, &oGLB.cp.factory, 0,
	                     pjsip_endpt_get_ioqueue(oGLB.endpt), pjsip_endpt_get_timer_heap(oGLB.endpt));

    ice_cfg.af = pj_AF_INET();
    ice_cfg.resolver = oGLB.dns_resolver;

    if(pj_sockaddr_has_addr(&oGLB.stun_srv))
    {
        pj_sockaddr_print(&oGLB.stun_srv, stunip, sizeof(stunip), 0);
    	ice_cfg.stun.server = pj_str(stunip);
    	ice_cfg.stun.port = pj_sockaddr_get_port(&oGLB.stun_srv);
    }
   	
    ice_cfg.stun.no_host_cands = PJ_FALSE;//pjsua_var.media_cfg.ice_no_host_cands;
        
    for(i=0; i<oSET.voip.rtp_max; i++)
    {
        // create ice media transport
        pjmedia_ice_cb ice_cb;
        char name[32];

        pj_bzero(&ice_cb, sizeof(pjmedia_ice_cb));
        ice_cb.on_ice_complete = &on_ice_complete;
        pj_ansi_snprintf(name, sizeof(name), "icetp%02d", i);

        status = pjmedia_ice_create(oGLB.med_endpt, name, 2,
                &ice_cfg, 
				&ice_cb,
                &oRTP[i]->g_med_transport);        

        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to create ICE",  status);
            return 1;
        }
        
    }

    return PJ_SUCCESS;
}
#endif

pj_status_t udp_media_transports_create(pj_sockaddr_in *p_pub_addr)
{
    pj_sockaddr_in p_pub_addr_rtp[2];        // for stun (rtp/rtcp)
    pj_status_t status;
    int i, j;
    
    // Init oGLB.stun_cfg
    //pj_stun_config_init(&oGLB.stun_cfg, &oGLB.cp.factory, 0,  pjsip_endpt_get_ioqueue(oGLB.endpt), pjsip_endpt_get_timer_heap(oGLB.endpt));
    
    // Create UDP media transport         
    j = 0;
    for(i=0; i<((oSET.voip.rtp_max)*2); i+=2)
    {        
        pj_sockaddr_in *ipv4_rtp;
        pj_sockaddr_in *ipv4_rtcp;

        ipv4_rtp = &oRTP[j]->g_med_skinfo.rtp_addr_name.ipv4;
        ipv4_rtcp = &oRTP[j]->g_med_skinfo.rtcp_addr_name.ipv4;

        // Init socket
        pj_sockaddr_in_init(ipv4_rtp, pjsip_endpt_name(oGLB.endpt), (pj_uint16_t)(oSET.voip.rtp_start + i));
        pj_sockaddr_in_init(ipv4_rtcp, pjsip_endpt_name(oGLB.endpt), (pj_uint16_t)(oSET.voip.rtp_start + i + 1));
        ipv4_rtp->sin_addr.s_addr = PJ_INADDR_ANY;
        ipv4_rtcp->sin_addr.s_addr = PJ_INADDR_ANY;
        ipv4_rtp->sin_port = pj_htons((pj_uint16_t)(oSET.voip.rtp_start + i));
        ipv4_rtcp->sin_port = pj_htons((pj_uint16_t)(oSET.voip.rtp_start + i + 1));
                
        // Change bind address (public ip, no stun!)
        if(strlen(oSET.voip.hostname) && !strlen(oSET.voip.stun)) {
            ipv4_rtp->sin_addr.s_addr  = p_pub_addr->sin_addr.s_addr;
            ipv4_rtcp->sin_addr.s_addr = p_pub_addr->sin_addr.s_addr;
        }
        
        // Create RTP/RTCP sockets (this will be bound to public or internal ip)
        status = pj_sock_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, &oRTP[j]->g_med_skinfo.rtp_sock);
        PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);        
    
        status = pj_sock_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, &oRTP[j]->g_med_skinfo.rtcp_sock);
        PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);

        // Bind RTP/RTCP socket        
        status = pj_sock_bind_in(oRTP[j]->g_med_skinfo.rtp_sock, pj_ntohl(ipv4_rtp->sin_addr.s_addr), (pj_uint16_t)(oSET.voip.rtp_start + i));
        if (status != PJ_SUCCESS) {
            app_err(THIS_FILE, "(%d) Unable to bind RTP socket",  status);
            return 1;
        }
       
        status = pj_sock_bind_in(oRTP[j]->g_med_skinfo.rtcp_sock, pj_ntohl(ipv4_rtcp->sin_addr.s_addr), (pj_uint16_t)(oSET.voip.rtp_start + i + 1));
        if (status != PJ_SUCCESS) {
            app_err(THIS_FILE, "(%d) Unable to bind RTCP socket",  status);
            return 1;
        }

        // Use STUN-Server?
        if(strlen(oSET.voip.stun))
        {        
             pj_sock_t socks[2];
             pj_str_t stun1 = pj_str(oSET.voip.stun);
             pj_str_t stun2 = pj_str(oSET.voip.stun);
             socks[0] = oRTP[j]->g_med_skinfo.rtp_sock;
             socks[1] = oRTP[j]->g_med_skinfo.rtcp_sock;
             
             status = pjstun_get_mapped_addr(&oGLB.cp.factory, 2, socks,
                             &stun1, 
                             oSET.voip.stun_port,
                             &stun2, 
                             oSET.voip.stun_port,
                             p_pub_addr_rtp);

             if(status == PJ_SUCCESS)
             {                 
                  // Modify RTP/RTCP address
                  oRTP[j]->g_med_skinfo.rtp_sock = socks[0];
                  pj_memcpy(&oRTP[j]->g_med_skinfo.rtp_addr_name, &p_pub_addr_rtp[0], sizeof(pj_sockaddr_in));
            
                  oRTP[j]->g_med_skinfo.rtcp_sock = socks[1];
                  pj_memcpy(&oRTP[j]->g_med_skinfo.rtcp_addr_name, &p_pub_addr_rtp[1], sizeof(pj_sockaddr_in));
             }
        } 
              
        // Attach UDP port
        status = pjmedia_transport_udp_attach(oGLB.med_endpt, NULL, &oRTP[j]->g_med_skinfo, 0, &oRTP[j]->g_med_transport);
        if (status != PJ_SUCCESS) 
        {
            app_err(THIS_FILE, "(%d) Unable to create media transport!", status, oSET.voip.rtp_start + i, oSET.voip.rtp_start + i + 1);
            return 1;
        }                   
        // Get socket info from every media transport 
        else
        {        
            //pjmedia_sock_info udp_info;
            pjmedia_transport_info udp_info;
            pjmedia_transport_get_info(oRTP[j]->g_med_transport, &udp_info);
            app_dbg(THIS_FILE, "RTP/RTCP: %s:%d-%d", pj_inet_ntoa(udp_info.sock_info.rtp_addr_name.ipv4.sin_addr), pj_ntohs(udp_info.sock_info.rtp_addr_name.ipv4.sin_port), pj_ntohs(udp_info.sock_info.rtcp_addr_name.ipv4.sin_port));
        }

        j++;
    }

    app_msg(THIS_FILE, "RTP/RTCP media transport created", oSET.voip.rtp_start, oSET.voip.rtp_start + i - 1 );    

    return PJ_SUCCESS;
}
#endif

int voip_close()
{    
    int i;

    #if DTMFBOX_USE_VOIP
        if(oSET.voip_active)
            app_msg(THIS_FILE, "Destroy VoIP ...");
    #endif            
    
    // disconnect every active connection
    if(oGLB.conf.bridge)
    {
        for(i=0; i<oSET.max_connections; i++)
        {
            if(oCONS[i].info->type == VOIP)        
                voip_hangup(&oCONS[i]);            
                    
            reset_connection(&oCONS[i]);
        }
    }
        
    if(oGLB.endpt != NULL) 
    {
        // Cancel keep alive timer
        pjsip_endpt_cancel_timer(oGLB.endpt, &oGLB.keep_alive_timer);
        pjsip_endpt_cancel_timer(oGLB.endpt, &oGLB.registrar_refresh_timer);

        // Unregister (all) credentials
        voip_client_register_deinit(-1);
    }

    // Close soundcard port
    if(oCONS != NULL)
        if(oCONS[0].ports->custom != NULL) {
            port_soundcard_close(&oCONS[0]);            
        }

    // Destroy process communication thread
    if(oGLB.signal_handler_thread != NULL) {
        pj_thread_join(oGLB.signal_handler_thread);
        pj_thread_destroy(oGLB.signal_handler_thread);
    }
    
    #if (DTMFBOX_USE_NULL_PORT == 1)
        // close null port
        if(oGLB.audio.flow_null_port != NULL)
            port_null_close();
    #endif
        
    // Destroy master port    
    if(oGLB.conf.master_port)
        pjmedia_master_port_destroy(oGLB.conf.master_port, 0);

    // Destroy master null port
    if(oGLB.audio.master_null_port)
        pjmedia_port_destroy(oGLB.audio.master_null_port);

    // Destroy conference
    if(oGLB.conf.bridge)
        pjmedia_conf_destroy(oGLB.conf.bridge);

    // Destroy transports
    if(oRTP != NULL)
        for(i=0; i<oSET.voip.rtp_max; i++)
        {
            // Close ICE media session
            #if DTMFBOX_USE_ICE
            if(oRTP[i]->g_med_transport) {
                pjmedia_transport_media_stop(oRTP[i]->g_med_transport);
            }
            #endif

            // Close UDP media socket      
            if(!oSET.voip.ice && oRTP[i]->g_med_transport)
                pjmedia_transport_close(oRTP[i]->g_med_transport);        
        }

    // Free connection memory pools
    if(oCONS != NULL)
        for(i=0; i<oSET.max_connections; i++)
        {
            if(oCONS[i].pools->con_pool)
                pj_pool_release(oCONS[i].pools->con_pool);
        }

    // Deinit codecs
    #if DTMFBOX_USE_VOIP
        
         // Codecs:
        for(i=0; i<6; i++)
        {
            // iLBC codec            
            #if defined(PJMEDIA_HAS_ILBC_CODEC) && PJMEDIA_HAS_ILBC_CODEC!=0
                if(!strcmp(oSET.voip.codec_order[i], "ilbc"))
                    pjmedia_codec_ilbc_deinit();
            #endif

            // GSM codec
            #if defined(PJMEDIA_HAS_GSM_CODEC) && PJMEDIA_HAS_GSM_CODEC!=0            
                if(!strcmp(oSET.voip.codec_order[i], "gsm"))
                    pjmedia_codec_gsm_deinit();
            #endif

            // PCMU/PCMA (G.711) codec 
            #if defined(PJMEDIA_HAS_G711_CODEC) && PJMEDIA_HAS_G711_CODEC!=0            
                if(!strcmp(oSET.voip.codec_order[i], "g711"))
                    pjmedia_codec_g711_deinit();            
            #endif
            
            // L16 codec        
            #if defined(PJMEDIA_HAS_L16_CODEC) && PJMEDIA_HAS_L16_CODEC!=0                                    
                if(!strcmp(oSET.voip.codec_order[i], "l16"))
                    pjmedia_codec_l16_deinit();
            #endif

            // Speex codec        
            #if defined(PJMEDIA_HAS_SPEEX_CODEC) && PJMEDIA_HAS_SPEEX_CODEC!=0                                    
                if(!strcmp(oSET.voip.codec_order[i], "speex"))
                    pjmedia_codec_speex_deinit();
            #endif
        }

    #endif    

    // Destroy media endpoint    
    if(oGLB.med_endpt)
        pjmedia_endpt_destroy(oGLB.med_endpt);
    
    #if DTMFBOX_USE_VOIP
        // Stop VoIP listen loop        
        if(oGLB.voip_listen_thread != NULL)
        {
            oGLB.closing = 2;
            pj_thread_join(oGLB.voip_listen_thread);
            pj_thread_destroy(oGLB.voip_listen_thread);
        }
    #endif

    // Destroy global endpoint
    if(oGLB.endpt) 
        pjsip_endpt_destroy(oGLB.endpt);
    
    // Destroy mutexes
    if (oGLB.mutex) {        
	    pj_mutex_destroy(oGLB.mutex);
	    oGLB.mutex = NULL;
    }
    if(oGLB.acc_mutex) {
        pj_mutex_destroy(oGLB.acc_mutex);
        oGLB.acc_mutex = NULL;
    }

    // Free memory pool!
    if(oGLB.pool) {
        pj_pool_release(oGLB.pool);       
        pj_caching_pool_destroy(&oGLB.cp);
    }
    
    #if DTMFBOX_USE_VOIP
        if(oSET.voip_active)
            app_msg(THIS_FILE, "VoIP destroyed!");
    #endif

    return 0;
}


void voip_handle_events(int msec)
{
    pj_time_val timeout;
    unsigned int count = 0;
    
    timeout.msec = msec;
    timeout.sec = 0;

    pjsip_endpt_handle_events2(oGLB.endpt, &timeout, &count); 

   // if(count > 1)
     //   app_msg(THIS_FILE, "COUNT : %d", count);
}

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//
// voip_listen loop (thread)
//
// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void voip_listen()
{
    #if DTMFBOX_USE_VOIP
        if(oSET.voip_active)
            app_dbg(THIS_FILE, "Start VoIP listen loop ..."); 
    #endif

    // Loop & handle messages    
    while(oGLB.closing != 2 && oGLB.endpt != NULL && oGLB.pool != NULL) {
        voip_handle_events(1);  // 1ms :)
    }
    
    #if DTMFBOX_USE_VOIP
        if(oSET.voip_active)
            app_dbg(THIS_FILE, "VoIP listen loop stopped!"); 
    #endif
}
