/*********************************************************************************
 * voip_events.c ($Revision: 250 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "voip_events.c"

#if DTMFBOX_USE_VOIP

// DTMF callback from the stream
void dtmf_callback(pjmedia_stream *strm, void *user_data, int digit)
{
    unsigned id = (int)user_data;
    PJ_UNUSED_ARG(strm);

    voip_get_dtmf((char)digit, 0, id);
}

//
// Callback when INVITE session state has changed.
// This callback is registered when the invite session module is initialized.
// Here we get the information about the session state (disconnected, connecting, confirmed, early)
//
void call_on_state_changed( pjsip_inv_session *inv, pjsip_event *e)
{
    char info[256];
    int i, j;
    
    PJ_UNUSED_ARG(e);

    // Get the invite session from connections
    i = connection_by_invite(inv);

    // Connection-Info
    if(i != -1)
        connection_print_info(&oCONS[i], info, sizeof(info));
    else
        memset(info, 0, sizeof(info));

    switch(inv->state)
    {       
        // ----------------------------------------------------------
        // State: CONNECTING
        // ----------------------------------------------------------
        case PJSIP_INV_STATE_CONNECTING:
            app_msg(THIS_FILE, "%s Call CONNECTING (reason=%d, %s)", info, inv->cause, pjsip_get_status_text(inv->cause)->ptr);
            break;

        // ----------------------------------------------------------
        // State: EARLY
        // ----------------------------------------------------------
        case PJSIP_INV_STATE_EARLY:
            app_msg(THIS_FILE, "%s Call EARLY (reason=%d, %s)", info, inv->cause, pjsip_get_status_text(inv->cause)->ptr);
            
            // Script event 
            if(i != -1) {                
                // no client (CONNECT), incoming only
                if(!(oCONS[i].info->direction & CLIENT)) {
                    script_event_add(VOIP, CONNECT, oCONS[i].script->file, oCONS[i].id, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");
                }

                // client (DDI/EARLY), outgoing only
                else if((oCONS[i].info->direction & CLIENT) && !(oCONS[i].info->direction & INCOMING)) 
                {
                    unsigned j;
                    char to[MAX_URI_LENGTH];

                    // CONNECT
                    script_event_add(VOIP, CONNECT, oCONS[i].script->file, i, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");

                    // DDI
                    strncpy(to, oCONS[i].info->no_target, sizeof(to));
                    sip_strip_user(to);
                    for(j=0; j<strlen(to); j++)
                        voip_get_dtmf(to[j], 1, i);                

                    // EARLY
                    if(!oCONS[i].capi->SetupEarly) {
                        script_event_add(VOIP, EARLY, oCONS[i].script->file, oCONS[i].id, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");
                        oCONS[i].capi->SetupEarly = 1;
                    }
                                                         
                    //pj_thread_sleep(2000);
                }
            }
            break;

        // ----------------------------------------------------------
        // State: CONFIRMED
        // ----------------------------------------------------------
        case PJSIP_INV_STATE_CONFIRMED:
            app_msg(THIS_FILE, "%s Call CONFIRMED (reason=%d, %s)", info, inv->cause, pjsip_get_status_text(inv->cause)->ptr);
            if(i != -1)
            {
                // Status: CONFIRMED
                oCONS[i].info->status = 2;

                // Initialize new tone port
                if(oCONS[i].ports->tone == NULL)            
                    port_tone_init(&oCONS[i], 0, 0, 0, 0, 0);
                 
                for(j=0; j<INCONF_PORTS; j++)
                {
                    int in_conf = oCONS[i].in_conf[j];

                    // Registrar-Mode: hook up client?
                    if(in_conf != -1) 
                    {         
                        if(oCONS[i].info->direction & CLIENT || (oCONS[in_conf].info->direction & CLIENT && oCONS[i].info->active))
                        {                     
                            if(oCONS[in_conf].info->type == VOIP)
                            {
                                if(!(oCONS[in_conf].info->direction & OUTGOING) ||
                                     oCONS[i].info->status == 2 && !(oCONS[i].info->direction & CLIENT)) 
                                {
                                    voip_confirm_call(&oCONS[in_conf]);
                                } 
                            }
                       
                            if(oCONS[in_conf].info->type == CAPI) {
                                capi_confirm_call(&oCONS[in_conf]);
                            }

                            // disconnect all inactive clients
                            if(!(oCONS[in_conf].info->direction & CLIENT)) 
                                voip_registrar_disconnect_clients(&oCONS[in_conf], 1);
                            if(!(oCONS[i].info->direction & CLIENT)) 
                                voip_registrar_disconnect_clients(&oCONS[i], 1);
                        }                       
                    }
               }
                
               // SCRIPT: "CONFIRMED"
               script_event_add(VOIP, CONFIRMED, oCONS[i].script->file, i, oCONS[i].in_conf[0], oCONS[i].info->no_source, oCONS[i].info->no_target, "");

               if(oCONS[i].info->auto_hookupdown)
                    script_auto_hookup(i);
            }
            break;

        // ----------------------------------------------------------
        // State: DISCONNECTED
        // ----------------------------------------------------------
        case PJSIP_INV_STATE_DISCONNECTED:
            app_msg(THIS_FILE, "%s Call DISCONNECTED (reason=%d, %s)", info, inv->cause, pjsip_get_status_text(inv->cause)->ptr);

            if(i!=-1)
            {        
                // Registrar-Mode: disconenct every client
                voip_registrar_disconnect_clients(&oCONS[i], 0);

                // hangup current call                                
                voip_hangup(&oCONS[i]);
                
                // Reset transport
                oCONS[i].rtp->g_inv = NULL;
                oCONS[i].rtp->g_med_session = NULL;
                oCONS[i].rtp = NULL;

                if(oCONS[i].info->auto_hookupdown)
                    script_auto_hookdown(i);

                // Reset current connection
                reset_connection(&oCONS[i]);                                    
            }

            break;

        // ----------------------------------------------------------
        // State: unknown
        // ----------------------------------------------------------
        default:
            app_msg(THIS_FILE, "%s Call state changed to %s", info, pjsip_inv_state_name(inv->state));
            break;
    }
}


//
// Callback when SDP negotiation has completed.
// This callback initialize the media as soon the negotation has completed.
//
void call_on_media_update(pjsip_inv_session *inv, pj_status_t status)
{
    pjmedia_session_info sess_info;
    const pjmedia_sdp_session *local_sdp;
    const pjmedia_sdp_session *remote_sdp;
    unsigned int i, j;

    app_dbg(THIS_FILE, "On media update");

    if (status != PJ_SUCCESS)
    {
        app_err(THIS_FILE, "(%d) SDP negotiation has failed", status);        
        return;
    }

    // Get invite session from connections
    i = connection_by_invite(inv);
    if(i == -1)
    {
        app_err(THIS_FILE, "Can not find invite session");        
        return;
    }

    // mhh...
    if(oCONS[i].info->active == 1)
    {
        // app_err(THIS_FILE, "(%d) already connected", status);
        // return;
    }

    // Get local and remote SDP.
    status = pjmedia_sdp_neg_get_active_local(inv->neg, &local_sdp);
    status = pjmedia_sdp_neg_get_active_remote(inv->neg, &remote_sdp);

    // Create session info
    status = pjmedia_session_info_from_sdp(inv->pool, oGLB.med_endpt, 1, &sess_info, local_sdp, remote_sdp);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create media session", status);        
        return;
    }


#if DTMFBOX_USE_ICE
     status = pjmedia_transport_media_start(oCONS[i].rtp->g_med_transport, inv->pool, local_sdp, remote_sdp, 0); 
     if(status != PJ_SUCCESS) {
            app_err(THIS_FILE, "(%d) Unable to start media transport", status);
     }
#endif


    // Stream settings
    for(j=0; j<sess_info.stream_cnt; j++)
    {
        char info[512];
        int len;

        // Display current codec
        pj_memset(info, 0, sizeof(info));
        len = pj_ansi_snprintf(info, sizeof(info), "stream #%d: %.*s (%d/%d)", j, (int)sess_info.stream_info[j].fmt.encoding_name.slen, sess_info.stream_info[j].fmt.encoding_name.ptr, sess_info.stream_info[j].fmt.clock_rate, sess_info.stream_info[j].fmt.channel_cnt);
        app_dbg(THIS_FILE, "%s", info);

        // Disable VAD for each stream (silence detection)
        if(!oSET.voip.use_vad)
        {
            app_msg(THIS_FILE, "stream #%d: disable VAD [port=%d]", j, i);
            sess_info.stream_info[j].param->setting.vad = 0;
        }
/*
// PTIME
        sess_info.stream_info[j].param->setting.frm_per_pkt = (pj_uint8_t)
		1;
	    if (sess_info.stream_info[j].param->setting.frm_per_pkt == 0)
		sess_info.stream_info[j].param->setting.frm_per_pkt = 1;
*/
        // >> CHANGE OTHER CODEC AND STREAM SETTINGS HERE <<
    }


    // Create new media session. It is active immediately.
    status = pjmedia_session_create(oGLB.med_endpt, &sess_info, &oCONS[i].rtp->g_med_transport, NULL, &oCONS[i].rtp->g_med_session);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create media session", status);        
        return;
    }
    
    // Get the media port interface from the first stream in the session, so we
    // have a basicly struct containing get_frame/put_frame.
    // This media port can be attached to the conference bridge or to a sound player/recorder device.
    // It is also needed by capi_streaming to stream capi<->voip.
    status = pjmedia_session_get_port(oCONS[i].rtp->g_med_session, 0, &oCONS[i].ports->media);
    
    if(status == PJ_SUCCESS)
    {    
        // Connection is ACTIVE and CONFIRMED
        oCONS[i].info->active = 1;
        oCONS[i].info->status = 2;
        
        // Wait for ringing thread 	
        if(oCONS[i].tone_thread != NULL && oCONS[i].ports->tone != NULL) {
            pj_thread_join(oCONS[i].tone_thread);
            pj_thread_destroy(oCONS[i].tone_thread);
            oCONS[i].tone_thread = NULL;
        }
                        
        // Initialize custom voip_port (second time) (for putframe/getframe)            
        // Now we have a media_port!
        port_voip_init(&oCONS[i]);

        // Connect conference partner, if there is one (media-streaming: voip<->voip)
        for(j=0; j<INCONF_PORTS; j++)
        {
          if(oCONS[i].in_conf[j] != -1)
          {
              mixer_conf_port_connect(&oCONS[i], &oCONS[oCONS[i].in_conf[j]], 1);
              mixer_conf_port_connect(&oCONS[oCONS[i].in_conf[j]], &oCONS[i], 1);
          }
        }        
    }
    else {
       app_err(THIS_FILE, "(%d) Unable to get media port!", status);       
       return;
    }

    // Set DTMF callback
    if(oCONS[i].rtp->g_med_session) 
    {
        status = pjmedia_session_set_dtmf_callback(oCONS[i].rtp->g_med_session, 0, &dtmf_callback, (void*)i);
     
        if(status != PJ_SUCCESS)
            app_err(THIS_FILE, "(%d) Unable to initialize dtmf callback!", status);            
    }
}   

// This callback is called when dialog has forked.
void call_on_forked(pjsip_inv_session *inv, pjsip_event *e)
{
    PJ_UNUSED_ARG(inv);
    PJ_UNUSED_ARG(e);
}

/*
// This callback is called, when tsx session-state has changed.
// At the moment, we use it for a workaround to receive DTMF by PJSIP_OTHER_METHOD
void call_on_tsx_changed(pjsip_inv_session *inv, pjsip_transaction *tsx, pjsip_event *e)
{    
    pjsip_method *method;
    int con_id = 0;
    pj_status_t status;

    PJ_UNUSED_ARG(inv);

    method = &tsx->method;
    con_id = connection_by_invite(inv);

    // DTMF workaround (some providers)
    //
    // we receive OTHER method with status_code=0 on DTMF so PRI disconnects :(
    // we send 200 response and call the dtmf-callback manually...
    if(method->id == PJSIP_OTHER_METHOD && tsx->status_code == 0)
    {
        char info_msg[128];
        pjsip_dialog *dlg;
        pjsip_tx_data *tdata;
        const pj_str_t reason = { "DTMF", 4 };

        // Get dialog
        dlg = pjsip_tsx_get_dlg(tsx);

        // Send 200 Ok response (so we do not get disconnected)
        if(e->body.rx_msg.rdata->msg_info.msg->type == PJSIP_REQUEST_MSG)
        {
            tsx->status_code = 200;
            status = pjsip_dlg_create_response(dlg, e->body.rx_msg.rdata, 200, &reason, &tdata);
            if (status == PJ_SUCCESS) {
                status = pjsip_dlg_send_response(dlg, tsx, tdata);
            }

            // Fetch body and read signal
            // pass dtmf to callback-function
            pjsip_print_text_body(e->body.rx_msg.rdata->msg_info.msg->body, (char*)&info_msg, sizeof(info_msg));
            if(strstr(info_msg, "Signal=") && strlen(info_msg) >= 8)
            {
                char dtmf[2];
                dtmf[0] = info_msg[7];
                dtmf_callback(NULL, (void*)con_id, (int)dtmf[0]);
            }
        }
    }    
}
*/

void registrar_connect_outgoing(void* args)
{   
    char SourceNo[MAX_URI_LENGTH];
    char TargetNo[MAX_URI_LENGTH];    
    int con_id;
    int acc_id;
    int client_id;
    int con_id_client;

    // Get data from temporary connection
    con_id = (int)args;
    pj_memcpy(SourceNo, oCONS[con_id].info->no_source, sizeof(SourceNo));    
    pj_memcpy(TargetNo, oCONS[con_id].info->no_target, sizeof(TargetNo));    
    acc_id = oCONS[con_id].info->account_id;
    client_id = oCONS[con_id].info->client_id;
    con_id_client = oCONS[con_id].in_conf[0];
    
    if(con_id == -1 || con_id_client == -1 || client_id == -1)
        return;

    // Wait some time
    pj_thread_sleep(2000);
    
    // Connection not up? Then resume calling...
    if(!oCONS[con_id_client].info->active && oCONS[con_id_client].info->status == 1) 
    {
        // Free temporary connection
        reset_connection(&oCONS[con_id]);

        // Make call...
        voip_registrar_connect_clients(&oCONS[con_id_client], SourceNo, TargetNo, acc_id, client_id, oSET.accounts[acc_id].controller.device, OUTGOING);
    }
}

//
// Callback when incoming requests outside any transactions and any
// dialogs are received.
// We're only interested to hande incoming INVITE, REGISTER and OTHER
// request, and we'll reject any other requests with 500 response.
//
pj_bool_t on_rx_request(pjsip_rx_data *rdata)
{   
    char from[MAX_URI_LENGTH];
    char to[MAX_URI_LENGTH];
    int accountid = -1;
    int clientid = -1;
    int incoming_direct = 0;
    int incoming_account = 0;
    int incoming_registrar = 0;
    int outgoing_registrar = 0;
   
    // Status
    app_dbg(THIS_FILE, "RX-REQUEST %d bytes %s from %s %s:%d", rdata->msg_info.len, pjsip_rx_data_get_info(rdata), rdata->tp_info.transport->type_name, rdata->pkt_info.src_name, rdata->pkt_info.src_port);       
    app_dbg2(THIS_FILE, "\n\n%.*s\n", rdata->msg_info.len, rdata->msg_info.msg_buf);

    // Get connection type 
    if(voip_check_type(rdata, &accountid, &clientid, &incoming_direct, &incoming_account, &incoming_registrar, &outgoing_registrar, from, to))
        return PJ_FALSE;
    
    // Incoming or outgoing registrar request? Then abort!        
    if(incoming_registrar == 1 || outgoing_registrar == 1)
        return PJ_FALSE;

    // Alert call (initialize connection / send 180 response)
    if(voip_alert_call(rdata) == NULL)
        return PJ_FALSE;

    // ok!
    return PJ_TRUE;
}

#endif

