/*********************************************************************************
 * voip_funcs.c ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "voip_funcs.c"

// Make an outgoing VOIP call
int voip_call(connection_t *mCON, char *no_source, char *no_target, int call_anonymous, char *scriptfile, char *delimiter)
{
    #if DTMFBOX_USE_VOIP

        connection_t *CUR_CON = (connection_t*)mCON;
        connection_transport_t *rtp = NULL;
        char local_uri[MAX_URI_LENGTH];
        char dest_uri[MAX_URI_LENGTH];         
        pjsip_dialog *dlg;
        pjmedia_sdp_session *local_sdp;
        pjsip_tx_data *tdata;
        pj_str_t dst_uri, loc_uri, loc_contact;
        pj_status_t status;
        int id = -1;

        // Get array-id
        id = CUR_CON->id;
        if(id == -1)
            return -1;
        
        // copy parameters to struct
        pj_ansi_snprintf(CUR_CON->info->no_source, sizeof(CUR_CON->info->no_source), "%s", no_source);
        pj_ansi_snprintf(CUR_CON->info->no_target, sizeof(CUR_CON->info->no_target), "%s", no_target);                
        pj_ansi_snprintf(CUR_CON->script->file, sizeof(CUR_CON->script->file), "%s", scriptfile);
        pj_ansi_snprintf(CUR_CON->script->delimiter, sizeof(CUR_CON->script->delimiter), "%s", delimiter);
        CUR_CON->info->status = 0; 
        CUR_CON->info->direction |= OUTGOING;

        // Source- & Target-No.        
        pj_ansi_snprintf(local_uri, sizeof(local_uri), "%s", CUR_CON->info->no_source);
        pj_ansi_snprintf(dest_uri, sizeof(dest_uri), "%s", CUR_CON->info->no_target);

        // Set account name (in front of number).
        // (not when registrar-client is called)
        if(CUR_CON->info->account_id >= 0 && !(CUR_CON->info->direction & CLIENT))
        {
            int acc_id = CUR_CON->info->account_id;

            // set the account name as display name
            if(strlen(oSET.accounts[acc_id].name) != 0 && !call_anonymous)
                pj_ansi_snprintf(local_uri, sizeof(local_uri), "\"%s\" %s", oSET.accounts[acc_id].name, CUR_CON->info->no_source);

            // set "anonymous" as display name
            if(call_anonymous)
                pj_ansi_snprintf(local_uri, sizeof(local_uri), "\"anonymous\" %s", CUR_CON->info->no_source);
        } 

        loc_uri = pj_str(local_uri);
        dst_uri = pj_str(dest_uri);
        
        // Create UAC dialog
        app_msg(THIS_FILE, "Initialize outgoing call from '%s' to '%s'", CUR_CON->info->no_source, CUR_CON->info->no_target);
        
        status = voip_create_uac_contact(CUR_CON->pools->con_pool, &loc_contact, CUR_CON->info->account_id, &dst_uri);
        if(status == PJ_SUCCESS)
        {
            status = pjsip_dlg_create_uac(pjsip_ua_instance(),
                            &loc_uri,     // local URI
                            &loc_contact, // local Contact
                            &dst_uri,     // remote URI
                            &dst_uri,     // remote target
                            &dlg);        // dialog                

        } else {
            status = pjsip_dlg_create_uac(pjsip_ua_instance(),
                &loc_uri,    // local URI
                NULL,        // local Contact
                &dst_uri,    // remote URI
                &dst_uri,    // remote target
                &dlg);       // dialog                
        }

        if (status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to create UAC dialog!", status);
            reset_connection(CUR_CON);                        
            return -1;
        }

        // Get a free transport
        if(CUR_CON->rtp == NULL)
        {
            int transport_id = transport_by_free();
            if(transport_id == -1)
            {
                app_err(THIS_FILE, "No free transports left. Max. %d!", oSET.voip.rtp_max);
                reset_connection(CUR_CON);
                return -1;
            } 
            else 
            {
                CUR_CON->rtp = oRTP[transport_id];
                rtp = CUR_CON->rtp;
            }
        }

        // Get the SDP body to be put in the outgoing INVITE, by asking
        // media endpoint to create one for us. The SDP will contain all
        // codecs that have been registered to it, plus telephony event.        
        app_dbg2(THIS_FILE, "Create SDP ...");                          
        if (rtp->g_med_transport == NULL) {
            reset_connection(CUR_CON);            
	        return -1;
        }

        // Get media socket info for transport!
        {
            pjmedia_transport_info tpinfo;

            pjmedia_transport_get_info(rtp->g_med_transport, &tpinfo);    
            pj_memcpy(&rtp->g_med_skinfo, &tpinfo.sock_info, sizeof(pjmedia_sock_info));
        }

        status = pjmedia_endpt_create_sdp(oGLB.med_endpt,         // the media endpt
                           dlg->pool,                             // pool
                           1,                                     // # of streams
                           &rtp->g_med_skinfo,                    // RTP sock info
                           &local_sdp);                           // the SDP result
        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to create SDP!", status);
            reset_connection(CUR_CON);            
            return -1;
        }
        
        #if DTMFBOX_USE_ICE
            status = pjmedia_transport_media_create(rtp->g_med_transport, CUR_CON->pools->con_pool, 0, local_sdp, 0);
            if(status != PJ_SUCCESS)
                app_err(THIS_FILE, "Can not create media transport!");                    
        #endif

        // Create the INVITE session
        app_dbg2(THIS_FILE, "Create invite session ...");        
        status = pjsip_inv_create_uac(dlg, local_sdp, PJSIP_INV_SUPPORT_100REL, &rtp->g_inv);
        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to create invite session!", status);
            reset_connection(CUR_CON);            
            return -1;
        }

        // Set Proxy ?
        // Note that Route URI SHOULD have an ";lr" parameter!
        if(CUR_CON->info->account_id >= 0)
        {
            int acc_id = CUR_CON->info->account_id;

            // Set dialog Route-Set (Proxy):             
            if (!pj_list_empty(&oSET.accounts[acc_id].route_set) && strlen(oSET.accounts[acc_id].proxy))               
            {
                // do NOT set route, when this is an incoming CLIENT connection (Registrar-Mode: DTMFBOX->CLIENT)
                if(!(CUR_CON->info->direction & CLIENT && CUR_CON->info->direction & INCOMING))
                    pjsip_dlg_set_route_set(dlg, &oSET.accounts[acc_id].route_set);
            }

            // Credentials for outgoing connection
            // get authorization data from oSET struct
            // do NOT set credentials, when this is an incoming CLIENT connection (Registrar-Mode: DTMFBOX->CLIENT)
            if(strlen(oSET.accounts[acc_id].realm) != 0 &&
               strlen(oSET.accounts[acc_id].registrar) != 0 &&
               oSET.accounts[acc_id].controller.device != CAPI &&
               oSET.accounts[acc_id].active != 0 &&
               oSET.voip_active &&
               !(CUR_CON->info->direction & CLIENT && CUR_CON->info->direction & INCOMING))
            {
                pjsip_cred_info cred[1];
                
                cred[0].realm     = pj_str(oSET.accounts[acc_id].realm);
                cred[0].username  = pj_str(oSET.accounts[acc_id].username);
                cred[0].data_type = PJSIP_CRED_DATA_PLAIN_PASSWD;
                cred[0].scheme    = pj_str("digest");
                cred[0].data      = pj_str(oSET.accounts[acc_id].password);

                pjsip_auth_clt_set_credentials(&dlg->auth_sess, 1, cred);
            }
        }

        // Create initial INVITE request
        status = pjsip_inv_invite(rtp->g_inv, &tdata);
        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to create initial invite request!", status);
            reset_connection(CUR_CON);            
            return -1;
        }

        // Send it!
        status = pjsip_inv_send_msg(rtp->g_inv, tdata);
        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to send invite request!", status);
            reset_connection(CUR_CON);            
            return -1;
        }
        
        // Initialize custom voip_port the first time (for ringing)        
        status = port_voip_init(CUR_CON);
        if(status != PJ_SUCCESS) {
            app_err(THIS_FILE, "(%d) Unable to create VoIP port!", status);
            reset_connection(CUR_CON);            
            return -1;
        }

        // STATUS: CONNECTING
        CUR_CON->info->status = 1;

        // Script event
        script_event_add(VOIP, CONNECT, CUR_CON->script->file, CUR_CON->id, CUR_CON->in_conf[0], CUR_CON->info->no_source, CUR_CON->info->no_target, "");

        // Start tone thread?
        if(CUR_CON->in_conf[0] != -1 && oSET.audio.earlyb3_ring)
        {
            // no client connections!! (Registrar-Mode)
            //if(!(CUR_CON->info->direction & CLIENT) && !(oCONS[CUR_CON->in_conf[0]].info->direction & CLIENT))
            if(oCONS[CUR_CON->in_conf[0]].info->active)
            {
                int args[2];
                args[0] = CUR_CON->in_conf[0];                     
                args[1] = CUR_CON->id;  
                pj_thread_create(CUR_CON->pools->con_pool, "fake_ring_%p", (pj_thread_proc*)&port_tone_ring_thread, (void*)&args, 0, (unsigned)NULL, &CUR_CON->tone_thread);
                pj_thread_sleep(25);
            }
        }
                
        return id;
    #else
        PJ_UNUSED_ARG(delimiter);
        PJ_UNUSED_ARG(scriptfile);
        PJ_UNUSED_ARG(call_anonymous);
        PJ_UNUSED_ARG(no_target);
        PJ_UNUSED_ARG(no_source);        
        PJ_UNUSED_ARG(mCON);        

        return -1;
    #endif
}


// Hangup a VoIP call
int voip_hangup(connection_t *mCON)
{    
    #if DTMFBOX_USE_VOIP

        connection_t *CUR_CON = (connection_t*)mCON;        
        connection_transport_t *rtp;
        pjsip_tx_data *tdata;
        pj_status_t status;        
        int ret = 0;        

        if(CUR_CON->info->type != VOIP || CUR_CON->info->status == 3) {        
            return -1;
        }

        rtp = CUR_CON->rtp;
        if(rtp->g_inv != NULL)
            snprintf(CUR_CON->script->dtmf, sizeof(CUR_CON->script->dtmf), "%d", rtp->g_inv->cause);
        else
            snprintf(CUR_CON->script->dtmf, sizeof(CUR_CON->script->dtmf), "%d", 200);
                 
        // SCRIPT: "DISCONNECT"
        if(CUR_CON->info->status != 3)
            script_event_add(VOIP, DISCONNECT, CUR_CON->script->file, CUR_CON->id, CUR_CON->in_conf[0], CUR_CON->info->no_source, CUR_CON->info->no_target, CUR_CON->script->dtmf);

        // stop capi transmit!
        CUR_CON->info->active = 0;        
        if(rtp->g_inv != NULL)
        {   
            int code = PJSIP_SC_REQUEST_TERMINATED;
            
            if (rtp->g_inv->state == PJSIP_INV_STATE_CONFIRMED)
	            code = PJSIP_SC_OK;
	        else if (rtp->g_inv->role == PJSIP_ROLE_UAS)
	            code = PJSIP_SC_DECLINE;
	        else
	            code = PJSIP_SC_REQUEST_TERMINATED;

            // Destroy invite session (when not already done!)
            status = pjsip_inv_end_session(rtp->g_inv, code, NULL, &tdata);
            if(status != PJ_SUCCESS || tdata == NULL)
                goto remove_session;
            
            // Send DISCONNECT!            
            if(rtp->g_inv != NULL)
            {
                status = pjsip_inv_send_msg(rtp->g_inv, tdata);
                if(status != PJ_SUCCESS)
                {
                    app_err(THIS_FILE, "(%d) Failed to send end session message", status);                
                    goto remove_session;
                }
            }
        }        

remove_session:
        // disconnect and remove ports from conference bridge
        voip_remove_media_session(CUR_CON);

        #if DTMFBOX_USE_ICE
            pjmedia_transport_media_stop(rtp->g_med_transport);            
        #endif

        return ret;
    #else
        PJ_UNUSED_ARG(mCON);
        return -1;
    #endif
}

void voip_confirm_call(connection_t *mCON)
{
    #if DTMFBOX_USE_VOIP

    connection_t *CUR_CON = (connection_t*)mCON;
    connection_transport_t *rtp;
    pjsip_tx_data *tdata;
    pj_status_t status;
    int statuscode = 200;
    
    if(CUR_CON->info->type != VOIP || CUR_CON->info->status != 1) {
        return;
    }
    
    //
    // Now create 200 response. (CONFIRM)
    //
    rtp = CUR_CON->rtp;
    if(rtp->g_inv != NULL && CUR_CON->info->status == 1 &&
       !(CUR_CON->info->direction & INCOMING && CUR_CON->info->direction & CLIENT))
    {        
        // set "CONFIRMED"
        CUR_CON->info->status = 2;

        app_dbg(THIS_FILE, "Respond stateless with code %d", statuscode);
        status = pjsip_inv_answer(rtp->g_inv, statuscode, NULL, NULL, &tdata);
        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Cannot create stateless respond with status %d!", status, statuscode);            
            return;
        }
        
        // Confirm!
        if(rtp->g_inv != NULL)
        {
            status = pjsip_inv_send_msg(rtp->g_inv, tdata);
            if(status != PJ_SUCCESS)
            {
                app_err(THIS_FILE, "(%d) Cannot send stateless respond with status %d!", status, statuscode);                
                return;
            }
        }
    }
    #else
    PJ_UNUSED_ARG(mCON);            
    #endif
}

// TODO: hold voip call
void voip_hold_call(connection_t *mCON, int hold)
{
    PJ_UNUSED_ARG(mCON);
    PJ_UNUSED_ARG(hold);
}

int voip_check_type(pjsip_rx_data *rdata, int *accountid, int *clientid, int *incoming_direct, int *incoming_account, int *incoming_registrar, int *outgoing_registrar, char from[MAX_URI_LENGTH], char to[MAX_URI_LENGTH])
{
    #if DTMFBOX_USE_VOIP

    char from_hdr[MAX_URI_LENGTH];
    char to_hdr[MAX_URI_LENGTH];    
    char tmp[MAX_URI_LENGTH];

    pj_str_t reason;
    pj_status_t status;
    pjsip_msg *msg = rdata->msg_info.msg;
    
    pjsip_fromto_hdr *hdr_from;
    pjsip_fromto_hdr *hdr_to;
    pjsip_sip_uri *uri_from;
    pjsip_sip_uri *uri_to;
   
    int i;    
    unsigned int options = PJSIP_INV_SUPPORT_100REL;

    // Init
    *clientid = -1;
    *accountid = -1;
    *incoming_direct = 0;
    *incoming_account = 0;    
    *incoming_registrar = 0;
    *outgoing_registrar = 0;
    pj_memset(from, 0, sizeof(from));
    pj_memset(to, 0, sizeof(to));

    // --------------------------------------------------------------------------------------------------------
    // Check, if we can handle the request at this point.     
    // --------------------------------------------------------------------------------------------------------
    // Only INVITE-Requests. Any other requests will be rejected...
    if(rdata->msg_info.msg->line.req.method.id != PJSIP_INVITE_METHOD)
    {
        app_dbg2(THIS_FILE, "Request %d unsupported!", rdata->msg_info.msg->line.req.method.id);
        return -1;
    }

    // Got dialog or tsx?
    if (pjsip_rdata_get_dlg(rdata) || pjsip_rdata_get_tsx(rdata))
    {
        app_err(THIS_FILE, "Unable to locate dialog and tsx!");
        return -2;
    }
	    
    // Verify that we can handle this invite (100REL)    
    status = pjsip_inv_verify_request(rdata, &options, NULL, NULL, oGLB.endpt, NULL);
    if (status != PJ_SUCCESS)
    {
        app_err(THIS_FILE, "(%d) Unable to verify invite request!", status);
        return -3;
    }

    // --------------------------------------------------------------------------------------------------------
    // Check, if this is an incoming or outgoing connection.
    // The FROM-Header contains the address of the client.
    // We search for the address in the list of all registred clients (oGLB.clients).
    // When the client was found, the connection is an outgoing connection (through this registrar).
    // --------------------------------------------------------------------------------------------------------        
    // Search for client
    for(i=0; i<oSET.voip.max_clients; i++)
    {        
        pj_ansi_snprintf(tmp, sizeof(tmp), "%s:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port);
        
        if(!strcmp(oGLB.clients[i].addr, tmp))
        {            
            // Outgoing request (Registrar-Mode only!)
            // Address was found in the clients list!
            *clientid = i;                        
            break;
        }
    }
    
    if(*clientid == -1)
    {
        // An incoming request
        // Either a normal incoming/outgoing call or an outgoing client call (registrar)
    }

    // --------------------------------------------------------------------------------------------------------
    // The source and destination address of the sip message must be changed:
    //
    // On outgoing connections, we use the "real" registrar as destination address.
    // On incoming connections, we use the client's address as destination. The phone rings...
    // --------------------------------------------------------------------------------------------------------
    hdr_from = (pjsip_fromto_hdr*) pjsip_msg_find_hdr(msg, PJSIP_H_FROM, NULL);
    hdr_to = (pjsip_fromto_hdr*) pjsip_msg_find_hdr(msg, PJSIP_H_TO, NULL);
    uri_from = (pjsip_sip_uri*) pjsip_uri_get_uri(hdr_from->uri);
    uri_to = (pjsip_sip_uri*) pjsip_uri_get_uri(hdr_to->uri);
    pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, uri_from, from_hdr, MAX_URI_LENGTH);
    pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, uri_to, to_hdr, MAX_URI_LENGTH);

    sip_strip_user((char*)&from_hdr);
    sip_strip_user((char*)&to_hdr);

    // INCOMING
    if(*clientid == -1)
    {     
        // Search for account
        *accountid = voip_get_acc_incoming(rdata, from, to);            
        
        // When this account is used as registrar account, search for any registered clients!
        if(*accountid != -1)
        {            
            int client_accountid = -1;

            for(i=0; i<MAX_ACCOUNTS; i++)
            {                    
               client_accountid = client_account_by_account(i, *accountid);
               if(client_accountid != -1) 
                   break;                   
            }

            if(client_accountid != -1) 
            {
                // a client found! Set as incoming registrar                
                *incoming_registrar = 1;                    
            }
            else 
            {
                // no client found! Just set as incoming client                
                *incoming_account = 1;
            }
        }
        else
        {           
            // direct IP call (no account found)
            *incoming_direct = 1;  
            
            // unknown accounts allowed?
            if(!oSET.voip.allow_unknown_acc)
            {
                reason = pj_str("Unknown number!");                
                app_msg(THIS_FILE, "Incoming call for unknown account [%s -> %s] (ignoring)", from, to);
                return -4;       
            }
        }

        app_msg(THIS_FILE, "Incoming call from %s to %s [account #%d]", from, to, accountid);
    }

    // OUTGOING
    else
    {
        *accountid = voip_get_acc_outgoing(rdata, from, to);
        if(*accountid == -1) 
        {
            app_err(THIS_FILE, "Unauthorized client request!");
            return -5;
        }        

        *outgoing_registrar = 1;
    }
    
#else
        PJ_UNUSED_ARG(rdata);
        PJ_UNUSED_ARG(accountid);
        PJ_UNUSED_ARG(clientid);
        PJ_UNUSED_ARG(incoming_direct);
        PJ_UNUSED_ARG(incoming_account);
        PJ_UNUSED_ARG(incoming_registrar);
        PJ_UNUSED_ARG(outgoing_registrar);
        PJ_UNUSED_ARG(from);
        PJ_UNUSED_ARG(to);
    #endif

    return 0;
}

// Initialize connection and alert call (send 180 response)
connection_t* voip_alert_call(pjsip_rx_data *rdata)
{    
    #if DTMFBOX_USE_VOIP

    char from[MAX_URI_LENGTH];
    char to[MAX_URI_LENGTH];
    
    pjmedia_transport_info tpinfo;
    pj_str_t reason;
    pj_str_t contact;
    pj_status_t status;
    pjsip_dialog *dlg;
    pjmedia_sdp_session *local_sdp;
    pjsip_tx_data *tdata;
   
    int conid = -1;
    int rtpid = -1;
    int clientid = -1;
    int accountid = -1;

    int incoming_direct = 0;
    int incoming_account = 0;    
    int incoming_registrar = 0;
    int outgoing_registrar = 0;

    if(voip_check_type(rdata, &accountid, &clientid, &incoming_direct, &incoming_account, &incoming_registrar, &outgoing_registrar, from, to))
        goto hell;

    // --------------------------------------------------------------------------------------------------------
    // Get a free connection for this request.
    // --------------------------------------------------------------------------------------------------------    
    DTMFBOX_ACC_LOCK();
    conid = connection_by_free();
    if(conid == -1) {
        DTMFBOX_ACC_UNLOCK();
        goto hell;
    }
        
    // An INVITE-Session already exist? Abort!
    if (oCONS[conid].rtp)
    {
        app_err(THIS_FILE, "Another call is in progress!");
        reason = pj_str("Another call is in progress!");
        pjsip_endpt_respond_stateless(oGLB.endpt, rdata, 500, &reason, NULL, NULL);        
        DTMFBOX_ACC_UNLOCK();
        goto hell;
    } 
    else 
    {
        rtpid = transport_by_free();

        if(rtpid == -1)
        {
            app_err(THIS_FILE, "No free transports left. Max. %d!", oSET.voip.rtp_max);
            reason = pj_str("No free transports left!");
            pjsip_endpt_respond_stateless(oGLB.endpt, rdata, 500, &reason, NULL, NULL);        
            DTMFBOX_ACC_UNLOCK();
            goto hell;
        }        
        
        oCONS[conid].rtp = oRTP[rtpid];
        oCONS[conid].info->type = VOIP;
    }
    DTMFBOX_ACC_UNLOCK();
    
    // Fill connection struct ...            
    oCONS[conid].capi->NCCI = (unsigned)-1;             // Pseudo-NCCI (-1=VoIP)
    oCONS[conid].info->account_id = accountid;          // Account-ID
    oCONS[conid].info->status = 0;

    // Some user agents urlencode characters.
    // make sure, that # is decoded!!
    replace(from, from, sizeof(from), "%23", "#");
    replace(to, to, sizeof(to), "%23", "#");

    if(clientid == -1)
    {
        // Normal incoming call
        oCONS[conid].info->direction = INCOMING;
        pj_ansi_snprintf(oCONS[conid].info->no_source, sizeof(oCONS[conid].info->no_source), "%s", to);
        pj_ansi_snprintf(oCONS[conid].info->no_target, sizeof(oCONS[conid].info->no_target), "%s", from);

    } 
    else 
    {
        // Outgoing call for client (registrar mode)
        oCONS[conid].info->direction = CLIENT | OUTGOING;    
        pj_ansi_snprintf(oCONS[conid].info->no_source, sizeof(oCONS[conid].info->no_source), "%s", from);
        pj_ansi_snprintf(oCONS[conid].info->no_target, sizeof(oCONS[conid].info->no_target), "%s", to);
    }

    // Scriptfile defaults
    pj_ansi_snprintf(oCONS[conid].script->file, sizeof(oCONS[conid].script->file), "%s", oSET.scriptfile);
    pj_ansi_snprintf(oCONS[conid].script->delimiter, sizeof(oCONS[conid].script->delimiter), "%s", oSET.delimiter);

   
    // Create UAS dialog
    app_dbg2(THIS_FILE, "Create UAS dialog ...");    
    
    // get suitable contact
    status = voip_create_uas_contact(oCONS[conid].pools->con_pool, &contact, accountid, rdata); 
    if(status == PJ_SUCCESS)
        status = pjsip_dlg_create_uas(pjsip_ua_instance(), rdata, &contact, &dlg);
    else            
        status = pjsip_dlg_create_uas(pjsip_ua_instance(), rdata, NULL, &dlg);
    
    if (status != PJ_SUCCESS) 
    {
        app_err(THIS_FILE, "Unable to create UAS dialog!");
        pjsip_endpt_respond_stateless(oGLB.endpt, rdata, 500, NULL, NULL, NULL);
        goto hell;
    }
    
    // Create SDP capability and get media socket info!    
    pjmedia_transport_get_info(oCONS[conid].rtp->g_med_transport, &tpinfo);
    pj_memcpy(&oCONS[conid].rtp->g_med_skinfo, &tpinfo.sock_info, sizeof(pjmedia_sock_info));
        
    app_dbg2(THIS_FILE, "Create media capability ...");
    status = pjmedia_endpt_create_sdp(oGLB.med_endpt, rdata->tp_info.pool, 1, &oCONS[conid].rtp->g_med_skinfo, &local_sdp);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "Can not create media capability!");
        goto hell;
    }

    // ICE
    #if DTMFBOX_USE_ICE
        status = pjmedia_transport_media_create(oCONS[conid].rtp->g_med_transport, oCONS[conid].pools->con_pool, 0, local_sdp, 0);
        if(status != PJ_SUCCESS)
        {
            app_err(THIS_FILE, "(%d) Unable to create media transport!", status);
            goto hell;
        }
    #endif


    // Create invite session
    app_dbg2(THIS_FILE, "Create invite session ..."); 
    status = pjsip_inv_create_uas(dlg, rdata, local_sdp, PJSIP_INV_SUPPORT_100REL, &oCONS[conid].rtp->g_inv);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create invite session!", status);
        goto hell;
    }

    // Create dialog 100 response
    app_dbg2(THIS_FILE, "Respond 100 ...");
    status = pjsip_dlg_create_response(dlg, rdata, 100, NULL, &tdata);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create 100 respond!", status);
        goto hell;
    }
    
    // Send initial 100 response
    status = pjsip_dlg_send_response(dlg, pjsip_rdata_get_tsx(rdata), tdata);
    if(status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to send 100 respond!", status);
        goto hell;
    }

    // --------------------------------------------------------------------------------------------------------
    // First respond with code 180 to initialize invite (ring)
    // (pjsip_inv_initial_answer !!!)
    // --------------------------------------------------------------------------------------------------------
    app_dbg2(THIS_FILE, "Respond 180 (*ring*)..");
    status = pjsip_inv_initial_answer(oCONS[conid].rtp->g_inv, rdata, 180, NULL, NULL, &tdata);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to create 180 respond!", status);
        goto hell;
    }

    // Send the 180 response.    
    status = pjsip_inv_send_msg(oCONS[conid].rtp->g_inv, tdata);
    if (status != PJ_SUCCESS) {
        app_err(THIS_FILE, "(%d) Unable to send 180 respond!", status);
        goto hell;
    }

    return &oCONS[conid];

hell:
    // Reset connections
    if(conid >= 0)
        reset_connection(&oCONS[conid]);

    #else
    PJ_UNUSED_ARG(rdata);    
    #endif

    return NULL;
}

// Remove media session
void voip_remove_media_session(connection_t *mCON)
{
    #if DTMFBOX_USE_VOIP

        connection_t *CUR_CON = (connection_t*)mCON;
        connection_transport_t *rtp;
        pj_status_t status;
        int i;
                
        // remove any wave-player/-recorder!        
        for(i=0; i<INCONF_PORTS; i++) {
            int play_id = -1;
            port_filestream_player_recorder(CUR_CON, DEVNULL, DEVNULL, 0, 0, 0, 0, 0, &play_id);            
        }
        
        // close custom voip_port
        if(CUR_CON->ports->custom != NULL)                
            status = port_voip_close(CUR_CON);
        
        // destroy media-session
        rtp = CUR_CON->rtp;
        if(rtp != NULL)
        {
            if(rtp->g_med_session != NULL)
            {
                status = pjmedia_session_destroy(rtp->g_med_session); 
                if (status != PJ_SUCCESS) {
                    app_err(THIS_FILE, "(%d) Unable to destroy media session", status);                
                    return;
                }            
                rtp->g_med_session = NULL;
            }
        }  

    #else
        PJ_UNUSED_ARG(mCON);
    #endif
}

int voip_get_acc_incoming(pjsip_rx_data *rdata, char *hdr_from, char *hdr_to)
{    
    #if DTMFBOX_USE_VOIP

    char to[MAX_URI_LENGTH];    
    char host[MAX_URI_LENGTH];    
    char user[MAX_URI_LENGTH];    
    pjsip_msg *msg;
    pjsip_fromto_hdr *fromto_hdr_from;
    pjsip_fromto_hdr *fromto_hdr_to;
    pjsip_sip_uri *uri_from;
    pjsip_sip_uri *uri_to;     
    int acc_id = -1;
    int i;
    
    PJ_ASSERT_RETURN(rdata != NULL, 1);
    PJ_ASSERT_RETURN(hdr_from != NULL, 2);
    PJ_ASSERT_RETURN(hdr_to != NULL, 3);

    msg = rdata->msg_info.msg;

    // get FROM-, TO-header
    fromto_hdr_from = (pjsip_fromto_hdr*)pjsip_msg_find_hdr(msg, PJSIP_H_FROM, NULL);
    fromto_hdr_to = (pjsip_fromto_hdr*)pjsip_msg_find_hdr(msg, PJSIP_H_TO, NULL);
    uri_from = (pjsip_sip_uri*)pjsip_uri_get_uri(fromto_hdr_from->uri);
    uri_to = (pjsip_sip_uri*)pjsip_uri_get_uri(fromto_hdr_to->uri);
    pj_ansi_snprintf(to, sizeof(to), "%.*s@%.*s", uri_to->user.slen, uri_to->user.ptr, uri_to->host.slen, uri_to->host.ptr);
    pj_ansi_snprintf(user, sizeof(user), "%.*s", uri_to->user.slen, uri_to->user.ptr);
    pj_ansi_snprintf(host, sizeof(host), "%.*s", uri_to->host.slen, uri_to->host.ptr);    
    sip_strip_user(user);

    // Search over user and domain part...
    for(i=0; i<MAX_ACCOUNTS; i++)
    {        
        if(oSET.accounts[i].active && oSET.accounts[i].controller.device == VOIP)
        {
            // Search account by username@registrar
            if(!strcmp(user, oSET.accounts[i].username) &&
               !strcmp(host, oSET.accounts[i].registrar))
            {
                acc_id = i;
                break;
            }

            // Search account by phonenumber@registrar
            if(!strcmp(user, oSET.accounts[i].number) &&
               !strcmp(host, oSET.accounts[i].registrar))
            {            
                acc_id = i;
                break;
            }

            // Search account by account id
            if(!strcmp(to, oSET.accounts[i].id))
            {
                acc_id = i;
                break;
            }        
        }
    }

    // No matching account found? Just try over user part...
    if(acc_id == -1)
    {
        for(i=0; i<MAX_ACCOUNTS; i++)
        {
            if(oSET.accounts[i].active && oSET.accounts[i].controller.device == VOIP)
            {
                // Search account by username
                if(!strcmp(user, oSET.accounts[i].username))
                {
                    acc_id = i;
                    break;
                }

                // Search account by phonenumber
                if(!strcmp(user, oSET.accounts[i].number))
                {            
                    acc_id = i;
                    break;
                }

                // Search account by *phonenumber*
                if(strstr(user, oSET.accounts[i].number))
                {
                    acc_id = i;
                    break;
                }
            }
        }    
    }

    // No matching account found? Just try over domain part...
    if(acc_id == -1)
    {
        for(i=0; i<MAX_ACCOUNTS; i++)
        {
            if(oSET.accounts[i].active && oSET.accounts[i].controller.device == VOIP)
            {
                if(!strcmp(host, oSET.accounts[i].registrar))
                {
                    acc_id = i;
                    break;
                }
            }
        }
    }
        
    // Save FROM-,TO-Header
    pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, uri_from, hdr_from, MAX_URI_LENGTH);
    pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, uri_to, hdr_to, MAX_URI_LENGTH);

    return acc_id;

    #else
    PJ_UNUSED_ARG(hdr_to);
    PJ_UNUSED_ARG(hdr_from);
    PJ_UNUSED_ARG(rdata);

    return -1;
    #endif
}

int voip_get_acc_outgoing(pjsip_rx_data *rdata, char *hdr_from, char *hdr_to)
{
    #if DTMFBOX_USE_VOIP

    char tmp[MAX_URI_LENGTH];    
    char from[MAX_URI_LENGTH];    
    pjsip_msg *msg;
    pjsip_fromto_hdr *fromto_hdr_from;
    pjsip_fromto_hdr *fromto_hdr_to;
    pjsip_sip_uri *uri_from;
    pjsip_sip_uri *uri_to;        
    int i;
    int client_id = -1;
    int acc_id = -1;

    PJ_ASSERT_RETURN(rdata != NULL, 1);
    PJ_ASSERT_RETURN(hdr_from != NULL, 2);
    PJ_ASSERT_RETURN(hdr_to != NULL, 3);

    msg = rdata->msg_info.msg;

    // get FROM-, TO-header
    fromto_hdr_from = (pjsip_fromto_hdr*)pjsip_msg_find_hdr(msg, PJSIP_H_FROM, NULL);
    fromto_hdr_to = (pjsip_fromto_hdr*)pjsip_msg_find_hdr(msg, PJSIP_H_TO, NULL);
    uri_from = (pjsip_sip_uri*)pjsip_uri_get_uri(fromto_hdr_from->uri);
    uri_to = (pjsip_sip_uri*)pjsip_uri_get_uri(fromto_hdr_to->uri);
    pj_ansi_snprintf(from, sizeof(from), "%.*s@%.*s", uri_from->user.slen, uri_from->user.ptr, uri_from->host.slen, uri_from->host.ptr);
    
    // Search for client
    for(i=0; i<oSET.voip.max_clients; i++)
    {
        pj_ansi_snprintf(tmp, sizeof(tmp), "%s:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port);
        //if(!pj_strcmp2(&oGLB.clients[i].addr, (const char*)&tmp))
        if(!strcmp(oGLB.clients[i].addr, (const char*)tmp))
        {
            // Outgoing request!
            // Address was found in the clients list!
            client_id = i;
            break;
        }
    }
    
    // An incoming request! return!        
    if(client_id == -1)            
        return -1;
    
    // Search for account in client table
    for(i=0; i<MAX_ACCOUNTS; i++)
    {
        acc_id = oGLB.clients[client_id].account_id[i];
        if(acc_id != -1)
        {
            // registered account found!
            if(!pj_stricmp2(&uri_from->user, (const char*)oSET.accounts[acc_id].registrar_username))
                break;

            acc_id = -1;
        }
    }    

    if(acc_id == -1)
        return -1;

    // Save FROM-,TO-Header
    pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, uri_from, hdr_from, MAX_URI_LENGTH);
    pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, uri_to, hdr_to, MAX_URI_LENGTH);
    
    if(oSET.accounts[acc_id].controller.device == CAPI) {
        sip_strip_user(hdr_to);
    }

    if(oSET.accounts[acc_id].controller.device == VOIP) {
        sip_strip_user(hdr_to);
        pj_ansi_snprintf(hdr_to, MAX_URI_LENGTH, "%s@%s", hdr_to, oSET.accounts[acc_id].registrar);
    }

    return acc_id;

    #else
    PJ_UNUSED_ARG(hdr_to);
    PJ_UNUSED_ARG(hdr_from);
    PJ_UNUSED_ARG(rdata);

    return -1;
    #endif
}

pj_status_t voip_resolve_transport_address(char *addr, int *port)
{
    #if DTMFBOX_USE_VOIP

    char buf[MAX_URI_LENGTH];
    pj_pool_t *pool;
    pjsip_tpselector tp_sel;
    pjsip_transport_type_e tp_type = PJSIP_TRANSPORT_UDP;
    pj_status_t status = PJ_SUCCESS;
    pj_str_t local_addr;
    int local_port;    
    
    // init transport selector
    tp_sel.type = PJSIP_TPSELECTOR_TRANSPORT;
    tp_sel.u.transport = oGLB.transport;        

    // fetch local address + port
    pool = pj_pool_create_on_buf(NULL, buf, sizeof(buf));
    status = pjsip_tpmgr_find_local_addr(pjsip_endpt_get_tpmgr(oGLB.endpt),
					                    pool, tp_type, &tp_sel, 
					                    &local_addr, &local_port);
    if (status != PJ_SUCCESS)
	    return status;

    pj_ansi_snprintf(addr, MAX_URI_LENGTH, "%.*s", (int)local_addr.slen, local_addr.ptr);
    *port = local_port;

    return status;
    #else
    PJ_UNUSED_ARG(port);
    PJ_UNUSED_ARG(addr);
    return PJ_TRUE;
    #endif
}

pj_status_t voip_create_uac_contact(pj_pool_t *pool, pj_str_t *contact, int acc_id, const pj_str_t *suri)
{
    #if DTMFBOX_USE_VOIP

    char local_addr[MAX_URI_LENGTH];
    int local_port;
    settings_account_t *acc = NULL;
    pjsip_transport_type_e tp_type = PJSIP_TRANSPORT_UDP;
    pj_status_t status;

    PJ_UNUSED_ARG(suri);

    if(acc_id != -1)
    {
        acc = &oSET.accounts[acc_id];

        // If contact is configured, then use use it 
        if (strlen(acc->contact)) {
            contact->ptr = (char*)pj_pool_alloc(pool, MAX_URI_LENGTH);
            contact->slen = pj_ansi_snprintf(contact->ptr, MAX_URI_LENGTH,
		    	                            "%s%s<sip:%s>",
			                                acc->name,				     
			                                (strlen(acc->name)?" " : ""),				     
                                            acc->contact);            
	        return PJ_SUCCESS;
        }
    }

    // resolve address:port from current transport
    status = voip_resolve_transport_address(local_addr, &local_port);
    if(status != PJ_SUCCESS)
        return status;

    // Create the contact header 
    contact->ptr = (char*)pj_pool_alloc(pool, MAX_URI_LENGTH);
    if(acc_id != -1 && acc)
    {
        contact->slen = pj_ansi_snprintf(contact->ptr, MAX_URI_LENGTH,
		    	         "%s%s<sip:%s%s%s:%d;transport=%s>",
			             acc->name,				     
			             (strlen(acc->name)?" " : ""),				     
			             acc->username,
			            (strlen(acc->username)?"@":""),
			            local_addr,
			            local_port,
			            pjsip_transport_get_type_name(tp_type));
    } else {
        contact->slen = pj_ansi_snprintf(contact->ptr, MAX_URI_LENGTH,
		    	         "<sip:%s:%d;transport=%s>",
			            local_addr,
			            local_port,
			            pjsip_transport_get_type_name(tp_type));
    }

    return PJ_SUCCESS;
    #else
    PJ_UNUSED_ARG(suri);
    PJ_UNUSED_ARG(acc_id);
    PJ_UNUSED_ARG(contact);
    PJ_UNUSED_ARG(pool);

    return PJ_TRUE;
    #endif
}

pj_status_t voip_create_uas_contact( pj_pool_t *pool, pj_str_t *contact, int acc_id, pjsip_rx_data *rdata)
{
    #if DTMFBOX_USE_VOIP

    char local_addr[MAX_URI_LENGTH];
    int local_port;
    settings_account_t *acc = NULL;
    pjsip_transport_type_e tp_type = PJSIP_TRANSPORT_UDP;
    pj_status_t status;

    PJ_UNUSED_ARG(rdata);

    if(acc_id != -1)
    {
        acc = &oSET.accounts[acc_id];

        // If contact is configured, then use use it 
        if (strlen(acc->contact)) {
	        contact->ptr = (char*)pj_pool_alloc(pool, MAX_URI_LENGTH);
            contact->slen = pj_ansi_snprintf(contact->ptr, MAX_URI_LENGTH,
		    	                            "%s%s<sip:%s>",
			                                acc->name,				     
			                                (strlen(acc->name)?" " : ""),				     
                                            acc->contact);
	        return PJ_SUCCESS;
        }
    }

    // resolve address:port from current transport
    status = voip_resolve_transport_address(local_addr, &local_port);
    if(status != PJ_SUCCESS)
        return status;    

    // Create the contact header
    contact->ptr = (char*) pj_pool_alloc(pool, MAX_URI_LENGTH);    
    if(acc_id != -1 && acc)
    {
        contact->slen = pj_ansi_snprintf(contact->ptr, MAX_URI_LENGTH,
				     "%s%s<sip:%s%s%s:%d;transport=%s>",
				     acc->name,
				     (strlen(acc->name)?" " : ""),				     
				     acc->username,				     
				     (strlen(acc->username)?"@":""),
				     local_addr,
				     local_port,
                     pjsip_transport_get_type_name(tp_type));
    } else {
        contact->slen = pj_ansi_snprintf(contact->ptr, MAX_URI_LENGTH,
	    	         "<sip:%s:%d;transport=%s>",
		            local_addr,
		            local_port,
		            pjsip_transport_get_type_name(tp_type));
    }

    return PJ_SUCCESS;

    #else
    PJ_UNUSED_ARG(rdata);
    PJ_UNUSED_ARG(acc_id);
    PJ_UNUSED_ARG(contact);
    PJ_UNUSED_ARG(pool);

    return PJ_TRUE;
    #endif
}



// Send a DTMF digit 
void voip_send_dtmf(unsigned char dtmf, int con_id)
{    
    #if DTMFBOX_USE_VOIP
    
    if(con_id >= 0 && con_id <= oSET.max_connections)
    {            
        char info[256];
        char c_digit[2];
        
        pj_str_t digits;
        pj_status_t status;
        
        c_digit[0] = (char)dtmf;
        c_digit[1] = '\0';
        digits = pj_str(c_digit);

        if(oCONS[con_id].rtp != NULL) {
            if(oCONS[con_id].rtp->g_med_session != NULL)
            {
                status = pjmedia_session_dial_dtmf(oCONS[con_id].rtp->g_med_session, 0, &digits);
                if(status == PJ_SUCCESS)
                    app_dbg(THIS_FILE, "%s Outgoing DTMF (%c)", connection_print_info(&oCONS[con_id], info, sizeof(info)), (unsigned char)(dtmf));
            }
        }
    }
    #else
    PJ_UNUSED_ARG(dtmf);
    PJ_UNUSED_ARG(con_id);
    #endif
}

// Receive a DTMF digit, call script 
void voip_get_dtmf(unsigned char dtmf, int ddi, int con_id)
{
    #if DTMFBOX_USE_VOIP

    enum SCRIPT_EVENTS s_event = DTMF;
	int dtmf_pos = 0;
    int i;

    // abort when no dtmf received is empty
    if((int)dtmf == 0)
        return;

    // check if it is really a DTMF signal
    //if((int)dtmf >= 0 && (int)dtmf <= 255)
    {
        if(!isdigit((int)dtmf)) {
            if(dtmf != '#' && dtmf != '*')
                return;
        }
    }

    if(con_id >= 0 && con_id <= oSET.max_connections)
    {
        char info[256];

        dtmf_pos = oCONS[con_id].script->dtmf_pos;        
        if(ddi) { s_event = DDI; }

        app_msg(THIS_FILE, "%s Incoming DTMF (%c)", connection_print_info(&oCONS[con_id], info, sizeof(info)), (unsigned char)(dtmf));

        // Send DTMF Signal to conference partner (voip<->voip, voip<->capi)
        for(i=0; i<INCONF_PORTS; i++)
        {
            int conf_id = oCONS[con_id].in_conf[i];
            if(conf_id != -1)
            {
                if(oCONS[conf_id].info->type == VOIP)
                    voip_send_dtmf(dtmf, conf_id);

                if(oCONS[conf_id].info->type == CAPI)                
                    capi_send_dtmf(dtmf, conf_id);    
            }
        }
        
        // Save DTMF and call script when necessary        
        oCONS[con_id].script->dtmf[dtmf_pos] = (char)dtmf;                
        oCONS[con_id].script->dtmf_pos++;

        if((char)dtmf == oCONS[con_id].script->delimiter[0] ||
           (dtmf_pos >= sizeof(oCONS[con_id].script->dtmf)-1) ||
           oCONS[con_id].script->delimiter[0] == '\0')
        {
            script_event_add(VOIP, s_event, oCONS[con_id].script->file, oCONS[con_id].id, oCONS[con_id].in_conf[0], oCONS[con_id].info->no_source, oCONS[con_id].info->no_target, oCONS[con_id].script->dtmf);            
            oCONS[con_id].script->dtmf_pos = 0;
            pj_memset(oCONS[con_id].script->dtmf, 0, sizeof(oCONS[con_id].script->dtmf));
        }
    }
    else
    {
        app_err(THIS_FILE, "Connection not found (%d)!", con_id);
    }

    #else
    PJ_UNUSED_ARG(con_id);
    PJ_UNUSED_ARG(ddi);
    PJ_UNUSED_ARG(dtmf);
    #endif
}
