/*********************************************************************************
 * voip_registrar.c ($Revision: 261 $)
 *
 * Copyright (c) 2006-2009 Marco Zissen <maz@v3v.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *********************************************************************************/
#include "global_headers.h"
#include "global_objects.h"

#define THIS_FILE "voip_registrar.c"

int voip_registrar_connect_clients(connection_t *mCON, char *SourceNo, char *TargetNo, int acc_id, int client_id, connection_devtype_e oDevType, connection_dir_e direction)
{
    int i = -1;

    #if DTMFBOX_USE_VOIP
    
    connection_t *CUR_CON = (connection_t*)mCON;
    int status = -1;

    // Connection already active? Then abort!
    if(CUR_CON->info->active)
        return -1;
        
    // Get new connection id
    DTMFBOX_ACC_LOCK();
    i = connection_by_free();
    if(i != -1) oCONS[i].info->type = oDevType;    
    DTMFBOX_ACC_UNLOCK();

    if(i != -1)
    {        
        char user_src[MAX_URI_LENGTH];
        char no_source[MAX_URI_LENGTH];
        char no_dest[MAX_URI_LENGTH];

        // type: client!        
        oCONS[i].capi->NCCI  = (unsigned long)-1;
        oCONS[i].info->type = oDevType;    
        oCONS[i].info->status = 1;                                
        oCONS[i].info->account_id = acc_id;
        oCONS[i].info->client_id = client_id;
        oCONS[i].info->direction |= direction;
        
        // extract number from sip uri
        strncpy(user_src, SourceNo, MAX_URI_LENGTH);        
        sip_strip_user(user_src);        

        // setup source number
        if(oDevType == VOIP && strlen(oGLB.clients[client_id].contact_param)) {
            pj_ansi_snprintf(no_source, sizeof(no_source), "<%s;%s>", SourceNo, oGLB.clients[client_id].contact_param);
        } else {
            if(oDevType == VOIP)
                pj_ansi_snprintf(no_source, sizeof(no_source), "<%s>", SourceNo);
            else
                pj_ansi_snprintf(no_source, sizeof(no_source), "%s", SourceNo);
        }

        // setup target number
        if((direction & INCOMING) && oDevType == VOIP && strlen(oGLB.clients[client_id].contact_param)) {
            pj_ansi_snprintf(no_dest, sizeof(no_dest), "<%s;%s>", TargetNo, oGLB.clients[client_id].contact_param);
        } else {
            if(oDevType == VOIP)
                pj_ansi_snprintf(no_dest, sizeof(no_dest), "<%s>", TargetNo);
            else
                pj_ansi_snprintf(no_dest, sizeof(no_dest), "%s", TargetNo);
        }

        // setup scriptfile & delimiter
        pj_ansi_snprintf(oCONS[i].script->file, sizeof(oCONS[i].script->file), "%s", oSET.scriptfile);
        pj_ansi_snprintf(oCONS[i].script->delimiter, sizeof(oCONS[i].script->delimiter), "%s", oSET.delimiter);
        
        // put into conference array
        mixer_conf_array_add(CUR_CON, i);
        mixer_conf_array_add(&oCONS[i], CUR_CON->id);
        
        // other connection information
        switch(oDevType)
        {
            case CAPI:                
                oCONS[i].capi->PLCI = (unsigned long)-2;                
                oCONS[i].capi->controller = oSET.accounts[acc_id].controller.ctrl_out;
                app_msg(THIS_FILE, "Make CAPI call by client (%s -> %s) ...", no_source, no_dest);
                status = capi_call(&oCONS[i], no_source, no_dest, oCONS[i].capi->controller, 0, oSET.scriptfile, oSET.delimiter);
                break;

            case VOIP:                                                
                app_msg(THIS_FILE, "Make VoIP call by client (%s -> %s) ...", no_source, no_dest);
                status = voip_call(&oCONS[i], no_source, no_dest, 0, oSET.scriptfile, oSET.delimiter);
                break;            
        }  

        if(status == -1)
            i = -1;
    }

    #else
        PJ_UNUSED_ARG(mCON);
        PJ_UNUSED_ARG(direction);
        PJ_UNUSED_ARG(oDevType);
        PJ_UNUSED_ARG(client_id);
        PJ_UNUSED_ARG(acc_id);
        PJ_UNUSED_ARG(TargetNo);
        PJ_UNUSED_ARG(SourceNo);    
    #endif

    return i;
}

int voip_registrar_disconnect_clients(connection_t *mCON, int keep_active_client)
{   
    int status = 0;

    #if DTMFBOX_USE_VOIP

    connection_t *CUR_CON = (connection_t*)mCON;    
    int in_conf_cpy[INCONF_PORTS];    
    int cur_conid = CUR_CON->id;        
    int j;
        
    // make a copy of in_conf array.
    // this is because in_conf gets reordered after hook down!!
    pj_memcpy(in_conf_cpy, oCONS[cur_conid].in_conf, sizeof(in_conf_cpy));
    
    // Disconnect incoming call (registrar mode)
    for(j=0; j<INCONF_PORTS; j++)
    {
        int in_conf = in_conf_cpy[j]; 
        if(in_conf != -1)
        {           
            if(oCONS[in_conf].info->active && keep_active_client)
            {
            }
            else
            {
                if(oCONS[in_conf].info->type == CAPI && (oCONS[in_conf].info->direction & CLIENT || oCONS[cur_conid].info->direction & CLIENT))
                {         
                    // maybe CAPI connection has no valid PLCI yet.
                    // give threads some time
                    while((oCONS[in_conf].capi->PLCI == (unsigned)-1 || 
                          oCONS[in_conf].capi->PLCI == (unsigned)-2) &&
                          oCONS[in_conf].info->type == CAPI)
                    {
                        pj_thread_sleep(100);
                    }

                    mixer_conf_array_remove(&oCONS[in_conf], cur_conid);
                    mixer_conf_array_remove(&oCONS[cur_conid], in_conf);
                    capi_hangup(&oCONS[in_conf]);                    
                    
                    app_dbg(THIS_FILE, "CAPI client disconnected [%d]", in_conf);                                                
                    in_conf_cpy[j] = -1;
                }
                 
                if(oCONS[in_conf].info->type == VOIP && (oCONS[in_conf].info->direction & CLIENT || oCONS[cur_conid].info->direction & CLIENT))
                {         
                    mixer_conf_array_remove(&oCONS[in_conf], cur_conid);
                    mixer_conf_array_remove(&oCONS[cur_conid], in_conf);                                     
                    voip_hangup(&oCONS[in_conf]);                                                                
                    
                    app_dbg(THIS_FILE, "VoIP client disconnected [%d]", in_conf);                                                
                    in_conf_cpy[j] = -1;
                }
            }            
        }
    }

    #else
    PJ_UNUSED_ARG(mCON);
    PJ_UNUSED_ARG(keep_active_client);
    #endif

    return status;
}

int voip_registrar_authorization_challenge(pjsip_rx_data *rdata)
{
    int statuscode = 0;

    #if DTMFBOX_USE_VOIP
    
    pjsip_msg *msg = rdata->msg_info.msg;

    // --------------------------------------------------------------------------------------------------------
    // Client registering with dtmfbox (registrar mode)
    //
    // At first, we check if there is an authorization header in the sip message. If no auth. header
    // is found, the application sends a stateless respond (status 401) to receive the md5 encrypted auth-string.
    // The next message should contain an authorization header with the encrypted pass/user/realm.
    // When authorization is successful, the client will be added to oGLB.clients
    // --------------------------------------------------------------------------------------------------------
    if((rdata->msg_info.msg->line.req.method.id == PJSIP_REGISTER_METHOD && oSET.voip.registrar)) //  ||
       //rdata->msg_info.msg->line.req.method.id == PJSIP_OTHER_METHOD) && oSET.voip.registrar)
    {
        pjsip_authorization_hdr* hdr;
        pjsip_contact_hdr *src_contact;   
        int acc_id = -1;

        hdr = (pjsip_authorization_hdr*) pjsip_msg_find_hdr(msg, PJSIP_H_AUTHORIZATION, NULL);
        src_contact = (pjsip_contact_hdr*) pjsip_msg_find_hdr(msg, PJSIP_H_CONTACT, NULL);

        if(hdr)
        {
            // Got authentication data in header...
            pj_str_t statusOk = pj_str("OK");
            pj_str_t statusWrgAuth = pj_str("Authorization failed!");
            pj_str_t statusNotAuth = pj_str("Authorization aborted! Too many clients!");

            // Check received authentication data with oSET.accounts
            statuscode = voip_registrar_check_auth_header(rdata, &acc_id);

            // Authorize or Reject?
            if(statuscode == 200 && acc_id != -1)
            {
                char tmp[MAX_URI_LENGTH];
                char tmp2[MAX_URI_LENGTH];

                pjsip_hdr hdr_list;
                pjsip_contact_hdr *hdr_contact_global;
                pjsip_contact_hdr *hdr_contact_local;
                pjsip_contact_hdr *hdr_contact_remote;

                // Create contact header                
                hdr_contact_global = pjsip_contact_hdr_create(rdata->tp_info.pool);
                hdr_contact_local = pjsip_contact_hdr_create(rdata->tp_info.pool);
                hdr_contact_remote = pjsip_contact_hdr_create(rdata->tp_info.pool);

                // global contact (maybe changed by STUN)
                {
                    char addr[MAX_URI_LENGTH];
                    int port;
                    pj_status_t status;
                    
                    pj_memset(tmp, 0, sizeof(tmp));
                    status = voip_resolve_transport_address(addr, &port);
                    if(status == PJ_SUCCESS)                        
                        pj_ansi_snprintf(tmp, sizeof(tmp), "%s%s<sip:%s%s%s:%d;transport=UDP>",
                                                            oSET.accounts[acc_id].name, 
                                                            (strlen(oSET.accounts[acc_id].name)?" " : ""), 
                                                            oSET.accounts[acc_id].registrar_username,
			                                                (strlen(oSET.accounts[acc_id].registrar_username)?"@":""),
                                                            addr, port);
                    else
                        pj_ansi_snprintf(tmp, sizeof(tmp), "%s%s<sip:%s%s%s:%d;transport=UDP>",
                                                            oSET.accounts[acc_id].name, 
                                                            (strlen(oSET.accounts[acc_id].name)?" " : ""), 
                                                            oSET.accounts[acc_id].registrar_username,
			                                                (strlen(oSET.accounts[acc_id].registrar_username)?"@":""),
                                                            oGLB.ip_addr_global, oGLB.udp_port_global);

                    hdr_contact_global->expires = oSET.voip.register_interval;
                    hdr_contact_global->uri = pjsip_parse_uri(rdata->tp_info.pool, tmp, strlen(tmp), 0);
                }
                
                // local contact 
                pj_memset(tmp2, 0, sizeof(tmp2));                
                pj_ansi_snprintf(tmp2, sizeof(tmp2), "%s%s<sip:%s%s%s:%d;transport=UDP>",
                                    oSET.accounts[acc_id].name, 
                                    (strlen(oSET.accounts[acc_id].name)?" " : ""), 
                                    oSET.accounts[acc_id].registrar_username,
                                    (strlen(oSET.accounts[acc_id].registrar_username)?"@":""),
                                    oGLB.ip_addr_local, oGLB.udp_port_local);

                hdr_contact_local->expires = oSET.voip.register_interval;
                hdr_contact_local->uri = pjsip_parse_uri(rdata->tp_info.pool, tmp2, strlen(tmp2), 0);                

                // Add headers to hdr_list
                pj_list_init(&hdr_list);                
                pj_list_push_back(&hdr_list, (pjsip_hdr*)oGLB.useragent_hdr);
                pj_list_push_back(&hdr_list, (pjsip_contact_hdr*)hdr_contact_global);
                pj_list_push_back(&hdr_list, (pjsip_contact_hdr*)hdr_contact_local);

                // add remote contact header (if exist)
                {
                    pjsip_contact_hdr *hdr_contact_rem = (pjsip_contact_hdr*) pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);                    
                    if(hdr_contact_rem != NULL)
                    {
                        if(hdr_contact_rem->uri != NULL)
                        {
                            hdr_contact_remote->uri = pjsip_uri_clone(rdata->tp_info.pool, hdr_contact_rem->uri);
                            hdr_contact_remote->expires = oSET.voip.register_interval;
                            pj_list_push_back(&hdr_list, (pjsip_contact_hdr*)hdr_contact_remote);
                        }
                    }
                }
                               
                
                // Ok (Registered/Unregistered)!
                {   
                    pjsip_endpt_respond_stateless(oGLB.endpt, rdata, statuscode, &statusOk, &hdr_list, NULL);
                    app_dbg(THIS_FILE, "Status: %d (%s)", statuscode, statusOk);
                }

            }
            else if(statuscode == 401 || statuscode == 407)
            {
                // Authorization failed!
                pjsip_endpt_respond_stateless(oGLB.endpt, rdata, statuscode, &statusWrgAuth, NULL, NULL);
                app_dbg(THIS_FILE, "Status: %d (%s)", statuscode, statusWrgAuth);
            }
            else if(statuscode == 487)
            {
                // Too many clients!
                pjsip_endpt_respond_stateless(oGLB.endpt, rdata, statuscode, &statusNotAuth, NULL, NULL);
                app_dbg(THIS_FILE, "Status: %d (%s)", statuscode, statusNotAuth);
            }
            else
            {
                // Authorization failed!
                pjsip_endpt_respond_stateless(oGLB.endpt, rdata, statuscode, NULL, NULL, NULL);
                app_dbg(THIS_FILE, "Status: %d (Unknown)", statuscode);
            }

            return PJ_TRUE;
        }
        else
        {
            // Request authorization in header (when not received!)
            voip_registrar_req_auth_header(rdata);
            return PJ_TRUE;
        }
    }

    #else
    PJ_UNUSED_ARG(rdata);
    #endif

    return statuscode;
}


pj_bool_t voip_registrar_rx_request(pjsip_rx_data *rdata)
{
    #if DTMFBOX_USE_VOIP

    connection_t *CUR_CON = NULL;
    pj_status_t status;
    int i,j,x;
    int dst_conid = -1;
    pjsip_uri *uri;
    pjsip_sip_uri *sip_uri;
    char to[MAX_URI_LENGTH];
    char from[MAX_URI_LENGTH];
    char host[MAX_URI_LENGTH];    

    char SourceNo[MAX_URI_LENGTH]; 
    char TargetNo[MAX_URI_LENGTH];

    int incoming_direct = 0;
    int incoming_account = 0;
    int incoming_registrar = 0;
    int outgoing_registrar = 0;
    int accountid = -1;
    int clientid = -1;
    
    // Status
    //app_dbg(THIS_FILE, "RX-REQUEST %d bytes %s from %s %s:%d", rdata->msg_info.len, pjsip_rx_data_get_info(rdata), rdata->tp_info.transport->type_name, rdata->pkt_info.src_name, rdata->pkt_info.src_port);      
    //app_dbg2(THIS_FILE, "RX-SIP:\n\n%.*s\n", rdata->msg_info.len, rdata->msg_info.msg_buf);
    
    // REGISTRAR?
    if(voip_registrar_authorization_challenge(rdata))
        goto hell;
        
    // PUBLISH?
    if(rdata->msg_info.msg->line.req.method.id == PJSIP_OTHER_METHOD && oSET.voip.registrar)
    {            
        pj_str_t reason;
        int code = 405;
        
        app_err(THIS_FILE, "Method not allowed (%d)!", rdata->msg_info.msg->line.req.method.id);        
        reason = pj_str("Method not allowed!");
                
        pjsip_endpt_respond_stateless(oGLB.endpt, rdata, code, &reason, NULL, NULL);        
        return PJ_TRUE;        
    }

    if(voip_check_type(rdata, &accountid, &clientid, &incoming_direct, &incoming_account, &incoming_registrar, &outgoing_registrar, from, to))
        return PJ_FALSE;

    if((incoming_registrar == 0 && outgoing_registrar == 0))
        return PJ_FALSE;

    CUR_CON = voip_alert_call(rdata);
    if(CUR_CON == NULL) 
        goto hell;
    
    // --------------------------------------------------------------------------------------------------------
    // INCOMING/OUTGOING (Registrar-Mode):
    //
    // On incoming request, we call each registered client.
    // On outgoing request, we just make a call and wait for connect 
    // --------------------------------------------------------------------------------------------------------
    if(oSET.voip.registrar)
    {   
        // REGISTRAR: INCOMING CALL
        //if(!found)
        if(incoming_registrar)
        {               
            x = 0;

            // Search for clients
            for(i=0; i<oSET.voip.max_clients; i++)
            {
                // Search for account
                for(j=0; j<MAX_ACCOUNTS; j++)
                {                    
                    int acc_id = oGLB.clients[i].account_id[j];

                    // Account and client found? Call!                    
                    if(acc_id == accountid)
                    {
                        // Get sip-uri
                        uri = rdata->msg_info.to->uri;
                        sip_uri = (pjsip_sip_uri*)pjsip_uri_get_uri(uri);

                        if (PJSIP_URI_SCHEME_IS_SIP(uri))
                        {
                            char ip[MAX_URI_LENGTH];
                            int port;
                            
                            pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, sip_uri, to, MAX_URI_LENGTH);
                            pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, sip_uri, host, MAX_URI_LENGTH);
                            sip_strip_user((char*)&to);
                            sip_strip_host((char*)&host);
                            
                            pj_ansi_snprintf(ip, sizeof(ip), "%s", oGLB.clients[i].addr);                            
                            pj_ansi_snprintf(TargetNo, sizeof(TargetNo), "sip:%s@%s", oSET.accounts[acc_id].registrar_username, oGLB.clients[i].addr);
                                                        
                            // public or private ip?
                            // change host SourceNo, so that media flow is correctly (private/public ip)
                            {                                
                                strcpy(to, CUR_CON->info->no_target);
                                sip_strip_user(to);
                            
                                status = voip_resolve_transport_address(ip, &port);
                                if(status != PJ_SUCCESS)
                                {
                                    app_err(THIS_FILE, "(%d) Unable to resolve transport address", status);
                                    break;
                                }                                    
                                pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "sip:%s@%s:%d", to, ip, port);

                                dst_conid = voip_registrar_connect_clients(CUR_CON, SourceNo, TargetNo, accountid, i, VOIP, INCOMING | CLIENT);
                            }
                            break;
                        }
                    }
                }
            }
        }

        // REGISTRAR: OUTGOING CALL
        else if(outgoing_registrar)
        {               
            // INIT OUTGOING VOIP CALL -->
            if(oSET.accounts[accountid].controller.device == VOIP)
            {
                pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "sip:%s", oSET.accounts[accountid].id);
                pj_ansi_snprintf(TargetNo, sizeof(TargetNo), "sip:%s", to);                
            }
            
            // INIT OUTGOING CAPI CALL -->
            else if(oSET.accounts[accountid].controller.device == CAPI)
            {                
                sip_strip_user(to);
                pj_ansi_snprintf(SourceNo, sizeof(SourceNo), "%s", oSET.accounts[accountid].number);                
                pj_ansi_snprintf(TargetNo, sizeof(TargetNo), "%s", to);                         
            }
                     
            {
                int outgoing_id = -1;

                // Get temporary connection 
                DTMFBOX_ACC_LOCK();
                outgoing_id = connection_by_free();                
                DTMFBOX_ACC_UNLOCK();

                // Make call (by thread)
                if(outgoing_id != -1)
                {
                    pj_memcpy(&oCONS[outgoing_id].info->no_source, SourceNo, sizeof(oCONS[outgoing_id].info->no_source));
                    pj_memcpy(&oCONS[outgoing_id].info->no_target, TargetNo, sizeof(oCONS[outgoing_id].info->no_target));
                    oCONS[outgoing_id].info->client_id = clientid;
                    oCONS[outgoing_id].info->account_id = accountid;
                    oCONS[outgoing_id].in_conf[0] = CUR_CON->id;

                    pj_thread_create(CUR_CON->pools->con_pool, "registrar_connect_%p", (pj_thread_proc*)&registrar_connect_outgoing, (void*)outgoing_id, 0, (unsigned)NULL, &CUR_CON->tone_thread);
                }
            }            
        }
    }
    else
    {
        app_err(THIS_FILE, "Registrar- and client-mode inactive! You must activate one or both mode(s)!");
        goto hell;
    }

    // ok!
    return PJ_TRUE;


// Error! Reset connection(s)...
hell:
    if(CUR_CON != NULL)
        reset_connection(CUR_CON);

    if(dst_conid >= 0)
        reset_connection(&oCONS[dst_conid]);
    
    #else
        PJ_UNUSED_ARG(rdata);
    #endif

    return PJ_FALSE;    
}
